package de.lmu.ifi.pms.cep.dura.compiler.dura;

import java.io.File;
import java.io.IOException;
import java.util.Collection;

import org.antlr.runtime.ANTLRFileStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.tree.CommonTree;
import org.antlr.runtime.tree.CommonTreeNodeStream;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraCAstValidation;
import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraLexer;
import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraParser;
import de.lmu.ifi.pms.cep.dura.compiler.dura.desugaring.Desugarer;
import de.lmu.ifi.pms.cep.dura.compiler.durac.DuraCCompiler;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraCModuleNotFoundException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraParseException;
import de.lmu.ifi.pms.cep.tsa.TSA_Program;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;


public class DuraCompiler {
	public static CommonTree parse(final ANTLRFileStream stream) throws RecognitionException {
		DuraLexer lexer = new DuraLexer(stream);
		CommonTokenStream tokens = new CommonTokenStream(lexer);
		DuraParser parser = new DuraParser(tokens);
		
		return (CommonTree) parser.program().getTree();
	}
	
	public static void compile(final String filename) throws IOException, RecognitionException {
		ANTLRFileStream stream = new ANTLRFileStream(filename);
		CommonTree tree = parse(stream);
		compile(tree);
	}
	
	public static void compile(final CommonTree tree) throws IOException, RecognitionException {
		Desugarer.desugar(tree);
		
		System.out.println();
		
		CommonTreeNodeStream nodes = new CommonTreeNodeStream(tree);
		
		DuraCAstValidation walker = new DuraCAstValidation(nodes);
        try {
            walker.program();
        } catch (RecognitionException e) {
            e.printStackTrace();
        }
	}
	
	public static TSA_Program compile(final String namespace, final String program_name, final File file) throws IOException {
		return compile(namespace, program_name, file.getAbsolutePath());
	}
	
	public static TSA_Program compile(final String namespace, final String program_name, final String filename) throws IOException {
		ANTLRFileStream stream = new ANTLRFileStream(filename);
		
		try {
			CommonTree ast = parse(stream);
			Desugarer.desugar(ast);

			DuraCAstValidation walker = new DuraCAstValidation(new CommonTreeNodeStream(ast));
			
			String durac = walker.program();
			
			System.out.println(durac);
	         
	        return DuraCCompiler.compile(namespace, program_name, filename, durac);
		} catch (RecognitionException e) {
			throw new DuraParseException(e, filename);
		}
	}
	
	public static TSA_Program compile(final String seed, final Collection<String> sourceFiles) throws IOException {
		if (sourceFiles == null) {
			throw new DuraCModuleNotFoundException(seed);
		}

		String[] moduleName = TSA_SyntaxPatterns.programFullName(seed);
		if(sourceFiles.size() == 0){
			throw new IllegalArgumentException("No source file specified.");
		}
		else if(sourceFiles.size() == 1){
			return compile(moduleName[0], moduleName[1], sourceFiles.iterator().next());
		}
		else{
			throw new IllegalArgumentException("Only one source file can be specified.");
		}
		
	}
	
	public static void main(final String[] args) throws Exception {
//		String filename = new File("src/emili/dura/compiler/durac/antlr/DuraC.example").getAbsolutePath();
//		String filename = new File("src/de/lmu/ifi/pms/cep/dura/compiler/dura/Dura.example").getAbsolutePath();
		
		compile("","", args[0]);
	}
}
