package de.lmu.ifi.pms.cep.dura.compiler.dura.desugaring;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.Stack;

import org.antlr.runtime.CommonToken;
import org.antlr.runtime.tree.CommonTree;
import org.antlr.runtime.tree.Tree;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraParser;


public class ActionSpecifications extends Transformation {

	// This is now done in ReactiveRuleActionExecutionProvenanceInfoAugmentation!
	/*
	private CommonTree compositionAugmentationTerm;
	private String compositionAugmentationTermString = "_composition { child-index{ integer }, parent-action-id{ identifier } } }";
	
	private HashSet<String> compositionAugmentedActionSet = new HashSet<String>();

	public ActionSpecifications() {
		super();
		
		ANTLRStringStream stream = new ANTLRStringStream(compositionAugmentationTermString);
		DuraLexer lexer = new DuraLexer(stream);
		CommonTokenStream tokens = new CommonTokenStream(lexer);
		DuraParser parser = new DuraParser(tokens);
			
		try {
			compositionAugmentationTerm = (CommonTree) parser.internalSchemaDefinition().getTree();
			normalizeTerm(compositionAugmentationTerm);
		} catch (RecognitionException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	*/
	
	
	private HashSet<String> actionsWithSuccessSpecificationSet;  
	
	@Override
	public void apply(CommonTree tree) {
		actionsWithSuccessSpecificationSet = new HashSet<String>();
		
		CommonTree searchRoot;
		Tree temp = tree.getChild(0).getChild(1);
		
		searchRoot = (CommonTree) temp;
		
		for (int i = 0; i < searchRoot.getChildCount(); i++) {
			temp = searchRoot.getChild(i);
			
			if (temp.getType() == DuraParser.ACTION) {
				if (temp.getChild(2).getChildCount() > 0) {
					actionsWithSuccessSpecificationSet.add(temp.getChild(1).getChild(0).getText());
					continue;
				}
				
				if (temp.getChildCount() > 4) {
					if (temp.getChild(4).getType() == DuraParser.FOR) {
						CommonTree actSpec = (CommonTree) temp.getChild(4);
						
						CommonTree head = (CommonTree) actSpec.getChild(1);
						CommonTree concurrent = (CommonTree) head.getChild(0);
						
						CommonTree succeedsOn = (CommonTree) concurrent.getChild(concurrent.getChildCount() - 2);
						if (succeedsOn.getChildCount() > 0) {
							actionsWithSuccessSpecificationSet.add(temp.getChild(1).getChild(0).getText());
						}
					}
				}
			}
		}
		
		
		for (CommonTree actionSpecification : collectActionSpecificationNodes(tree)) {		
			CommonTree body = (CommonTree) actionSpecification.getChild(0);
			
			HashSet<String> actionIdentSet = new HashSet<String>();
			
			
			int nextId = getNextNameId();
			resetNextNameId();
			
			CommonTree parentTerm = (CommonTree) body.getChild(1);
			String parentIdent = null;
			if (body.getChild(0).getChildCount() > 0) {
				parentIdent = body.getChild(0).getChild(0).getText();
			} else {
				// parentIdent = createName();
				parentIdent = "_no_ident_parent";
			}
			actionIdentSet.add(parentIdent);
			
			
			String parentActionName = parentTerm.getText();
						
			if (!actionCompositionAugmentedActionSet.contains(parentActionName)) {
				addActionCompositionInformation(tree, parentActionName);
				
				actionCompositionAugmentedActionSet.add(parentActionName);
				
				if (EMIT_ANNOTATIONS) {
					CommonTree actionDefinition = (CommonTree) explicitActionSchemataMap.get(parentActionName).getParent().getParent();
					CommonTree annotation = new CommonTree(new CommonToken(DuraParser.ANNOTATION, "ANNOTATION"));
					actionDefinition.addChild(annotation);
					annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Result of transformation " + this.getClass().getSimpleName() + ".\"")));
					annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Addition of composition-provenance information to schema of this action due to action specification of action " + parentActionName + ".\"")));
				}
			}
			
			CommonTree head = (CommonTree) actionSpecification.getChild(1);
			CommonTree concurrent = (CommonTree) head.getChild(0);
				
			HashMap<String, Integer> childActionIdentToIndexMap = new HashMap<String, Integer>();
			ArrayList<CommonTree> childActionTerms = new ArrayList<CommonTree>();
			
			for (int i = 0; i < concurrent.getChildCount() - 3; i++) {
				CommonTree child = (CommonTree) concurrent.getChild(i);
				
				CommonTree childTerm = (CommonTree) child.getChild(1);
				String childIdent = null;
				if (child.getChild(0).getChildCount() > 0) {
					childIdent = child.getChild(0).getChild(0).getText();
					
					if (actionIdentSet.contains(childIdent)) {
						throw new RuntimeException("Action identifier " + childIdent + " has already been defined!");
					}
				} else {
					childIdent = createName();
				}
				
				
				actionIdentSet.add(childIdent);
					
				childActionIdentToIndexMap.put(childIdent, childActionTerms.size());
				
				
				childActionTerms.add(childTerm);
					
				String childActionName = childTerm.getText();
					
				if (!actionCompositionAugmentedActionSet.contains(childActionName)) {
					addActionCompositionInformation(tree, childActionName);
						
					actionCompositionAugmentedActionSet.add(childActionName);
					
					if (EMIT_ANNOTATIONS) {
						CommonTree actionDefinition = (CommonTree) explicitActionSchemataMap.get(childActionName).getParent().getParent();
						CommonTree annotation = new CommonTree(new CommonToken(DuraParser.ANNOTATION, "ANNOTATION"));
						actionDefinition.addChild(annotation);
						annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Result of transformation " + this.getClass().getSimpleName() + ".\"")));
						annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Addition of composition-provenance information to schema of this action due to action specification of action " + parentActionName + ".\"")));
					}
				}
			}
			
			setNextNameId(nextId);
			
			CommonTree execConstraints = (CommonTree) concurrent.getChild(concurrent.getChildCount() - 3);
			
			translateComposition(actionSpecification, parentIdent, parentTerm, actionIdentSet, execConstraints, childActionIdentToIndexMap, childActionTerms);
			
			CommonTree succeedsOn = (CommonTree) concurrent.getChild(concurrent.getChildCount() - 2);
			if (succeedsOn.getChildCount() > 0) {
				translateSucceedsFailsOn("$succeeded", (CommonTree) actionSpecification.getParent(), 0, succeedsOn, parentIdent, parentTerm, childActionIdentToIndexMap, childActionTerms);
			}
			
			CommonTree failsOn = (CommonTree) concurrent.getChild(concurrent.getChildCount() - 1);
			if (failsOn.getChildCount() > 0) {
				translateSucceedsFailsOn("$failed", (CommonTree) actionSpecification.getParent(), 2, failsOn, parentIdent, parentTerm, childActionIdentToIndexMap, childActionTerms);
			}
			
			actionSpecification.getParent().deleteChild(actionSpecification.childIndex);
		}
		
		/*
		// Debug
		System.out.println();
		
		// Debug
		for (CommonTree c : actionSchemataMap.values()) {
			System.out.println(c.toStringTree());
		}
		
		// Debug
		System.out.println();
		
		// Debug
		for (CommonTree c : eventSchemataMap.values()) {
			System.out.println(c.toStringTree());
		}
		*/
		
		
		temp = tree.getChild(0).getChild(1);
		
		searchRoot = (CommonTree) temp;
		
		HashMap<String, Integer> emptyIdentToIndexMap = new HashMap<String, Integer>();
		ArrayList<CommonTree> emptyActionTerms = new ArrayList<CommonTree>();
		
		for (int i = 0; i < searchRoot.getChildCount(); i++) {
			temp = searchRoot.getChild(i);
			
			if (temp.getType() == DuraParser.ACTION) {
				CommonTree schemaTerm = (CommonTree) temp.getChild(1).getChild(0);
				
				if (!actionCompositionAugmentedActionSet.contains(schemaTerm.getText())) {
					addActionCompositionInformation(tree, schemaTerm.getText());
						
					actionCompositionAugmentedActionSet.add(schemaTerm.getText());
					
					if (EMIT_ANNOTATIONS) {
						CommonTree actionDefinition = (CommonTree) explicitActionSchemataMap.get(schemaTerm.getText()).getParent().getParent();
						CommonTree annotation = new CommonTree(new CommonToken(DuraParser.ANNOTATION, "ANNOTATION"));
						actionDefinition.addChild(annotation);
						annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Result of transformation " + this.getClass().getSimpleName() + ".\"")));
						annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Addition of composition-provenance information to schema of this action.\"")));
					}
				}
				
				
				CommonTree succeedsOn = (CommonTree) temp.getChild(2);
				if (temp.getChild(2).getChildCount() > 0) {
					translateSucceedsFailsOn("$succeeded", (CommonTree) temp, 0, succeedsOn, "_no_ident_parent", schemaTerm, emptyIdentToIndexMap, emptyActionTerms);					
				}
				
				CommonTree failsOn = (CommonTree) temp.getChild(3);
				if (temp.getChild(3).getChildCount() > 0) {
					translateSucceedsFailsOn("$failed", (CommonTree) temp, 2, failsOn, "_no_ident_parent", schemaTerm, emptyIdentToIndexMap, emptyActionTerms);					
				}
				
				temp.deleteChild(2);
				temp.deleteChild(2);
			}
		}
	}
	
	private void translateSucceedsFailsOn(String labelSuffix, CommonTree actionDefinition, int addIndexOffset, CommonTree succeedsOnOrFailsOn, String parentIdent, CommonTree parentTerm,
			HashMap<String, Integer> childActionIdentToIndexMap, ArrayList<CommonTree> childActionTerms) {
		
		CommonTree defs = (CommonTree) actionDefinition.getParent();
		int addIndex = actionDefinition.getChildIndex() + 1 + addIndexOffset;

		CommonTree eventDefinition = new CommonTree(new CommonToken(DuraParser.EVENT, "EVENT"));
		defs.getChildren().add(addIndex, eventDefinition);
		addIndex++;
		
		eventDefinition.addChild(new CommonTree(new CommonToken(DuraParser.MODIFIERS, "MODIFIERS")));
		
		CommonTree schema = new CommonTree(new CommonToken(DuraParser.SCHEMA, "SCHEMA"));
		eventDefinition.addChild(schema);
		
		String label = parentTerm.getText() + labelSuffix;
		
		CommonTree temp = eventSchemataMap.get(label);
		int payloadIndex;
		for (payloadIndex = 0; payloadIndex < temp.getChildCount(); payloadIndex++) {
			if (temp.getChild(payloadIndex).getText().equals("payload")) {
				break;
			}
		}
		
		CommonTree payloadSchema = (CommonTree) temp.getChild(payloadIndex);
		
		CommonTree schemaTerm = new CommonTree(new CommonToken(DuraParser.TERM, label));
		schema.addChild(schemaTerm);
		
		schemaTerm.addChild(copyTree(payloadSchema));
		
		schema.addChild(new CommonTree(new CommonToken(DuraParser.SUPPLEMENT, "SUPPLEMENT")));
		
		String name = createName();
		CommonTree helperEvent = createSucceedsFailsEvent(name, payloadSchema, succeedsOnOrFailsOn, parentIdent, parentTerm, childActionIdentToIndexMap, childActionTerms);
		
		defs.getChildren().add(addIndex, helperEvent);
		
		
		CommonTree helperEventSchemaTerm = (CommonTree) helperEvent.getChild(1).getChild(0);
		CommonTree helperEventSchemaTermCopy = copyTree(helperEventSchemaTerm);
		
		joinTermCopiesWithTermChildren(implicitEventSchemaTerms, helperEventSchemaTermCopy);
		eventSchemataMap.put(name, helperEventSchemaTermCopy);
		
		explicitEventSchemataMap.put(name, helperEventSchemaTerm);
		
		
		
		// BEGIN
		
		// ^(DETECT ^(HEAD constructTerm (^(AGGREGATION ^(GROUPING binding+)))?) ^(BODY eventQuery))
		
		CommonTree eventSpec = new CommonTree(new CommonToken(DuraParser.DETECT, "DETECT"));
		eventDefinition.addChild(eventSpec);
		
		CommonTree head = new CommonTree(new CommonToken(DuraParser.HEAD, "HEAD"));
		eventSpec.addChild(head);
		
		
		CommonTree constructTerm = copyTree((CommonTree) helperEvent.getChild(2).getChild(0).getChild(0));
		head.addChild(constructTerm);
		
		constructTerm.getToken().setText(label);
		
		
		CommonTree body = new CommonTree(new CommonToken(DuraParser.BODY, "BODY"));
		eventSpec.addChild(body);
		
		CommonTree conjunction = new CommonTree(new CommonToken(DuraParser.CONJUNCTION, "CONJUNCTION"));
		body.addChild(conjunction);
		
		
		// Positive
		CommonTree positiveHelperEventQuery = new CommonTree(new CommonToken(DuraParser.EVENT_QUERY, "EVENT_QUERY"));
		conjunction.addChild(positiveHelperEventQuery);
		
		CommonTree positiveHelperEventQueryIdentifier = new CommonTree(new CommonToken(DuraParser.IDENTIFIER, "event"));
		positiveHelperEventQuery.addChild(positiveHelperEventQueryIdentifier);
		positiveHelperEventQueryIdentifier.addChild(new CommonTree(new CommonToken(DuraParser.ID, "a")));
		
		CommonTree query = new CommonTree(new CommonToken(DuraParser.QUERY, "QUERY"));
		positiveHelperEventQuery.addChild(query);
		
		CommonTree queryTerm = copyTree(constructTerm);
		queryTerm.getToken().setText(name);
		query.addChild(queryTerm);
		
		CommonTree idTerm = new CommonTree(new CommonToken(DuraParser.TERM, "id"));
		queryTerm.getChildren().add(0, idTerm);
		queryTerm.freshenParentAndChildIndexes();
		
		CommonTree positiveHelperEventQueryIdVariable = new CommonTree(new CommonToken(DuraParser.VARIABLE, "VARIABLE"));
		positiveHelperEventQueryIdVariable.addChild(new CommonTree(new CommonToken(DuraParser.ID, "A")));
		idTerm.addChild(positiveHelperEventQueryIdVariable);
		
		CommonTree querySupplement = new CommonTree(new CommonToken(DuraParser.SUPPLEMENT, "SUPPLEMENT"));
		positiveHelperEventQuery.addChild(querySupplement);		
		
		
		// Negative
		CommonTree negation = new CommonTree(new CommonToken(DuraParser.NEGATION, "NEGATION"));
		conjunction.addChild(negation);
		
		CommonTree negativeHelperEventQuery = copyTree(positiveHelperEventQuery);
		negation.addChild(negativeHelperEventQuery);
		
		CommonTree negativeHelperEventQueryIdentifier = (CommonTree) negativeHelperEventQuery.getChild(0);
		((CommonTree) negativeHelperEventQueryIdentifier.getChild(0)).getToken().setText("b");
		
		CommonTree negativeHelperEventQueryIdVariable = (CommonTree) negativeHelperEventQuery.getChild(1).getChild(0).getChild(0).getChild(0);
		((CommonTree) negativeHelperEventQueryIdVariable.getChild(0)).getToken().setText("B");
		
		
		
		CommonTree negationSupplement = new CommonTree(new CommonToken(DuraParser.SUPPLEMENT, "SUPPLEMENT"));
		negation.addChild(negationSupplement);
		
		CommonTree negationSupplementCondition = new CommonTree(new CommonToken(DuraParser.CONDITION, "CONDITION"));
		negationSupplement.addChild(negationSupplementCondition);
		
		CommonTree negationSupplementConjunction = new CommonTree(new CommonToken(DuraParser.CONJUNCTION, "CONJUNCTION"));
		negationSupplementCondition.addChild(negationSupplementConjunction);
		
		
		
		CommonTree constraint = new CommonTree(new CommonToken(DuraParser.ARITHMETIC_RELATION, "<="));
		negationSupplementConjunction.addChild(constraint);
		
		CommonTree left = new CommonTree(new CommonToken(DuraParser.INTERVAL_OP, "end"));
		constraint.addChild(left);
		left.addChild(copyTree(negativeHelperEventQueryIdentifier));
		
		CommonTree right = new CommonTree(new CommonToken(DuraParser.INTERVAL_OP, "end"));
		constraint.addChild(right);
		right.addChild(copyTree(positiveHelperEventQueryIdentifier));
		
		
		constraint = new CommonTree(new CommonToken(DuraParser.ARITHMETIC_RELATION, "<"));
		negationSupplementConjunction.addChild(constraint);
		
		left = copyTree(negativeHelperEventQueryIdVariable);
		constraint.addChild(left);
		
		right = copyTree(positiveHelperEventQueryIdVariable);
		constraint.addChild(right);
		
		
		
		CommonTree supplement = new CommonTree(new CommonToken(DuraParser.SUPPLEMENT, "SUPPLEMENT"));
		conjunction.addChild(supplement);
		
		
		// END
		
		if (EMIT_ANNOTATIONS) {
			CommonTree annotation = new CommonTree(new CommonToken(DuraParser.ANNOTATION, "ANNOTATION"));
			eventDefinition.addChild(annotation);
			annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Result of transformation " + this.getClass().getSimpleName() + ".\"")));
			annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Translation of " + succeedsOnOrFailsOn.getText() + " for action " + parentTerm.getText() + ".\"")));
			annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Maintains that the event is only generated once using helper event " + helperEvent.getChild(1).getChild(0).getText() + ".\"")));
			
			annotation = new CommonTree(new CommonToken(DuraParser.ANNOTATION, "ANNOTATION"));
			helperEvent.addChild(annotation);
			annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Result of transformation " + this.getClass().getSimpleName() + ".\"")));
			annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Translation of " + succeedsOnOrFailsOn.getText() + " for action " + parentTerm.getText() + ".\"")));
			annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Helper event! Does NOT maintain that this is event only detected once. Deductive rule for event " + label + " does this!\"")));
		}
		
		
		// TODO: Die eigtl. Anfrage muss noch einmal ausgelagert werden... sagen wir event _x.  Dann hierin:
		//       and {
		//         event a: _x{ id{ var A }, payload{ id{ var C }, <...> } },
		//         not { event b: _x{ id{ var B }, payload{ id{ var C } } } } where { end(event b) - end(event a) <= 0 ms, var A > var B }
		//       }
		//       Damit garantieren, dass nur EINMAL das succeeded event erzeugt wird.
		
		
		
		defs.freshenParentAndChildIndexes();
	}

	private CommonTree createSucceedsFailsEvent(String name, CommonTree payloadSchema, CommonTree succeedsOnOrFailsOn, String parentIdent, CommonTree parentTerm,
			HashMap<String, Integer> childActionIdentToIndexMap, ArrayList<CommonTree> childActionTerms) {
		int nextId = getNextNameId();
		resetNextNameId();

		CommonTree eventDefinition = new CommonTree(new CommonToken(DuraParser.EVENT, "EVENT"));
		
		eventDefinition.addChild(new CommonTree(new CommonToken(DuraParser.MODIFIERS, "MODIFIERS")));
		
		CommonTree schema = new CommonTree(new CommonToken(DuraParser.SCHEMA, "SCHEMA"));
		eventDefinition.addChild(schema);
		
		CommonTree schemaTerm = new CommonTree(new CommonToken(DuraParser.TERM, name));
		schema.addChild(schemaTerm);
		
		schemaTerm.addChild(copyTree(payloadSchema));
			
		schema.addChild(new CommonTree(new CommonToken(DuraParser.SUPPLEMENT, "SUPPLEMENT")));
		
		// ^(DETECT ^(HEAD constructTerm (^(AGGREGATION ^(GROUPING binding+)))?) ^(BODY eventQuery))
		
		CommonTree eventSpec = new CommonTree(new CommonToken(DuraParser.DETECT, "DETECT"));
		eventDefinition.addChild(eventSpec);
		
		CommonTree head = new CommonTree(new CommonToken(DuraParser.HEAD, "HEAD"));
		eventSpec.addChild(head);
		
		CommonTree constructTerm = copyTree(schemaTerm);
		head.addChild(constructTerm);
		
		CommonTree actionIdVariableNode = null;
		for (CommonTree type : collectTypes(constructTerm)) {
			Tree parent = type.getParent();
			parent.deleteChild(0);
			CommonTree v = createVariableNode("for" + name);
			parent.addChild(v);
			
			if (parent.getText().equals("id")) {
				actionIdVariableNode = v; 
			}
		}
		
		CommonTree body = new CommonTree(new CommonToken(DuraParser.BODY, "BODY"));
		eventSpec.addChild(body);
		
		CommonTree conjunction = new CommonTree(new CommonToken(DuraParser.CONJUNCTION, "CONJUNCTION"));
		body.addChild(conjunction);
		
		
		// ^(EVENT_QUERY ^(IDENTIFIER[$i] $j?) ^(QUERY ^(IDENTIFIER[$k] $l)? term) querySupplement)
		
		CommonTree initEventQuery = new CommonTree(new CommonToken(DuraParser.EVENT_QUERY, "EVENT_QUERY"));
		conjunction.addChild(initEventQuery);
		
		CommonTree identifier = new CommonTree(new CommonToken(DuraParser.IDENTIFIER, "event"));
		initEventQuery.addChild(identifier);
		
		CommonTree query = new CommonTree(new CommonToken(DuraParser.QUERY, "QUERY"));
		initEventQuery.addChild(query);
		
		CommonTree initQueryTerm = copyTree(constructTerm);
		initQueryTerm.getToken().setText(parentTerm.getText() + "$initiated");
		query.addChild(initQueryTerm);
		
		CommonTree querySupplement = new CommonTree(new CommonToken(DuraParser.SUPPLEMENT, "SUPPLEMENT"));
		initEventQuery.addChild(querySupplement);
		
		
		
		CommonTree eventQuery = new CommonTree(new CommonToken(DuraParser.EVENT_QUERY, "EVENT_QUERY"));
		conjunction.addChild(eventQuery);
		
		identifier = new CommonTree(new CommonToken(DuraParser.IDENTIFIER, "event"));
		eventQuery.addChild(identifier);
		
		query = new CommonTree(new CommonToken(DuraParser.QUERY, "QUERY"));
		eventQuery.addChild(query);
		
		
		
		
		
		CommonTree specialQueryConjunctionOrDisjunction = (CommonTree) succeedsOnOrFailsOn.getChild(0);
		query.addChild(specialQueryConjunctionOrDisjunction);
		
		querySupplement = new CommonTree(new CommonToken(DuraParser.SUPPLEMENT, "SUPPLEMENT"));
		eventQuery.addChild(querySupplement);
		
		
		CommonTree shortenedInitEventQuery = copyTree(initEventQuery);
		CommonTree shortenedInitEventQueryTermPayload = (CommonTree) shortenedInitEventQuery.getChild(1).getChild(0).getChild(0);
		
		for (int i = 0; i < shortenedInitEventQueryTermPayload.getChildCount();) {
			if (!shortenedInitEventQueryTermPayload.getChild(i).getText().equals("id")) {
				shortenedInitEventQueryTermPayload.deleteChild(i);
			} else {
				i++;
			}
		}
		
		IdentityHashMap<CommonTree, Boolean> markedEventQueriesMap = resolveActionInstanceReferences(shortenedInitEventQuery, actionIdVariableNode, eventQuery, parentIdent, parentTerm, childActionIdentToIndexMap, childActionTerms);
		
		
		
		
		// addShortenedInitEventQueries(actionIdVariableNode.getChild(0).getText(), shortenedInitEventQuery, eventQuery, markedEventQueriesMap);
		
		// Refer to the explanation above the addShortenedInitEventQueries method!
		if (markedEventQueriesMap.get(eventQuery) != null && !exposesVariable(eventQuery, actionIdVariableNode.getChild(0).getText())) {
			throw new RuntimeException(
					"The query expressed by the succeeds / fails on block ( " + eventQuery.toStringTree() + " ) does not expose the required bindings.\n" +
					"If such a block contains an action instance reference in a disjunction every other query in the disjunction needs to\n" +
					"contain an action instance reference (does not necessarily have to be the same) as well.\n" +
					"If a conjunction contains a negated query that contains an action instance reference the conjunction also needs to contain\n" +
					"a positive query that contains an action instance reference.");
		}
		
		
		
		
		CommonTree supplement = new CommonTree(new CommonToken(DuraParser.SUPPLEMENT, "SUPPLEMENT"));
		conjunction.addChild(supplement);
		
		
		setNextNameId(nextId);
		
		return eventDefinition;
	}
	
	private IdentityHashMap<CommonTree, Boolean> resolveActionInstanceReferences(CommonTree shortenedInitEventQuery, CommonTree actionIdVariableNode, CommonTree eventQuery, String parentIdent, CommonTree parentTerm,
			HashMap<String, Integer> childActionIdentToIndexMap, ArrayList<CommonTree> childActionTerms) {
		IdentityHashMap<CommonTree, Boolean> ammendedGroupBysMap = new IdentityHashMap<CommonTree, Boolean>();
		
		
		IdentityHashMap<CommonTree, Boolean> markedEventQueriesMap = new IdentityHashMap<CommonTree, Boolean>();
		
		LinkedList<CommonTree> nodeQueue = new LinkedList<CommonTree>();
		nodeQueue.add(eventQuery);
		
		while(!nodeQueue.isEmpty()) {
			CommonTree node = (CommonTree) nodeQueue.remove();			
			
			if (node.getType() == DuraParser.IDENTIFIER && node.getText().equals("action")) {
				markEventQueries(node, eventQuery, markedEventQueriesMap);
				
				String ident;
				if (node.getChildCount() > 0) {
					ident = node.getChild(0).getText();
				} else {
					ident = parentIdent;
				}
				
				CommonTree query = (CommonTree) node.getParent();
				CommonTree term = (CommonTree) query.getChild(1);
				
				if (!term.getText().equals("succeeded") && !term.getText().equals("initiated") && !term.getText().equals("failed")) {
					throw new RuntimeException("Invalid referenced action event query: " + query.toStringTree());
				}
				
				if (ident.equals(parentIdent)) {
					if (term.getText().equals("succeeded") || term.getText().equals("failed")) {
						throw new RuntimeException("Invalid referenced action event query: " + query.toStringTree() + ". For this action (parent) you can only query for initiation.");
					}
					
					term.getToken().setText(parentTerm.getText() + "$" + term.getText());
					
					int payloadIndex;
					for (payloadIndex = 0; payloadIndex < term.getChildCount(); payloadIndex++) {
						if (term.getChild(payloadIndex).getText().equals("payload")) {
							break;
						}
					}
					
					if (payloadIndex < term.getChildCount()) {
						// There is a payload term.
						
						CommonTree payloadTerm = (CommonTree) term.getChild(payloadIndex);
						
						int idIndex;
						for (idIndex = 0; payloadIndex < payloadTerm.getChildCount(); idIndex++) {
							if (payloadTerm.getChild(idIndex).getText().equals("id")) {
								break;
							}
						}
						
						if (idIndex < payloadTerm.getChildCount()) {
							CommonTree idTerm = (CommonTree) payloadTerm.getChild(idIndex);
							
							if (idTerm.getChildCount() > 0) {
								// There is an id term.
								
								CommonTree c = (CommonTree) idTerm.getChild(0);
								if (c.getType() == DuraParser.VARIABLE) {
									/*
									// Make this into a conjunction... important for later 
									
									CommonTree tempEventQuery = (CommonTree) query.getParent();
											
									CommonTree tempEventQueryCopy = copyTree(tempEventQuery);
									if (tempEventQueryCopy.getChild(0).getChildCount() > 0) {
										tempEventQueryCopy.getChild(0).deleteChild(0);
									}
									tempEventQueryCopy.getChild(1).deleteChild(0);
									
									
									query.deleteChild(0);
									
									CommonTree conjunction = new CommonTree(new CommonToken(DuraParser.CONJUNCTION, "CONJUNCTION"));
									query.addChild(conjunction);
									
									// Augmentation!
									conjunction.addChild(copyTree(shortenedInitEventQuery));
									
									
									conjunction.addChild(tempEventQueryCopy);								
									
									CommonTree supp = new CommonTree(new CommonToken(DuraParser.SUPPLEMENT, "SUPPLEMENT"));
									conjunction.addChild(supp);
									
									CommonTree where = new CommonTree(new CommonToken(DuraParser.CONDITION, "CONDITION"));
									CommonTree whereConjunction = new CommonTree(new CommonToken(DuraParser.CONJUNCTION, "CONJUNCTION"));
									where.addChild(whereConjunction);
									
									CommonTree equ = new CommonTree(new CommonToken(DuraParser.ARITHMETIC_RELATION, "="));
									whereConjunction.addChild(equ);
									
									equ.addChild(copyTree(c));
									equ.addChild(copyTree(actionIdVariableNode));
									
									supp.addChild(where);
									*/
									
									CommonTree par = (CommonTree) query.getParent();
									CommonTree supp = (CommonTree) par.getChild(par.getChildCount() - 1);
									
									CommonTree let = new CommonTree(new CommonToken(DuraParser.LET, "LET"));
									if (supp.getChildCount() == 0) {
										supp.addChild(let);
									} else {
										supp.getChildren().add(0, let);
										supp.freshenParentAndChildIndexes();
									}
									
									idTerm.deleteChild(0);
									idTerm.addChild(copyTree(actionIdVariableNode));
									
									CommonTree unification = createUnificationNode(new CommonToken(DuraParser.BASIC_TYPE, "identifier"), copyTree(c), copyTree(actionIdVariableNode));
									let.addChild(unification);
								} else {
									throw new RuntimeException("Illegal leaf: " + idTerm.toStringTree());
								}
							} else {
								// There is no id term.
								
								idTerm.addChild(copyTree(actionIdVariableNode));
							}
						}
					} else {
						// There is no payload term.
						
						CommonTree payloadTerm = (CommonTree) new CommonTree(new CommonToken(DuraParser.TERM, "payload"));
						term.addChild(payloadTerm);
						
						CommonTree idTerm = (CommonTree) new CommonTree(new CommonToken(DuraParser.TERM, "id"));
						payloadTerm.addChild(idTerm);
						
						idTerm.addChild(copyTree(actionIdVariableNode));
						
						normalizeTerm(term);
					}
				} else {
					Integer childIndex = childActionIdentToIndexMap.get(ident);
					if (childIndex == null) {
						throw new RuntimeException("No action identifier " + ident + " found: " + query.toStringTree());
					}
					
					CommonTree childTerm = childActionTerms.get(childIndex);
					
					int payloadIndex;
					for (payloadIndex = 0; payloadIndex < term.getChildCount(); payloadIndex++) {
						if (term.getChild(payloadIndex).getText().equals("payload")) {
							break;
						}
					}
					
					term.getToken().setText(childTerm.getText() + "$" + term.getText());
					
					CommonTree payloadTerm;
					if (payloadIndex < term.getChildCount()) {
						// There is a payload term.
						
						payloadTerm = (CommonTree) term.getChild(payloadIndex);
					} else {
						// There is no payload term.
						
						payloadTerm = (CommonTree) new CommonTree(new CommonToken(DuraParser.TERM, "payload"));
						term.addChild(payloadTerm);
						
						normalizeTerm(term);
					}
					
					CommonTree composition = copyTree(actionCompositionAugmentationTerm);
					payloadTerm.addChild(composition);
					
					
					// child-index
					composition.getChild(0).setChild(0, new CommonTree(new CommonToken(DuraParser.NUMBER, Integer.toString(childIndex))));
					
					// parent-action-id
					composition.getChild(1).setChild(0, copyTree(actionIdVariableNode));
					
					
					normalizeTerm(payloadTerm);
				}
				
				
				
				// Fixing references in group by blocks.
				if (query.getParent().getChild(0).getChildCount() > 0 ) {
					CommonTree eventQueryIdentifier = (CommonTree) query.getParent().getChild(0);
					
					CommonTree composition = (CommonTree) query.getParent().getParent();
					
					if (composition.getType() == DuraParser.CONJUNCTION) {
						CommonTree supplement = (CommonTree) composition.getChild(composition.getChildCount() - 1);
						
						for (int i = 0; i < supplement.getChildCount(); i++) {
							CommonTree aggregation = (CommonTree) supplement.getChild(i);
							if (aggregation.getType() == DuraParser.AGGREGATION) {
								CommonTree groupBy = (CommonTree) aggregation.getChild(0);
								
								if (ammendedGroupBysMap.get(groupBy) == null) {
									for (int j = 0; j < groupBy.getChildCount(); j++) {
										if (treesEqual((CommonTree) groupBy.getChild(j), eventQueryIdentifier)) {
											groupBy.addChild(copyTree(actionIdVariableNode));
											
											ammendedGroupBysMap.put(groupBy, true);
											break;
										}
									}
								}
							}
						}
					}
				}
				
				
				query.deleteChild(0);
			}
			
			if (node.getChildCount() > 0) {
				for (int i = node.getChildCount() - 1; i >= 0; i--) {
					CommonTree child = (CommonTree) node.getChild(i);
					
					if (child.getType() != DuraParser.SUPPLEMENT) {
						nodeQueue.add(child);
					}
				}
			}
		}
		
		return markedEventQueriesMap;
	}
	
	private void markEventQueries(CommonTree node, CommonTree stopNode, IdentityHashMap<CommonTree, Boolean> markedEventQueriesMap) {
		while (node != stopNode.getParent()) {
			if (markedEventQueriesMap.get(node) != null) {
				break;
			}
			
			if (node.getType() == DuraParser.EVENT_QUERY) {
				markedEventQueriesMap.put(node, true);
			}
			
			node = (CommonTree) node.getParent();
		}
	}

	/*
	 *  The issue with action instance references...
	 *  
	 *  Consider the following situation:
	 *  
	 *  FOR
	 *    action x: ...
	 *  DO
	 *    concurrent {
	 *      action y: ...
	 *      action z: ...
	 *    } succeeds on {
	 *      and {
	 *        event t: t{},
	 *        event u: and {
	 *          event v: k{},
	 *    not { event w: action y$succeeded{} } where { ... }
	 *        } where { event u after event t }
	 *      }
	 *    }
	 *  END
	 *  
	 *  How is the "succeeds on" block translated? Well, at the top we insert a reference to the instance of action x
	 *  from which we know the action ID thereof and we have to match the parent IDs (or action IDs) of ALL other instance
	 *  references to that ID...  this is achieved by a variable.
	 *  
	 *  But the negation does not create / expose the necessary binding to be used above... so this matching can never be achieved.
	 *  
	 *  Initially I thought that one could just add an action instance reference (action x$initiated) to those blocks
	 *  so as to make them expose the proper binding.
	 *  
	 *  However THIS DOES NOT WORK PROPERLY!
	 *  
	 *  Why?  Because it would change the timing (begin to end interval) of the composite event which is not foreseeable for a
	 *  programmer.  That's why the method is commented out!
	 * 
	 *  This is very similar to the issue we have with resolving complex subqueries of that kind, were "inserting the environment"
	 *  screws up this timing information.
	 *  
	 *  
	 *  If DuraC supported complex subqueries this limitation could be lifted (this means that we don't need to check and throw an error ...).
	 */
	/*
	private void addShortenedInitEventQueries(String variableName, CommonTree shortenedInitEventQuery, CommonTree eventQuery, IdentityHashMap<CommonTree, Boolean> markedEventQueriesMap) {
		assert(eventQuery.getType() == DuraParser.EVENT_QUERY);
		assert(eventQuery.getChild(1).getChild(0).getType() == DuraParser.CONJUNCTION || eventQuery.getChild(1).getChild(0).getType() == DuraParser.DISJUNCTION);
		
		LinkedList<CommonTree> nodeQueue = new LinkedList<CommonTree>();
		nodeQueue.add(eventQuery);
		
		// This is important!
		exposedVariableMemoizationMap.clear();
		
		while(!nodeQueue.isEmpty()) {
			CommonTree node = (CommonTree) nodeQueue.remove();			
			
			if (node.getType() == DuraParser.EVENT_QUERY) {
				CommonTree termConjunctionOrDisjunction = (CommonTree) node.getChild(1).getChild(0);
				
				// This works because we know that the initial eventQuery does always contain a conjunction or disjunction!
				
				
				if (!exposesVariable(node, variableName)) {
					if (markedEventQueriesMap.get(node) == null) {
						// Node is not marked (that it contains the variable in its descendants) ...
						// ... and it doesn't (yet) expose it.
						
						// Important: this case can only occur if the else case has added such nodes to the queue,
						//            i.e. this means that this node SHOULD expose it. 
						
						
						
						if (termConjunctionOrDisjunction.getType() == DuraParser.TERM ||
							termConjunctionOrDisjunction.getType() == DuraParser.DISJUNCTION) {
							CommonTree query = (CommonTree) node.getChild(1);
							
							CommonTree nodeCopy = copyTree(node);
							if (nodeCopy.getChild(0).getChildCount() > 0) {
								nodeCopy.getChild(0).deleteChild(0);
							}
								
							query.deleteChild(0);
								
							CommonTree conjunction = new CommonTree(new CommonToken(DuraParser.CONJUNCTION, "CONJUNCTION"));
							query.addChild(conjunction);
								
							conjunction.addChild(nodeCopy);
								
							CommonTree supp = new CommonTree(new CommonToken(DuraParser.SUPPLEMENT, "SUPPLEMENT"));
							conjunction.addChild(supp);
								
							
							conjunction.getChildren().add(0, copyTree(shortenedInitEventQuery));
							
							conjunction.freshenParentAndChildIndexes();
						} else if (termConjunctionOrDisjunction.getType() == DuraParser.CONJUNCTION) {
							termConjunctionOrDisjunction.getChildren().add(0, copyTree(shortenedInitEventQuery));
							
							termConjunctionOrDisjunction.freshenParentAndChildIndexes();
						}
					} else {
						// Node is marked (that it contains the variable in its descendants) ...
						// ... but it doesn't (yet) expose it.
						
						// This means... termConjunctionOrDisjunction must be a conjunction or disjunction.
						// Why? Because if it was a term, this node would be marked to contain the variable
						// but does not expose it, which is (trivially) not possible.
						
						
						for (int i = 0; i < termConjunctionOrDisjunction.getChildCount(); i++) {
							CommonTree c = (CommonTree) termConjunctionOrDisjunction.getChild(i);
							
							if (c.getType() == DuraParser.NEGATION) {
								nodeQueue.add((CommonTree) c.getChild(0));
							} else if (c.getType() != DuraParser.SUPPLEMENT && !exposesVariable(c, variableName)) {
								nodeQueue.add(c);
							}
						}
					}
				}
			}
		}
	}
	*/
	
	private IdentityHashMap<CommonTree, Boolean> exposedVariableMemoizationMap = new IdentityHashMap<CommonTree, Boolean>();
	private boolean exposesVariable(CommonTree eventQuery, String variableName) {
		Boolean temp = exposedVariableMemoizationMap.get(eventQuery);
		if (temp != null) {
			return temp;
		}
		
		CommonTree termConjunctionOrDisjunction = (CommonTree) eventQuery.getChild(1).getChild(0);
		if (termConjunctionOrDisjunction.getType() == DuraParser.TERM) {
			LinkedList<CommonTree> nodeQueue = new LinkedList<CommonTree>();
			nodeQueue.add(termConjunctionOrDisjunction);
			
			while(!nodeQueue.isEmpty()) {
				CommonTree node = (CommonTree) nodeQueue.remove();
				
				if (node.getType() == DuraParser.VARIABLE && node.getChild(0).getText().equals(variableName)) {
					exposedVariableMemoizationMap.put(node, true);
					return true;
				}
				
				
				for (int i = 0; i < node.getChildCount(); i++) {
					CommonTree child = (CommonTree) node.getChild(i);
						
					if (child.getType() != DuraParser.SUPPLEMENT) {
						nodeQueue.add(child);
					}
				}
			}
		} else if (termConjunctionOrDisjunction.getType() == DuraParser.CONJUNCTION) {
			for (int i = 0; i < termConjunctionOrDisjunction.getChildCount(); i++) {
				CommonTree c = (CommonTree) termConjunctionOrDisjunction.getChild(i);
				if (c.getType() != DuraParser.NEGATION && c.getType() != DuraParser.SUPPLEMENT && exposesVariable(c, variableName)) {
					exposedVariableMemoizationMap.put(eventQuery, true);
					return true;
				}
			}
		} else if (termConjunctionOrDisjunction.getType() == DuraParser.DISJUNCTION) {
			boolean exposedInAll = true;
			
			for (int i = 0; i < termConjunctionOrDisjunction.getChildCount(); i++) {
				CommonTree c = (CommonTree) termConjunctionOrDisjunction.getChild(i);
				
				if (c.getType() != DuraParser.NEGATION && c.getType() != DuraParser.SUPPLEMENT && !exposesVariable(c, variableName)) {
					exposedInAll = false;
				}
			}
			
			if (exposedInAll) {
				exposedVariableMemoizationMap.put(eventQuery, true);
				return true;
			}
		}
		
		
		exposedVariableMemoizationMap.put(eventQuery, false);
		return false;
	}

	private List<CommonTree> collectTypes(CommonTree term) {
		ArrayList<CommonTree> nodes = new ArrayList<CommonTree>();
		
		Stack<CommonTree> nodeStack = new Stack<CommonTree>();
		nodeStack.add(term);
		
		while(!nodeStack.isEmpty()) {
			CommonTree node = (CommonTree) nodeStack.pop();			
			
			if (node.getType() == DuraParser.BASIC_TYPE || node.getType() == DuraParser.COMPOSITE_TYPE) {
				nodes.add(node);
			}
			
			if (node.getChildCount() > 0) {
				for (int i = node.getChildCount() - 1; i >= 0; i--) {
					CommonTree child = (CommonTree) node.getChild(i);
					
					nodeStack.push(child);
				}
			}
		}
		return nodes;
	}

	private void translateComposition(CommonTree actionSpecification, String parentIdent, CommonTree parentTerm, HashSet<String> actionIdentSet, CommonTree execConstraints,
			HashMap<String, Integer> childActionIdentToIndexMap, ArrayList<CommonTree> childActionTerms) {
		int nextId = getNextNameId();
		resetNextNameId();
		
		HashMap<String, List<ActionExecutionDependency>> identToDependenciesMap = calculateDependencies(parentIdent, actionIdentSet, execConstraints, childActionIdentToIndexMap, childActionTerms);
		
		CommonTree defs = (CommonTree) actionSpecification.getParent().getParent();
		int addIndex = actionSpecification.getParent().getChildIndex() + 1;
		
		
		ArrayList<Integer> bundledExecutions = new ArrayList<Integer>();
		for (String ident : childActionIdentToIndexMap.keySet()) {
			if (identToDependenciesMap.get(ident).isEmpty()) {
				int index = childActionIdentToIndexMap.get(ident);
				bundledExecutions.add(index);
			}
		}
		
		Collections.sort(bundledExecutions);
		
		if (!bundledExecutions.isEmpty()) {
			CommonTree parentActionIdVariableNode = createVariableNode();
			resetNextNameId();
			
			int index = bundledExecutions.get(0);
			CommonTree reactiveRule = createReactiveRule(parentActionIdVariableNode, parentIdent, parentTerm, childActionTerms.get(index), index);
			
			
			CommonTree annotation = null;
			if (EMIT_ANNOTATIONS) {
				annotation = new CommonTree(new CommonToken(DuraParser.ANNOTATION, "ANNOTATION"));
				reactiveRule.addChild(annotation);
				annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Result of transformation " + this.getClass().getSimpleName() + ".\"")));
				annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Translation of bundled action execution " + childActionTerms.get(index).getText() + " within action specification for action " + parentTerm.getText() + ".\"")));
			}
			
			defs.getChildren().add(addIndex, reactiveRule);
			addIndex++;
			
			
			CommonTree concurrent = (CommonTree) reactiveRule.getChild(1).getChild(0);
			
			for (int i = 1; i < bundledExecutions.size(); i++) {
				index = bundledExecutions.get(i);
				CommonTree term = childActionTerms.get(index);
				
				CommonTree actionEx = new CommonTree(new CommonToken(DuraParser.ACTION_EXECUTION, "ACTION_EXECUTION"));
				concurrent.addChild(actionEx);
				
				actionEx.addChild(term);
				
				
				CommonTree composition = copyTree(actionCompositionAugmentationTerm);
				term.addChild(composition);
				
				normalizeTerm(term);
				
				// child-index
				composition.getChild(0).setChild(0, new CommonTree(new CommonToken(DuraParser.NUMBER, Integer.toString(index))));
				
				// parent-action-id
				composition.getChild(1).setChild(0, copyTree(parentActionIdVariableNode));
				
				if (EMIT_ANNOTATIONS) {
					annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Translation of bundled action execution " + term.getText() + " within action specification for action " + parentTerm.getText() + ".\"")));
				}
			}
		}
		
		
		
		ArrayList<Integer> ruleIndexes = new ArrayList<Integer>();
		HashMap<Integer, CommonTree> indexToRuleMap = new HashMap<Integer, CommonTree>();
		
		for (String ident : childActionIdentToIndexMap.keySet()) {
			if (!identToDependenciesMap.get(ident).isEmpty()) {
				CommonTree parentActionIdVariableNode = createVariableNode();
				resetNextNameId();
				
				int index = childActionIdentToIndexMap.get(ident);
				CommonTree reactiveRule = createReactiveRule(parentActionIdVariableNode, parentIdent, parentTerm, childActionTerms.get(index), index);
				
				
				ruleIndexes.add(index);
				indexToRuleMap.put(index, reactiveRule);
				
				
				appendDependencyInformation(parentIdent, reactiveRule, ident, identToDependenciesMap.get(ident), childActionIdentToIndexMap, childActionTerms);
				
				if (EMIT_ANNOTATIONS) {
					CommonTree annotation = new CommonTree(new CommonToken(DuraParser.ANNOTATION, "ANNOTATION"));
					reactiveRule.addChild(annotation);
					annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Result of transformation " + this.getClass().getSimpleName() + ".\"")));
					annotation.addChild(new CommonTree(new CommonToken(DuraParser.STRING, "\"Translation of action execution " + childActionTerms.get(index).getText() + " (action " + ident + ") within action specification for action " + parentTerm.getText() + ".\"")));
				}
			}
		}
		
		Collections.sort(ruleIndexes);
		
		for (int index : ruleIndexes) {
			defs.getChildren().add(addIndex, indexToRuleMap.get(index));
			addIndex++;
		}
		
		
		defs.freshenParentAndChildIndexes();
		
		setNextNameId(nextId);
	}

	private void appendDependencyInformation(String parentIdent, CommonTree reactiveRule, String actionToExecuteIdent, List<ActionExecutionDependency> dependencies,
			HashMap<String, Integer> childActionIdentToIndexMap, ArrayList<CommonTree> childActionTerms) {
		CommonTree outerConjunction = (CommonTree) reactiveRule.getChild(0).getChild(0); 
		
		CommonTree parentBeginQuery = (CommonTree) outerConjunction.getChild(0);
		
		CommonTree parentBeginQueryTerm = (CommonTree) parentBeginQuery.getChild(1).getChild(0);
		CommonTree parentBeginQueryPayload = (CommonTree) parentBeginQueryTerm.getChild(0);
		
		CommonTree parentPayloadId = null;
		for (int i = 0; i < parentBeginQueryPayload.getChildCount(); i++) {
			if (parentBeginQueryPayload.getChild(i).getText().equals("id")) {
				parentPayloadId = (CommonTree) parentBeginQueryPayload.getChild(i);
			}
		}
		
		CommonTree parentActionIdVariableNode = (CommonTree) parentPayloadId.getChild(0);
		
		
		int addIndex = parentBeginQuery.getChildIndex() + 1;
		
		CommonTree negation = new CommonTree(new CommonToken(DuraParser.NEGATION, "NEGATION"));
		outerConjunction.getChildren().add(addIndex, negation);
		
		CommonTree negationQuery = new CommonTree(new CommonToken(DuraParser.EVENT_QUERY, "EVENT_QUERY"));
		negation.addChild(negationQuery);
		CommonTree identifier = new CommonTree(new CommonToken(DuraParser.IDENTIFIER, "event"));
		negationQuery.addChild(identifier);
		CommonTree idToken = new CommonTree(new CommonToken(DuraParser.ID, actionToExecuteIdent + "_begin"));
		identifier.addChild(idToken);
		
		CommonTree query = new CommonTree(new CommonToken(DuraParser.QUERY, "QUERY"));
		negationQuery.addChild(query);
		
		String actionToExecuteName = childActionTerms.get(childActionIdentToIndexMap.get(actionToExecuteIdent)).getText();
		
		CommonTree term = new CommonTree(new CommonToken(DuraParser.TERM, actionToExecuteName + "$initiated" ));
		query.addChild(term);
		
		CommonTree negationQueryPayload = new CommonTree(new CommonToken(DuraParser.TERM, "payload"));
		term.addChild(negationQueryPayload);
		
		CommonTree composition = copyTree(actionCompositionAugmentationTerm);
		negationQueryPayload.addChild(composition);
		
		
		// child-index
		composition.getChild(0).setChild(0, new CommonTree(new CommonToken(DuraParser.NUMBER, Integer.toString(childActionIdentToIndexMap.get(actionToExecuteIdent)))));
		
		// parent-action-id
		composition.getChild(1).setChild(0, copyTree(parentActionIdVariableNode));
		
		
		CommonTree querySupplement = new CommonTree(new CommonToken(DuraParser.SUPPLEMENT, "SUPPLEMENT"));
		negationQuery.addChild(querySupplement);

		
		CommonTree negationSupplement = new CommonTree(new CommonToken(DuraParser.SUPPLEMENT, "SUPPLEMENT"));
		negation.addChild(negationSupplement);
		
		CommonTree negationSupplementCondition = new CommonTree(new CommonToken(DuraParser.CONDITION, "CONDITION"));
		negationSupplement.addChild(negationSupplementCondition);
		
		CommonTree negationSupplementConjunction = new CommonTree(new CommonToken(DuraParser.CONJUNCTION, "CONJUNCTION"));
		negationSupplementCondition.addChild(negationSupplementConjunction);
		
		HashSet<String> addedBeginDependencyQueries = new HashSet<String>();
		HashSet<String> addedEndDependencyQueries = new HashSet<String>();
		
		for (ActionExecutionDependency dependency : dependencies) {
			if (!dependency.getDependsOn().equals(parentIdent)) {
				if (dependency.getSuccess() && !addedEndDependencyQueries.contains(dependency.getDependsOn())) {
					CommonTree depQuery = createDependencyQuery(dependency.getDependsOn(), dependency.getSuccess(), parentActionIdVariableNode,
							childActionIdentToIndexMap, childActionTerms);
					
					addedEndDependencyQueries.add(dependency.getDependsOn());
					
					outerConjunction.getChildren().add(addIndex, depQuery);
					addIndex++;
				} else if (!dependency.getSuccess() && !addedBeginDependencyQueries.contains(dependency.getDependsOn())) {
					CommonTree depQuery = createDependencyQuery(dependency.getDependsOn(), dependency.getSuccess(), parentActionIdVariableNode,
							childActionIdentToIndexMap, childActionTerms);
					
					addedBeginDependencyQueries.add(dependency.getDependsOn());
					
					outerConjunction.getChildren().add(addIndex, depQuery);
					addIndex++;
				}
			}
			
			CommonTree negationConstraint = createConstraint(actionToExecuteIdent + "_begin", dependency);
			
			negationSupplementConjunction.addChild(negationConstraint);
		}
		
		outerConjunction.freshenParentAndChildIndexes();
	}
	
	private CommonTree createConstraint(String negatedEventQueryIdentifier, ActionExecutionDependency dependency) {
		String rel = dependency.getStrict() ? "<=" : "<";
		
		CommonTree constraint = new CommonTree(new CommonToken(DuraParser.ARITHMETIC_RELATION, rel));
		
		CommonTree left = new CommonTree(new CommonToken(DuraParser.ARITHMETIC_OP, "-"));
		constraint.addChild(left);		

		CommonTree leftLeft = new CommonTree(new CommonToken(DuraParser.INTERVAL_OP, "end"));
		left.addChild(leftLeft);
		
		CommonTree leftLeftIdentifier = new CommonTree(new CommonToken(DuraParser.IDENTIFIER, "event"));
		leftLeft.addChild(leftLeftIdentifier);
		
		leftLeftIdentifier.addChild(new CommonTree(new CommonToken(DuraParser.ID, negatedEventQueryIdentifier)));

		
		CommonTree leftRight = new CommonTree(new CommonToken(DuraParser.INTERVAL_OP, "end"));
		left.addChild(leftRight);
		
		CommonTree leftRightIdentifier = new CommonTree(new CommonToken(DuraParser.IDENTIFIER, "event"));
		leftRight.addChild(leftRightIdentifier);
		
		String suffix = dependency.getSuccess() ? "_end" : "_begin";
		leftRightIdentifier.addChild(new CommonTree(new CommonToken(DuraParser.ID, dependency.getDependsOn() + suffix)));
		
		CommonTree right = dependency.getWaitDuration();
		constraint.addChild(right);
			
		return constraint;
	}

	private CommonTree createDependencyQuery(String actionIdent, boolean success, CommonTree parentActionIdVariableNode,
			HashMap<String, Integer> childActionIdentToIndexMap, ArrayList<CommonTree> childActionTerms) {
		CommonTree dependencyQuery = new CommonTree(new CommonToken(DuraParser.EVENT_QUERY, "EVENT_QUERY"));
		
		CommonTree identifier = new CommonTree(new CommonToken(DuraParser.IDENTIFIER, "event"));
		dependencyQuery.addChild(identifier);
		
		String suffix = success ? "_end" : "_begin";
		CommonTree idToken = new CommonTree(new CommonToken(DuraParser.ID, actionIdent + suffix));
		identifier.addChild(idToken);
		
		CommonTree query = new CommonTree(new CommonToken(DuraParser.QUERY, "QUERY"));
		dependencyQuery.addChild(query);
		
		String actionToExecuteName = childActionTerms.get(childActionIdentToIndexMap.get(actionIdent)).getText();
		
		suffix = success ? "$succeeded" : "$initiated";
		CommonTree term = new CommonTree(new CommonToken(DuraParser.TERM, actionToExecuteName + suffix ));
		query.addChild(term);
		
		CommonTree negationQueryPayload = new CommonTree(new CommonToken(DuraParser.TERM, "payload"));
		term.addChild(negationQueryPayload);
		
		CommonTree composition = copyTree(actionCompositionAugmentationTerm);
		negationQueryPayload.addChild(composition);
		
		
		// child-index
		composition.getChild(0).setChild(0, new CommonTree(new CommonToken(DuraParser.NUMBER, Integer.toString(childActionIdentToIndexMap.get(actionIdent)))));
		
		// parent-action-id
		composition.getChild(1).setChild(0, copyTree(parentActionIdVariableNode));
		
		
		CommonTree querySupplement = new CommonTree(new CommonToken(DuraParser.SUPPLEMENT, "SUPPLEMENT"));
		dependencyQuery.addChild(querySupplement);
		
		return dependencyQuery;
	}

	private CommonTree createReactiveRule(CommonTree parentActionIdVariableNode, String parentIdent, CommonTree parentTerm, CommonTree childTerm, int childIndex) {
		CommonTree rule = new CommonTree(new CommonToken(DuraParser.DO, "DO"));
		
		CommonTree body = new CommonTree(new CommonToken(DuraParser.BODY, "BODY"));
		rule.addChild(body);
		CommonTree conjunction = new CommonTree(new CommonToken(DuraParser.CONJUNCTION, "CONJUNCTION"));
		body.addChild(conjunction);
		
		CommonTree eventQuery = new CommonTree(new CommonToken(DuraParser.EVENT_QUERY, "EVENT_QUERY"));
		conjunction.addChild(eventQuery);
		CommonTree identifier = new CommonTree(new CommonToken(DuraParser.IDENTIFIER, "event"));
		eventQuery.addChild(identifier);
		CommonTree idToken = new CommonTree(new CommonToken(DuraParser.ID, parentIdent + "_begin"));
		identifier.addChild(idToken);
		
		CommonTree query = new CommonTree(new CommonToken(DuraParser.QUERY, "QUERY"));
		eventQuery.addChild(query);
		
		
		CommonTree parentBeginQueryTerm = new CommonTree(new CommonToken(DuraParser.TERM, parentTerm.getText() + "$initiated" ));
		query.addChild(parentBeginQueryTerm);
		
		CommonTree parentBeginQueryPayload = copyTree(parentTerm);
		parentBeginQueryTerm.addChild(parentBeginQueryPayload);
		parentBeginQueryPayload.getToken().setText("payload");
		
		// Due to variable name reuse, it's important to remove the _composition sub term,
		// which is not required for this part of the reactive rule query.
		for (int i = 0; i < parentBeginQueryPayload.getChildCount(); i++) {
			if (parentBeginQueryPayload.getChild(i).getText().equals("_composition")) {
				parentBeginQueryPayload.deleteChild(i);
				break;
			}
		}
		
		
		CommonTree parentPayloadId = new CommonTree(new CommonToken(DuraParser.TERM, "id"));
		parentBeginQueryPayload.addChild(parentPayloadId);
		
		normalizeTerm(parentBeginQueryPayload);
		
		
		parentPayloadId.addChild(copyTree(parentActionIdVariableNode));

		
		CommonTree supplement = new CommonTree(new CommonToken(DuraParser.SUPPLEMENT, "SUPPLEMENT"));
		eventQuery.addChild(supplement);
		
		supplement = new CommonTree(new CommonToken(DuraParser.SUPPLEMENT, "SUPPLEMENT"));
		conjunction.addChild(supplement);
		
		
		
		CommonTree head = new CommonTree(new CommonToken(DuraParser.HEAD, "HEAD"));
		rule.addChild(head);
		CommonTree concurrent = new CommonTree(new CommonToken(DuraParser.CONCURRENT, "CONCURRENT"));
		head.addChild(concurrent);
		CommonTree actionEx = new CommonTree(new CommonToken(DuraParser.ACTION_EXECUTION, "ACTION_EXECUTION"));
		concurrent.addChild(actionEx);
		
		actionEx.addChild(childTerm);
		
		
		CommonTree composition = copyTree(actionCompositionAugmentationTerm);
		childTerm.addChild(composition);
		
		normalizeTerm(childTerm);
		
		// child-index
		composition.getChild(0).setChild(0, new CommonTree(new CommonToken(DuraParser.NUMBER, Integer.toString(childIndex))));
		
		// parent-action-id
		composition.getChild(1).setChild(0, copyTree(parentActionIdVariableNode));
		
		return rule;
	}

	/*
	private void addCompositionInformation(CommonTree tree, String actionName) {
		CommonTree actionSchemaTerm = actionSchemataMap.get(actionName);
		
		if (actionSchemaTerm == null) {	
			throw new RuntimeException("No action of name " + actionName + " has been defined!");
		}
		
		CommonTree explicitActionSchemaTerm = explicitActionSchemataMap.get(actionName);
		
		if (explicitActionSchemaTerm == null) {
			// Extension of external events that are not explicitly represented in the code?
			// Siehe E-Mail vom 11. Okt. 2011.
			
			Tree defs = tree.getChild(0).getChild(1);
			
			assert(defs.getType() == DuraParser.DEFINITIONS);
			
			CommonTree def = new CommonTree(new CommonToken(DuraParser.ACTION, actionName));
			def.addChild(new CommonTree(new CommonToken(DuraParser.MODIFIERS)));
			
			CommonTree schema = new CommonTree(new CommonToken(DuraParser.SCHEMA));
			
			explicitActionSchemaTerm = copyTree(actionSchemaTerm);
			augmentWithCompositionTerm(explicitActionSchemaTerm);
			
			schema.addChild(explicitActionSchemaTerm);
			schema.addChild(new CommonTree(new CommonToken(DuraParser.SUPPLEMENT)));
			
			def.addChild(schema);
			
			explicitActionSchemataMap.put(actionName, explicitActionSchemaTerm);
			
			// TODO: Has to be tested.
		}
		
		augmentWithCompositionTerm(explicitActionSchemaTerm);
		augmentWithCompositionTerm(actionSchemaTerm);
		
		CommonTree initiatedSchema = eventSchemataMap.get(actionName + "$initiated");
		CommonTree suceededSchema = eventSchemataMap.get(actionName + "$succeeded");
		CommonTree failedSchema = eventSchemataMap.get(actionName + "$failed");
		
		int payloadIndex; 
		for (payloadIndex = 0; payloadIndex < initiatedSchema.getChildCount(); payloadIndex++) {
			if (initiatedSchema.getChild(payloadIndex).getText().equals("payload")) {
				break;
			}
		}
		
		augmentWithCompositionTerm((CommonTree) initiatedSchema.getChild(payloadIndex));
		augmentWithCompositionTerm((CommonTree) suceededSchema.getChild(payloadIndex));
		augmentWithCompositionTerm((CommonTree) failedSchema.getChild(payloadIndex));
	}

	private void augmentWithCompositionTerm(CommonTree term) {
		CommonTree addTerm = copyTree(compositionAugmentationTerm);
		
		if (term.getChildCount() == 0) {
			term.addChild(addTerm);
		} else {
			int i;
			
			for (i = 0; i < term.getChildCount(); i++) {
				CommonTree childTerm = (CommonTree) term.getChild(i);
				
				int comp = childTerm.getText().compareTo(addTerm.getText());
				if (comp == 0) {
					throw new RuntimeException("This should not happen. Cannot join (" + addTerm.toStringTree() + ") with children of (" + term.toStringTree() + ").");
				} else if (comp > 0) {
					break;
				}
			}
			
			term.getChildren().add(i, addTerm);
			term.freshenParentAndChildIndexes();
		}
	}
	*/

	@SuppressWarnings("unchecked")
	private HashMap<String, List<ActionExecutionDependency>> calculateDependencies(String parentIdent, HashSet<String> actionIdentSet, CommonTree execConstraints,
			HashMap<String, Integer> childActionIdentToIndexMap, ArrayList<CommonTree> childActionTerms) {
		HashMap<String, List<ActionExecutionDependency>> actionDependencies = new HashMap<String, List<ActionExecutionDependency>>();
		
		ArrayList<Integer>[] timePointDependencies = new ArrayList[actionIdentSet.size() * 2];
		
		HashMap<String, Integer> identToBeginIndexMap = new HashMap<String, Integer>();
		HashMap<String, Integer> identToEndIndexMap = new HashMap<String, Integer>();
		
		String[] indexToNames = new String[timePointDependencies.length];
		
		timePointDependencies[0] = new ArrayList<Integer>();
		identToBeginIndexMap.put(parentIdent, 0);
		timePointDependencies[1] = new ArrayList<Integer>();
		identToEndIndexMap.put(parentIdent, 1);
		
		int index = 2;
		for (String ident : actionIdentSet) {
			if (!ident.equals(parentIdent)) {
				timePointDependencies[index] = new ArrayList<Integer>();
				identToBeginIndexMap.put(ident, index);
				timePointDependencies[index + 1] = new ArrayList<Integer>();
				identToEndIndexMap.put(ident, index + 1);
				
				indexToNames[index] = "begin(" + ident + ")";
				indexToNames[index + 1] = "end(" + ident + ")";
				
				// Implicit dependencies
				timePointDependencies[index + 1].add(index);
				timePointDependencies[index].add(0);
				
				actionDependencies.put(ident, new ArrayList<ActionExecutionDependency>());
					
				index += 2;
			}
		}
		
		for (int i = 0; i < execConstraints.getChildCount(); i ++) {
			CommonTree constraint = (CommonTree) execConstraints.getChild(i);
			
			if (constraint.getType() == DuraParser.ARITHMETIC_RELATION) {		
				if (constraint.getText().equals("<") || constraint.getText().equals("<=")) {
					CommonTree left = (CommonTree) constraint.getChild(0);
					CommonTree right = (CommonTree) constraint.getChild(1);
							
					// Right side must be negative or zero duration.  
					if (right.getText().equals("-")) {
						if (right.getChildCount() > 1) {
							throw new RuntimeException("Invalid execution constraint (" + constraint.toStringTree() + ").");
						}
						
						right = (CommonTree) right.getChild(0);
						
						if (right.getType() != DuraParser.DURATION) {
							throw new RuntimeException("Invalid execution constraint (" + constraint.toStringTree() + ").");
						}
					} else {
						if (right.getType() != DuraParser.DURATION) {
							throw new RuntimeException("Invalid execution constraint (" + constraint.toStringTree() + ").");
						}
						
						for (int j = 0; j < right.getChildCount(); j ++) {
							if (Integer.valueOf(right.getChild(j).getChild(0).getText()) > 0) {
								throw new RuntimeException("Invalid execution constraint (" + constraint.toStringTree() + ").");
							}
						}
					}
					
					// Left side must be begin() - begin() or end() - begin()   
					if (!left.getText().equals("-")) {
						throw new RuntimeException("Invalid execution constraint (" + constraint.toStringTree() + ").");
					}
					
					if (left.getChildCount() < 2) {
						throw new RuntimeException("Invalid execution constraint (" + constraint.toStringTree() + ").");
					}
					
					CommonTree leftLeft = (CommonTree) left.getChild(0);
					if (leftLeft.getType() != DuraParser.INTERVAL_OP ||
						leftLeft.getChild(0).getType() != DuraParser.IDENTIFIER ||
						!leftLeft.getChild(0).getText().equals("action") ||
						( leftLeft.getText().equals("end") && leftLeft.getChild(0).getChild(0).getText().equals(parentIdent) )
						) {
						throw new RuntimeException("Invalid execution constraint (" + constraint.toStringTree() + ").");
					}
					
					CommonTree leftRight = (CommonTree) left.getChild(1);
					if (leftRight.getType() != DuraParser.INTERVAL_OP ||
						leftRight.getText().equals("end") ||
						leftRight.getChild(0).getType() != DuraParser.IDENTIFIER ||
						!leftRight.getChild(0).getText().equals("action") ||
						leftRight.getChild(0).getChild(0).getText().equals(parentIdent)
						) {
						throw new RuntimeException("Invalid execution constraint (" + constraint.toStringTree() + ").");
					}
					
					
					boolean dependsOnSuccess = false;
					
					String leftLeftActionIdent = leftLeft.getChild(0).getChild(0).getText();
					int leftLeftTimePoint;
					if (leftLeft.getText().equals("end")) {
						leftLeftTimePoint = identToEndIndexMap.get(leftLeftActionIdent);
						
						dependsOnSuccess = true;
					} else {
						leftLeftTimePoint = identToBeginIndexMap.get(leftLeftActionIdent);
					}
					
					String leftRightActionIdent = leftRight.getChild(0).getChild(0).getText();
					int leftRightTimePoint = identToBeginIndexMap.get(leftRightActionIdent);
					
					if (!actionIdentSet.contains(leftLeftActionIdent) || !actionIdentSet.contains(leftRightActionIdent)) {
						throw new RuntimeException("Invalid execution constraint (" + constraint.toStringTree() + ").");
					}
					
					timePointDependencies[leftRightTimePoint].add(leftLeftTimePoint);
					
					boolean dependsOnStrict = constraint.getText().equals("<");
					
					
					
					// Futile wait detection
					int tempIndex = childActionIdentToIndexMap.get(leftLeftActionIdent);
					String tempName = childActionTerms.get(tempIndex).getText();
					if (dependsOnSuccess && !actionsWithSuccessSpecificationSet.contains(tempName)) {
						throw new RuntimeException("Futile wait detected due to constraint (" + constraint.toStringTree() + "). Success of action " + tempName + " is not specified.");
					}
					
					
					actionDependencies.get(leftRightActionIdent).add(new ActionExecutionDependency(leftLeftActionIdent, dependsOnSuccess, dependsOnStrict, right));
				} else {
					throw new RuntimeException("Invalid execution constraint (" + constraint.toStringTree() + ").");
				}
			}
		}
		
		/*
		// Debug
		for (String s : identToBeginIndexMap.keySet()) {
			System.out.println(identToBeginIndexMap.get(s) + " = begin(" + s + ")");
			System.out.println(identToEndIndexMap.get(s) + " = end(" + s + ")");
		}
		System.out.println();
		
		// Debug
		for (int i = 0; i < timePointDependencies.length; i++) {
			System.out.print(i + " depends on ");
			for (int p : timePointDependencies[i]) {
				System.out.print(p + " ");
			}
			System.out.println();
		}
		*/
		
		ArrayList<Integer> detectedCycles = new ArrayList<Integer>();
		for (int i = 0; i < timePointDependencies.length; i++) {
			boolean[] visited = new boolean[timePointDependencies.length];
			
			Queue<Integer> indexQueue = new LinkedList<Integer>();
			
			for (int j : timePointDependencies[i]) {
				indexQueue.add(j);
			}
			
			while(!indexQueue.isEmpty()) {
				int ix = indexQueue.remove();
				
				visited[ix] = true;
				
				//System.out.println("(" + i + ", " + ix + ")");
				
				if (ix == i) {
					detectedCycles.add(i);
					break;
				}
				
				if (timePointDependencies[ix].size() > 0) {
					for (int j : timePointDependencies[ix]) {
						if (!visited[j]) {
							indexQueue.add(j);
						}
					}
				}
			}
		}
		
		if (!detectedCycles.isEmpty()) {
			StringBuffer cyclesDetectedMessage = new StringBuffer("Cycles detected for controllable time points: ");
			for (int i : detectedCycles) {
				if (identToBeginIndexMap.containsValue(i)) {
					cyclesDetectedMessage.append(indexToNames[i]);
					cyclesDetectedMessage.append(" ");
				}
			}
			
			throw new RuntimeException(cyclesDetectedMessage.toString());
		}
		
		/*
		for (String ident : actionDependencies.keySet()) {
			System.out.println("Execution of action " + ident);
			
			for (ActionExecutionDependency dep : actionDependencies.get(ident)) {
				System.out.println(dep.toString());
			}
			
			System.out.println();
		}
		*/
		
		return actionDependencies;
	}
	
	private class ActionExecutionDependency {
		private String dependsOn;
		private boolean success;
		private boolean strict;
		private CommonTree waitDuration;
		
		ActionExecutionDependency(String dependsOn, boolean success, boolean strict, CommonTree waitDuration) {
			this.dependsOn = dependsOn;
			this.success = success;
			this.strict = strict;
			this.waitDuration = waitDuration;
		}
		
		public String getDependsOn() {
			return dependsOn;
		}
		
		public boolean getSuccess() {
			return success;
		}
		
		public boolean getStrict() {
			return strict;
		}
		
		public CommonTree getWaitDuration() {
			return waitDuration;
		}
		
		/*
		public String toString() {
			return "  waits for " + (success ? "successful end" : "initiation") + " of action " + dependsOn + " and " + (strict ? "more than" : "at least") + " " + waitDuration.toStringTree();
		}
		*/
	}
}
