package de.lmu.ifi.pms.cep.dura.compiler.durac;

import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.*;

import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraLexer.*;
import static de.lmu.ifi.pms.cep.tsa.conditions.TupleFunctionID.USER;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Stack;

import org.antlr.runtime.tree.Tree;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.utils.DuraTree;
import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.ConstDefinitions;
import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.DuraSchema;
import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.Preamble;
import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.QueryVariables;
import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.TypeDefinitions;
import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.UnambiguousQueryVariables;
import de.lmu.ifi.pms.cep.dura.compiler.durac.formula_transformations.FormulaNormalization;
import de.lmu.ifi.pms.cep.dura.compiler.durac.queries.AtomicEventQuery;
import de.lmu.ifi.pms.cep.dura.compiler.durac.queries.CompositeEventQuery;
import de.lmu.ifi.pms.cep.dura.compiler.durac.queries.StatefulObjectQuery;
import de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraAggregationFunctionMapping;
import de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraBasicTypesMapping;
import de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping;
import de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraMathFunctionMapping;
import de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraRelationConstraintMapping;
import de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraTimeUnitMapping;
import de.lmu.ifi.pms.cep.dura.compiler.durac.tsa.RelativeTimerOp;
import de.lmu.ifi.pms.cep.dura.compiler.durac.tsa.TimeIntervalOp;
import de.lmu.ifi.pms.cep.dura.compiler.durac.tsa.UntypedIdentifier;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraAmbiguousMappingException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraCompilationException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraUnspecificCompliationException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.UnexpectedTokenException;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.UnsupportedTokenException;
import de.lmu.ifi.pms.cep.tsa.TSA_CompositeExpression;
import de.lmu.ifi.pms.cep.tsa.TSA_Expression;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.conditions.AggregationFunction;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeValue;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonOperatorID;
import de.lmu.ifi.pms.cep.tsa.conditions.BooleanValue;
import de.lmu.ifi.pms.cep.tsa.conditions.CastFunction;
import de.lmu.ifi.pms.cep.tsa.conditions.ClockFunction;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.ConjunctiveConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.DisjunctiveConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.DoubleValue;
import de.lmu.ifi.pms.cep.tsa.conditions.DurationValue;
import de.lmu.ifi.pms.cep.tsa.conditions.FloatValue;
import de.lmu.ifi.pms.cep.tsa.conditions.IntValue;
import de.lmu.ifi.pms.cep.tsa.conditions.LongValue;
import de.lmu.ifi.pms.cep.tsa.conditions.NamedConstant;
import de.lmu.ifi.pms.cep.tsa.conditions.NamedTypeVariable;
import de.lmu.ifi.pms.cep.tsa.conditions.NegatedConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.StringValue;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.TupleFunction;
import de.lmu.ifi.pms.cep.tsa.conditions.TupleFunctionID;
import de.lmu.ifi.pms.cep.tsa.conditions.TypedAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.UntypedAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.UntypedVariable;
import de.lmu.ifi.pms.cep.tsa.conditions.UserRelationFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.UserTupleFunction;
import de.lmu.ifi.pms.cep.tsa.conditions.Variable;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_SchemaException;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_TypeException;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_CrossproductOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_GroupingAndAggregationOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_ImbeddingOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_ProjectionOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_SelectionOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_SetDifferenceOperator;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.schema.NoType;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.Type;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.filter.AbstractFilter;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public class DuraCQueryCompiler {
	
	public static CompositeEventQuery compileConjunctiveQuery(final DuraTree query, final Preamble preamble, final DuraSchema schema) {
		assert(query.getType() == CONJUNCTION);

		HashMap<String, String> identifierMapping = new HashMap<String, String>();
		
		ArrayList<TSA_Expression> positiveQueries = new ArrayList<TSA_Expression>();
		ArrayList<AtomicEventQuery> negativeQueries = new ArrayList<AtomicEventQuery>();
		
		ArrayList<UnambiguousQueryVariables> queryVariables = new ArrayList<UnambiguousQueryVariables>();
		
		final HashSet<String> negativeQueryIdentifier = new HashSet<String>();
		
		ArrayList<ConditionFormula> queryConditions = new ArrayList<ConditionFormula>();
		
		
		int childCount = query.getChildCount();
		DuraTree supplement = query.getChild(childCount-1);
		
		assert(supplement.getType() == SUPPLEMENT);
		
		for (int i=0; i<childCount-1; i++) {
			DuraTree flatQuery = query.getChild(i);
			
			switch (flatQuery.getType()) {
			case NEGATION:
				DuraTree child = flatQuery.getChild(0);
				
				switch (child.getType()) {
				case EVENT_QUERY:
					AtomicEventQuery eventQuery = new AtomicEventQuery(child, preamble, schema);
					
					queryVariables.add(eventQuery.getVariables());
					
					negativeQueries.add(eventQuery);
					negativeQueryIdentifier.add(eventQuery.identifier);
					
					break;
					
				case STATE_QUERY:
					StatefulObjectQuery soQuery = new StatefulObjectQuery(child, preamble, schema);
					
					if (soQuery.queriesStaticObject) {
						queryVariables.add(soQuery.variables);
						
						identifierMapping.putAll(soQuery.identifierMappig);

						negativeQueries.add(soQuery.positiveQuery);

						negativeQueryIdentifier.add(soQuery.positiveQuery.identifier);
						negativeQueryIdentifier.add(soQuery.negativeQuery.identifier);			//although the negative query part is not used, its schema information is required to filter attributes in the supplement that refer to it 
					} else {
						throw new DuraCompilationException("Currently, only queries for events and static stateful objects can be negated.", flatQuery);
					}
					
					break;
				
				default:
					throw new UnexpectedTokenException(flatQuery);
				}
				
				DuraTree negationSupplement = flatQuery.getChild(1);
				
				if (negationSupplement.getChildCount() == 1) {
					assert (negationSupplement.getChild(0).getType() == CONDITION);

					queryConditions.add(compileFormula(negationSupplement.getChild(0).getChild(0)));
				}
				
				break;

			case STATE_QUERY:
				StatefulObjectQuery soQuery = new StatefulObjectQuery(flatQuery, preamble, schema);
				
				queryVariables.add(soQuery.variables);
				
				positiveQueries.add(soQuery.positiveQuery.getExpression());
				identifierMapping.putAll(soQuery.identifierMappig);
				negativeQueryIdentifier.add(soQuery.negativeQuery.identifier);

				if (! soQuery.queriesStaticObject) {
					negativeQueries.add(soQuery.negativeQuery);
					queryConditions.add(soQuery.joinCondition);
				}
				
				break;
				
			case EVENT_QUERY:
				AtomicEventQuery eventQuery = new AtomicEventQuery(flatQuery, preamble, schema);

				positiveQueries.add(eventQuery.getExpression());
				queryVariables.add(eventQuery.getVariables());

				break;
				
			default:
				throw new UnexpectedTokenException(flatQuery);
			}
		}

		
		QueryVariables ambiguousQueryVariables = QueryVariables.unionUnambiguousMappings(queryVariables);
		
		//this includes also join conditions for negative and positive query parts
		queryConditions.add(ambiguousQueryVariables.compileImplicitVariableUnifications());
		
		//FIXME: filter implied and purely negative join conditions: pos = neg1, pos = neg2, neg1 = neg2   -->   pos = neg1, pos = neg2 (cf FilteredAttributes5)
		
		//if the attributes of a variable have positive and negative polarity, keep only the attributes with positive polarity 
		final UnambiguousQueryVariables uniquePolarityQueryVariables = ambiguousQueryVariables.filterAmbiguosMatchingMappings(negativeQueryIdentifier).dropAmbiguousMappings();

		
		TSA_Expression positiveCrossProduct;
		if (positiveQueries.size() > 1) {
			try{
				positiveCrossProduct = new TSA_CompositeExpression(new TSA_CrossproductOperator(query.getPositionDescription(), positiveQueries.size()), new FixedArray<TSA_Expression>(positiveQueries));
			} catch (TSA_Exception e) {
				throw new DuraCompilationException(e, query);
			}
		} else {
			assert positiveQueries.size() == 1;
			
			positiveCrossProduct = positiveQueries.get(0);
		}

		
		TemporalStreamSchema positiveAndNegativeSchema = positiveCrossProduct.schema();
		for (AtomicEventQuery negativeQuery : negativeQueries) {			
			try {
				positiveAndNegativeSchema = positiveAndNegativeSchema.extend(negativeQuery.getExpression().schema().attributes);
			} catch (TSA_SchemaException e) {
				throw new DuraCompilationException(e, query);
			}
		}

		
		FormulaNormalization bodyRefiner = new FormulaNormalization(uniquePolarityQueryVariables, positiveAndNegativeSchema, identifierMapping, schema.constants, preamble);

		
		if (supplement.getChildCount()>0 && supplement.getChild(0).getType()==CONDITION) {
			queryConditions.add(DuraCQueryCompiler.compileFormula(supplement.getChild(0).getChild(0)));
		}

		
		ConditionFormula combinedQueryCondition;
		try {
			combinedQueryCondition = bodyRefiner.apply(new ConjunctiveConditionFormula(new FixedArray<ConditionFormula>(queryConditions)));
		} catch (DuraUnspecificCompliationException e) {
			throw new DuraCompilationException(e, supplement.getChild(0));
		}
		
		
		TSA_Expression positiveQueryPart;
		try {
			AbstractFilter<AtomicConditionFormula> filterNegativeConditions = new AbstractFilter<AtomicConditionFormula>() {
				
				@Override
				public boolean accepts(final AtomicConditionFormula element) {
					FixedArrayNavigableSet<String> names = element.AttributeNames();
					
					for (String attr : names) {
						if (negativeQueryIdentifier.contains(TSA_SyntaxPatterns.splitAttributeName(attr)[0])) {
							return false;
						}
					}
					
					return true;
				}
			};

			ConditionFormula positiveQueryConditions = combinedQueryCondition.focus(filterNegativeConditions);
			
			positiveQueryPart = new TSA_CompositeExpression(new TSA_SelectionOperator(supplement.getPositionDescription(), positiveQueryConditions), positiveCrossProduct);
		} catch (TSA_Exception e) {
			throw new DuraCompilationException(e, supplement);
		}

		
		TSA_Expression positiveAndNegativeQueryPart = positiveQueryPart;

		if (negativeQueries.size() > 0) {
			for (AtomicEventQuery negativeQuery : negativeQueries) {
					
				positiveAndNegativeQueryPart = compileAntiSemiJoin(positiveAndNegativeQueryPart, negativeQuery, ambiguousQueryVariables, combinedQueryCondition);
			}
		}
		
		UnambiguousQueryVariables positiveVariables = ambiguousQueryVariables.dropMappingsMatchingPrefix(negativeQueryIdentifier).dropAmbiguousMappings();

		// if there is a negated query, the first supplement must be a where part and it hence has already been considered
		return compileSupplement(supplement, negativeQueries.size()>0 ? 1 : 0, positiveAndNegativeQueryPart, positiveVariables, identifierMapping, schema.constants, preamble);
	}
	
	public static CompositeEventQuery compileSupplement(final DuraTree supplement, final TSA_Expression expression, final UnambiguousQueryVariables variables, final ConstDefinitions constants, final Preamble preamble) {
		return compileSupplement(supplement, 0, expression, variables, new HashMap<String, String>(), constants, preamble);
	}
	
	public static CompositeEventQuery compileSupplement(final DuraTree supplement, final int offset, TSA_Expression expression, UnambiguousQueryVariables variables, final Map<String, String> identifierMapping, final ConstDefinitions constants, final Preamble preamble) {
		assert(supplement.getType() == SUPPLEMENT);
		
		int childCount = supplement.getChildCount();
		
		FormulaNormalization refiner = new FormulaNormalization(variables, expression.schema(), identifierMapping, constants, preamble);

		for (int i=offset; i<childCount; i++) {
			DuraTree child = supplement.getChild(i);
			
			try {
				switch (child.getType()) {
				case CONDITION:
					ConditionFormula where = refiner.apply(DuraCQueryCompiler.compileFormula(child.getChild(0)));
					TSA_SelectionOperator selection = new TSA_SelectionOperator(child.getPositionDescription(), where);
					
					expression = new TSA_CompositeExpression(selection, expression);
					
					break;
					
				case LET:
					Unification unification = compileUnifications(child, refiner);
					TSA_ImbeddingOperator imbeding = new TSA_ImbeddingOperator(child.getPositionDescription(), unification.definitions);
					
					expression = new TSA_CompositeExpression(imbeding, expression);
					variables = QueryVariables.unionUnambiguousMappings(variables, unification.variables).convertToUnambiguousMapping();
					refiner = new FormulaNormalization(variables, expression.schema(), constants, preamble);
					
					break;
					
				case AGGREGATION:
					CompositeEventQuery aggregation = compileAggregation(child, expression, variables, refiner);
					
					expression = aggregation.getExpression();
					variables = aggregation.getVariables();
					refiner = new FormulaNormalization(variables, expression.schema(), constants, preamble);
					
					break;

				default:
					throw new UnexpectedTokenException(child);

				}
			} catch (TSA_Exception e) {
				throw new DuraCompilationException(e, child);
			} catch (DuraUnspecificCompliationException e) {
				throw new DuraCompilationException(e, child);
			}
		}
		
		return new CompositeEventQuery(expression, variables, identifierMapping);
	}
	
	
	private static Unification compileUnifications(final DuraTree unifications, final FormulaNormalization refiner) {
		assert(unifications == null || unifications.getType() == LET || unifications.getType() == AGGREGATION);
		
		UnambiguousQueryVariables newVariables = new UnambiguousQueryVariables();
		HashMap<TypedAttribute<? extends Type>, Term<? extends Type>> attributeDefinitions = new HashMap<TypedAttribute<? extends Type>, Term<? extends Type>>();

		if (unifications != null) {
			for (int i=0; i<unifications.getChildCount(); i++) {
				DuraTree unification = unifications.getChild(i);
				
				assert(unification.getType() == UNIFICATION);
				
				try {
					Term<? extends Type> value = refiner.apply(DuraCQueryCompiler.compileTerm(unification.getChild(1)));
					UntypedVariable var = (UntypedVariable) DuraCQueryCompiler.compileTerm(unification.getChild(0));
					
					newVariables = QueryVariables.unionUnambiguousMappings(newVariables, QueryVariables.generateImplicitAttribute(var, value.type)).convertToUnambiguousMapping();
	
					TypedAttribute<? extends Type> attribute = (TypedAttribute<? extends Type>) newVariables.getVariableAttribute(var);
					
					attributeDefinitions.put(attribute, value);
				} catch (DuraUnspecificCompliationException e) {
					throw new DuraCompilationException(e, unification);
				}
			}
		}

		return new Unification(newVariables, new FixedArrayNavigableMap<TypedAttribute<? extends Type>, Term<? extends Type>>(attributeDefinitions));
	}
	
	private static CompositeEventQuery compileAggregation(final DuraTree aggregation, final TSA_Expression expression, final UnambiguousQueryVariables variables, final FormulaNormalization refiner) {
		assert(aggregation.getType() == AGGREGATION);
		
		DuraTree groupings = aggregation.getChild(0);
		DuraTree unifications = aggregation.getChild(1);
		
		UnambiguousQueryVariables groupingVariables = new UnambiguousQueryVariables();
		Unification unification = compileUnifications(unifications, refiner);

		int childCount = groupings.getChildCount();
		ArrayList<String> groupingAttributes = new ArrayList<String>();

		for (int i=0; i<childCount; i++) {
			DuraTree child = groupings.getChild(i);
			
			try {
				Term<?> term = DuraCQueryCompiler.compileTerm(child);
				TypedAttribute<? extends Type> attribute = (TypedAttribute<? extends Type>) refiner.apply(term);
				
				if (attribute.type.equals(BasicTypeID.IDENTIFIER.type)) {
					//implicitly group by all attributes of the according stream
					
					String streamPrefix = TSA_SyntaxPatterns.splitAttributeName(attribute.name)[0];
					groupingAttributes.addAll(expression.schema().getAttribute(streamPrefix).type.basicAttributes(streamPrefix).keys().javaCollection());
					
					groupingVariables = QueryVariables.unionUnambiguousMappings(groupingVariables, variables.getMappingsMatchingPrefix(streamPrefix)).convertToUnambiguousMapping();
				} else {
					groupingAttributes.add(attribute.name);
					groupingVariables = QueryVariables.unionUnambiguousMappings(groupingVariables, variables.getMappingsMatchingPrefix(attribute.name)).convertToUnambiguousMapping();
				}
			} catch (DuraUnspecificCompliationException e) {
				throw new DuraCompilationException(e, child);
			} catch (TSA_TypeException e) {
				throw new DuraCompilationException(e, child);
			}
		}
		
		TSA_GroupingAndAggregationOperator grouping = new TSA_GroupingAndAggregationOperator(aggregation.getPositionDescription(), new FixedArray<String>(groupingAttributes), unification.definitions);

		 try {
			 return new CompositeEventQuery(new TSA_CompositeExpression(grouping, expression), QueryVariables.unionUnambiguousMappings(groupingVariables, unification.variables).convertToUnambiguousMapping());
		} catch (TSA_Exception e) {
			throw new DuraCompilationException(e, aggregation);
		} catch (DuraAmbiguousMappingException e) {
			throw new DuraCompilationException(e, aggregation);
		}
	}

	private static TSA_Expression compileAntiSemiJoin(final TSA_Expression positiveStream, final AtomicEventQuery negativeQuery, final QueryVariables ambiguousVariables, final ConditionFormula supplementCondition) {		
		try {
			TSA_Expression crossProduct = new TSA_CompositeExpression(new TSA_CrossproductOperator(negativeQuery.origin.getPositionDescription(), 2), new FixedArray<TSA_Expression>(negativeQuery.getExpression(), positiveStream));
			
			AbstractFilter<AtomicConditionFormula> filterPurelyPositiveConditions = new AbstractFilter<AtomicConditionFormula>() {
				
				@Override
				public boolean accepts(final AtomicConditionFormula element) {
					FixedArrayNavigableSet<String> names = element.AttributeNames();
					
					for (String attr : names) {
						if (negativeQuery.identifier.equals(TSA_SyntaxPatterns.splitAttributeName(attr)[0])) {
							return true;
						}
					}
					
					return false;
				}
			};
			
			ConditionFormula relevantSupplementCondition = supplementCondition.focus(filterPurelyPositiveConditions);
			
			TSA_Expression negationSupplement = new TSA_CompositeExpression(new TSA_SelectionOperator(negativeQuery.origin.getPositionDescription(), relevantSupplementCondition), crossProduct);
			TSA_Expression negationProjection = new TSA_CompositeExpression(new TSA_ProjectionOperator(negativeQuery.origin.getPositionDescription(), positiveStream.schema().basicAttributes.keys()), negationSupplement);
			
			return new TSA_CompositeExpression(new TSA_SetDifferenceOperator(negativeQuery.origin.getPositionDescription()), new FixedArray<TSA_Expression>(positiveStream, negationProjection));
		} catch (TSA_Exception e) {
			String msg = e.getMessage();
			String attribute = msg.substring(msg.indexOf(": ")+2, msg.indexOf("\n"));
			
			throw new DuraCompilationException(String.format("found variables or attributes that are not visible in the negated part of the query: %s",  ambiguousVariables.getVariablesMatchingAttribute(attribute)), negativeQuery.origin);
		}
	}

	public static Term<?> compileTerm(final DuraTree term) {
		return compileTerm(term, false);
	}
	
	public static Term<?> compileTerm(final DuraTree term, final boolean identifierContext) {
		Term<?> left, right;
		
		int childCount = term.getChildCount();
		
		switch (term.getType()) {
		case VARIABLE:
//		case TYPED_VARIABLE:
			return QueryVariables.variableFromToken(term);
			
		case IDENTIFIER:
			Variable<?> var = QueryVariables.variableFromToken(term);
			
			if (identifierContext) {
				return NamedTypeVariable.create(var.name, TypeDefinitions.TIME_INTERVAL);
			} else {
				return var;
			}
			
		case CONSTANT:
			return new NamedConstant(term.getChild(0).getText());

		case INTERVAL_OP:
			assert(childCount == 1);
			
			left = compileTerm(term.getChild(0), true);
			
			return new TimeIntervalOp(term.getText(), left);
			
		case IDENTIFIER_OP:
			assert(childCount == 1);
			
			String name;
			
			left = compileTerm(term.getChild(0));
			
			if (left instanceof Variable<?>) {
				name = ((Variable<?>) left).name;
				name = name.substring(name.lastIndexOf(' ')+1);
			} else if (left instanceof UntypedAttribute) {
				name = ((UntypedAttribute) left).name;
			} else {
				throw new UnexpectedTokenException(term.getChild(0));
			}
			
			return UntypedAttribute.create(TSA_SyntaxPatterns.attributeName(name, DuraDefaultAttributeNameMapping.identifierAttribute));
			
		case AGGREGATION_OP:
			assert(childCount == 1);
			
			left = compileTerm(term.getChild(0), identifierContext);
			
			return AggregationFunction.apply(DuraAggregationFunctionMapping.fromToken(term), left);

		case RELATIVE_TIMER_OP:
			assert(childCount == 2);
			
			left = compileTerm(term.getChild(0), true);
			right = compileTerm(term.getChild(1), true);
			
			return new RelativeTimerOp(term.getText(), left, right);

		case PATH:
			String path = term.getText();
			
			if (identifierContext && TSA_SyntaxPatterns.splitAttributeName(path).length==1) {
				return UntypedIdentifier.create(path);
			} else {
				return Term.attr(term.getText());
			}
			
		case ARITHMETIC_OP:
			TupleFunctionID id = DuraMathFunctionMapping.fromToken(term);
			Term<?>[] arguments = new Term[childCount];
			
			for (int i=0; i<childCount; i++) {
				arguments[i] = compileTerm(term.getChild(i), identifierContext);
			}
			
			if(id == USER){
				throw new UnsupportedTokenException(term);
			}
			
			return TupleFunction.apply(id, arguments);
			
		case STRING:
			String tokenText = term.getText();
			
			return new StringValue(tokenText.substring(1, tokenText.length()-1));
			
		case NUMBER:
			return compileNumber(term);
			
		case DURATION:
			return new DurationValue(DuraTimeUnitMapping.tokenToDuration(term));
			
		case IDENTIFIERS:
			assert(identifierContext);

			Term<?>[] identifiers =  new Term[childCount];
			
			for (int i=0; i<childCount; i++) {
				DuraTree child = term.getChild(i);
				
				identifiers[i] = compileTerm(child, identifierContext);
			}
			
			return new UserTupleFunction<NoType>("identifier-set", NoType.instance, identifiers);
			
		case CLOCK_FUNCTION:
			if (term.getText().equals(DuraDefaultAttributeNameMapping.identifierClockDuraName)) {
				return identifierClock;
			} else if (term.getText().equals(DuraDefaultAttributeNameMapping.systemTimeClockDuraName)) {
				return timeClock;
			} else {
				throw new UnexpectedTokenException(term);
			}
		
		case BOOLEAN:
			return new BooleanValue(Boolean.parseBoolean(term.getText()));
			
		case CAST_OP:
			return new CastFunction(DuraBasicTypesMapping.fromDuraName(term.getText()).type, compileTerm(term.getChild(0)));
			
		default:
			throw new UnexpectedTokenException(term);
		}
	}
	
	private static BasicTypeValue compileNumber(final Tree number) {
		assert(number.getType() == NUMBER);
		
		String value = number.getText();
		String digits = value.substring(0, value.length()-1);
		
		switch (value.charAt(value.length()-1)) {
		case 'd':
			return new DoubleValue(Double.parseDouble(digits));

		case 'f':
			return new FloatValue(Float.parseFloat(digits));
			
		case 'l':
			return new LongValue(Long.parseLong(digits));
			
		default:
			if (value.contains(".")) {
				return new DoubleValue(Double.parseDouble(value));
			} else {
				return new IntValue(Integer.parseInt(value));
			}
		}
	}
	
	
	public static ConditionFormula compileFormula(final DuraTree formula) {
		int childCount = formula.getChildCount();
		ConditionFormula[] subformulas = new ConditionFormula[childCount];
		
		if (formula.getType() == CONJUNCTION || formula.getType() == DISJUNCTION || formula.getType() == NEGATION) {
			for (int i=0; i<childCount; i++) {
				DuraTree subformula = formula.getChild(i);
				
				switch (subformula.getType()) {
				case BOOLEAN:
					if (subformula.getText().equals("true")) {
						subformulas[i] = ConditionFormula.Top;
					} else if (subformula.getText().equals("false")) {
						subformulas[i] = ConditionFormula.Bottom;
					} else {
						throw new UnexpectedTokenException(subformula);
					}
					
					break;
				
				case CONJUNCTION:
				case DISJUNCTION:
				case NEGATION:
					subformulas[i] = compileFormula(subformula);
					
					break;
	
				case ARITHMETIC_RELATION:
				case TIME_RELATION:
				case CONSTRAINT:
					subformulas[i] = compileAtomicFormula(subformula);
					
					break;
					
				default: throw new UnexpectedTokenException(subformula);
				}
			}
		}
		
		switch (formula.getType()) {
		case CONJUNCTION:
			return new ConjunctiveConditionFormula(subformulas);
			
		case DISJUNCTION:
			return new DisjunctiveConditionFormula(subformulas);
			
		case NEGATION:
			assert(childCount == 1);
			
			return new NegatedConditionFormula(subformulas[0]);
			
		default:
			return compileAtomicFormula(formula);
		}
	}

	private static ConditionFormula compileAtomicFormula(final DuraTree formula) {
		Term<?> left, right;
		
		switch (formula.getType()) {
		case ARITHMETIC_RELATION:
			BinaryComparissonOperatorID operator = BinaryComparissonOperatorID.fromToken(formula);
			
			left = compileTerm(formula.getChild(0));
			right = compileTerm(formula.getChild(1));
			
			return new BinaryComparissonFormula(operator, left, right);

		case TIME_RELATION:
			left = compileTerm(formula.getChild(0), true);
			right = compileTerm(formula.getChild(1), true);
			
			return new UserRelationFormula(formula.getText(), left, right);
			
		case CONSTRAINT:
			return DuraRelationConstraintMapping.fromToken(formula);
			
		default:
			throw new UnexpectedTokenException(formula);
		}
	}
	
	public static String joinPath(final Stack<String> path) {
		StringBuilder sb = new StringBuilder();

		for (Iterator<String> it = path.iterator(); it.hasNext();) {
			sb.append(it.next());
			
			if (it.hasNext())
				sb.append(".");
		}

		return sb.toString();
	}
	
}

class Unification {
	public final UnambiguousQueryVariables variables;
	public final FixedArrayNavigableMap<TypedAttribute<? extends Type>, Term<? extends Type>> definitions;
	
	public Unification(
			final UnambiguousQueryVariables variables,
			final FixedArrayNavigableMap<TypedAttribute<? extends Type>, Term<? extends Type>> definitions) {
		this.variables = variables;
		this.definitions = definitions;
	}
}