package de.lmu.ifi.pms.cep.dura.compiler.durac.formula_transformations;

import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.intervalBeginAttribute;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.intervalEndAttribute;
import static de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonOperatorID.EQUAL;
import static de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonOperatorID.GREATER_EQUAL;
import static de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonOperatorID.LESS;
import static de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonOperatorID.LESS_EQUAL;
import static de.lmu.ifi.pms.cep.tsa.conditions.TupleFunction.apply;
import static de.lmu.ifi.pms.cep.tsa.conditions.TupleFunctionID.GREATEST;
import static de.lmu.ifi.pms.cep.tsa.conditions.TupleFunctionID.LEAST;
import static de.lmu.ifi.pms.cep.tsa.conditions.TupleFunctionID.MINUS;
import static de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID.DURATION;
import static de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID.TIMESTAMP;

import static de.lmu.ifi.pms.cep.dura.compiler.durac.stateful_object_mappings.StatefulObjectMappings.*;

import java.util.HashMap;

import de.lmu.ifi.pms.cep.dura.compiler.durac.collections.TypeDefinitions;
import de.lmu.ifi.pms.cep.dura.compiler.durac.tsa.TimeInterval;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraCExceptionWrapper;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraUnspecificCompliationException;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.ConjunctiveConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.DisjunctiveConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.DurationValue;
import de.lmu.ifi.pms.cep.tsa.conditions.Minus;
import de.lmu.ifi.pms.cep.tsa.conditions.NamedTypeAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.TimestampAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.UserRelationFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.UserTupleFunction;
import de.lmu.ifi.pms.cep.tsa.conditions.adaptor.MapReplaceAtomicTerm;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.schema.CompositeType;
import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

class MapRewriteTimeRelations implements Adapter<AtomicConditionFormula, ConditionFormula> {

	public static final MapRewriteTimeRelations instance = new MapRewriteTimeRelations();
	
	private static final HashMap<String, ConditionFormula> binaryIntervalRelations = new HashMap<String, ConditionFormula>();

	private static final TimestampAttribute i_begin = TimestampAttribute.create(TSA_SyntaxPatterns.prefixAttribute("i", intervalBeginAttribute));
	private static final TimestampAttribute i_end = TimestampAttribute.create(TSA_SyntaxPatterns.prefixAttribute("i", intervalEndAttribute));
	private static final TimestampAttribute j_begin = TimestampAttribute.create(TSA_SyntaxPatterns.prefixAttribute("j", intervalBeginAttribute));
	private static final TimestampAttribute j_end = TimestampAttribute.create(TSA_SyntaxPatterns.prefixAttribute("j", intervalEndAttribute));

	static {
		binaryIntervalRelations.put("before", new BinaryComparissonFormula(LESS, i_end, j_begin));
		binaryIntervalRelations.put("after", new BinaryComparissonFormula(LESS, j_end, i_begin));
		binaryIntervalRelations.put("during", new ConjunctiveConditionFormula(new BinaryComparissonFormula(LESS, j_begin, i_begin), new BinaryComparissonFormula(LESS, i_end, j_end)));
		binaryIntervalRelations.put("contains", new ConjunctiveConditionFormula(new BinaryComparissonFormula(LESS, i_begin, j_begin), new BinaryComparissonFormula(LESS, j_end, i_end)));
		binaryIntervalRelations.put("overlaps", new ConjunctiveConditionFormula(new BinaryComparissonFormula(LESS, j_begin, i_begin), new BinaryComparissonFormula(LESS, i_begin, j_end), new BinaryComparissonFormula(LESS, j_end, i_end)));
		binaryIntervalRelations.put("overlaped-by", new ConjunctiveConditionFormula(new BinaryComparissonFormula(LESS, i_begin, j_begin), new BinaryComparissonFormula(LESS, j_begin, i_end), new BinaryComparissonFormula(LESS, i_end, j_end)));
		binaryIntervalRelations.put("meets", new BinaryComparissonFormula(EQUAL, i_end, j_end));
		binaryIntervalRelations.put("met-by", new BinaryComparissonFormula(EQUAL, j_end, i_begin));
		binaryIntervalRelations.put("starts", new ConjunctiveConditionFormula(new BinaryComparissonFormula(EQUAL, i_begin, j_begin), new BinaryComparissonFormula(LESS, i_end, j_end)));
		binaryIntervalRelations.put("started-by", new ConjunctiveConditionFormula(new BinaryComparissonFormula(EQUAL, j_begin, i_begin), new BinaryComparissonFormula(LESS, j_end, i_end)));
		binaryIntervalRelations.put("finishes", new ConjunctiveConditionFormula(new BinaryComparissonFormula(LESS, j_begin, i_begin), new BinaryComparissonFormula(EQUAL, i_end, j_end)));
		binaryIntervalRelations.put("finished-by", new ConjunctiveConditionFormula(new BinaryComparissonFormula(LESS, i_begin, j_begin), new BinaryComparissonFormula(EQUAL, j_end, i_end)));
		binaryIntervalRelations.put("equals", new ConjunctiveConditionFormula(new BinaryComparissonFormula(EQUAL, i_begin, j_begin), new BinaryComparissonFormula(EQUAL, j_end, i_end)));

		binaryIntervalRelations.put("while", new ConjunctiveConditionFormula(new BinaryComparissonFormula(LESS_EQUAL, i_begin, j_begin), new BinaryComparissonFormula(LESS_EQUAL, j_end, i_end)));
		binaryIntervalRelations.put("valid-during",	new ConjunctiveConditionFormula(new BinaryComparissonFormula(LESS, i_begin, j_end), new BinaryComparissonFormula(LESS_EQUAL, i_end, j_begin)));
	}
	
	@Override
	public ConditionFormula adapt(final AtomicConditionFormula source) {
		if (source instanceof UserRelationFormula) {
			UserRelationFormula relation = (UserRelationFormula) source;
			
			try {
				if(binaryIntervalRelations.containsKey(relation.name)) {
					TimeInterval left = toTimeInterval(relation.term(0));
					TimeInterval right = toTimeInterval(relation.term(1));
					
					final FixedArrayNavigableMap<AtomicTerm<?>, Term<?>> mapping = new FixedArrayNavigableMap<AtomicTerm<?>, Term<?>>(
								new AtomicTerm[] {     i_begin,       i_end,      j_begin,        j_end },
	     							  new Term[] { left.begin(), left.end(), right.begin(), right.end() }
							);
					
					return binaryIntervalRelations.get(relation.name).replaceAtomicTerm(new MapReplaceAtomicTerm(mapping));
				} else if (relation.name.equals("valid-at")) {
					TimeInterval interval = toTimeInterval(relation.term(0));
					Term<?> timepoint = relation.term(1);
					
					if (! (timepoint.isTyped && timepoint.type.equals(BasicTypeID.TIMESTAMP.type))) {
						throw new DuraUnspecificCompliationException(String.format("second argument of %s needs to be of type timepoint, found: %s", relation.name, timepoint));
					}

					return new ConjunctiveConditionFormula(
							new BinaryComparissonFormula(LESS, interval.begin(), timepoint),
							new BinaryComparissonFormula(LESS_EQUAL, interval.end(), timepoint)	
						);
					
				} else if (relation.name.equals("valid-at-distance")) {
						TimeInterval interval = toTimeInterval(relation.term(0));
						Term<?> timepoint = relation.term(1);
						
						if (! (timepoint.isTyped && timepoint.type.equals(BasicTypeID.TIMESTAMP.type))) {
							throw new DuraUnspecificCompliationException(String.format("second argument of %s needs to be of type timepoint, found: %s", relation.name, timepoint));
						}

						return new ConjunctiveConditionFormula(
								new BinaryComparissonFormula(LESS, interval.begin(), timepoint),
								new BinaryComparissonFormula(LESS_EQUAL, interval.end(), apply(MINUS, TIMESTAMP.type, timepoint, new DurationValue(Long.parseLong(minDurationDigits))))			//this part of the condition is implicitly negated
							);
						
				} else if (relation.name.equals("within")) {
					FixedArray<Term<?>> identifiers = ((UserTupleFunction<?>) relation.term(0)).arguments();
					Term<?> duration = relation.term(1);
					
					if (! (duration.isTyped && duration.type.equals(BasicTypeID.DURATION.type))) {
						throw new DuraUnspecificCompliationException(String.format("second argument of %s needs to be of type duration, found: %s", relation.name, duration));
					}
					
					int identifierCount = identifiers.size();
					
					Term<?>[] begins = new Term[identifierCount];
					Term<?>[] ends = new Term[identifierCount];
					
					for (int i=0; i<identifierCount; i++) {
						TimeInterval interval = toTimeInterval(identifiers.get(i));
						
						begins[i] = interval.begin();
						ends[i] = interval.end();
					}
					
					return new BinaryComparissonFormula(LESS, apply(MINUS, DURATION.type, apply(GREATEST, TIMESTAMP.type, ends), apply(LEAST, TIMESTAMP.type, begins)), duration);
					
				} else if (relation.name.equals("apart")) {
					FixedArray<Term<?>> identifiers = ((UserTupleFunction<?>) relation.term(0)).arguments();
					Term<?> duration = relation.term(1);
					
					if (! (duration.isTyped && duration.type.equals(BasicTypeID.DURATION.type))) {
						throw new DuraUnspecificCompliationException(String.format("second argument of %s needs to be of type duration, found: %s", relation.name, duration));
					}
					
					assert(identifiers.size() == 2);
					
					TimeInterval first =  toTimeInterval(identifiers.get(0));
					TimeInterval second = toTimeInterval(identifiers.get(1));
	
					return new DisjunctiveConditionFormula(
							new BinaryComparissonFormula(GREATER_EQUAL, apply(MINUS, DURATION.type, first.end(), second.begin()), duration),
							new BinaryComparissonFormula(GREATER_EQUAL, apply(MINUS, DURATION.type, second.end(), first.begin()), duration)
						);
				}
			} catch (DuraUnspecificCompliationException e) {
				throw new DuraCExceptionWrapper(e);
			}
		}
		
		return source;
	}
	
	public static TimeInterval toTimeInterval(final Term<?> term) throws DuraUnspecificCompliationException {
		if (term instanceof TimeInterval) {
			return (TimeInterval) term;
		} else if (term instanceof NamedTypeAttribute<?>) {
			assert ((NamedTypeAttribute<?>) term).type.equals(TypeDefinitions.TIME_INTERVAL);
			
			return new TimeInterval((NamedTypeAttribute<CompositeType>) term);
		}
		
		throw new DuraUnspecificCompliationException(String.format("term of type time interval expected, found %s", term));
	}
}