package de.lmu.ifi.pms.cep.dura.compiler.durac.stateful_object_mappings;

import static de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraLexer.*;
import static de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings.DuraDefaultAttributeNameMapping.*;

import java.io.IOException;
import java.io.InputStream;
import java.util.Scanner;

import org.antlr.runtime.ANTLRStringStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.RecognitionException;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraC;
import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraLexer;
import de.lmu.ifi.pms.cep.dura.compiler.antlr.utils.DuraImplicitTree;
import de.lmu.ifi.pms.cep.dura.compiler.antlr.utils.DuraTree;
import de.lmu.ifi.pms.cep.dura.compiler.exceptions.DuraTransformationException;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;

public class StatefulObjectMappings {
	public static final String minDurationUnit = "ms";
	public static final String minDurationDigits = "3000";

	public static final String minDurationString = String.format("%s %s", minDurationDigits, minDurationUnit);

	public static DuraImplicitTree getGenericStatefulObjectRules(final DuraTree soDefinition, final DuraTree conflictResolution) {
		try {
			DuraTree schema = soDefinition.getChild(1);
			
			assert (schema.getType() == SCHEMA);
			
			String soName = schema.getChild(0).getText();
			
			String specificRules;
			String path = "de/lmu/ifi/pms/cep/dura/compiler/durac/stateful_object_mappings/GenericStatefulObjectRules.dura";
			{
				InputStream is = StatefulObjectMappings.class.getClassLoader().getResourceAsStream(path);
				if(is != null){
					specificRules = new Scanner(is).useDelimiter("\\Z").next();
				}
				else{
					throw new IOException("Could not find resource \"" + path + "\".");
				}
			}
			
			specificRules = specificRules.replaceAll("operation-mode", soName);
			specificRules = specificRules.replace("const min-dur", minDurationString);
			specificRules = specificRules.replace("false //conflict resolution", translateConflictResolution(conflictResolution));
			
			ANTLRStringStream stream = new ANTLRStringStream(specificRules);
			DuraC parser = new DuraC(new CommonTokenStream(new DuraLexer(stream)));
			DuraImplicitTree tree = new DuraImplicitTree(parser.eventSpecifications().getTree(), soDefinition);

			return tree;
		} catch (IOException e) {
			throw new Error(e);
		} catch (RecognitionException e) {
			throw new DuraTransformationException(e, soDefinition);
		}
	}
	
	private static String translateConflictResolution(final DuraTree conflictResolution) {
		StringBuilder sb = new StringBuilder();
		
		int childCount = conflictResolution==null ? 0 : conflictResolution.getChildCount();

		for (int i=0; i<childCount; i++) {
			DuraTree select = conflictResolution.getChild(i);
						
			assert select.getType()==SELECT;
			
			String path = select.getChild(0).getText();
			String expandedPath = path.startsWith(setLabel+".") ? TSA_SyntaxPatterns.attributeName(payloadLabel, path) : path;
			
			String firstPath = TSA_SyntaxPatterns.attributeName("first-u-s", expandedPath);
			String otherPath = TSA_SyntaxPatterns.attributeName("other-u-s", expandedPath);

			sb.append("or{ ");
			sb.append(firstPath);
			
			if (select.getText().equals("min")) {
				sb.append(" > ");
			} else if (select.getText().equals("max")) {
				sb.append(" < ");				
			}
			
			sb.append(otherPath);
			sb.append(", ");
			
			sb.append("and{ ");
			sb.append(firstPath);
			sb.append(" = ");
			sb.append(otherPath);
			
			sb.append(", ");
		}
		
		sb.append("false");
		
		for (int i=0; i<childCount*2; i++) {
			sb.append(" }");
		}
		
		return sb.toString();
	}
}
