package de.lmu.ifi.pms.cep.dura.compiler.durac.token_mappings;

import static de.lmu.ifi.pms.cep.tsa.DTSA_BufferType.INPUT;
import static de.lmu.ifi.pms.cep.tsa.DTSA_BufferType.LOG;
import static de.lmu.ifi.pms.cep.tsa.DTSA_BufferType.OUTPUT;

import java.util.HashSet;
import java.util.Set;

import org.antlr.runtime.tree.Tree;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.DuraLexer;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


public class DuraExternalModifierMapping {
	private static final FixedArrayNavigableMap<String, DTSA_BufferType> fromDuraMap = new FixedArrayNavigableMap<String, DTSA_BufferType>(
			new         String[]{"input", "output", "log"}, 
			new DTSA_BufferType[]{  INPUT,   OUTPUT,   LOG}
	);
	
	public static Set<DTSA_BufferType> fromToken(final Tree modifier) {
		assert(modifier.getType() == DuraLexer.MODIFIERS);
		
		Set<DTSA_BufferType> result = new HashSet<DTSA_BufferType>();
		
		for (int i=0; i<modifier.getChildCount(); i++) {
			String name = modifier.getChild(i).getText();
			
			if (isExternalModifier(name)) {
				result.add(fromDuraMap.get(name));
			}
		}
		
		return result;
	}
	
	public static boolean isExternalModifier(String name) {
		return fromDuraMap.has(name);
	}
	
	public static boolean containsStaticModifier(final Tree modifier) {
		assert(modifier.getType() == DuraLexer.MODIFIERS);

		for (int i=0; i<modifier.getChildCount(); i++) {
			if (modifier.getChild(i).getText().equals("static")) {
				return true;
			}
		}

		return false;
	}
}