package de.lmu.ifi.pms.cep.dura.compiler.exceptions;

import java.io.File;

import de.lmu.ifi.pms.cep.dura.compiler.antlr.utils.DuraTree;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_AbstractException;

/**
 * Exceptions that are referring to a position in a dura program file.  
 */
public class DuraCompilationException extends RuntimeException {
	
	DuraCompilationException(final String msg) {
		super(msg);
	}
	
	public DuraCompilationException(final String msg, final String positionDescription, final Throwable cause) {
		super(String.format("%s\n%s", msg, positionDescription), cause);
	}
	
	public DuraCompilationException(final String msg, final String positionDescription) {
		super(String.format("%s\n%s", msg, positionDescription));
	}
	
	public DuraCompilationException(final TSA_AbstractException e, final DuraTree cause) {
		this(e.getSimpleMessage(), cause.getPositionDescription(), e);
	}
	
	public DuraCompilationException(final String message, final Throwable e) {
		super(message, e);
	}

	public DuraCompilationException(final Throwable e, final DuraTree cause) {
		this(e.getMessage(), cause.getPositionDescription(), e);
	}

	public DuraCompilationException(final String msg, final DuraTree cause) {
		this(msg, cause.getPositionDescription());
	}
	
	public DuraCompilationException(final String msg, final String filename, final int line) {
		this(msg, String.format("\tat %s:%d", filename==null ? "virtual file" : new File(filename).getAbsolutePath(), line));
	}
	
}
