package de.lmu.ifi.pms.cep.event_mill.configuration;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.ElementMap;
import org.simpleframework.xml.Path;
import org.simpleframework.xml.Root;
import org.simpleframework.xml.Serializer;
import org.simpleframework.xml.core.Persister;
import org.simpleframework.xml.strategy.ExplicitCycleStrategy;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.COMMON_CONSTANTS;
import de.lmu.ifi.pms.cep.compilation.CompilerFactory;
import de.lmu.ifi.pms.cep.event_mill.execution.program.StreamBufferStoreFactory;
import de.lmu.ifi.pms.cep.event_mill.execution.scheduler.SchedulerFactory;
import de.lmu.ifi.pms.cep.tsa.DTSA_Program;
import de.lmu.ifi.pms.cep.tsa.TSA_Program;

import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


@Root(name="configuration")
public class Configuration {
	
	@Element(name="stream-buffer-store")
	public final StreamBufferStoreFactory<?,?> streamBufferStore;
	
	@Element(name="scheduler")
	public final SchedulerFactory scheduler;
	
	
	
	public final FixedArrayNavigableMap<String, CompilerFactory<?, ?>> compilers;

	@Path(value="compilers")
	@ElementList(inline=true, name="compiler", required=false, type=CompilerFactory.class)
	java.util.ArrayList<CompilerFactory<?, ?>> getCompilers(){
		return new java.util.ArrayList<CompilerFactory<?, ?>>(this.compilers.values().javaCollection());
	}
	
	@Path(value="compilers")
	@Attribute(name="default-tsa-compiler", required=false)
	public final String defaultTSACompiler;
	public final CompilerFactory<String[], TSA_Program> defaultTSACompilerFactory;

	@Path(value="compilers")
	@Attribute(name="default-dtsa-compiler", required=false)
	public final String defaultDTSACompiler;
	public final CompilerFactory<TSA_Program, DTSA_Program> defaultDTSACompilerFactory;
	
	protected static FixedArrayNavigableMap<String, CompilerFactory<?, ?>> compilers(final java.util.ArrayList<CompilerFactory<?, ?>> compilers){
		CompilerFactory<?, ?>[] compilerArray = compilers.toArray(new CompilerFactory<?, ?>[0]);
		String[] identifierArray = new String[compilerArray.length];
		for(int index = 0; index < compilerArray.length; ++index){
			identifierArray[index] = compilerArray[index].identifier;
		}
		return new FixedArrayNavigableMap<String, CompilerFactory<?,?>>(identifierArray, compilerArray);
	}
	
	public final FixedArrayNavigableMap<String, CompilerFactory<?,?>> aliases;
	
	@Path(value="compilers")
	@ElementMap(inline=true, required=false, entry="alias", key="name", attribute=true, value="compiler")
	java.util.HashMap<String, CompilerFactory<?,?>> getAliases(){
		if(this.aliases.size() == 0){
			return null;
		}
		else{
			return new java.util.HashMap<String, CompilerFactory<?,?>>(this.aliases.javaMap());
		}
	}
	
	
	public Configuration(@Element(name="stream-buffer-store") final StreamBufferStoreFactory<?,?> streamBufferStore
			, @Element(name="scheduler") final SchedulerFactory scheduler
			, @Path(value="compilers") @ElementList(inline=true, name="compiler", required=false, type=CompilerFactory.class) final java.util.ArrayList<CompilerFactory<?,?>> compilers
			, @Path(value="compilers") @Attribute(name="default-tsa-compiler", required=false) final String defaultTSACompiler
			, @Path(value="compilers") @Attribute(name="default-dtsa-compiler", required=false) final String defaultDTSACompiler
			, @Path(value="compilers") @ElementMap(inline=true, required=false, entry="alias", key="name", attribute=true, value="compiler") final java.util.HashMap<String, CompilerFactory<?,?>> aliases
	) {
		this(streamBufferStore, scheduler, (compilers == null ? new FixedArrayNavigableMap<String, CompilerFactory<?, ?>>() : compilers(compilers)), defaultTSACompiler, defaultDTSACompiler, (aliases == null ? new FixedArrayNavigableMap<String, CompilerFactory<?, ?>>(): new FixedArrayNavigableMap<String, CompilerFactory<?, ?>>(aliases)));
	}
	
	@SuppressWarnings("unchecked")
	public Configuration(final StreamBufferStoreFactory<?,?> streamBufferStore, final SchedulerFactory scheduler, final FixedArrayNavigableMap<String, CompilerFactory<?, ?>> compilers, final String defaultTSACompiler, final String defaultDTSACompiler, final FixedArrayNavigableMap<String, CompilerFactory<?,?>> aliases) {
		super();
		this.streamBufferStore = streamBufferStore;
		this.scheduler = scheduler;
		this.compilers = compilers;
		this.defaultTSACompiler = defaultTSACompiler;
		this.defaultDTSACompiler = defaultDTSACompiler;
		if(this.defaultTSACompiler != null){
			if(this.compilers.has(this.defaultTSACompiler)){
				this.defaultTSACompilerFactory = (CompilerFactory<String[], TSA_Program>) compilers.get(defaultTSACompiler);
				if(!this.defaultTSACompilerFactory.sourceProgramType.isAssignableFrom(String[].class)){
					throw new IllegalArgumentException("Default tsa compiler has incompatible source type.");
				}
				if(!TSA_Program.class.isAssignableFrom(this.defaultTSACompilerFactory.destinationProgramType)){
					throw new IllegalArgumentException("Default tsa compiler has incompatible destination type.");
				}
			}
			else{
				throw new IllegalArgumentException("Default tsa compiler is not defined.");
			}
		}
		else{
			this.defaultTSACompilerFactory = DEFAULTS.defaultTSACompiler;
		}
		if(this.defaultDTSACompiler != null){
			if(this.compilers.has(this.defaultDTSACompiler)){
				this.defaultDTSACompilerFactory = (CompilerFactory<TSA_Program, DTSA_Program>) compilers.get(defaultDTSACompiler);
				if(!this.defaultDTSACompilerFactory.sourceProgramType.isAssignableFrom(TSA_Program.class)){
					throw new IllegalArgumentException("Default dtsa compiler has incompatible source type.");
				}
				if(!DTSA_Program.class.isAssignableFrom(this.defaultDTSACompilerFactory.destinationProgramType)){
					throw new IllegalArgumentException("Default dtsa compiler has incompatible destination type.");
				}
			}
			else{
				throw new IllegalArgumentException("Default dtsa compiler is not defined.");
			}
		}
		else{
			this.defaultDTSACompilerFactory = DEFAULTS.defaultDTSACompiler;
		}
		this.aliases = aliases;
	}

	public static final Logger logger = LoggerFactory.getLogger(Configuration.class);
	
	public static boolean exists(){
		return existsResource(DEFAULTS.defaultConfigResourcePath);
	}
	
	public static Configuration read() throws IOException{
		return readResource(DEFAULTS.defaultConfigResourcePath);
	}
	
	public static String defaultConfigFilePath(){
		return DEFAULTS.resourceFilePath(DEFAULTS.defaultConfigResourcePath);
	}
	
	public static void write(final Configuration configuration){
		write(defaultConfigFilePath(), configuration);
	}
	
	public static boolean existsFile(final String path){
		File configFile = new File(path);
		return configFile.exists();
	}
	
	public static boolean existsResource(final String path){		
		return Configuration.class.getClassLoader().getResource(path) != null;
	}
	
	public static Configuration readResource(final String path) throws IOException{
		InputStream is = Configuration.class.getClassLoader().getResourceAsStream(path);
		if(is != null){
			return read(is);
		}
		else{
			throw new IOException("Could not find resource \"" + path + "\".");
		}
	}
	
	public static Configuration readFile(final String path) throws IOException{
		return read(new FileInputStream(path));
	}
	
	public static Configuration read(final InputStream input) throws IOException{
		Serializer serializer = new Persister(new ExplicitCycleStrategy("iri","ref", LiberiesLoader.classLoader), COMMON_CONSTANTS.SERIALIZATION_FORMAT);
		try {
			return serializer.read(Configuration.class, input);
		} catch (RuntimeException e) {
			throw e;
		} catch (IOException e) {
			throw e;
		} catch (Exception e) {
			logger.error("Unable to read configuration file!", e);
			throw new RuntimeException(e);
		}
	}
	
	public static void write(final String path, final Configuration configuration){
		Serializer serializer = new Persister(new ExplicitCycleStrategy("iri","ref", LiberiesLoader.classLoader), COMMON_CONSTANTS.SERIALIZATION_FORMAT);
		File configFile = new File(path);
		try {
			if(!configFile.exists()){
				configFile.createNewFile();
			}
			serializer.write(configuration, configFile);
		} catch (Exception e) {
			logger.error("Unable to write configuration file!", e);
			throw new RuntimeException(e);
		}
	}
	
	public static void main(final String[] args) throws Exception{
		System.out.println(defaultConfigFilePath());
		Configuration conf = Configuration.read();
		Serializer s = new Persister(new ExplicitCycleStrategy("iri","ref", LiberiesLoader.classLoader), COMMON_CONSTANTS.SERIALIZATION_FORMAT);
		s.write(conf, System.out);
	}

}
