package de.lmu.ifi.pms.cep.event_mill.console;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.event_mill.configuration.Configuration;
import de.lmu.ifi.pms.cep.event_mill.configuration.IOConfiguration;
import de.lmu.ifi.pms.cep.event_mill.configuration.LiberiesLoader;
import de.lmu.ifi.pms.cep.event_mill.execution.engine.Engine;
import de.lmu.ifi.pms.cep.event_mill.execution.program.ExecutableProgram;
import de.lmu.ifi.pms.cep.event_mill.execution.scheduler.SchedulerFactory;
import de.lmu.ifi.pms.cep.event_mill.io.MessageSink;
import de.lmu.ifi.pms.cep.event_mill.io.MessageSource;
import de.lmu.ifi.pms.cep.event_mill.io.PATTERNS;
import de.lmu.ifi.pms.cep.event_mill.io.bus.MessageBus;
import de.lmu.ifi.pms.cep.event_mill.io.connect.BusToStreamConnector;
import de.lmu.ifi.pms.cep.event_mill.io.connect.StreamToBusConnector;
import de.lmu.ifi.pms.cep.event_mill.io.format.MessageFormatProvider;
import de.lmu.ifi.pms.cep.event_mill.io.stream.StreamIO;

import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.fixed.empty.Empty;
import de.lmu.ifi.pms.java.util.map.Map;



public class ControlThread extends AbstractMarkedConsoleThread {
	
	static final String streamBufferStoreIdentifier = "de.lmu.ifi.pms.cep.event-mill.prototype";
	static final String ownIdentity = "de.lmu.ifi.pms.cep.event_mill.console.ControlThread";//FIXME
	
	final SchedulerFactory scheduler;
	IOConfiguration ioConfiguration;
	MessageFormatProvider<String> inputFormat;
	MessageFormatProvider<String> outputFormat;
	public final Engine engine;
	MessageSource<String> source;
	MessageSink<String> sink;
	StreamIO stream; 
	BusToStreamConnector<?> inputConnector;
	StreamToBusConnector<?> outputConnector;
	
	public ControlThread(final Map<String, String, ?> variables, final SchedulerFactory scheduler, final IOConfiguration ioConfiguration, final ExecutableProgram<?, ?, ?> program) {
		this(variables, scheduler, ioConfiguration, program, new BufferedReader(new InputStreamReader(System.in)),System.out);
	}
	
	public ControlThread(final Map<String, String, ?> variables, final SchedulerFactory scheduler, final IOConfiguration ioConfiguration, final ExecutableProgram<?, ?, ?> program, final InputStream inputStream, final PrintStream outputStream) {
		this(variables, scheduler, ioConfiguration, program, new BufferedReader(new InputStreamReader(inputStream)),outputStream);
	}
	
	public ControlThread(final Map<String, String, ?> variables, final SchedulerFactory scheduler, final IOConfiguration ioConfiguration, final ExecutableProgram<?, ?, ?> program, final BufferedReader inputStream, final PrintStream outputStream) {
		this(variables, scheduler, ioConfiguration, program, Empty.<Iterator<String>>instance() , inputStream, outputStream);
	}
	
	public ControlThread(final Map<String, String, ?> variables, final SchedulerFactory scheduler, final ExecutableProgram<?, ?, ?> program, final Collection<Iterator<String>> immediateCommands, final BufferedReader inputStream, final PrintStream outputStream) {
		this(variables, scheduler, null, program, immediateCommands, inputStream, outputStream);
	}
	
	public ControlThread(final Map<String, String, ?> variables, final SchedulerFactory scheduler, final IOConfiguration ioConfiguration, final ExecutableProgram<?, ?, ?> program, final Collection<Iterator<String>> immediateCommands, final BufferedReader inputStream, final PrintStream outputStream) {
		this(variables, scheduler, ioConfiguration, new Engine(program), immediateCommands, inputStream, outputStream);
	}
	
	public ControlThread(final Map<String, String, ?> variables, final SchedulerFactory scheduler, final IOConfiguration ioConfiguration, final Engine engine, final Collection<Iterator<String>> immediateCommands, final BufferedReader inputStream, final PrintStream outputStream) {
		super(engine.marker.getName() + ": ControlThread", variables, immediateCommands, inputStream, outputStream, engine.marker);
		this.scheduler = scheduler;
		this.engine = engine;
		
		this.config(ioConfiguration);
	}
	
	static final String parseCommandRegEx = "parse[ \\t]+(.*+)";
	static final Pattern parseCommandPattern = Pattern.compile(parseCommandRegEx);
	
	void config(final IOConfiguration ioConfiguration){
		//FIXME check compatibility
		this.closeIO();
		this.ioConfiguration = ioConfiguration;
	}
	
	@Override
	public void run() {
		int exitCode = 0;
		try{
			String command;
			boolean running = true;
			
			this.outputStream.println("For leaving type \"quit\"");
			this.outputStream.println("Proposed step: \"init\", \"start\"");
			
			if(this.ioConfiguration == null){
				this.info("Loading default I/O configuration.");						
				try{
					if(Configuration.exists()){
						this.info("Default I/O configuration exists. Loading default I/O configuration.");
						
						try {
							this.config(IOConfiguration.read());
							this.outputStream.println("Loaded default I/O configuration");
							this.info("Loaded default I/O configuration.");
						} catch (IOException e) {
							this.warn("Unable to load default I/O configuration.", e);
							this.skipRemainigCommands();
						}
	
					}
					else{
						this.outputStream.println("Default I/O configuration does not exist.");
					}
				} catch (Exception e) {
					this.warn("Unable to load default I/O configuration.", e);
					this.outputStream.println("Unable to load I/O default configuration.");
					this.skipRemainigCommands();
				}
			}
			
			while(running && this.ioConfiguration == null){
				this.outputStream.println("Proposed step: \"config\"");	
				try{
					command = this.nextCommand().trim();
				} catch (EndOfInputException e) {
					running = false;
					this.info("End of input.",e);
					break;
				}
				this.debug("Executing command: " + command);
				if(command.equals("")){
					continue;
				}
				else if(command.startsWith("set")){
					Matcher matcher = variableDefinitionPattern.matcher(command);
					
					if(matcher.matches()){
						String var = matcher.group(1);
						String value = matcher.group(2);
						this.setVariable(var, value);
					}
					else{
						this.outputStream.println("Illegal argument for set command.");
						this.skipRemainigCommands();
					}
				}
				else if("quit".equalsIgnoreCase(command)){
					running = false;
					
					this.debug("Stopping...");
				}
				else if(command.startsWith("config")){
					this.info("Starting configuration.");
					
					Matcher matcher = MainThread.configCommandPattern.matcher(command);
					
					if(matcher.matches()){
						String configFile = matcher.group(1);
						this.info("Loading configuration.");						
						try{
							if(Configuration.existsFile(configFile)){
								this.info("Configuration exists. Loading configuration.");
								
								try {
									this.config(IOConfiguration.readFile(configFile));
									this.outputStream.println("Loaded configuration");
									this.info("Loaded configuration.");
								} catch (IOException e) {
									this.outputStream.println("Unable to load configuration.");
									this.warn("Unable to load configuration.", e);
									this.skipRemainigCommands();
								}
								
							}
							else{
								this.outputStream.println("Configuration file does not exist: " + new File(configFile).getAbsolutePath());
								this.info("Configuration file does not exist: {}", new File(configFile).getAbsolutePath());
								this.skipRemainigCommands();
							}
						} catch (Exception e) {
							this.outputStream.println("Unable to load configuration.");
							this.warn("Unable to load configuration.", e);
							this.skipRemainigCommands();
						}
					}
					else{
						this.outputStream.println("Illegal argument for config command.");
						this.skipRemainigCommands();
					}
					
					this.info("Finished configuration.");
				}
				else{
					this.info("Unknown command: {}", command);
					
					this.outputStream.println("Sorry, I do not know that command! Please enter another one.");
				}
			}
		
			while(running){
				try{
					command = this.nextCommand().trim();
				} catch (EndOfInputException e) {
					running = false;
					this.info("End of input.",e);
					break;
				}
				this.debug("Executing command: " + command);
				if(command.equals("")){
					
				}
				else if(command.startsWith("config")){
					this.info("Starting configuration.");
					
					Matcher matcher = MainThread.configCommandPattern.matcher(command);
					
					if(matcher.matches()){
						String configFile = matcher.group(1);
						this.info("Loading configuration.");						
						try{
							if(Configuration.existsFile(configFile)){
								this.info("Configuration exists. Loading configuration.");
								
								try {
									this.config(IOConfiguration.readFile(configFile));
									this.outputStream.println("Loaded configuration");
									this.info("Loaded configuration.");
								} catch (IOException e) {
									this.outputStream.println("Unable to load configuration.");
									this.warn("Unable to load configuration.", e);
									this.skipRemainigCommands();
								}
								
							}
							else{
								this.outputStream.println("Configuration file does not exist: " + new File(configFile).getAbsolutePath());
								this.info("Configuration file does not exist: {}", new File(configFile).getAbsolutePath());
								this.skipRemainigCommands();
							}
						} catch (Exception e) {
							this.outputStream.println("Unable to load configuration.");
							this.warn("Unable to load configuration.", e);
							this.skipRemainigCommands();
						}
					}
					else{
						this.outputStream.println("Illegal argument for config command.");
						this.skipRemainigCommands();
					}
					
					this.info("Finished configuration.");
				}
				else if(command.startsWith("inject")){
					Matcher matcher = injectCommandPattern.matcher(command);
					if(matcher.matches()){
						String file = matcher.group(1);
						try {
							this.inject(file);
						} catch (FileNotFoundException e) {
							this.outputStream.println("Unable to load file \"" + new File(file).getAbsolutePath() + "\".");
							this.info("Unable to load file.", e);
							this.skipRemainigCommands();
						}
					}
					else{
						this.outputStream.println("Illegal argument for pause command.");
						this.skipRemainigCommands();
					}
				}
				else if(command.startsWith("pause")){
					Matcher matcher = pauseCommandPattern.matcher(command);
					if(matcher.matches()){
						if(matcher.group(1) == null){
							this.outputStream.println("Press [Enter] to continue.");
							this.pause();
						}
						else{
							long milies = Long.parseLong(matcher.group(1));
							this.outputStream.println("Pausing for " + milies + " millies.");
							this.pause(milies);
						}
					}
					else{
						this.outputStream.println("Illegal argument for pause command.");
						this.skipRemainigCommands();
					}
				}
				else if(command.equalsIgnoreCase("init")){
					this.engine.init(this.scheduler);
					
					this.info("Initialized");
					
					this.outputStream.println("Initialized");
					this.outputStream.println("Proposed step: \"start\"");
				}
				else if(command.equalsIgnoreCase("start")){
					if(!this.engine.isInitialized()){
						this.engine.init(this.scheduler);
						
						this.info("Initialized");
						
						this.outputStream.println("Initialized");
					}
					
					this.engine.start();
					
					this.info("Started");
					
					this.outputStream.println("Started");
					this.outputStream.println("Possible step: \"stop\", \"quit\"");
				}
				else if(command.equalsIgnoreCase("stop")){
					if(this.engine.isRunning()){
						this.engine.stop();
						
						this.info("Stopped");
						
						this.outputStream.println("Stopped");
						this.outputStream.println("Proposed steps: \"start\", \"quit\"");
					}
					else{
						this.outputStream.println("Invalid command: Program is not running.");
						this.skipRemainigCommands();
					}
				}
				else if(command.startsWith("set")){
					Matcher matcher = variableDefinitionPattern.matcher(command);
					
					if(matcher.matches()){
						String var = matcher.group(1);
						String value = matcher.group(2);
						this.setVariable(var, value);
					}
					else{
						this.outputStream.println("Illegal argument for set command.");
						this.skipRemainigCommands();
					}
				}
				else if(command.equalsIgnoreCase("quit")){
					if(this.engine.isRunning()){
						this.engine.stop();					
					}
					running = false;
					
				}
				else if(command.equalsIgnoreCase("write")){
					if(this.inputConnector == null){
						this.initIO();
					}
					this.inputConnector.nextBlock();
				}
				else if(command.equalsIgnoreCase("read")){
					if(this.outputConnector == null){
						this.initIO();
					}
					this.outputConnector.nextBlock();
				}
				else if(command.equalsIgnoreCase("print")){
					if(this.source == null){
						this.initIO();
					}
					String message;
					while(true){//TODO stop after a certain number
						message = this.source.read();
						if(message != null){
							this.outputStream.println("Read " + message + " from message bus.");
						}
						else{
							break;
						}
					}
				}
				else if(command.startsWith("parse")){
					if(this.sink == null){
						this.initIO();
					}
					Matcher matcher = parseCommandPattern.matcher(command);
					if(matcher.matches()){
						String message = matcher.group(1);
	
						this.sink.write(message);
											
						this.outputStream.println("Send " + message + " to message bus.");
					}
					else{
						this.outputStream.println("Illegal argument for config command.");
						this.skipRemainigCommands();
					}
				}
				else{
					this.info("Unknown command");
					
					this.outputStream.println("Sorry, I do not understand what you want! Please try again.");
					this.skipRemainigCommands();
				}
			}
		}
		catch(Throwable t){
			this.error("",t);
			this.outputStream.println(t.toString());
			exitCode = 1;
		}
		finally{
			this.engine.close();
			this.closeIO();
		}
	}
	
	protected void closeIO(){
		this.inputFormat = null;
		if(this.stream != null){
			this.stream.close();
			this.stream = null;
		}
		if(this.inputConnector != null){
			this.inputConnector.close();
			this.inputConnector = null;
		}
		if(this.outputConnector != null){
			this.outputConnector.close();
			this.outputConnector = null;
		}
		if(this.sink != null){
			this.sink.close();
			this.sink = null;
		}
		if(this.source != null){
			this.source.close();
			this.source = null;
		}
	}
	
	protected void initIO(){
		if(this.inputFormat == null){
			try{
				this.inputFormat = this.ioConfiguration.format(String.class, "de.lmu.ifi.pms.cep.event-mill.console.in");
			}
			catch(NoSuchElementException e){
				try {
					@SuppressWarnings("unchecked")
					MessageFormatProvider<String> inputFormat = (MessageFormatProvider<String>) Class.forName("de.lmu.ifi.pms.cep.event-mill.io.format.impl.semi_structured.SemiStructuredMessageFormatProvider", true, LiberiesLoader.classLoader).newInstance();
					assert String.class.isAssignableFrom(inputFormat.type);
					this.inputFormat = inputFormat;
				} catch (Exception e1) {
					throw e;
				}
			}
		}
		if(this.outputFormat == null){
			try{
				this.outputFormat = this.ioConfiguration.format(String.class, "de.lmu.ifi.pms.cep.event-mill.console.out");
			}
			catch(NoSuchElementException e){
				try {
					@SuppressWarnings("unchecked")
					MessageFormatProvider<String> outputFormat = (MessageFormatProvider<String>) Class.forName("de.lmu.ifi.pms.cep.event-mill.io.format.impl.semi_structured.SemiStructuredMessageFormatProvider", true, LiberiesLoader.classLoader).newInstance();
					assert String.class.isAssignableFrom(outputFormat.type);
					this.outputFormat = outputFormat;
				} catch (Exception e1) {
					throw e;
				}
			}
		}	
		if(this.stream == null){
			this.stream = this.ioConfiguration.streamProvider(streamBufferStoreIdentifier).connect(PATTERNS.programIdentifier(streamBufferStoreIdentifier, this.engine.executableProgram.fullName));
		}
		if(this.inputConnector == null){
			this.inputConnector = BusToStreamConnector.create(this.ioConfiguration.bus(PATTERNS.inputBusIdentifier(streamBufferStoreIdentifier, this.engine.executableProgram.fullName), String.class, this.inputFormat, PATTERNS.programIdentifier(streamBufferStoreIdentifier, this.engine.executableProgram.fullName)), this.stream);
		}
		if(this.outputConnector == null){
			this.outputConnector = StreamToBusConnector.create(this.ioConfiguration.bus(PATTERNS.outputBusIdentifier(streamBufferStoreIdentifier, this.engine.executableProgram.fullName), String.class, this.outputFormat, PATTERNS.programIdentifier(streamBufferStoreIdentifier, this.engine.executableProgram.fullName)), this.stream);
		}
		if(this.sink == null){
			MessageBus<?> messageBus = this.ioConfiguration.bus(PATTERNS.inputBusIdentifier(streamBufferStoreIdentifier, this.engine.executableProgram.fullName), String.class, this.inputFormat, ownIdentity + ":" + this.engine.executableProgram.fullName);
			if(messageBus.type.isAssignableFrom(String.class)){
				@SuppressWarnings("unchecked")
				MessageBus<String> stringMessageBus = (MessageBus<String>) messageBus;
				this.sink = stringMessageBus.sink();
			}
		}
		if(this.source == null){
			MessageBus<?> messageBus = this.ioConfiguration.bus(PATTERNS.outputBusIdentifier(streamBufferStoreIdentifier, this.engine.executableProgram.fullName), String.class, this.outputFormat, ownIdentity + ":" + this.engine.executableProgram.fullName);
			if(messageBus.type.isAssignableFrom(String.class)){
				@SuppressWarnings("unchecked")
				MessageBus<String> stringMessageBus = (MessageBus<String>) messageBus;
				this.source = stringMessageBus.source();
			}
		}
	}
	
	//logging
	protected static final Logger logger = LoggerFactory.getLogger(ControlThread.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}



}
