package de.lmu.ifi.pms.cep.event_mill.execution.dummy;

import de.lmu.ifi.pms.cep.event_mill.execution.program.ExecutableProgram;
import de.lmu.ifi.pms.cep.event_mill.execution.program.ExecutableProgramState;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.DTSA_QueryType;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArraySet;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;
import de.lmu.ifi.pms.java.util.map.fixed.FixedEnumMap;

public class DummyExecutableProgram extends ExecutableProgram<DummyStreamBufferStore, DummyStreamBuffer, DummyExecutableQuery>
{
	
	//final FixedArrayNavigableMap<String, Clock> clocks;
	protected ExecutableProgramState lastState;
	
	FixedEnumMap<DTSA_BufferType,FixedArrayNavigableMap<String, DummyStreamBuffer>> streamBuffers;
	FixedArraySet<DummyExecutableQuery>[] queries;
	
	public DummyExecutableProgram(final String namespace, final String simple_name, final String instance, final DummyStreamBufferStore bufferStore){
		this(namespace, simple_name, instance, ExecutableProgramState.CREATED, bufferStore);
	}
	
	@SuppressWarnings("unchecked")
	DummyExecutableProgram(final String namespace, final String simple_name, final String instance, final ExecutableProgramState state, final DummyStreamBufferStore bufferStore)
	{
		super(namespace, simple_name, instance, bufferStore);
		
		//this.clocks = new FixedArrayNavigableMap<String, Clock>(new String[]{EMILI.SystemClockName, SQL_Translation.tupleIdSeqenceName}, new Clock[]{new SystemClock(), new TupleIdSequenceClock(this.bufferStore, this.meta_schema, SQL_Translation.tupleIdSeqenceName, this.usePreparedStatements)});
		
		FixedArrayNavigableMap<String, DummyStreamBuffer>[] streamBuffers = new FixedArrayNavigableMap[DTSA_BufferType.count];
		for(int ordinal = 0; ordinal < DTSA_BufferType.count; ++ordinal){
			streamBuffers[ordinal] = new FixedArrayNavigableMap<String, DummyStreamBuffer>();
		}
		this.streamBuffers = new FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String,DummyStreamBuffer>>(DTSA_BufferType.class, streamBuffers);
		
		this.queries = (FixedArraySet<DummyExecutableQuery>[]) new FixedArraySet<?>[DTSA_QueryType.values().length];
		java.util.Arrays.fill(this.queries, new FixedArraySet<DummyExecutableQuery>());
		
		assert state != null;
		this.lastState = state;
		this.info("{}",this.lastState);
	}
	
	@Override
	public FixedEnumMap<DTSA_BufferType,FixedArrayNavigableMap<String, DummyStreamBuffer>> StreamBuffers()
	{
		return this.streamBuffers;
	}
	
	synchronized void register(final DummyStreamBuffer buffer){
		if(!this.isLocked()){			
			String bufferName = buffer.name();
			DTSA_BufferType bufferType = buffer.type;
			FixedArrayNavigableMap<String, DummyStreamBuffer> streamBuffersOfSameType = this.streamBuffers.get(bufferType);
			int index = streamBuffersOfSameType.keys().ceilingIndex(bufferName);
			if(index < streamBuffersOfSameType.size() && bufferName.equals(streamBuffersOfSameType.keys().get(index))){
				Error e = new Error("Invalid call to register. ALready registered buffer with name: " + bufferName);
				
				this.error("", e); //TODO better log message
			}
			else{
				String[] bufferNames = new String[streamBuffersOfSameType.size() + 1];
				DummyStreamBuffer[] buffers = new DummyStreamBuffer[bufferNames.length];
				
				streamBuffersOfSameType.keys().copyTo(0, bufferNames, 0, index);
				streamBuffersOfSameType.values().copyTo(0, buffers, 0, index);
				bufferNames[index] = bufferName;
				buffers[index] = buffer;
				streamBuffersOfSameType.keys().copyTo(index, bufferNames, index + 1, streamBuffersOfSameType.size() - index);
				streamBuffersOfSameType.values().copyTo(index, buffers, index + 1, streamBuffersOfSameType.size() - index);

				@SuppressWarnings("unchecked")
				FixedArrayNavigableMap<String, DummyStreamBuffer>[] streamBuffers = new FixedArrayNavigableMap[DTSA_BufferType.count];
				for(int ordinal = 0; ordinal < DTSA_BufferType.count; ++ordinal){
					if(ordinal == bufferType.ordinal()){
						streamBuffers[ordinal] = new FixedArrayNavigableMap<String, DummyStreamBuffer>(new FixedArrayNavigableSet<String>(false, false, bufferNames), new FixedArray<DummyStreamBuffer>(false, buffers));
					}
					else{
						streamBuffers[ordinal] = this.streamBuffers.values().get(ordinal);
					}
				}
				
				this.streamBuffers = new FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String,DummyStreamBuffer>>(DTSA_BufferType.class, streamBuffers);
				
				this.debug("Registered Buffer {} {}", buffer.id, buffer.name()); //TODO better log message
			}
		}
		else{
			throw new IllegalStateException();
		}
	}

	@Override
	public FixedEnumMap<DTSA_QueryType, FixedArraySet<DummyExecutableQuery>> Queries()
	{
		return new FixedEnumMap<DTSA_QueryType, FixedArraySet<DummyExecutableQuery>>(DTSA_QueryType.class, java.util.Arrays.copyOf(this.queries, this.queries.length));
	}
	
	synchronized void register(final DummyExecutableQuery query){
		if(!this.isLocked()){
			DummyExecutableQuery[] queryTypeQueries = new DummyExecutableQuery[this.queries[query.type.ordinal()].size() + 1];
			for(int index = 0; index < this.queries[query.type.ordinal()].size(); ++index){
				if(query == this.queries[query.type.ordinal()].get(index)){//TODO better test?
					Error e = new Error("Invalid call to register.");
					
					this.error("", e); //TODO better log message
				}
				queryTypeQueries[index] = this.queries[query.type.ordinal()].get(index);
			}
			queryTypeQueries[queryTypeQueries.length - 1] = query;
			this.queries[query.type.ordinal()] = new FixedArraySet<DummyExecutableQuery>(false, false, queryTypeQueries);
			
			this.debug("Registered Query {}", query.id); //TODO better log message
		}
		else{
			throw new IllegalStateException();
		}
	}
	
	@Override
	protected synchronized ExecutableProgramState readState() {
		if(this.state != null){
			return this.state;
		}
		else{
			return this.lastState;
		}
	}
	
	@Override
	protected synchronized boolean aquireLock() {
		return this.state == null;
	}
	
	@Override
	protected synchronized void releaseLock() {
		
	}
	
	@Override
	public synchronized void unlock() {
		this.lastState = this.state;
		super.unlock();
	}
	
//	protected static HashMap<String, HashMap<String,WeakReference<SQL_StreamBufferStore>>> stores = new HashMap<String, HashMap<String,WeakReference<SQL_StreamBufferStore>>>();
//	
//	protected static synchronized void remove(final String url, final String catalog, final String meta_schema_name){
//		stores.get(url).remove(encode(catalog, meta_schema_name));
//		if(stores.get(url).isEmpty()){
//			stores.remove(url);;
//		}
//	}
//	
//	protected static synchronized void put(final SQL_StreamBufferStore store){
//		if(!stores.containsKey(store.url)){
//			stores.put(store.url, new HashMap<String, WeakReference<SQL_StreamBufferStore>>());
//		}
//		stores.get(store.url).put(encode(store.catalog, store.meta_schema_name), new WeakReference<SQL_StreamBufferStore>(store));
//	}

}
