package de.lmu.ifi.pms.cep.event_mill.execution.scheduler;

import de.lmu.ifi.pms.cep.event_mill.execution.engine.Engine;
import de.lmu.ifi.pms.cep.event_mill.execution.program.ExecutableProgram;
import de.lmu.ifi.pms.cep.event_mill.execution.program.ExecutableQuery;
import de.lmu.ifi.pms.java.math.Math;
import de.lmu.ifi.pms.java.util.stream.io.BlockingPipe;

public class BasicScheduler extends SchedulerThread
{

	long delayInMillies;
	final ExecutableProgram<?, ?, ?> executableProgram;
	final QuerySequencer sequencer;
	final Thread thread;
	final BlockingPipe<Iterable<? extends ExecutableQuery<?, ?>>> rounds = new BlockingPipe<Iterable<? extends ExecutableQuery<?,?>>>(1);
	
	public BasicScheduler(final Engine engine, final long delayInMillies, final QuerySequencer sequencer)
	{
		super(engine);
		this.delayInMillies = Math.max(delayInMillies,1);
		this.executableProgram = engine.executableProgram;
		this.sequencer = sequencer;
		this.thread = new Thread(this.marker.getName() + ": SchedulerHelperThread"){
			@Override
			public void run()
			{
				long time = System.currentTimeMillis();
				//initialization
				
				Iterable<? extends ExecutableQuery<?,?>> round = BasicScheduler.this.sequencer.order(BasicScheduler.this.executableProgram);
				try{
					//repeat fixed round until execution is stopped
					while(BasicScheduler.this.engine.isRunning()){
						try{
							BasicScheduler.this.rounds.writeNext(round);
						}
						catch (IllegalStateException e) {
							break;
						}
						long newTime = System.currentTimeMillis();
						long waitTime = time + BasicScheduler.this.delayInMillies - newTime;
						time = newTime;
						if (waitTime > 0){
							try
							{
								Thread.sleep(waitTime);
							}
							catch (InterruptedException e)
							{
								// TODO better catch block
								e.printStackTrace();
							}
						}
					}
				}
				catch (final Throwable e) {
					BasicScheduler.this.error("Uncatched Exception:", e);
					new Thread(BasicScheduler.this.getName() + " Uncatched Exception Handler"){
						
						@Override
						public void run() {
							BasicScheduler.this.engine.close(e);
						};
						
					}.start();					
				}
				finally{
					BasicScheduler.this.rounds.close();
				}
			}
		};
	}
	
	@Override
	void close() {
		this.rounds.close();
	}
	
	@Override
	public void run()
	{
		try{
			this.thread.start();
			while(this.thread.isAlive()){
				try
				{
					this.thread.join();
				}
				catch (InterruptedException e)
				{
					
				}
			}
		}
		catch (Throwable e) {
			this.error("Uncatched Exception:", e);
			this.engine.close(e);					
		}
		finally{
			BasicScheduler.this.rounds.close();
		}
	}

	@Override
	Iterable<Iterable<? extends ExecutableQuery<?, ?>>> getRounds()
	{
		return this.rounds;
	}

}
