package de.lmu.ifi.pms.cep.event_mill.execution.scheduler;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.Marker;

import de.lmu.ifi.pms.cep.event_mill.execution.engine.Engine;
import de.lmu.ifi.pms.cep.event_mill.execution.program.ExecutableQuery;


public class QueryThread extends Thread {

	final Engine engine;
	final SchedulerThread scheduler;
	final Iterable<Iterable<? extends ExecutableQuery<?,?>>> rounds;
	
	//logging
	protected static final Logger logger = LoggerFactory.getLogger(QueryThread.class);	
	public final Marker marker;
	
	public QueryThread(final SchedulerThread scheduler)
	{
		super(scheduler.engine.marker.getName() + ": QueryThread");
		this.engine = scheduler.engine;
		this.scheduler = scheduler;
		this.rounds = this.scheduler.getRounds();
		
		this.marker = this.engine.marker;
	}
	
	@Override
	public void run()
	{
		try{
		
			this.info("Started");
			//TODO reporting; e.g. log-thread writing to the data base
			for(Iterable<? extends ExecutableQuery<?,?>> round: this.rounds)
			{
				this.info("Started Round");
				for(ExecutableQuery<?,?> q: round)
				{
					q.execute();
				}
				this.info("Finished Round");
			}
			this.info("Terminated");
		
		} catch (final Throwable e) {
			this.error("Uncatched Exception:", e);
			new Thread(QueryThread.this.getName() + " Uncatched Exception Handler"){
				
				@Override
				public void run() {
					QueryThread.this.scheduler.close();
					QueryThread.this.engine.close(e);
				};
				
			}.start();
		}
	}
	
	protected Logger logger(){
		return logger;
	}

	protected final void trace(final String msg) {
		this.logger().trace(this.marker, msg);
	}

	protected final void trace(final String format, final Object arg) {
		this.logger().trace(this.marker, format, arg);
	}

	protected final void trace(final String format, final Object arg1, final Object arg2) {
		this.logger().trace(this.marker, format, arg1, arg2);
	}

	protected final void trace(final String format, final Object[] argArray) {
		this.logger().trace(this.marker, format, argArray);
	}

	protected final void trace(final String msg, final Throwable t) {
		this.logger().trace(this.marker, msg, t);
	}

	protected final void debug(final String msg) {
		this.logger().debug(this.marker, msg);
	}

	protected final void debug(final String format, final Object arg) {
		this.logger().debug(this.marker, format, arg);
	}

	protected final void debug(final String format, final Object arg1, final Object arg2) {
		this.logger().debug(this.marker, format, arg1, arg2);
	}

	protected final void debug(final String format, final Object[] argArray) {
		this.logger().debug(this.marker, format, argArray);
	}

	protected final void debug(final String msg, final Throwable t) {
		this.logger().debug(this.marker, msg, t);
	}

	protected final void info(final String msg) {
		this.logger().info(this.marker, msg);
	}

	protected final void info(final String format, final Object arg) {
		this.logger().info(this.marker, format, arg);
	}

	protected final void info(final String format, final Object arg1, final Object arg2) {
		this.logger().info(this.marker, format, arg1, arg2);
	}

	protected final void info(final String format, final Object[] argArray) {
		this.logger().info(this.marker, format, argArray);
	}

	protected final void info(final String msg, final Throwable t) {
		this.logger().info(this.marker, msg, t);
	}

	protected final void warn(final String msg) {
		this.logger().warn(this.marker, msg);
	}

	protected final void warn(final String format, final Object arg) {
		this.logger().warn(this.marker, format, arg);
	}

	protected final void warn(final String format, final Object arg1, final Object arg2) {
		this.logger().warn(this.marker, format, arg1, arg2);
	}

	protected final void warn(final String format, final Object[] argArray) {
		this.logger().warn(this.marker, format, argArray);
	}

	protected final void warn(final String msg, final Throwable t) {
		this.logger().warn(this.marker, msg, t);
	}

	protected final void error(final String msg) {
		this.logger().error(this.marker, msg);
	}

	protected final void error(final String format, final Object arg) {
		this.logger().error(this.marker, format, arg);
	}

	protected final void error(final String format, final Object arg1, final Object arg2) {
		this.logger().error(this.marker, format, arg1, arg2);
	}

	protected final void error(final String format, final Object[] argArray) {
		this.logger().error(this.marker, format, argArray);
	}

	protected final void error(final String msg, final Throwable t) {
		this.logger().error(this.marker, msg, t);
	}

}
