package de.lmu.ifi.pms.cep.event_mill.io;

import java.util.concurrent.ConcurrentLinkedQueue;


public abstract class BufferedMessageSource<MessageType> extends MessageSource<MessageType>{
	
	final ConcurrentLinkedQueue<java.util.ArrayList<MessageType>> blockBufferPool = new ConcurrentLinkedQueue<java.util.ArrayList<MessageType>>();
	
	public BufferedMessageSource(final Class<MessageType> type){
		super(type);
	}
	
	public abstract boolean nextBlock();

	public abstract boolean next();
	
	public abstract MessageType current();
	
	@Override
	public MessageType read() {
		if(!this.next()){
			this.nextBlock();
			this.next();
		}
		return this.current();
	}
	
	//FIXME instead of block maybe pucket principle. get Bucket -> use content of bucket -> return bucket -> get filled bucket
	
	@Override
	public MessageType[] readBlock() {
		boolean hasNext = this.next();
		if(!hasNext){
			this.nextBlock();
			hasNext = this.next();
		}
		if(hasNext){
			int remainingBlockCapacity = 1024; //FIXME
			java.util.ArrayList<MessageType> block = this.blockBufferPool.peek();
			if(block == null){
				block =	new java.util.ArrayList<MessageType>(1024);
			}
			do{
				--remainingBlockCapacity;
				block.add(this.current());
				if(remainingBlockCapacity == 0){
					break;
				}
			} while(this.next());
			MessageType[] blockArray = block.toArray(this.newBlock(block.size()));
			
			block.clear();
			this.blockBufferPool.add(block);
			
			return blockArray;
		}
		else{
			return this.emptyBlock;
		}
	}
	
}
