package de.lmu.ifi.pms.cep.event_mill.io.bus;

import java.util.Arrays;

import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.ElementListUnion;

import de.lmu.ifi.pms.java.util.collection.Collection;


public class CompositeMessageBusProviderCollection extends MessageBusProviderCollection {
	
	protected final MessageBusProviderCollection[] messageBusProviders;
	
	@ElementListUnion({
		@ElementList(entry="composite", inline=true, type=CompositeMessageBusProviderCollection.class, required=false, empty=false),
		@ElementList(entry="explicit", inline=true, type=ExplicitMessageBusProviderCollection.class, required=false, empty=false),
		@ElementList(entry="heap-broadcast", inline=true, type=HeapBroadcastMessageBusProviderCollection.class, required=false, empty=false),
		@ElementList(entry="buses", type=MessageBusProviderCollection.class, inline=true, required=false, empty=false)
	})
	protected java.util.ArrayList<MessageBusProviderCollection> getBuses(){
		if(this.messageBusProviders != null){
			return new java.util.ArrayList<MessageBusProviderCollection>(java.util.Arrays.asList(this.messageBusProviders));
		}
		else{
			return null;
		}
	}
	
	public CompositeMessageBusProviderCollection() {
		super();
		this.messageBusProviders = new MessageBusProviderCollection[0];
	}
	
	public CompositeMessageBusProviderCollection(final MessageBusProviderCollection[] messageBusProviders) {
		super();
		this.messageBusProviders = Arrays.copyOf(messageBusProviders, messageBusProviders.length);
	}
	
	public CompositeMessageBusProviderCollection(final Collection<MessageBusProviderCollection> messageBusProviders) {
		this(messageBusProviders.javaCollection());
	}
	
	public CompositeMessageBusProviderCollection(
			@ElementList(entry="buses", inline=true, type=MessageBusProvider.class, required=false, empty=false)
			final java.util.ArrayList<MessageBusProviderCollection> messageBusProviders) {
		this((java.util.Collection<MessageBusProviderCollection>) messageBusProviders);
	}
	
	public CompositeMessageBusProviderCollection(final java.util.Collection<MessageBusProviderCollection> messageBusProviders) {
		super();
		if(messageBusProviders == null){
			this.messageBusProviders = new MessageBusProviderCollection[0];
		}
		else{
			this.messageBusProviders = messageBusProviders.toArray(new MessageBusProviderCollection[messageBusProviders.size()]);
		}
	}

	@Override
	public boolean contains(final String identifier) {
		for(MessageBusProviderCollection provider: this.messageBusProviders){
			if(provider.contains(identifier)){
				return true;
			}
		}
		return false;
	}

	@Override
	public MessageBusProvider create(final String identity) {
		for(MessageBusProviderCollection provider: this.messageBusProviders){
			if(provider.contains(identity)){
				return provider.create(identity);
			}
		}
		throw new IllegalArgumentException();
	}

}
