package de.lmu.ifi.pms.cep.event_mill.io.bus;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.event_mill.io.MessageAdapter;
import de.lmu.ifi.pms.cep.event_mill.io.MessageSink;
import de.lmu.ifi.pms.cep.event_mill.io.MessageSource;
import de.lmu.ifi.pms.cep.event_mill.io.catalog.DataCatalog;
import de.lmu.ifi.pms.cep.event_mill.io.format.MessageFormatProvider;

public class WrapperMessageBus<MessageType, BusMessageType> extends MessageBus<MessageType>{
	
	protected final DataCatalog dataCatalog;
	
	protected final MessageFormatProvider<MessageType> formatter;//needs bijective formatter
	protected final MessageBus<BusMessageType> bus;
	protected final MessageFormatProvider<BusMessageType> busFormatter;
	
	protected final MessageAdapter<BusMessageType, MessageType> readAdapter;
	protected final MessageAdapter<MessageType, BusMessageType> writeAdapter;
	
	public WrapperMessageBus(final WrapperMessageBusProvider<MessageType> busProvider, final MessageBus<BusMessageType> bus) {
		this(busProvider, busProvider.formatter, busProvider.dataCatalog.create(busProvider.identifier), bus);
	}
	
	public WrapperMessageBus(final MessageFormatProvider<MessageType> formatter, final DataCatalog dataCatalog, final MessageBus<BusMessageType> bus) {
		this(bus.providerReference, formatter, dataCatalog, bus);
	}
	
	protected WrapperMessageBus(final MessageBusProvider provider, final MessageFormatProvider<MessageType> formatter, final DataCatalog dataCatalog, final MessageBus<BusMessageType> bus) {
		super(formatter.type, provider);
		this.dataCatalog = dataCatalog;
		this.formatter = formatter;
		this.bus = bus;
		this.busFormatter = this.bus.formatter();
		
		this.readAdapter = new MessageAdapter<BusMessageType,MessageType>(this.dataCatalog, bus.formatter(), this.formatter);
		this.writeAdapter = new MessageAdapter<MessageType,BusMessageType>(this.dataCatalog, this.formatter, bus.formatter());
	}

	@Override
	public MessageFormatProvider<MessageType> formatter() {
		return this.formatter;
	}

	@Override
	public MessageSink<MessageType> sink() {
		return new WrapperMessageSink<MessageType, BusMessageType>(this.writeAdapter, this.bus.sink());
	}

	@Override
	public MessageSource<MessageType> source() {
		return new WrapperMessageSource<BusMessageType, MessageType>(this.readAdapter, this.bus.source());
	}
	
	protected static final Logger logger = LoggerFactory.getLogger(WrapperMessageBus.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}

}
