package de.lmu.ifi.pms.cep.event_mill.io.catalog;

import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementUnion;

import de.lmu.ifi.pms.cep.event_mill.io.FixedDataCatalog;
import de.lmu.ifi.pms.cep.event_mill.io.PATTERNS;
import de.lmu.ifi.pms.cep.event_mill.io.stream.EmptyStreamIOProvider;
import de.lmu.ifi.pms.cep.event_mill.io.stream.StreamIOProvider;
import de.lmu.ifi.pms.cep.event_mill.io.stream.StreamIOProviderCollection;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;

public class StreamIODataCatalogProvider extends DataCatalogProvider {
	
	@ElementUnion({
		@Element(name="empty-stream", type=EmptyStreamIOProvider.class),
		@Element(name="composite-stream", type=StreamIOProviderCollection.class),
		@Element(name="stream", type=StreamIOProvider.class)
	})
	final StreamIOProvider stream;

	public StreamIODataCatalogProvider(@Element(name="stream") final StreamIOProvider stream){
		super();
		this.stream = stream;
	}
	
	protected String programIdentifier(final String busIdentifier){
		String program_instance_full_name = PATTERNS.busIdentifier(busIdentifier)[1];
		return PATTERNS.programIdentifier(this.stream.identifier, program_instance_full_name);
	}
	
	protected DTSA_BufferType type(final String identifier){
		String type  = PATTERNS.busIdentifier(identifier)[2];
		return DTSA_BufferType.valueOf(type.toUpperCase());
	}
	
	@Override
	public boolean accepts(final String identifier) {
		return identifier.matches(PATTERNS.programBusIdentifierRegEx);
	}

	@Override
	public DataCatalog create(final String identifier) {
		return new FixedDataCatalog(this.stream.connect(this.programIdentifier(identifier)), this.type(identifier));
	}

}
