package de.lmu.ifi.pms.cep.event_mill.io.connect;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.event_mill.io.MessageSink;
import de.lmu.ifi.pms.cep.event_mill.io.SpecificBufferedMessageSource;
import de.lmu.ifi.pms.cep.event_mill.io.bus.MessageBus;
import de.lmu.ifi.pms.cep.event_mill.io.format.MessageFormatProvider;
import de.lmu.ifi.pms.cep.event_mill.io.stream.SourceStream;
import de.lmu.ifi.pms.cep.event_mill.io.stream.StreamIO;

import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

public class StreamToBusConnector<MessageType> extends Connector{
	
	public final Class<MessageType> type;
	final MessageFormatProvider<MessageType> formatter;
	final MessageSink<MessageType> sink;
	final FixedArrayNavigableMap<String, SpecificBufferedMessageSource<MessageType>> messageSources;
	
	public StreamToBusConnector(final MessageBus<MessageType> messageBus, final StreamIO io) {
		this(messageBus.formatter(), messageBus.sink(), io.sourceStreams());
	}

	public StreamToBusConnector(final MessageFormatProvider<MessageType> formatter, final MessageSink<MessageType> sink, final SourceStream... sources) {
		super();
		this.type = sink.type;
		this.formatter = formatter;
		this.sink = sink;
		String[] messageTypes = new String[sources.length];
		@SuppressWarnings("unchecked")
		SpecificBufferedMessageSource<MessageType>[] messageSources = (SpecificBufferedMessageSource<MessageType>[]) new SpecificBufferedMessageSource<?>[sources.length];
		for(int messageType = 0; messageType < messageTypes.length; ++messageType){
			messageTypes[messageType] = sources[messageType].name;
			messageSources[messageType] = this.formatter.messageSource(sources[messageType]);
		}
		this.messageSources = new FixedArrayNavigableMap<String, SpecificBufferedMessageSource<MessageType>>(messageTypes, messageSources);
	}
	
	@Override
	public void nextBlock(){
		for(SpecificBufferedMessageSource<MessageType> messageSource: this.messageSources.values()){
			if(messageSource.nextBlock()){
				this.sink.writeBlock(messageSource.readBlock());
			}
		}
	}
	
	
	protected static final Logger logger = LoggerFactory.getLogger(BusToStreamConnector.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}
	
	public static <MessageType> StreamToBusConnector<MessageType> create(final MessageBus<MessageType> messageBus, final StreamIO io){
		return new StreamToBusConnector<MessageType>(messageBus, io);
	}
	
	@Override
	public void close(){
		this.sink.close();
		for(SpecificBufferedMessageSource<MessageType> source: this.messageSources.values()){
			source.close();
		}
	}
	
	@Override
	protected void finalize() throws Throwable {
		this.close();
	}

}
