package de.lmu.ifi.pms.cep.event_mill.io.format;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Root;

import de.lmu.ifi.pms.cep.event_mill.io.BufferedMessageSink;
import de.lmu.ifi.pms.cep.event_mill.io.SpecificBufferedMessageSource;
import de.lmu.ifi.pms.cep.event_mill.io.stream.MessageSourceStream;
import de.lmu.ifi.pms.cep.event_mill.io.stream.SinkStream;
import de.lmu.ifi.pms.cep.event_mill.io.stream.SourceStream;
import de.lmu.ifi.pms.cep.event_mill.io.stream.StreamMessageSink;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;

import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;

@Root(name="format")
public abstract class MessageFormatProvider<MessageType> {
	
	public final Class<MessageType> type;

	@Attribute(name="iri", required=true)
	public final String identifier;

	public MessageFormatProvider(final Class<MessageType> type, final String identifier) {
		super();
		this.type = type;
		this.identifier = identifier;
	}

	public abstract MessageGenerator<MessageType> messageGenerator(String messageType, final FixedArrayNavigableMap<String, BasicTypeID> types);
	
	public abstract MessageReader<MessageType> messageReader(String messageType, final FixedArrayNavigableMap<String, BasicTypeID> types);
	
	public abstract String messageType(MessageType message);
	
	public BufferedMessageSink<MessageType> messageSink(final SinkStream sinkStream){
		return new StreamMessageSink<MessageType>(sinkStream, this);
	}
	
	public SpecificBufferedMessageSource<MessageType> messageSource(final SourceStream sourceStream){
		return new MessageSourceStream<MessageType>(sourceStream, this);
	}
	
}
