package de.lmu.ifi.pms.cep.event_mill.io.format;

import org.slf4j.Logger;

import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.java.LoggingObject;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


public abstract class ParameterObject extends LoggingObject{
	
	public final FixedArray<String> parameters;
	public final FixedArrayNavigableSet<String> sortedParameters;
	protected final BasicTypeID[] types;
	final int[] positions;
	
	public ParameterObject(final FixedArrayNavigableMap<String, BasicTypeID> types){
		this.sortedParameters = types.keys();
		this.parameters = this.sortedParameters;
		this.types = types.values().toArray(new BasicTypeID[this.parameters.size()]);
		this.positions = new int[this.types.length];
		for(int parameter = 0; parameter < this.positions.length; ++parameter){
			this.positions[parameter] = parameter + 1;
		}
	}
	
	public ParameterObject(final FixedArray<String> parameters, final FixedArray<BasicTypeID> types) {
		super();
		if(parameters.size() != types.size()){
			throw new IllegalArgumentException("Number of parameters and number of specified types do not match.");
		}
		this.parameters = parameters;
		this.types = types.toArray(new BasicTypeID[this.parameters.size()]);
		String[] sortedParameters = parameters.toArray(new String[parameters.size()]);
		this.positions = new int[parameters.size()];
		{
			int[] parameterNewPositions = new int[parameters.size()];
			int sortedSize = de.lmu.ifi.pms.java.util.collection.array.Arrays.sort(sortedParameters, parameterNewPositions, false);
			if(sortedSize != parameters.size()){
				throw new IllegalArgumentException("Dublicate parameter!");
			}
			for(int parameter = 0; parameter < this.positions.length; ++parameter){
				this.positions[parameterNewPositions[parameter]] = parameter + 1;
				assert this.positions[parameterNewPositions[parameter]] >= 1;
			}
		}
		
		this.sortedParameters = new FixedArrayNavigableSet<String>(false, false, sortedParameters);		
	}
	
	public int indexOf(final String parameter){
		int index = this.sortedParameters.indexOf(parameter);
		if(index < 0){
			return -1;
		}
		else{
			assert this.positions[index] >= 1;
			return this.positions[index];
		}
	}

	
//	public boolean isFirst() {
//
//	}
//
//	
//	public boolean isLast() {
//
//	}

	
//	public void setFetchSize(final int rows) {
//
//	}
//
//	
//	public int getFetchSize() {
//
//	}
	
	@Override
	protected abstract Logger logger();

}