package de.lmu.ifi.pms.cep.event_mill.io.stream;

import de.lmu.ifi.pms.cep.event_mill.io.SpecificBufferedMessageSink;
import de.lmu.ifi.pms.cep.event_mill.io.format.MessageFormatProvider;
import de.lmu.ifi.pms.cep.event_mill.io.format.MessageReader;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;

public class StreamMessageSink<MessageType> extends SpecificBufferedMessageSink<MessageType>{
	
	final MessageReader<MessageType> messageReader;
	final SinkStream sinkStream;
	public final int numberOfAttributes;
	final BasicTypeID[] types;
	
	public StreamMessageSink(final String name, final StreamIO io, final MessageFormatProvider<MessageType> formatProvider){
		this(io.sinkStream(name), formatProvider);
	}
	
	public StreamMessageSink(final String name, final StreamIO io, final MessageReader<MessageType> messageReader){
		this(io.sinkStream(name), messageReader);
	}
	
	public StreamMessageSink(final SinkStream sinkStream, final MessageFormatProvider<MessageType> formatProvider){
		this(sinkStream, formatProvider.messageReader(sinkStream.name, sinkStream.types));
	}
	
	public StreamMessageSink(final SinkStream sinkStream, final MessageReader<MessageType> messageReader){
		super(messageReader.type, sinkStream.name);
		this.sinkStream = sinkStream;
		this.messageReader = messageReader;//FIXME remove attributes with "." at the beginning and void setting attributes that are not used in the pattern
		this.numberOfAttributes = sinkStream.types.size();
		this.types = sinkStream.types.values().toArray(new BasicTypeID[this.numberOfAttributes]);
	}
	
	
	@Override
	public void add(final MessageType message){
    	BasicTypeID attributeType;
    	this.messageReader.decompose(message);
    	for (int attribute = 1; attribute <= this.numberOfAttributes; ++attribute){
    		attributeType = this.types[attribute - 1];
    		switch (attributeType) {
			case BOOLEAN:
				this.sinkStream.setBoolean(attribute, this.messageReader.getBoolean(attribute));
				break;
			case STRING:
				this.sinkStream.setString(attribute, this.messageReader.getString(attribute));
				break;
			case INT:
				this.sinkStream.setInt(attribute, this.messageReader.getInt(attribute));
				break;
			case LONG:
				this.sinkStream.setLong(attribute, this.messageReader.getLong(attribute));
				break;				
			case FLOAT:
				this.sinkStream.setFloat(attribute, this.messageReader.getFloat(attribute));
				break;
			case DOUBLE:
				this.sinkStream.setDouble(attribute, this.messageReader.getDouble(attribute));
				break;
			case TIMESTAMP:
				this.sinkStream.setTimestamp(attribute, this.messageReader.getTimestamp(attribute));
				break;				
			case DURATION:
				this.sinkStream.setDuration(attribute, this.messageReader.getDuration(attribute));
				break;				
			case IDENTIFIER:
				this.sinkStream.setIdentifier(attribute, this.messageReader.getIdentifier(attribute));
				break;
			case REFERENCE: throw new IllegalArgumentException("Type \"reference\" not supported.");
			default:
				break;
			}
    	}
    	this.messageReader.clearMessage();
	}
	
	@Override
	public boolean next() {
		return this.sinkStream.next();
	}
	
	@Override
	public boolean nextBlock() {
		return this.sinkStream.nextBlock();
	}
	
	@Override
	public void close() {
		this.sinkStream.close();
	}
	
}
