package de.lmu.ifi.pms.cep.event_mill.sql.compiler;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.compilation.Compiler;
import de.lmu.ifi.pms.cep.event_mill.sql.SQL_Program_Spec;
import de.lmu.ifi.pms.cep.tsa.DTSA_Program;
import de.lmu.ifi.pms.cep.tsa.optimization.DTSA_Optimizer;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;


public class DTSA_SQL_Compiler extends Compiler<DTSA_Program, SQL_Program_Spec>{
	
	public final FixedArray<DTSA_Optimizer> preWriteOptimizers;
	
	public final DTSA_SQL_Writer dtsa_SQL_Writer;
	
	public DTSA_SQL_Compiler(final DTSA_SQL_Writer dtsa_SQL_Writer) {
		this(new FixedArray<DTSA_Optimizer>(), dtsa_SQL_Writer);
	}
	
	public DTSA_SQL_Compiler(final DTSA_Optimizer[] preWriteOptimizers, final DTSA_SQL_Writer dtsa_SQL_Writer) {
		this(new FixedArray<DTSA_Optimizer>(preWriteOptimizers), dtsa_SQL_Writer);
	}
	
	public DTSA_SQL_Compiler(final FixedArray<DTSA_Optimizer> preWriteOptimizers, final DTSA_SQL_Writer dtsa_SQL_Writer) {
		super(DTSA_Program.class, SQL_Program_Spec.class);
		this.preWriteOptimizers = preWriteOptimizers;
		this.dtsa_SQL_Writer = dtsa_SQL_Writer;
	}

	@Override
	public SQL_Program_Spec compile(DTSA_Program dtsa_program) {
		for(DTSA_Optimizer optimizer: this.preWriteOptimizers){
			dtsa_program = optimizer.optimize(dtsa_program);
		}
		return this.dtsa_SQL_Writer.compile(dtsa_program);
	}
	
	protected static final Logger logger = LoggerFactory.getLogger(DTSA_SQL_Compiler.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}
		
}
