package de.lmu.ifi.pms.cep.event_mill.sql.io;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.event_mill.io.stream.SourceStream;
import de.lmu.ifi.pms.cep.event_mill.sql.SQL_CONSTANTS;
import de.lmu.ifi.pms.cep.event_mill.sql.SQL_EXCEPTIONS;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;

public class SQL_SourceStream extends SourceStream{
	
	//FIXME buffer ResultSet
	
	final SQL_EngineIO io;
	final String readMaxIdSql;
	PreparedStatement readMaxId;
	final String readSql;
	PreparedStatement read;
	ResultSet maxId;
	ResultSet currentResults;
	long lastReadId;
	long currentReadId;
	int delay = 1000;//FIXME
	long earliestNextResultSet = Long.MIN_VALUE;
	
	public SQL_SourceStream(final String bufferName, final SQL_EngineIO io) {
		this(bufferName, SQL_CONSTANTS.BIGINT.MIN_VALUE, io);
	}
	
	public SQL_SourceStream(final String buffer_name, final long lastReadId, final SQL_EngineIO io) {
		super(buffer_name, io.attributeTypes(DTSA_BufferType.OUTPUT, buffer_name));
		this.io = io;
		this.readMaxIdSql = io._o_Id_MaxReadValueSQL(buffer_name);
		this.readSql = io.outputSQL(buffer_name);
	}
	
	void reconnect(){
		try{
			this.io.reconnect(this.read.getConnection());
		} catch (SQLException e) {
			this.error("", e);
			throw new RuntimeException(e);
		}
	}
	
	protected void initStatements(){
		while(true){
			try{
				this.readMaxId = this.io.prepareStatement(this.readMaxIdSql);
				this.read = this.io.prepareStatement(this.readSql);
				break;
			} catch (SQLException e) {
				if(SQL_EXCEPTIONS.isConnectionFailure(e)){
					this.debug("Connection failure. Trying again.", e);
					this.reconnect();
					continue;
				}
				else{
					this.error("", e);
					throw new RuntimeException(e);
				}
			}
		}
	}
	
	@Override
	public boolean nextBlock(){
		if(this.read == null){
			this.initStatements();
		}
		if(this.isOpen){		
			while(true){
				try {
					this.maxId = this.readMaxId.executeQuery();
					boolean returnedValue = this.maxId.next();
					assert returnedValue;
					this.currentReadId = this.maxId.getLong(1);
					try{this.maxId.close();} catch(SQLException e){};
					this.maxId = null;
					break;
				} catch (SQLException e) {
					if(SQL_EXCEPTIONS.isConnectionFailure(e)){
						this.debug("Connection failure. Trying again.", e);
						this.reconnect();
						this.initStatements();
						continue;
					}
					else if(SQL_EXCEPTIONS.isRollback(e)){
						this.debug("Statement rolled back. Trying again.", e);
					}
					else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
						this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
						this.initStatements();
					}
					else{
						throw new RuntimeException(e);
					}
				}
			}
	
			while(true){
				try {
					this.read.setLong(1, this.lastReadId);
					this.read.setLong(2, this.currentReadId);
					
					this.currentResults = this.read.executeQuery();
	
					this.lastReadId = this.currentReadId;//FIXME set at beginning of method, i.e. at next call to method?
					break;
				} catch (SQLException e) {
					if(SQL_EXCEPTIONS.isConnectionFailure(e)){
						this.debug("Connection failure. Trying again.", e);
						this.reconnect();//FIXME current connection as parameter, otherwise cicular reconnect over several result streams
						this.initStatements();
						continue;
					}
					else if(SQL_EXCEPTIONS.isRollback(e)){
						this.debug("Statement rolled back. Trying again.", e);
					}
					else if(SQL_EXCEPTIONS.isPreparedStatementUnknownHandleFailure(e)){
						this.debug("Handle of set state statement is invalid. Repreparing statement and trying again.", e);
						this.initStatements();
					}
					else{
						throw new RuntimeException(e);
					}
				}
			}
			return true;
		}
		else{
			return false;
		}
	}
	
	@Override
	public boolean next() {
		try {
			return this.isOpen && this.currentResults != null && this.currentResults.next();
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
	}

	
	@Override
	public void close() {
		super.close();
		try{ if(this.currentResults != null){ this.currentResults.close(); } } catch (SQLException e){}
		try{ if(this.readMaxId != null){ this.readMaxId.close(); } } catch (SQLException e){}
		try{ if(this.read != null){ this.read.close(); } } catch (SQLException e){}
	}
	
	@Override
	public String getString(final int columnIndex) {
		try {
			return this.currentResults.getString(columnIndex);
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
	}

	
	@Override
	public boolean getBoolean(final int columnIndex) {
		try {
			return this.currentResults.getBoolean(columnIndex);
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
	}

	
	@Override
	public byte getByte(final int columnIndex) {
		try {
			return this.currentResults.getByte(columnIndex);
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
	}

	
	@Override
	public short getShort(final int columnIndex) {
		try {
			return this.currentResults.getShort(columnIndex);
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
	}

	
	@Override
	public int getInt(final int columnIndex) {
		try {
			return this.currentResults.getInt(columnIndex);
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
	}

	
	@Override
	public long getLong(final int columnIndex) {
		try {
			return this.currentResults.getLong(columnIndex);
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
	}
	
	@Override
	public long getTimestamp(final int columnIndex) {
		try {
			return this.currentResults.getLong(columnIndex);
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
	}
	
	@Override
	public long getDuration(final int columnIndex) {
		try {
			return this.currentResults.getLong(columnIndex);
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
	}
	
	@Override
	public long getIdentifier(final int columnIndex) {
		try {
			return this.currentResults.getLong(columnIndex);
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
	}

	
	@Override
	public float getFloat(final int columnIndex) {
		try {
			return this.currentResults.getFloat(columnIndex);
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
	}

	
	@Override
	public double getDouble(final int columnIndex) {
		try {
			return this.currentResults.getDouble(columnIndex);
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
	}

	
	@Override
	public String getString(final String columnLabel) {
		try {
			return this.currentResults.getString(this.findColumn(columnLabel));
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
	}

	
	@Override
	public boolean getBoolean(final String columnLabel) {
		try {
			return this.currentResults.getBoolean(this.findColumn(columnLabel));
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
	}

	
	@Override
	public byte getByte(final String columnLabel) {
		try {
			return this.currentResults.getByte(this.findColumn(columnLabel));
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
	}

	
	@Override
	public short getShort(final String columnLabel) {
		try {
			return this.currentResults.getShort(this.findColumn(columnLabel));
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
	}

	
	@Override
	public int getInt(final String columnLabel) {
		try {
			return this.currentResults.getInt(this.findColumn(columnLabel));
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
	}

	
	@Override
	public long getLong(final String columnLabel) {
		try {
			return this.currentResults.getLong(this.findColumn(columnLabel));
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
	}
	
	@Override
	public long getTimestamp(final String columnLabel) {
		try {
			return this.currentResults.getLong(this.findColumn(columnLabel));
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
	}
	
	@Override
	public long getDuration(final String columnLabel) {
		try {
			return this.currentResults.getLong(this.findColumn(columnLabel));
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
	}
	
	@Override
	public long getIdentifier(final String columnLabel) {
		try {
			return this.currentResults.getLong(this.findColumn(columnLabel));
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
	}

	
	@Override
	public float getFloat(final String columnLabel) {
		try {
			return this.currentResults.getFloat(this.findColumn(columnLabel));
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
	}

	
	@Override
	public double getDouble(final String columnLabel) {
		try {
			return this.currentResults.getDouble(this.findColumn(columnLabel));
		} catch (SQLException e) {
			throw new RuntimeException(e);
		}
	}
	
	public boolean isBeforeFirst(){
		return this.lastReadId == SQL_CONSTANTS.BIGINT.MIN_VALUE;
	}

	
//	public boolean isFirst() {
//
//	}
//
//	
//	public boolean isLast() {
//
//	}

	
//	public void setFetchSize(final int rows) {
//
//	}
//
//	
//	public int getFetchSize() {
//
//	}

	
	public int getType(){
		return ResultSet.TYPE_FORWARD_ONLY;
	}
	
	protected static final Logger logger = LoggerFactory.getLogger(SQL_SourceStream.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}

}