package de.lmu.ifi.pms.cep.tsa;

import org.simpleframework.xml.Root;

import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.collection.fixed.FixedCollection;


@Root(name="expression")
public abstract class DTSA_Expression {
	
	public final boolean isStatic;
	public final boolean isMonotonic;
	
	public final FixedCollection<String> origin;
	
	public DTSA_Expression(final boolean isStatic, final boolean isMonotonic, final FixedCollection<String> origin) {
		super();
		this.isStatic = isStatic;
		this.isMonotonic = isMonotonic;
		this.origin = origin;
	}

	public abstract TemporalStreamSchema schema();
	
	public abstract DTSA_Expression replaceSchema(TemporalStreamSchema newSchema) throws TSA_Exception;
	
	public FixedArray<DTSA_StreamBuffer> inputs(){
		ArrayList<DTSA_StreamBuffer> inputs = new ArrayList<DTSA_StreamBuffer>();
		this.inputs(inputs);
		return new FixedArray<DTSA_StreamBuffer>(false, inputs.toArray(new DTSA_StreamBuffer[inputs.size()]));
	}
	
	protected abstract void inputs(final ArrayList<DTSA_StreamBuffer> inputs);
	
	public boolean isValid(){
		try{
			this.check();
			//TODO maybe more efficient implementation as isValid does not need to collect exceptions
			return true;
		}
		catch(TSA_Exception e){
			return false;
		}
	}
	
	public abstract void check() throws TSA_Exception;
	
	public abstract void writeTo(final StringBuilder sb, final int indentDepthFirstLine, final int indentDepth);
	
	public void writeTo(final StringBuilder sb, final int indentDepth){
		this.writeTo(sb, indentDepth, indentDepth);
	}
	
	public void writeTo(final StringBuilder sb){
		this.writeTo(sb, 0);
	}
	
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		this.writeTo(sb);
		return sb.toString();
	}

}
