 package de.lmu.ifi.pms.cep.tsa;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.ElementMap;

import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.fixed.FixedCollection;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


public class DTSA_Query
{
	@Attribute(name="package", required=true)
	public final String prog_package;
	@Attribute(name="type", required=true)
	public final DTSA_QueryType type;
	
	
	public final DTSA_BufferType input_type;
	public final FixedArrayNavigableMap<String, DTSA_StreamBuffer> inputs;
	//public final FixedArrayNavigableSet<String> inputStreamBounds;
	//final long[][][] streamBoundDistances;
	
	@Element(name="output", required=true)
	public final DTSA_StreamBuffer output;
	@Element(name="expression", required=true)
	public final DTSA_Expression expression;
	
	public final boolean isStatic;	
	public final boolean isMonotonic;
	
	
	public final FixedArrayNavigableMap<String, Term<BasicType>> streamBoundFunctions;
	
	@ElementMap(entry="stream-bound-function", inline=true, key="stream-bound-identifier", attribute=true, required=false)
	final java.util.HashMap<String, Term<BasicType>> getStreamBoundFunctions(){
		return new java.util.HashMap<String, Term<BasicType>>(this.streamBoundFunctions.javaMap());
	}
	
	public final FixedArrayNavigableMap<String, Term<BasicType>> streamBoundClockIncreaseFunctions;
	
	@ElementMap(entry="stream-bound-clock-increase-function", inline=true, key="stream-bound-identifier", attribute=true, required=false)
	final java.util.HashMap<String, Term<BasicType>> getStreamBoundClockIncreaseFunctions(){
		return new java.util.HashMap<String, Term<BasicType>>(this.streamBoundClockIncreaseFunctions.javaMap());
	}

	public final FixedArrayNavigableMap<String, Term<BasicType>> relevanceBoundFunctions;
	
	@ElementMap(entry="relevance-bound-function", inline=true, key="stream-bound-identifier", attribute=true, required=false)
	final java.util.HashMap<String, Term<BasicType>> getRelevanceBoundFunctions(){
		return new java.util.HashMap<String, Term<BasicType>>(this.relevanceBoundFunctions.javaMap());
	}
	
	public final FixedArrayNavigableMap<String, Term<BasicType>> relevanceBoundClockIncreaseFunctions;
	
	@ElementMap(entry="relevance-bound-clock-increase-function", inline=true, key="stream-bound-identifier", attribute=true, required=false)
	final java.util.HashMap<String, Term<BasicType>> getRelevanceBoundClockIncreaseFunctions(){
		return new java.util.HashMap<String, Term<BasicType>>(this.relevanceBoundClockIncreaseFunctions.javaMap());
	}
	
	@Element(name="description", required=false)
	public final String description;
	
	public final FixedCollection<String> origin;
	
	@ElementList(entry="origin", inline=true, required=false)
	java.util.ArrayList<String> getOriginList(){
		return new java.util.ArrayList<String>(this.origin.javaCollection());
	}
	
	public DTSA_Query(@Attribute(name="package", required=true) final String prog_package
			, @Element(name="expression", required=true) final DTSA_Expression expression
			, @Element(name="output", required=true) final DTSA_StreamBuffer output
			, @Attribute(name="type", required=true) final DTSA_QueryType type
			, @ElementMap(entry="stream-bound-function", inline=true, key="stream-bound-identifier", attribute=true, required=false) final java.util.HashMap<String, Term<BasicType>> streamBoundFunctions
			, @ElementMap(entry="stream-bound-clock-increase-function", inline=true, key="stream-bound-identifier", attribute=true, required=false) final java.util.HashMap<String, Term<BasicType>> streamBoundClockIncreaseFunctions
			, @ElementMap(entry="relevance-bound-function", inline=true, key="stream-bound-identifier", attribute=true, required=false) final java.util.HashMap<String, Term<BasicType>> relevanceBoundFunctions
			, @ElementMap(entry="relevance-bound-clock-increase-function", inline=true, key="stream-bound-identifier", attribute=true, required=false) final java.util.HashMap<String, Term<BasicType>> relevanceBoundClockIncreaseFunctions
			, @Element(name="description", required=false) final String description
			, @ElementList(entry="origin", inline=true, required=false) final java.util.ArrayList<String> origin
		) throws TSA_Exception {
		this(prog_package, expression, output, type
				, (streamBoundFunctions == null ? new FixedArrayNavigableMap<String, Term<BasicType>>() : new FixedArrayNavigableMap<String, Term<BasicType>>(streamBoundFunctions))
				, (streamBoundClockIncreaseFunctions == null ? new FixedArrayNavigableMap<String, Term<BasicType>>() : new FixedArrayNavigableMap<String, Term<BasicType>>(streamBoundClockIncreaseFunctions))
				, (relevanceBoundFunctions == null ? new FixedArrayNavigableMap<String, Term<BasicType>>() : new FixedArrayNavigableMap<String, Term<BasicType>>(relevanceBoundFunctions))
				, (relevanceBoundClockIncreaseFunctions == null ? new FixedArrayNavigableMap<String, Term<BasicType>>() : new FixedArrayNavigableMap<String, Term<BasicType>>(relevanceBoundClockIncreaseFunctions))
				, (description == null ? "" : description)
				, (origin == null ? new FixedArray<String>(): new FixedArray<String>(origin)));
	}

	@SuppressWarnings("unchecked")
	public DTSA_Query(final String prog_package, final DTSA_Expression expression,
			final DTSA_StreamBuffer output, final DTSA_QueryType type,
			final FixedArrayNavigableMap<String, ? extends Term<BasicType>> streamBoundFunctions,
			final FixedArrayNavigableMap<String, ? extends Term<BasicType>> streamBoundClockIncreaseFunctions,
			final FixedArrayNavigableMap<String, ? extends Term<BasicType>> relevanceBoundFunctions,
					final FixedArrayNavigableMap<String, ? extends Term<BasicType>> relevanceBoundClockIncreaseFunctions,
			final String description, final FixedCollection<String> origin) throws TSA_Exception {
		super();
		this.prog_package = prog_package;
 
		assert output.type == DTSA_BufferType.WORKING | output.type == DTSA_BufferType.OUTPUT | output.type == DTSA_BufferType.LOG | output.type == DTSA_BufferType.BLACK_HOLE_INPUT | output.type == DTSA_BufferType.BLACK_HOLE_WORKING;
		this.type = type;
		this.output = output;
		this.expression = expression;
		
		FixedArray<DTSA_StreamBuffer> inputs = this.expression.inputs();
		String[] inputNamesArray = new String[inputs.size()];
		if(inputs.size() > 0){		
			this.input_type = inputs.get(0).type;
			inputNamesArray[0] = inputs.get(0).full_name;
			for(int input = 1; input < inputs.size(); ++input){
				if(!(inputs.get(input).type == this.input_type)){
					throw new TSA_Exception(origin, "All inputs of a query must be of the same type. Mixing type \"" + this.input_type + "\" and type \"" + inputs.get(input).type + "\" is not allowed.");
				}
				inputNamesArray[input] = inputs.get(input).full_name; 
			}
		}
		else{
			this.input_type = DTSA_BufferType.INPUT;
		}
		FixedArrayNavigableSet<String> inputNames = new FixedArrayNavigableSet<String>(inputNamesArray);
		DTSA_StreamBuffer[] inputBuffers = new DTSA_StreamBuffer[inputNames.size()];
		for(DTSA_StreamBuffer input: inputs){
			int index = inputNames.indexOf(input.full_name);
			assert index >= 0;
			
			if(inputBuffers[index] == null){
				inputBuffers[index] = input;
			}
			else{
				if(!inputBuffers[index].equals(input)){
					throw new TSA_Exception(origin);
				}
			}
		}
		this.inputs = new FixedArrayNavigableMap<String, DTSA_StreamBuffer>(inputNames, new FixedArray<DTSA_StreamBuffer>(false, inputBuffers));
		
		this.isStatic = this.expression.isStatic;
		this.isMonotonic = expression.isMonotonic & !this.type.isGarbageQuery;
		this.streamBoundFunctions = (FixedArrayNavigableMap<String, Term<BasicType>>) streamBoundFunctions;
		this.relevanceBoundFunctions = (FixedArrayNavigableMap<String, Term<BasicType>>) relevanceBoundFunctions;
		this.streamBoundClockIncreaseFunctions = (FixedArrayNavigableMap<String, Term<BasicType>>) streamBoundClockIncreaseFunctions;
		this.relevanceBoundClockIncreaseFunctions = (FixedArrayNavigableMap<String, Term<BasicType>>) relevanceBoundClockIncreaseFunctions;
		if(streamBoundFunctions.size() != this.streamBoundClockIncreaseFunctions.size()){
			throw new IllegalArgumentException();
			//TODO better exception
		}
		this.description = description;
		
		this.origin = origin;
	}
	
	public void writeTo(final StringBuilder sb, final int indentDepth){
		indentTab(sb, indentDepth);
		sb.append("DERIVE ");
		sb.append(this.output.type.toString());
		sb.append(" ");
		sb.append(this.output.full_name);
		sb.append(" BY\n");
		this.expression.writeTo(sb, indentDepth + 1);
		sb.append("\n");
		indentTab(sb, indentDepth);
		sb.append("END");
	}
	
	public void writeTo(final StringBuilder sb){
		this.writeTo(sb, 0);
	}
	
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		this.writeTo(sb);
		return sb.toString();
	}

}
