package de.lmu.ifi.pms.cep.tsa.compiler;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;
import org.simpleframework.xml.ElementList;

import de.lmu.ifi.pms.cep.compilation.Compiler;
import de.lmu.ifi.pms.cep.compilation.CompilerFactory;
import de.lmu.ifi.pms.cep.tsa.DTSA_Program;
import de.lmu.ifi.pms.cep.tsa.TSA_Program;
import de.lmu.ifi.pms.cep.tsa.optimization.DTSA_Optimizer;
import de.lmu.ifi.pms.cep.tsa.optimization.DTSA_OptimizerFactory;
import de.lmu.ifi.pms.cep.tsa.optimization.TSA_Optimizer;
import de.lmu.ifi.pms.cep.tsa.optimization.TSA_OptimizerFactory;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;


public class TSA_DTSA_CompilerFactory extends CompilerFactory<TSA_Program, DTSA_Program>{
	
	public static final String defaultIdentifier = "de.lmu.ifi.pms.cep.tsa.compiler.dtsa";
	
	public final FixedArray<TSA_OptimizerFactory> preDifferentiationOptimizers;
	
	@ElementList(inline=true, entry="pre-optimizer", type=TSA_OptimizerFactory.class, required=false)
	java.util.ArrayList<TSA_OptimizerFactory> getPreOptimizers(){
		return new java.util.ArrayList<TSA_OptimizerFactory>(this.preDifferentiationOptimizers.javaCollection());
	}
	
	@Element(name="differencer")
	public final TSA_DifferencerFactory differencer;
	
	public final FixedArray<DTSA_OptimizerFactory> postDifferentiationOptimizers;
	
	@ElementList(inline=true, entry="post-optimizer", type=DTSA_OptimizerFactory.class, required=false)
	java.util.ArrayList<DTSA_OptimizerFactory> getPostOptimizers(){
		return new java.util.ArrayList<DTSA_OptimizerFactory>(this.postDifferentiationOptimizers.javaCollection());
	}
	
	public TSA_DTSA_CompilerFactory(final String identifier, final TSA_DifferencerFactory differencer) {
		this(identifier, new FixedArray<TSA_OptimizerFactory>(), differencer, new FixedArray<DTSA_OptimizerFactory>());
	}
	
	TSA_DTSA_CompilerFactory(@Attribute(name="iri", required=true) final String identifier, @ElementList(inline=true, entry="pre-optimizer", type=TSA_OptimizerFactory.class, required=false) final java.util.ArrayList<TSA_OptimizerFactory> preDifferenciationOptimizers, @Element(name="differencer") final TSA_DifferencerFactory differencer, @ElementList(inline=true, entry="post-optimizer", type=DTSA_OptimizerFactory.class, required=false) final java.util.ArrayList<DTSA_OptimizerFactory> postDifferenciationOptimizers) {
		this(identifier, (preDifferenciationOptimizers == null ? new FixedArray<TSA_OptimizerFactory>() : new FixedArray<TSA_OptimizerFactory>(preDifferenciationOptimizers)), differencer, (postDifferenciationOptimizers == null ? new FixedArray<DTSA_OptimizerFactory>() : new FixedArray<DTSA_OptimizerFactory>(postDifferenciationOptimizers)));
	}
	
	public TSA_DTSA_CompilerFactory(final TSA_OptimizerFactory[] preDifferenciationOptimizers, final TSA_DifferencerFactory differencer, final DTSA_OptimizerFactory[] postDifferenciationOptimizers) {
		this(defaultIdentifier, preDifferenciationOptimizers, differencer, postDifferenciationOptimizers);
	}
	
	public TSA_DTSA_CompilerFactory(final String identifier, final TSA_OptimizerFactory[] preDifferenciationOptimizers, final TSA_DifferencerFactory differencer, final DTSA_OptimizerFactory[] postDifferenciationOptimizers) {
		this(identifier, new FixedArray<TSA_OptimizerFactory>(preDifferenciationOptimizers), differencer, new FixedArray<DTSA_OptimizerFactory>(postDifferenciationOptimizers));
	}
	
	public TSA_DTSA_CompilerFactory(final FixedArray<TSA_OptimizerFactory> preDifferenciationOptimizers, final TSA_DifferencerFactory differencer, final FixedArray<DTSA_OptimizerFactory> postDifferenciationOptimizers) {
		this(defaultIdentifier, preDifferenciationOptimizers, differencer, postDifferenciationOptimizers);
	}
	
	public TSA_DTSA_CompilerFactory(final String identifier, final FixedArray<TSA_OptimizerFactory> preDifferenciationOptimizers, final TSA_DifferencerFactory differencer, final FixedArray<DTSA_OptimizerFactory> postDifferenciationOptimizers) {
		super(TSA_Program.class, DTSA_Program.class, identifier);
		this.preDifferentiationOptimizers = preDifferenciationOptimizers;
		this.differencer = differencer;
		this.postDifferentiationOptimizers = postDifferenciationOptimizers;
	}

	@Override
	public Compiler<TSA_Program, DTSA_Program> make() {
		final TSA_Optimizer[] preDifferentiationOptimizers = new TSA_Optimizer[this.preDifferentiationOptimizers.size()];
		for(int optimizer = 0; optimizer < preDifferentiationOptimizers.length; ++optimizer){
			preDifferentiationOptimizers[optimizer] = this.preDifferentiationOptimizers.get(optimizer).make();
		}
		final DTSA_Optimizer[] postDifferentiationOptimizers = new DTSA_Optimizer[this.postDifferentiationOptimizers.size()];
		for(int optimizer = 0; optimizer < postDifferentiationOptimizers.length; ++optimizer){
			postDifferentiationOptimizers[optimizer] = this.postDifferentiationOptimizers.get(optimizer).make();
		}
		return new TSA_DTSA_Compiler(preDifferentiationOptimizers, this.differencer.make(), postDifferentiationOptimizers);
	}
		
}
