package de.lmu.ifi.pms.cep.tsa.compiler;

import static de.lmu.ifi.pms.cep.tsa.conditions.AtomicTermID.*;
import static de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonOperatorID.*;
import static de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula.*;
import static de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormulaID.*;
import static de.lmu.ifi.pms.cep.tsa.conditions.Term.*;
import static de.lmu.ifi.pms.cep.tsa.conditions.TupleFunctionID.*;
import static de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID.*;
import static de.lmu.ifi.pms.cep.tsa.time.analysis.ArithmeticAnalyzer.PLUS_INFTY;
import static de.lmu.ifi.pms.java.util.filter.AbstractFilter.*;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.lmu.ifi.pms.cep.COMMON_CONSTANTS;
import de.lmu.ifi.pms.cep.event_mill.configuration.DEFAULTS;
import de.lmu.ifi.pms.cep.tsa.DTSA_BufferType;
import de.lmu.ifi.pms.cep.tsa.DTSA_CompositeExpression;
import de.lmu.ifi.pms.cep.tsa.DTSA_Expression;
import de.lmu.ifi.pms.cep.tsa.DTSA_Program;
import de.lmu.ifi.pms.cep.tsa.DTSA_Query;
import de.lmu.ifi.pms.cep.tsa.DTSA_QueryType;
import de.lmu.ifi.pms.cep.tsa.DTSA_StreamBuffer;
import de.lmu.ifi.pms.cep.tsa.TSA_CompositeExpression;
import de.lmu.ifi.pms.cep.tsa.TSA_Expression;
import de.lmu.ifi.pms.cep.tsa.TSA_Program;
import de.lmu.ifi.pms.cep.tsa.TSA_Query;
import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.TSA_TemporalStream;
import de.lmu.ifi.pms.cep.tsa.TSA_TemporalStreamExpression;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicStreamBoundFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.AtomicTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.Attribute;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeValue;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeVariable;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.BooleanAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.BooleanTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.ClockFunction;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormulaID;
import de.lmu.ifi.pms.cep.tsa.conditions.ConjunctiveConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.DefaultDefinition;
import de.lmu.ifi.pms.cep.tsa.conditions.DisjunctiveConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.DurationValue;
import de.lmu.ifi.pms.cep.tsa.conditions.Parameter;
import de.lmu.ifi.pms.cep.tsa.conditions.SequenceFunction;
import de.lmu.ifi.pms.cep.tsa.conditions.StreamBound;
import de.lmu.ifi.pms.cep.tsa.conditions.StreamBoundConstraint;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.TypedAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.Variable;
import de.lmu.ifi.pms.cep.tsa.conditions.adaptor.RenameAttributes;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_RuntimeException;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_SchemaException;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_BasicOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_ComplexOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_ImbeddingOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_ProjectionOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_RenameOperator;
import de.lmu.ifi.pms.cep.tsa.operators.TSA_SelectionOperator;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.Type;
import de.lmu.ifi.pms.cep.tsa.schema.TypeID;
import de.lmu.ifi.pms.cep.tsa.time.analysis.ArithmeticAnalyzer;
import de.lmu.ifi.pms.cep.tsa.time.analysis.ArithmeticAnalyzerFactory;
import de.lmu.ifi.pms.cep.tsa.time.analysis.ArithmeticRelationFormulaExtractor;
import de.lmu.ifi.pms.cep.tsa.time.analysis.CombinedArithmeticAnalyzerFactory;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.adaption.CollectionAdapter;
import de.lmu.ifi.pms.java.util.collection.AbstractCollection;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.array.Arrays;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.collection.fixed.FixedCollection;
import de.lmu.ifi.pms.java.util.filter.AbstractFilter;
import de.lmu.ifi.pms.java.util.filter.Filter;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;
import de.lmu.ifi.pms.java.util.map.fixed.FixedEnumMap;
import de.lmu.ifi.pms.java.util.tupel.Single;


public class TSA_DefaultDifferencer extends TSA_Differencer {
	
	static final BasicTypeAttribute _o_id_attribute = attr(DEFAULTS._o_id, IDENTIFIER);
	static final BasicTypeAttribute _o_block_attribute = attr(DEFAULTS._o_block, IDENTIFIER);
	//static final String id = DuraDefaultAttributeNameMapping.identifierName;
	static final ClockFunction tuple_id_seq = new SequenceFunction(COMMON_CONSTANTS.TUPLE_ID_CLOCK_NAME, IDENTIFIER); 
	
	static final String isNewAttributeName = ".tsa.isNew";
	static final String isNewAttributeAliasName = ".tsa.isNew.old";
	static final BooleanAttribute isNewAttribute = BooleanAttribute.create(isNewAttributeName);
	static final BooleanAttribute isNewAliasAttribute = BooleanAttribute.create(isNewAttributeAliasName);
	
	static final AbstractFilter<ConditionFormula> monotonicStreamBoundFilterWithoutVariables = new AbstractFilter<ConditionFormula>() {
		
		@Override
		public boolean accepts(final ConditionFormula element) {
			if(element instanceof AtomicStreamBoundFormula){
				AtomicStreamBoundFormula sbf = (AtomicStreamBoundFormula) element;
				if(sbf.isMonotonic){
					if(sbf.atomicTerm instanceof BasicTypeAttribute){
						BasicTypeAttribute attribute = (BasicTypeAttribute) sbf.atomicTerm;
						return attribute.ID == sbf.bound.type.ID;
					}
				}
			}
			return false;
		}
		
	};
	
	static final AbstractFilter<ConditionFormula> monotonicStreamBoundFilterWithVariables = new AbstractFilter<ConditionFormula>() {
		
		@Override
		public boolean accepts(final ConditionFormula element) {
			if(element instanceof AtomicStreamBoundFormula){
				AtomicStreamBoundFormula sbf = (AtomicStreamBoundFormula) element;
				if(sbf.isMonotonic){
					if(sbf.atomicTerm instanceof BasicTypeAttribute){
						BasicTypeAttribute attribute = (BasicTypeAttribute) sbf.atomicTerm;
						return attribute.ID == sbf.bound.type.ID;
					}
					if(sbf.atomicTerm instanceof BasicTypeVariable){
						BasicTypeVariable variable = (BasicTypeVariable) sbf.atomicTerm;
						return variable.ID == sbf.bound.type.ID;
					}
				}
			}
			return false;
		}
		
	};
	
	final ArithmeticAnalyzerFactory analyserFactory;
	
	public TSA_DefaultDifferencer(){
		this(new CombinedArithmeticAnalyzerFactory());
	}
	
	public TSA_DefaultDifferencer(final ArithmeticAnalyzerFactory analyserFactory) {
		this.analyserFactory = analyserFactory;
	}
	
	@Override
	public DTSA_Program differenciate(final TSA_Program program) {
		try {
			return this.differenciateImpl(program);
		} catch (TSA_Exception e) {
			throw new TSA_RuntimeException(e);
		}
	}

	@SuppressWarnings("unchecked")
	public DTSA_Program differenciateImpl(final TSA_Program program) throws TSA_Exception {
		try{
			assert program.isValid();
			
			FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers = this.differentiateBuffers(program.temporalStreams.values());
			
			FixedEnumMap<DTSA_QueryType, FixedArray<TSA_Query>> implicitQueries = this.implicitQueries(streamBuffers);
			
			FixedEnumMap<DTSA_QueryType, FixedArray<TSA_Query>> explicitQueries = this.explicitQueries(program.queries, streamBuffers);
			
			FixedArray<TSA_Query>[] allQueriesArray = new FixedArray[DTSA_QueryType.values().length];
			for(int type = 0; type < allQueriesArray.length; ++type){
				allQueriesArray[type] = FixedArray.concat(implicitQueries.values().get(type), explicitQueries.values().get(type));
			}
			
			FixedEnumMap<DTSA_QueryType, FixedArray<DTSA_Query>> queries = this.differentiateQueries(streamBuffers, new FixedEnumMap<DTSA_QueryType, FixedArray<TSA_Query>>(DTSA_QueryType.class, allQueriesArray));
			
			return new DTSA_Program(program.namespace, program.simple_name, program.description, streamBuffers, queries, program.origin);
		}
		catch (TSA_Exception e) {
			throw e;
		}
		catch (RuntimeException e) {
			throw new TSA_Exception(program.origin, e);
		}
	}
	
	public FixedArrayNavigableMap<TypedAttribute<?>, Term<? extends Type>> implicitAttributeSpecifications(final FixedArrayNavigableMap<String, Term<? extends Type>> implicitAttributeSpecifications){
		TypedAttribute<?>[] attributes = new TypedAttribute<?>[implicitAttributeSpecifications.size()];
		for(int attribute = 0; attribute < attributes.length; ++attribute){
			attributes[attribute] = attr(implicitAttributeSpecifications.keys().get(attribute), implicitAttributeSpecifications.values().get(attribute).type);
		}
		return new FixedArrayNavigableMap<TypedAttribute<?>, Term<? extends Type>>(new FixedArrayNavigableSet<TypedAttribute<?>>(false, false, attributes), implicitAttributeSpecifications.values());
	}
	
	public FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> differentiateBuffers(final FixedArray<TSA_TemporalStream> temporalStreams) throws TSA_Exception{
		ArrayList<String> inputStreamBufferNames = new ArrayList<String>();
		ArrayList<DTSA_StreamBuffer> inputStreamBuffers = new ArrayList<DTSA_StreamBuffer>();
		ArrayList<String> workingStreamBufferNames = new ArrayList<String>();
		ArrayList<DTSA_StreamBuffer> workingStreamBuffers = new ArrayList<DTSA_StreamBuffer>();
		ArrayList<String> outputStreamBufferNames = new ArrayList<String>();
		ArrayList<DTSA_StreamBuffer> outputStreamBuffers = new ArrayList<DTSA_StreamBuffer>();
		ArrayList<String> logStreamBufferNames = new ArrayList<String>();
		ArrayList<DTSA_StreamBuffer> logStreamBuffers = new ArrayList<DTSA_StreamBuffer>();
		
		for(final TSA_TemporalStream temporalStream: temporalStreams){

			try{ 

				//TODO treatment of definitions for composite attributes?
				FixedArrayNavigableMap<String, Term<? extends Type>> implicitAttributeSpecifications = temporalStream.schema.implicitAttributeSpecifications;
				if(temporalStream.types.contains(DTSA_BufferType.INPUT)){
	
					final TemporalStreamSchema schema = temporalStream.schema;					
					
					DTSA_StreamBuffer inputStream;
					TemporalStreamSchema newSchema;
					if(implicitAttributeSpecifications.size() > 0){
						final FixedArrayNavigableSet<String> explicitAttributeNames = FixedArrayNavigableSet.difference(schema.basicAttributes.keys(), implicitAttributeSpecifications.keys());
						final AbstractFilter<AtomicConditionFormula> explicitAttributeConstraintsFilter = new AbstractFilter<AtomicConditionFormula>(){
							@Override
							public boolean accepts(final AtomicConditionFormula atom) {
								if(explicitAttributeNames.containsAll(atom.AttributeNames())){
									return true;
								}
								return false;
							}
						};
						if(explicitAttributeNames.size() > 0){
							ConditionFormula newTupleConstraints = new ConjunctiveConditionFormula(schema.tupleConstraints()).focus(explicitAttributeConstraintsFilter).flaten();
							ConditionFormula newRelationConstraints = new ConjunctiveConditionFormula(schema.relationConstraints()).focus(notFilter(DefaultDefinition.implicitAttributeValuesSpecifications)).focus(explicitAttributeConstraintsFilter).flaten();
							try {
								newSchema = TemporalStreamSchema.create(schema.ID, schema.basicAttributes, (newTupleConstraints instanceof ConjunctiveConditionFormula ? ((ConjunctiveConditionFormula)newTupleConstraints).subformulas() : new FixedArray<ConditionFormula>(newTupleConstraints)), (newRelationConstraints instanceof ConjunctiveConditionFormula ? ((ConjunctiveConditionFormula)newRelationConstraints).subformulas() : new FixedArray<ConditionFormula>(newRelationConstraints)));
								newSchema = newSchema.retain(explicitAttributeNames);
							} catch (TSA_SchemaException e) {
								throw new TSA_Exception(temporalStream.origin, e);
							}
						}
						else{
							try {
								newSchema = TemporalStreamSchema.create(schema.ID, new FixedArrayNavigableMap<String, Type>(), new FixedArray<ConditionFormula>(), new FixedArray<ConditionFormula>());
							} catch (TSA_SchemaException e) {
								throw new TSA_Exception(temporalStream.origin, e);
							} 
						}
						
						if(!temporalStream.isStatic){
							try{
								newSchema = newSchema.extend(new FixedArrayNavigableMap<String, Type>(new String[]{DEFAULTS._o_block, DEFAULTS._o_id}, new Type[]{IDENTIFIER.type, IDENTIFIER.type})).constraint(new FixedArray<ConditionFormula>(new StreamBoundConstraint(_o_block_attribute, true)));
							} catch (TSA_SchemaException e) {
								throw new TSA_Exception(temporalStream.origin, e);
							} 
						}
						
						inputStream = this.differentiateBuffer(temporalStream.full_name, temporalStream.description, DTSA_BufferType.INPUT, newSchema, temporalStream.origin);
					}
					else{
						if(!temporalStream.schema.isStatic){
							try{
								newSchema = schema.extend(new FixedArrayNavigableMap<String, Type>(new String[]{DEFAULTS._o_block, DEFAULTS._o_id}, new Type[]{IDENTIFIER.type, IDENTIFIER.type})).constraint(new FixedArray<ConditionFormula>(new StreamBoundConstraint(_o_block_attribute, true)));
							} catch (TSA_SchemaException e) {
								throw new TSA_Exception(temporalStream.origin, e);
							} 
						}
						else{
							newSchema = schema;
						}
						inputStream = this.differentiateBuffer(temporalStream.full_name, temporalStream.description, DTSA_BufferType.INPUT, newSchema, temporalStream.origin);
					}
					
					inputStreamBufferNames.add(inputStream.full_name);
					inputStreamBuffers.add(inputStream);
				}
				
				if(temporalStream.types.contains(DTSA_BufferType.WORKING)){			
					DTSA_StreamBuffer workingStream = this.differentiateBuffer(temporalStream.full_name, temporalStream.description, DTSA_BufferType.WORKING, temporalStream.schema, temporalStream.origin);
					workingStreamBufferNames.add(workingStream.full_name);
					workingStreamBuffers.add(workingStream);
				}
				
				if(temporalStream.types.contains(DTSA_BufferType.OUTPUT)){
					DTSA_StreamBuffer outputStreamBuffer;
					try {
						TemporalStreamSchema newSchema = temporalStream.schema.extend(new FixedArrayNavigableMap<String, Type>(new String[]{DEFAULTS._o_id}, new Type[]{IDENTIFIER.type}));
						newSchema = newSchema.constraint(new FixedArray<ConditionFormula>(new DefaultDefinition(_o_id_attribute, tuple_id_seq), new StreamBoundConstraint(_o_id_attribute, true)));
						
						outputStreamBuffer = this.differentiateBuffer(temporalStream.full_name, temporalStream.description, DTSA_BufferType.OUTPUT, newSchema, temporalStream.origin);
					} catch (TSA_SchemaException e) {
						throw new TSA_Exception(temporalStream.origin, e);
					}
					outputStreamBufferNames.add(outputStreamBuffer.full_name);
					outputStreamBuffers.add(outputStreamBuffer);
				}
				
				if(temporalStream.types.contains(DTSA_BufferType.LOG)){
					DTSA_StreamBuffer logStreamBuffer;
					try {
						TemporalStreamSchema newSchema = temporalStream.schema.extend(new FixedArrayNavigableMap<String, Type>(new String[]{DEFAULTS._o_id}, new Type[]{IDENTIFIER.type}));
						newSchema = newSchema.constraint(new FixedArray<ConditionFormula>(new DefaultDefinition(_o_id_attribute, tuple_id_seq), new StreamBoundConstraint(_o_id_attribute, true)));
						
						logStreamBuffer = this.differentiateBuffer(temporalStream.full_name, temporalStream.description, DTSA_BufferType.LOG, newSchema, temporalStream.origin);
					} catch (TSA_SchemaException e) {
						throw new TSA_Exception(temporalStream.origin, e);
					}
					logStreamBufferNames.add(logStreamBuffer.full_name);
					logStreamBuffers.add(logStreamBuffer);
				}
			}
			catch (TSA_Exception e) {
				throw e;
			}
			catch (RuntimeException e) {
				throw new TSA_Exception(temporalStream.origin, e);
			}
		}
		
		@SuppressWarnings("unchecked")
		FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers = new FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String,DTSA_StreamBuffer>>(DTSA_BufferType.class,
				new FixedArrayNavigableMap<String,DTSA_StreamBuffer>(new FixedArrayNavigableSet<String>(false, false, inputStreamBufferNames.toArray(new String[0])), new FixedArray<DTSA_StreamBuffer>(false, inputStreamBuffers.toArray(new DTSA_StreamBuffer[0]))),		
				new FixedArrayNavigableMap<String,DTSA_StreamBuffer>(new FixedArrayNavigableSet<String>(false, false, workingStreamBufferNames.toArray(new String[0])), new FixedArray<DTSA_StreamBuffer>(false, workingStreamBuffers.toArray(new DTSA_StreamBuffer[0]))),
				new FixedArrayNavigableMap<String,DTSA_StreamBuffer>(new FixedArrayNavigableSet<String>(false, false, outputStreamBufferNames.toArray(new String[0])), new FixedArray<DTSA_StreamBuffer>(false, outputStreamBuffers.toArray(new DTSA_StreamBuffer[0]))),
				new FixedArrayNavigableMap<String,DTSA_StreamBuffer>(new FixedArrayNavigableSet<String>(false, false, logStreamBufferNames.toArray(new String[0])), new FixedArray<DTSA_StreamBuffer>(false, logStreamBuffers.toArray(new DTSA_StreamBuffer[0]))),
				new FixedArrayNavigableMap<String,DTSA_StreamBuffer>(),
				new FixedArrayNavigableMap<String,DTSA_StreamBuffer>()
		);
		
		return streamBuffers;
	}
	
	public TSA_Expression extendExpression(final TSA_Expression expression, final DTSA_StreamBuffer output_buffer) throws TSA_Exception{
		FixedArrayNavigableSet<String> missingAttributes = FixedArrayNavigableSet.difference(output_buffer.schema.basicAttributes.keys(), expression.schema().basicAttributes.keys());
		
		if(missingAttributes.size() > 0){
			FixedArrayNavigableMap<String, Term<? extends Type>> implicitAttributeSpecifications = output_buffer.implicitAttributeSpecifications;
			
			//FIXME what happens if both, default for attribute and subattributes?
			
			//extend to non basic attributes
			
			ArrayList<TypedAttribute<?>> implicitAttributes = new ArrayList<TypedAttribute<?>>(implicitAttributeSpecifications.size());
			ArrayList<Term<?>> definitions = new ArrayList<Term<?>>(implicitAttributeSpecifications.size());
			
			for (int specification = 0; specification < implicitAttributeSpecifications.size(); specification++) {
				String name = implicitAttributeSpecifications.keys().get(specification);
				Term<? extends Type> definition = implicitAttributeSpecifications.values().get(specification);
				assert definition.isTyped;
	
				if(!expression.schema().hasAttribute(name)){
					TypedAttribute<?> implicitAttribute = Term.attr(name, definition.type);
					implicitAttributes.add(implicitAttribute);								
					definitions.add(definition);
				}
			}
			
			TSA_ImbeddingOperator imbeddingOperator = new TSA_ImbeddingOperator(output_buffer.origin, new FixedArrayNavigableMap<TypedAttribute<?>, Term<?>>(implicitAttributes.toArray(new TypedAttribute<?>[0]), definitions.toArray(new Term[0])));
			return new TSA_CompositeExpression(imbeddingOperator, expression);
		}
		else{
			return expression;
		}
	}
	
	FixedEnumMap<DTSA_QueryType, FixedArray<TSA_Query>> implicitQueries(final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers) throws TSA_Exception{
		@SuppressWarnings("unchecked")
		ArrayList<TSA_Query>[] implicitQueriesListsArray = (ArrayList<TSA_Query>[]) new ArrayList<?>[DTSA_QueryType.values().length];
		for(int type = 0; type < implicitQueriesListsArray.length; ++type){
			implicitQueriesListsArray[type] = new ArrayList<TSA_Query>();
		}		
		FixedEnumMap<DTSA_QueryType, ArrayList<TSA_Query>> implicitQueriesLists = new FixedEnumMap<DTSA_QueryType, ArrayList<TSA_Query>>(DTSA_QueryType.class, implicitQueriesListsArray);
		
		for(DTSA_StreamBuffer inputBuffer: streamBuffers.get(DTSA_BufferType.INPUT).values()){
			try{
				DTSA_StreamBuffer workingBuffer = streamBuffers.get(DTSA_BufferType.WORKING).get(inputBuffer.full_name);
				
				if(workingBuffer != null){					
					TSA_Expression query_expression = this.extendExpression(new TSA_CompositeExpression(new TSA_ProjectionOperator(workingBuffer.origin, FixedArrayNavigableSet.intersect(inputBuffer.schema.basicAttributes.keys(), workingBuffer.schema.basicAttributes.keys())), new TSA_TemporalStreamExpression(inputBuffer.full_name, inputBuffer.type, inputBuffer.schema, inputBuffer.origin)), workingBuffer);
					
					if(query_expression.isStatic){
						if(workingBuffer.isStatic){
							implicitQueriesLists.get(DTSA_QueryType.STATIC).add(new TSA_Query(workingBuffer.prog_package, workingBuffer.full_name, workingBuffer.schema, DTSA_BufferType.WORKING, query_expression, "Static input query for type " + workingBuffer.full_name, workingBuffer.origin));
						}
						else{
							implicitQueriesLists.get(DTSA_QueryType.INIT).add(new TSA_Query(workingBuffer.prog_package, workingBuffer.full_name, workingBuffer.schema, DTSA_BufferType.WORKING, query_expression, "Initial input query for type " + workingBuffer.full_name, workingBuffer.origin));
						}
					}
					else{
						implicitQueriesLists.get(DTSA_QueryType.INPUT).add(new TSA_Query(workingBuffer.prog_package, workingBuffer.full_name, workingBuffer.schema, DTSA_BufferType.WORKING, query_expression, "Continous input query for type " + workingBuffer.full_name, workingBuffer.origin));
					}
				}
			}
			catch (TSA_Exception e) {
				throw e;
			}
			catch (RuntimeException e) {
				throw new TSA_Exception(inputBuffer.origin, e);
			}
		}
		
		for(DTSA_StreamBuffer outputBuffer: streamBuffers.get(DTSA_BufferType.OUTPUT).values()){
			try{
				DTSA_StreamBuffer workingBuffer = streamBuffers.get(DTSA_BufferType.WORKING).get(outputBuffer.full_name);
				
				if(workingBuffer != null){
				
					TSA_Expression query_expression = new TSA_TemporalStreamExpression(workingBuffer.full_name, workingBuffer.type, workingBuffer.schema, workingBuffer.origin);
					query_expression = new TSA_CompositeExpression(new TSA_ImbeddingOperator(outputBuffer.origin, new FixedArrayNavigableMap<BasicTypeAttribute, Term<?>>(new BasicTypeAttribute[]{_o_id_attribute}, new Term<?>[]{tuple_id_seq})), query_expression);

					if(query_expression.isStatic){
						assert !outputBuffer.isStatic;
						implicitQueriesLists.get(DTSA_QueryType.INIT).add(new TSA_Query(outputBuffer.prog_package, outputBuffer.full_name, outputBuffer.schema, DTSA_BufferType.OUTPUT, query_expression, "Initial output query for type " + outputBuffer.full_name, outputBuffer.origin));
					}
					else{
						implicitQueriesLists.get(DTSA_QueryType.OUTPUT).add(new TSA_Query(outputBuffer.prog_package, outputBuffer.full_name, outputBuffer.schema, DTSA_BufferType.OUTPUT, query_expression, "Continous output query for type " + outputBuffer.full_name, outputBuffer.origin));
					}
				
				}
				else{
					//FIXME
					//System.err.println("Type \"" + outputBuffer.full_name + "\" declared as output type but not as working. No output will ever be produced.");
					
				}
			}
			catch (TSA_Exception e) {
				throw e;
			}
			catch (RuntimeException e) {
				throw new TSA_Exception(outputBuffer.origin, e);
			}
		}
		
		for(DTSA_StreamBuffer logBuffer: streamBuffers.get(DTSA_BufferType.LOG).values()){
			try{
				DTSA_StreamBuffer workingBuffer = streamBuffers.get(DTSA_BufferType.WORKING).get(logBuffer.full_name);
				
				if(workingBuffer != null){
				
					TSA_Expression query_expression = new TSA_TemporalStreamExpression(workingBuffer.full_name, workingBuffer.type, workingBuffer.schema, workingBuffer.origin);
					query_expression = new TSA_CompositeExpression(new TSA_ImbeddingOperator(logBuffer.origin, new FixedArrayNavigableMap<BasicTypeAttribute, Term<?>>(new BasicTypeAttribute[]{_o_id_attribute}, new Term<?>[]{tuple_id_seq})), query_expression);
					
					if(query_expression.isStatic){
						assert !logBuffer.isStatic;
						implicitQueriesLists.get(DTSA_QueryType.INIT).add(new TSA_Query(logBuffer.prog_package, logBuffer.full_name, logBuffer.schema, DTSA_BufferType.LOG, query_expression, "Initial logging query for type " + logBuffer.full_name, logBuffer.origin));
					}
					else{
						implicitQueriesLists.get(DTSA_QueryType.LOG).add(new TSA_Query(logBuffer.prog_package, logBuffer.full_name, logBuffer.schema, DTSA_BufferType.LOG, query_expression, "Continous logging query for type " + logBuffer.full_name, logBuffer.origin));
					}
				}
				else{
					//FIXME
					//System.err.println("Type \"" + logBuffer.full_name + "\" declared as log type but not as working. No output will ever be produced.");
					
				}
			}
			catch (TSA_Exception e) {
				throw e;
			}
			catch (RuntimeException e) {
				throw new TSA_Exception(logBuffer.origin, e);
			}
		}
		
		@SuppressWarnings("unchecked")
		FixedArray<TSA_Query>[] implicitQueriesArray = (FixedArray<TSA_Query>[]) new FixedArray<?>[implicitQueriesListsArray.length];
		for(int type = 0; type < implicitQueriesArray.length; ++type){
			implicitQueriesArray[type] = new FixedArray<TSA_Query>(implicitQueriesListsArray[type]);
		}
		FixedEnumMap<DTSA_QueryType, FixedArray<TSA_Query>> implicitQueries = new FixedEnumMap<DTSA_QueryType, FixedArray<TSA_Query>>(DTSA_QueryType.class,implicitQueriesArray);
		return implicitQueries;
	}
	
	public FixedEnumMap<DTSA_QueryType, FixedArray<TSA_Query>> explicitQueries(final FixedArray<TSA_Query> queries, final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers) throws TSA_Exception{
		@SuppressWarnings("unchecked")
		ArrayList<TSA_Query>[] explicitQueriesListsArray = (ArrayList<TSA_Query>[]) new ArrayList<?>[DTSA_QueryType.values().length];
		for(int type = 0; type < explicitQueriesListsArray.length; ++type){
			explicitQueriesListsArray[type] = new ArrayList<TSA_Query>();
		}		
		FixedEnumMap<DTSA_QueryType, ArrayList<TSA_Query>> explicitQueriesLists = new FixedEnumMap<DTSA_QueryType, ArrayList<TSA_Query>>(DTSA_QueryType.class, explicitQueriesListsArray);
		
		for(TSA_Query query: queries){
			DTSA_StreamBuffer output_buffer = streamBuffers.get(query.output_type).get(query.output_full_name);
			if(output_buffer == null){
				throw new TSA_Exception(query.origin, "The query writes to the non-existing stream buffer \"" + query.output_type + " " + query.output_full_name + "\"");
			}
			
			TSA_Expression extendedExpression = this.extendExpression(query.expression, output_buffer);
			
			if(extendedExpression != query.expression){
				query = new TSA_Query(query.prog_package, query.output_full_name, query.output_schema, query.output_type, extendedExpression, query.description, query.origin);
			}
			
			if(query.isStatic){
				if(query.output_isStatic){
					explicitQueriesLists.get(DTSA_QueryType.STATIC).add(query);
				}
				else{
					explicitQueriesLists.get(DTSA_QueryType.INIT).add(query);	
				}
			}
			else{
				explicitQueriesLists.get(DTSA_QueryType.WORKING).add(query);
			}
		}
		
		@SuppressWarnings("unchecked")
		FixedArray<TSA_Query>[] explicitQueriesArray = (FixedArray<TSA_Query>[]) new FixedArray<?>[explicitQueriesListsArray.length];
		for(int type = 0; type < explicitQueriesArray.length; ++type){
			explicitQueriesArray[type] = new FixedArray<TSA_Query>(explicitQueriesListsArray[type]);
		}
		FixedEnumMap<DTSA_QueryType, FixedArray<TSA_Query>> explicitQueries = new FixedEnumMap<DTSA_QueryType, FixedArray<TSA_Query>>(DTSA_QueryType.class,explicitQueriesArray);
		return explicitQueries;
	}
	
	public FixedEnumMap<DTSA_QueryType, FixedArray<DTSA_Query>> differentiateQueries(final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers, final FixedEnumMap<DTSA_QueryType, FixedArray<TSA_Query>> queries) throws TSA_Exception{	
		@SuppressWarnings("unchecked")
		ArrayList<DTSA_Query>[] differentiatedQueriesListsArray = (ArrayList<DTSA_Query>[]) new ArrayList<?>[DTSA_QueryType.values().length];
		for(int type = 0; type < differentiatedQueriesListsArray.length; ++type){
			differentiatedQueriesListsArray[type] = new ArrayList<DTSA_Query>();
		}		
		FixedEnumMap<DTSA_QueryType, ArrayList<DTSA_Query>> differentiatedQueriesLists = new FixedEnumMap<DTSA_QueryType, ArrayList<DTSA_Query>>(DTSA_QueryType.class, differentiatedQueriesListsArray);
		
		for(DTSA_QueryType type: queries.keys()){
			ArrayList<DTSA_Query> differentiatedQueriesList = differentiatedQueriesLists.get(type);
			for(TSA_Query query: queries.get(type)){
				DTSA_Query differentiatedQuery = this.differentiateQuery(query, type, streamBuffers);
				if(differentiatedQuery != null){
					differentiatedQueriesList.add(differentiatedQuery);
				}
			}
		}
		
		@SuppressWarnings("unchecked")
		FixedArray<DTSA_Query>[] differentiatedQueriesArray = (FixedArray<DTSA_Query>[]) new FixedArray<?>[differentiatedQueriesListsArray.length];
		for(int type = 0; type < differentiatedQueriesArray.length; ++type){
			differentiatedQueriesArray[type] = new FixedArray<DTSA_Query>(differentiatedQueriesListsArray[type]);
		}
		FixedEnumMap<DTSA_QueryType, FixedArray<DTSA_Query>> differentiatedQueries = new FixedEnumMap<DTSA_QueryType, FixedArray<DTSA_Query>>(DTSA_QueryType.class,differentiatedQueriesArray);
		return differentiatedQueries;
	}
	

	
/*	protected ConditionFormula outputMinTimeDefinitionFormula(final StreamBound outputMinTime, final ConditionFormula minTimeFormula, final ArithmeticAnalyser temporalAnalyser, final FixedArrayNavigableMap<String, StreamBound> referencedAttributes, final FixedArrayNavigableMap<String, StreamBound> referencedVariables){
		assert !minTimeFormula.containsNegation();
		String outputAttributeName = outputMinTime.name;
		
		final BasicTypeID type = outputMinTime.type.ID;
		Filter<ConditionFormula> typeFilter = new AbstractFilter<ConditionFormula>() {
			
			@Override
			public boolean accepts(final ConditionFormula element) {
				assert element instanceof AtomicStreamBoundFormula;
				AtomicStreamBoundFormula sbf = (AtomicStreamBoundFormula) element;				
				assert sbf.streamBound.type.equals(sbf.atomicTerm.type);
				return sbf.streamBound.type.ID == type;
			}
			
		};
		
		final ConditionFormula typeFilteredMinTimeFormula = not(minTimeFormula).focus(typeFilter).subformula().flaten();
		
		if(typeFilteredMinTimeFormula.ID == ConditionFormulaID.TOP){
			return Top;
		}
		else{		
			for(ConjunctiveConditionFormula possibleStreamBounds: typeFilteredMinTimeFormula.DNFiteration()){
				boolean isStreamBound = true;
				ArrayList<ConditionFormula> outputMinTimeDefinitionFormulaPart = new ArrayList<ConditionFormula>();
				for(ConditionFormula atom: possibleStreamBounds.subformulas()){
					assert atom instanceof AtomicStreamBoundFormula;
					AtomicStreamBoundFormula streamBound = (AtomicStreamBoundFormula) atom;
					boolean isAttribute = (streamBound.atomicTerm instanceof Attribute<?>);
					String streamBoundAttribute;
					StreamBound streamBoundMinTime;
					if(isAttribute){
						streamBoundAttribute = ((Attribute<?>) streamBound.atomicTerm).name;
						assert referencedAttributes.has(streamBoundAttribute);
						streamBoundMinTime = referencedAttributes.get(streamBoundAttribute);
					}
					else{
						assert streamBound.atomicTerm instanceof Variable<?>;
						streamBoundAttribute = ((Variable<?>) streamBound.atomicTerm).name;
						assert referencedVariables.has(streamBoundAttribute);
						streamBoundMinTime = referencedVariables.get(streamBoundAttribute);
					}
					
					long distance = temporalAnalyser.distance(streamBoundAttribute, isAttribute, outputAttributeName, true);
					if(distance < ArithmeticAnalyser.PLUS_INFTY){
						long constant = ArithmeticAnalyser.constant(distance);
						if(constant == 0){
							outputMinTimeDefinitionFormulaPart.add(new BinaryComparissonFormula(LESS_EQUAL, outputMinTime, streamBoundMinTime));
						}
						else{
							outputMinTimeDefinitionFormulaPart.add(new BinaryComparissonFormula(LESS_EQUAL, outputMinTime, apply(MINUS, TIMESTAMP.type, streamBoundMinTime, new DurationValue(constant))));
						}
					}
					else{
						isStreamBound = false;
						break;
					}
				}
				if(isStreamBound){
					//FIXME alternatives! now just the first possible one
					return and(outputMinTimeDefinitionFormulaPart).flaten();
				}
			}
			return Bottom;
		}
	}*/
	
	@SuppressWarnings("unchecked")
	protected Term<BasicType>[] streamBoundFunctions(final FixedArrayNavigableMap<String, ConditionFormula> streamBoundFunctionFormulas){				
		Term<BasicType>[] streamBoundFunctions = (Term<BasicType>[]) new Term<?>[streamBoundFunctionFormulas.size()]; 
		
		for(int outputMinTime = 0; outputMinTime < streamBoundFunctions.length; ++outputMinTime){
			String streamBound = streamBoundFunctionFormulas.keys().get(outputMinTime);
			ConditionFormula flatenedStreamBoundFormula = streamBoundFunctionFormulas.values().get(outputMinTime).flaten();
			streamBoundFunctions[outputMinTime] = this.streamBoundFunction(streamBound, flatenedStreamBoundFormula);
		}
		
		return streamBoundFunctions;
	}
	
	@SuppressWarnings("unchecked")
	protected Term<BasicType> streamBoundFunction(final String streamBound, final ConditionFormula outputStreamBoundFormula){
		assert outputStreamBoundFormula.isFlatened;
		switch(outputStreamBoundFormula.ID){
			case ATOMIC: assert outputStreamBoundFormula instanceof BinaryComparissonFormula;
				BinaryComparissonFormula bcf = (BinaryComparissonFormula) outputStreamBoundFormula;
				assert bcf.operator == LESS_EQUAL;
				assert bcf.secondTerm().isTyped;
				assert ((Term<Type>)bcf.secondTerm()).type.ID == TypeID.BASIC;
				assert bcf.firstTerm() instanceof StreamBound; 
				assert((StreamBound)bcf.firstTerm()).qualified_name.equals(streamBound);
				return (Term<BasicType>) bcf.secondTerm();
			case CONJUNCTIVE:{
				ConjunctiveConditionFormula conj = (ConjunctiveConditionFormula) outputStreamBoundFormula;
				Term<BasicType>[] subDefinitions = (Term<BasicType>[])new Term<?>[conj.subformulas().size()];
				for(int subFormula = 0; subFormula < subDefinitions.length; ++subFormula){
					subDefinitions[subFormula] = this.streamBoundFunction(streamBound, conj.subformulas().get(subFormula));
				}
				assert subDefinitions.length > 0;
				BasicType type = subDefinitions[0].type;
				return apply(LEAST, type, subDefinitions);
			}
			case DISJUNCTIVE:{
				DisjunctiveConditionFormula disj = (DisjunctiveConditionFormula) outputStreamBoundFormula;
				Term<BasicType>[] subDefinitions = (Term<BasicType>[])new Term<?>[disj.subformulas().size()];
				for(int subFormula = 0; subFormula < subDefinitions.length; ++subFormula){
					subDefinitions[subFormula] = this.streamBoundFunction(streamBound, disj.subformulas().get(subFormula));
				}
				assert subDefinitions.length > 0;
				BasicType type = subDefinitions[0].type;
				return apply(GREATEST, type, subDefinitions);
			}
			case NEGATED: throw new IllegalArgumentException("Invalid case. ");
			case BOTTOM: return Bottom;
			case TOP: return Top;
		}
		throw new Error("Unexpected Case: " + outputStreamBoundFormula);
	}
	
	@SuppressWarnings("unchecked")
	protected Term<BasicType>[] streamBoundFunctionsClockIncrease(final FixedArrayNavigableMap<String, ConditionFormula> streamBoundFunctionFormulas){				
		Term<BasicType>[] streamBoundFunctions = (Term<BasicType>[]) new Term<?>[streamBoundFunctionFormulas.size()]; 
		
		for(int outputMinTime = 0; outputMinTime < streamBoundFunctions.length; ++outputMinTime){
			String streamBound = streamBoundFunctionFormulas.keys().get(outputMinTime);
			ConditionFormula flatenedStreamBoundFormula = streamBoundFunctionFormulas.values().get(outputMinTime).flaten();
			streamBoundFunctions[outputMinTime] = this.streamBoundFunctionClockIncrease(streamBound, flatenedStreamBoundFormula);
		}
		
		return streamBoundFunctions;
	}
	
	@SuppressWarnings("unchecked")
	protected Term<BasicType> streamBoundFunctionClockIncrease(final String streamBound, final ConditionFormula outputStreamBoundFormula){
		assert outputStreamBoundFormula.isFlatened;
		switch(outputStreamBoundFormula.ID){
			case ATOMIC: assert outputStreamBoundFormula instanceof BinaryComparissonFormula;
				BinaryComparissonFormula bcf = (BinaryComparissonFormula) outputStreamBoundFormula;
				assert bcf.operator == GREATER_EQUAL;
				assert bcf.secondTerm().isTyped;
				assert ((Term<Type>)bcf.secondTerm()).type.ID == TypeID.BASIC;
				assert bcf.firstTerm() instanceof StreamBound; 
				assert((StreamBound)bcf.firstTerm()).qualified_name.equals(streamBound);
				return (Term<BasicType>) bcf.secondTerm();
			case CONJUNCTIVE:{
				ConjunctiveConditionFormula conj = (ConjunctiveConditionFormula) outputStreamBoundFormula;
				Term<BasicType>[] subDefinitions = (Term<BasicType>[])new Term<?>[conj.subformulas().size()];
				for(int subFormula = 0; subFormula < subDefinitions.length; ++subFormula){
					subDefinitions[subFormula] = this.streamBoundFunction(streamBound, conj.subformulas().get(subFormula));
				}
				assert subDefinitions.length > 0;
				BasicType type = subDefinitions[0].type;
				return apply(GREATEST, type, subDefinitions);
			}
			case DISJUNCTIVE:{
				DisjunctiveConditionFormula disj = (DisjunctiveConditionFormula) outputStreamBoundFormula;
				Term<BasicType>[] subDefinitions = (Term<BasicType>[])new Term<?>[disj.subformulas().size()];
				for(int subFormula = 0; subFormula < subDefinitions.length; ++subFormula){
					subDefinitions[subFormula] = this.streamBoundFunction(streamBound, disj.subformulas().get(subFormula));
				}
				assert subDefinitions.length > 0;
				BasicType type = subDefinitions[0].type;
				return apply(LEAST, type, subDefinitions);
			}
			case NEGATED: throw new IllegalArgumentException("Invalid case. ");
			case BOTTOM: return Top;
			case TOP: return Bottom;
		}
		throw new Error("Unexpected Case: " + outputStreamBoundFormula);
	}
	
	protected Term<BasicType>[] streamBoundFunctionsClockIncrease(final FixedArrayNavigableMap<String, StreamBound> streamBounds, final ArithmeticAnalyzer temporalAnalyser, final ConditionFormula outputSBF, final FixedArrayNavigableMap<String, ClockFunction> clockAttributes, final FixedArrayNavigableMap<String, ClockFunction> clockVariables){
		return this.streamBoundFunctionsClockIncrease(this.streamBoundFunctionClockIncreaseFormulas(streamBounds, temporalAnalyser, outputSBF, clockAttributes, clockVariables));
	}
	
	protected FixedArrayNavigableMap<String, ConditionFormula> streamBoundFunctionClockIncreaseFormulas(final FixedArrayNavigableMap<String, StreamBound> streamBounds, final ArithmeticAnalyzer temporalAnalyser, final ConditionFormula outputSBF, final FixedArrayNavigableMap<String, ClockFunction> clockAttributes, final FixedArrayNavigableMap<String, ClockFunction> clockVariables){
		assert TemporalStreamSchema.isValidBufferStreamBoundFormula(outputSBF);
		//TODO stream bound assigned to multiple attributes
		ConditionFormula[] streamBoundFunctionClockFormulas = new ConditionFormula[streamBounds.size()];
	
		Arrays.fill(streamBoundFunctionClockFormulas, Top);
		for(ConditionFormula formula: outputSBF.collectFormulas(atomicFormulaFilter)){
			AtomicStreamBoundFormula streamBoundFormula = (AtomicStreamBoundFormula) formula;
			assert streamBoundFormula.bound instanceof StreamBound;
			StreamBound streamBound = (StreamBound) streamBoundFormula.bound;
			
			int index = streamBounds.keys().indexOf(streamBound.qualified_name);
			if(index >= 0){
				assert streamBounds.values().get(index).equals(streamBound);
							
				String outputAttributeName = ((BasicTypeAttribute) streamBoundFormula.atomicTerm).name;
				
				ArrayList<BinaryComparissonFormula> increasingClocks = new ArrayList<BinaryComparissonFormula>();
				for(String clockAttribute: clockAttributes.keys()){
					long distance = temporalAnalyser.distance(outputAttributeName, true, clockAttribute, true);
					if(distance < PLUS_INFTY){
						long constant = ArithmeticAnalyzer.constant(distance);
						ClockFunction clock = clockAttributes.get(clockAttribute);
						if(constant == 0){
							increasingClocks.add(new BinaryComparissonFormula(GREATER_EQUAL, streamBound, clock));
						}
						else{
							increasingClocks.add(new BinaryComparissonFormula(GREATER_EQUAL, streamBound, apply(MINUS, clock.type, clock, new DurationValue(constant))));
						}
					}
				}
				for(String clockVariable: clockVariables.keys()){
					long distance = temporalAnalyser.distance(outputAttributeName, true, clockVariable, false);
					if(distance < ArithmeticAnalyzer.PLUS_INFTY){
						long constant = ArithmeticAnalyzer.constant(distance);
						ClockFunction clock = clockVariables.get(clockVariable);
						if(constant == 0){
							increasingClocks.add(new BinaryComparissonFormula(GREATER_EQUAL, streamBound, clock));
						}
						else{
							increasingClocks.add(new BinaryComparissonFormula(GREATER_EQUAL, streamBound, apply(MINUS, clock.type, clock, new DurationValue(constant))));
						}
					}
				}
				
				if(increasingClocks.size() == 1){
					streamBoundFunctionClockFormulas[index] =  increasingClocks.first();
				}
				else{
					streamBoundFunctionClockFormulas[index] =  and(increasingClocks);//FIXME check which direction is used above	
				}
			}
		}		
		
		return new FixedArrayNavigableMap<String, ConditionFormula>(streamBounds.keys(), new FixedArray<ConditionFormula>(false, streamBoundFunctionClockFormulas));
	}
	
	protected final static Filter<ConditionFormula> atomicFormulaFilter = new AbstractFilter<ConditionFormula>() {
		@Override
		public boolean accepts(final ConditionFormula element) {
			return element.ID == ConditionFormulaID.ATOMIC;
		}
	};
	
	public DTSA_Query differentiateQuery(final TSA_Query tsa_query, final DTSA_QueryType type, final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers) throws TSA_Exception{
		try{
			assert (tsa_query.isStatic & (type == DTSA_QueryType.STATIC | type == DTSA_QueryType.INIT)) | (!tsa_query.isStatic & type != DTSA_QueryType.STATIC & type != DTSA_QueryType.INIT); 
			
			final DTSA_BufferType output_type = tsa_query.output_type;
			
			final DTSA_StreamBuffer output = streamBuffers.get(output_type).get(tsa_query.output_full_name);
			final ConditionFormula outputStreamBoundFormula = this.streamBoundFormula(output.schema.relationConstraints(), output.origin);
			assert TemporalStreamSchema.isValidBufferStreamBoundFormula(outputStreamBoundFormula);
			
			final boolean isMonotonic = tsa_query.isMonotonic();
			
			final boolean monotonicInputStreamBounds = not(not(this.streamBoundFormula(tsa_query.expression.schema().relationConstraints(), tsa_query.origin)).focus(monotonicStreamBoundFilterWithVariables)).flaten().ID != BOTTOM;
			assert !(type == DTSA_QueryType.INPUT) | monotonicInputStreamBounds;
			
			final DTSA_Expression differentiatedExpression = this.differentiateExpression(tsa_query.expression, type, isMonotonic, monotonicInputStreamBounds, streamBuffers);
			
			ConditionFormula differenciatedExpressionStreamBoundFormula = this.streamBoundFormula(differentiatedExpression.schema().relationConstraints(), tsa_query.origin);
			assert tsa_query.isStatic | differenciatedExpressionStreamBoundFormula.ID != ConditionFormulaID.TOP: "Top is only allowed for static queries";			
			
			ConditionFormula differentiatedExpressionTemporalRelationFormula = and(differentiatedExpression.schema().tupleConstraints()).flaten();//FIXME only temporal relation constraints including clocks

			differenciatedExpressionStreamBoundFormula = differenciatedExpressionStreamBoundFormula.replaceAtomicTerm(ArithmeticRelationFormulaExtractor.enhaceClockAnalysisVariablePrefixer);
			differentiatedExpressionTemporalRelationFormula = ArithmeticRelationFormulaExtractor.enhaceClockAnalysis(differentiatedExpressionTemporalRelationFormula.replaceAtomicTerm(ArithmeticRelationFormulaExtractor.enhaceClockAnalysisVariablePrefixer));
			
			
			final FixedArrayNavigableMap<String, StreamBound> outputStreamBounds = output.streamBounds;
			
			
			final FixedArrayNavigableMap<String, DTSA_StreamBuffer> inputs;
			{
				FixedArray<DTSA_StreamBuffer> inputBuffers = differentiatedExpression.inputs();
				String[] inbutBufferNames = new String[inputBuffers.size()];
				
				for(int bufferIndex = 0; bufferIndex < inbutBufferNames.length; ++bufferIndex){
					inbutBufferNames[bufferIndex] = inputBuffers.get(bufferIndex).full_name;
				}	
				
				inputs = new FixedArrayNavigableMap<String, DTSA_StreamBuffer>(new FixedArray<String>(false, inbutBufferNames), inputBuffers);
			}
			
			final FixedArrayNavigableMap<String, StreamBound> inputStreamBounds;			
			{
				ArrayList<FixedArrayNavigableMap<String, StreamBound>> inputStreamBoundsList = new ArrayList<FixedArrayNavigableMap<String,StreamBound>>();
				for(DTSA_StreamBuffer input: inputs.values()){
					inputStreamBoundsList.add(input.streamBounds);
				}			
				inputStreamBounds = FixedArrayNavigableMap.merge(inputStreamBoundsList);
			}
			
			final FixedArrayNavigableMap<String, BasicTypeAttribute> outputReferencedAttributes;
			{					
				FixedArrayNavigableSet<Attribute<?>> attributes = outputStreamBoundFormula.Attributes();
				
				BasicTypeAttribute[] outputReferencedAttributesArray = new BasicTypeAttribute[attributes.size()];
				String[] outputReferencedAttributeNamesArray = new String[outputReferencedAttributesArray.length];
				
				for(int attribute = 0; attribute < outputReferencedAttributesArray.length; ++attribute){
					outputReferencedAttributesArray[attribute] = (BasicTypeAttribute) attributes.get(attribute);
					outputReferencedAttributeNamesArray[attribute] = attributes.get(attribute).name;
				}
				
				outputReferencedAttributes = new FixedArrayNavigableMap<String, BasicTypeAttribute>(new FixedArrayNavigableSet<String>(false, false, outputReferencedAttributeNamesArray), new FixedArrayNavigableSet<BasicTypeAttribute>(false, false, outputReferencedAttributesArray));
			}
			
			final FixedArrayNavigableMap<String, BasicTypeAttribute> expressionReferencedAttributes;
			{
				FixedArrayNavigableSet<Attribute<?>> attributes = differenciatedExpressionStreamBoundFormula.Attributes();
				
				BasicTypeAttribute[] expressionReferencedAttributesArray = new BasicTypeAttribute[attributes.size()];
				String[] expressionReferencedAttributeNamesArray = new String[expressionReferencedAttributesArray.length];
				
				for(int attribute = 0; attribute < expressionReferencedAttributesArray.length; ++attribute){
					expressionReferencedAttributesArray[attribute] = (BasicTypeAttribute) attributes.get(attribute);
					expressionReferencedAttributeNamesArray[attribute] = attributes.get(attribute).name;
				}
				
				expressionReferencedAttributes = new FixedArrayNavigableMap<String, BasicTypeAttribute>(new FixedArrayNavigableSet<String>(false, false, expressionReferencedAttributeNamesArray), new FixedArrayNavigableSet<BasicTypeAttribute>(false, false, expressionReferencedAttributesArray));
			}
			final FixedArrayNavigableMap<String, BasicTypeVariable> expressionReferencedVariables;
			{
				FixedArrayNavigableSet<Variable<?>> variables = differenciatedExpressionStreamBoundFormula.Variables();
				
				BasicTypeVariable[] expressionReferencedVariablesArray = new BasicTypeVariable[variables.size()];
				String[] expressionReferencedVariableNamesArray = new String[expressionReferencedVariablesArray.length];
				
				for(int variable = 0; variable < expressionReferencedVariablesArray.length; ++variable){
					expressionReferencedVariablesArray[variable] = (BasicTypeVariable) variables.get(variable);
					expressionReferencedVariableNamesArray[variable] = variables.get(variable).name;
				}
				
				expressionReferencedVariables = new FixedArrayNavigableMap<String, BasicTypeVariable>(new FixedArrayNavigableSet<String>(false, false, expressionReferencedVariableNamesArray), new FixedArrayNavigableSet<BasicTypeVariable>(false, false, expressionReferencedVariablesArray));
			}
			
			final FixedArrayNavigableMap<String, BasicTypeAttribute> clockReferencedAttributes;
			final FixedArrayNavigableMap<String, ClockFunction> clockAttributes;
			final FixedArrayNavigableMap<String, BasicTypeVariable> clockReferencedVariables;
			final FixedArrayNavigableMap<String, ClockFunction> clockVariables;
			{//FIXME more general treatment of definitions that contain a clock. More general terms on both sided (increment AND selection). Solve in propagation
				Filter<ConditionFormula> clockFunctionsFilter = new AbstractFilter<ConditionFormula>() {
					
					@Override
					public boolean accepts(final ConditionFormula element) {
						if(element instanceof BinaryComparissonFormula){
							BinaryComparissonFormula bcf = (BinaryComparissonFormula) element;
							if(bcf.operator == EQUAL & bcf.secondTerm() instanceof ClockFunction){
								return true;
							}
						}
						return false;
					}
					
				};
				
				AbstractCollection<ConditionFormula> clockEquations = differentiatedExpressionTemporalRelationFormula.collectFormulas(clockFunctionsFilter);
				
				ArrayList<String> clockAttributeNamesList = new ArrayList<String>();
				ArrayList<BasicTypeAttribute> clockAttributesList = new ArrayList<BasicTypeAttribute>();
				ArrayList<ClockFunction> clockAttributeClockFunctionList = new ArrayList<ClockFunction>();
				ArrayList<String> clockVariableNamesList = new ArrayList<String>();
				ArrayList<BasicTypeVariable> clockVariablesList = new ArrayList<BasicTypeVariable>();
				ArrayList<ClockFunction> clockVariableClockFunctionList = new ArrayList<ClockFunction>();
				
				for(ConditionFormula minTimeInequationFormula: clockEquations){
					BinaryComparissonFormula clockEquation = (BinaryComparissonFormula) minTimeInequationFormula;
					ClockFunction clock = (ClockFunction) clockEquation.secondTerm();
					if(clockEquation.firstTerm() instanceof BasicTypeAttribute){
						BasicTypeAttribute attribute = (BasicTypeAttribute) clockEquation.firstTerm();
						if(attribute.ID == IDENTIFIER | attribute.ID == TIMESTAMP){
							clockAttributeNamesList.add(attribute.name);
							clockAttributesList.add(attribute);
							clockAttributeClockFunctionList.add(clock);
						}
					}
					else{assert clockEquation.firstTerm() instanceof BasicTypeVariable;					
						BasicTypeVariable variable = (BasicTypeVariable) clockEquation.firstTerm();
						if(variable.ID == IDENTIFIER | variable.ID == TIMESTAMP){
							clockVariableNamesList.add(variable.name);
							clockVariablesList.add(variable);
							clockVariableClockFunctionList.add(clock);
						}
					}
				}
				
				clockAttributes = new FixedArrayNavigableMap<String, ClockFunction>(clockAttributeNamesList.toArray(new String[0]), clockAttributeClockFunctionList.toArray(new ClockFunction[0]));
				clockReferencedAttributes = new FixedArrayNavigableMap<String, BasicTypeAttribute>(clockAttributes.keys(), new FixedArray<BasicTypeAttribute>(false, clockAttributesList.toArray(new BasicTypeAttribute[0])));
				clockVariables = new FixedArrayNavigableMap<String, ClockFunction>(clockVariableNamesList.toArray(new String[0]), clockVariableClockFunctionList.toArray(new ClockFunction[0]));
				clockReferencedVariables = new FixedArrayNavigableMap<String, BasicTypeVariable>(clockVariables.keys(), new FixedArray<BasicTypeVariable>(false, clockVariablesList.toArray(new BasicTypeVariable[0])));;
			}
			
			@SuppressWarnings("unchecked")
			FixedArrayNavigableMap<String, BasicTypeAttribute> referencedAttributes = FixedArrayNavigableMap.merge(outputReferencedAttributes, expressionReferencedAttributes, clockReferencedAttributes);
			FixedArrayNavigableMap<String, BasicTypeVariable> referencedVariables = FixedArrayNavigableMap.merge(expressionReferencedVariables, clockReferencedVariables);
			
			@SuppressWarnings("unchecked")
			final Term<BasicType>[] outputStreamBoundFunctionsAllArray = (Term<BasicType>[]) new Term<?>[outputStreamBounds.size()];
			@SuppressWarnings("unchecked")
			final Term<BasicType>[] inputRelevanceBoundFunctionsAllArray = (Term<BasicType>[]) new Term<?>[inputStreamBounds.size()];
			
			@SuppressWarnings("unchecked")
			final Term<BasicType>[] outputStreamBoundFunctionsClockIncreaseArray = (Term<BasicType>[]) new Term<?>[outputStreamBounds.size()];
			@SuppressWarnings("unchecked")
			final Term<BasicType>[] inputRelevanceBoundFunctionsClockIncreaseArray = (Term<BasicType>[]) new Term<?>[inputStreamBounds.size()];
			
			de.lmu.ifi.pms.java.util.collection.array.Arrays.fill(outputStreamBoundFunctionsAllArray, Top);
			de.lmu.ifi.pms.java.util.collection.array.Arrays.fill(outputStreamBoundFunctionsClockIncreaseArray, Top);
			
			de.lmu.ifi.pms.java.util.collection.array.Arrays.fill(inputRelevanceBoundFunctionsAllArray, Top);
			de.lmu.ifi.pms.java.util.collection.array.Arrays.fill(inputRelevanceBoundFunctionsClockIncreaseArray, Top);
			
			Collection<ConditionFormula> trfCases = new Single<ConditionFormula, ConditionFormula>(differentiatedExpressionTemporalRelationFormula);
			
			boolean isConsistent = false;
			for(ConditionFormula trfCase: trfCases){
				ArithmeticAnalyzer analyser = this.analyserFactory.make(trfCase, referencedAttributes.keys(), referencedVariables.keys());
				if(analyser.isConsistent()){
					isConsistent = true;
					if(type == DTSA_QueryType.STATIC){
						break;
					}
					else{
						final Term<BasicType>[] caseOutputStreamBoundFunctionsArray = this.streamBoundFunctions(outputStreamBounds, analyser, differenciatedExpressionStreamBoundFormula, outputStreamBoundFormula);
						final Term<BasicType>[] caseOutputStreamBoundFunctionsClockIncreaseArray = this.streamBoundFunctionsClockIncrease(outputStreamBounds, analyser, outputStreamBoundFormula, clockAttributes, clockVariables);
						
						//FIXME input stream bounds instead of output stream bounds
						final Term<BasicType>[] caseInputRelevanceBoundFunctionsArray = this.relevanceBoundFunctions(inputStreamBounds, analyser, differenciatedExpressionStreamBoundFormula, outputStreamBoundFormula);
						final Term<BasicType>[] caseInputRelevanceBoundFunctionsClockIncreaseArray = this.relevanceBoundFunctionsClockIncrease(inputStreamBounds, analyser, differenciatedExpressionStreamBoundFormula, clockAttributes, clockVariables);
						
						for(int outputStreamBoundIndex = 0; outputStreamBoundIndex < outputStreamBoundFunctionsAllArray.length; ++outputStreamBoundIndex){
							outputStreamBoundFunctionsAllArray[outputStreamBoundIndex] = TSA_DefaultDifferencer.minBoundFunctions(outputStreamBoundFunctionsAllArray[outputStreamBoundIndex], caseOutputStreamBoundFunctionsArray[outputStreamBoundIndex]);
							outputStreamBoundFunctionsClockIncreaseArray[outputStreamBoundIndex] = TSA_DefaultDifferencer.minBoundFunctions(outputStreamBoundFunctionsClockIncreaseArray[outputStreamBoundIndex], caseOutputStreamBoundFunctionsClockIncreaseArray[outputStreamBoundIndex]);
							
						}
						
						for(int inputRelevanceBoundIndex = 0; inputRelevanceBoundIndex < inputRelevanceBoundFunctionsAllArray.length; ++inputRelevanceBoundIndex){
							inputRelevanceBoundFunctionsAllArray[inputRelevanceBoundIndex] = TSA_DefaultDifferencer.minBoundFunctions(inputRelevanceBoundFunctionsAllArray[inputRelevanceBoundIndex], caseInputRelevanceBoundFunctionsArray[inputRelevanceBoundIndex]);
							inputRelevanceBoundFunctionsClockIncreaseArray[inputRelevanceBoundIndex] = TSA_DefaultDifferencer.minBoundFunctions(inputRelevanceBoundFunctionsClockIncreaseArray[inputRelevanceBoundIndex], caseInputRelevanceBoundFunctionsClockIncreaseArray[inputRelevanceBoundIndex]);
						}
					}
				}
			}
 
			//check definitions of output stream bound functions
			for(int outputStreamBoundIndex = 0; outputStreamBoundIndex < outputStreamBoundFunctionsAllArray.length; ++outputStreamBoundIndex){
				Term<BasicType> streamBoundFunction = outputStreamBoundFunctionsAllArray[outputStreamBoundIndex];
				assert type != DTSA_QueryType.STATIC;
				assert type != DTSA_QueryType.INIT || ((streamBoundFunction instanceof ConditionFormula) && ((ConditionFormula)streamBoundFunction).ID == TOP);
				assert !isConsistent || !(streamBoundFunction instanceof ConditionFormula) || type == DTSA_QueryType.INIT || ((ConditionFormula)streamBoundFunction).ID != TOP;
				
				Term<BasicType> streamBoundClockIncreaseFunction = outputStreamBoundFunctionsClockIncreaseArray[outputStreamBoundIndex];
	
				if(streamBoundFunction instanceof ConditionFormula && ((ConditionFormula)streamBoundFunction).ID == ConditionFormulaID.BOTTOM){
					if(streamBoundClockIncreaseFunction instanceof ConditionFormula && ((ConditionFormula)streamBoundClockIncreaseFunction).ID == ConditionFormulaID.BOTTOM){
						throw new TSA_Exception(tsa_query.origin , "Output stream bound " + outputStreamBounds.keys().get(outputStreamBoundIndex) + " is invalid.");
					}
				}
			}
			
			FixedArrayNavigableMap<String, Term<BasicType>> outputStreamBoundFunctions = new FixedArrayNavigableMap<String, Term<BasicType>>(outputStreamBounds.keys(), new FixedArray<Term<BasicType>>(false, outputStreamBoundFunctionsAllArray));
			FixedArrayNavigableMap<String, Term<BasicType>> outputStreamBoundFunctionsClock = new FixedArrayNavigableMap<String, Term<BasicType>>(outputStreamBounds.keys(), new FixedArray<Term<BasicType>>(false, outputStreamBoundFunctionsClockIncreaseArray));
			
			/*FIXME
			//increase relevance bound functions for monotonic queries with a single input (particularly input queries) 
			assert !(type == DTSA_QueryType.INPUT) || differentiatedExpression.isMonotonic && differentiatedExpression.inputs().size() == 1;
			if(differentiatedExpression.isMonotonic && differentiatedExpression.inputs().size() == 1){
				for(int inputStreamBoundIndex = 0; inputStreamBoundIndex < inputRelevanceBoundFunctionsAllArray.length; ++inputStreamBoundIndex){
					
					
					inputRelevanceBoundFunctionsAllArray[inputStreamBoundIndex] = TSA_DefaultDifferencer.maxBoundFunctions(inputRelevanceBoundFunctionsAllArray[inputStreamBoundIndex], inputStreamBounds.values().get(inputStreamBoundIndex));							
				}
			}
			*/
			
			{
				//check definitions of input relevance bound functions
				
				int inputStreamBoundIndex = 0;
				int inputStreamBoundMaxIndex = 0;
					
				for(int inputIndex = 0; inputIndex < inputs.size(); ++inputIndex){
					DTSA_StreamBuffer input = inputs.values().get(inputIndex); 
					inputStreamBoundMaxIndex = inputStreamBoundIndex + input.streamBounds.size();
					assert inputStreamBoundMaxIndex <= inputRelevanceBoundFunctionsAllArray.length;
					
					boolean impliesRelevanceCondition = (type == DTSA_QueryType.STATIC | type == DTSA_QueryType.INIT | type == DTSA_QueryType.INPUT) | input.isStatic;
					for(; inputStreamBoundIndex < inputStreamBoundMaxIndex; ++inputStreamBoundIndex){
						if(inputRelevanceBoundFunctionsAllArray[inputStreamBoundIndex] != Bottom | inputRelevanceBoundFunctionsClockIncreaseArray[inputStreamBoundIndex] != Bottom){
							impliesRelevanceCondition = true;
						}
					}
					if(!impliesRelevanceCondition){
						//FIXME warning
						System.err.println(new TSA_Exception(tsa_query.origin, "The query does not imply a garbage collection condition for buffer \"" + input.type + " " + input.full_name + "\"").getMessage());
					}
				}
			}
			
			FixedArrayNavigableMap<String, Term<BasicType>> inputRelevanceBoundFunctions = new FixedArrayNavigableMap<String, Term<BasicType>>(inputStreamBounds.keys(), new FixedArray<Term<BasicType>>(false, inputRelevanceBoundFunctionsAllArray));
			FixedArrayNavigableMap<String, Term<BasicType>> inputRelevanceBoundFunctionsClock = new FixedArrayNavigableMap<String, Term<BasicType>>(inputStreamBounds.keys(), new FixedArray<Term<BasicType>>(false, inputRelevanceBoundFunctionsClockIncreaseArray));
			
			DTSA_Expression incrementExpression;
			
			if(!isConsistent){
				//FIXME warning
				System.err.println(new TSA_Exception(tsa_query.origin, "Program contains inconsistent query.").getMessage());
				return null;
			}
			else if(type == DTSA_QueryType.STATIC | type == DTSA_QueryType.INIT){
				incrementExpression = differentiatedExpression;
			}
			else if(isMonotonic & monotonicInputStreamBounds){		
				
				boolean hasIsNewAttribute = differentiatedExpression.schema().hasAttribute(isNewAttributeName);
				ConditionFormula isNewFormula = this.isNewFormula(hasIsNewAttribute, not(not(differenciatedExpressionStreamBoundFormula).focus(monotonicStreamBoundFilterWithVariables)).focus(monotonicStreamBoundFilterWithoutVariables).flaten());
				assert isNewFormula.ID != ConditionFormulaID.TOP;
				assert isNewFormula.ID != ConditionFormulaID.BOTTOM;
				final TSA_SelectionOperator isNewSelection = new TSA_SelectionOperator(tsa_query.origin, isNewFormula);
	
				if(hasIsNewAttribute){
					incrementExpression =  new DTSA_CompositeExpression(new TSA_ProjectionOperator(tsa_query.origin, tsa_query.expression.schema().basicAttributes.keys()), new DTSA_CompositeExpression(isNewSelection, differentiatedExpression));
				}
				else{
					incrementExpression = new DTSA_CompositeExpression(isNewSelection, differentiatedExpression);
				}
				
	
			}
			else{
				assert type != DTSA_QueryType.STATIC;
				assert type != DTSA_QueryType.INIT;
				assert type != DTSA_QueryType.INPUT;
				assert !isMonotonic | !monotonicInputStreamBounds;
				assert !tsa_query.isStatic;		
				
				TSA_SelectionOperator incrementSelection = new TSA_SelectionOperator(tsa_query.origin, this.incrementFormula(outputStreamBoundFormula, outputStreamBoundFunctions, tsa_query.origin));
	
				incrementExpression = new DTSA_CompositeExpression(incrementSelection, differentiatedExpression);
			}
	
			return new DTSA_Query(tsa_query.prog_package, incrementExpression, output, type, outputStreamBoundFunctions, outputStreamBoundFunctionsClock, inputRelevanceBoundFunctions, inputRelevanceBoundFunctionsClock, tsa_query.description, tsa_query.origin);
		}
		catch (TSA_Exception e) {
			throw e;
		}
		catch (RuntimeException e) {
			throw new TSA_Exception(tsa_query.origin, e);
		}
	}
	
	public static Term<BasicType> maxBoundFunctions(final Term<BasicType> function1, final Term<BasicType> function2){
		if((function1 instanceof ConditionFormula && ((ConditionFormula)function1).ID == TOP) | (function2 instanceof ConditionFormula && ((ConditionFormula)function2).ID == TOP)){
			//FIXME XML replace by function1 == Top | function2 == Top 
			return Top;
		}
		else if(function1 instanceof ConditionFormula && ((ConditionFormula)function1).ID == BOTTOM){//FIXME XML replace by function1 == Bottom
			return function2;
		}
		else if(function2 instanceof ConditionFormula && ((ConditionFormula)function2).ID == BOTTOM){//FIXME XML replace by function2 == Bottom
			return function1;
		}
		else{
			assert function1.type.equals(function2.type);
			return apply(GREATEST, function1.type, function1, function2);
		}
	}
	
	public static Term<BasicType> minBoundFunctions(final Term<BasicType> function1, final Term<BasicType> function2){
		if((function1 instanceof ConditionFormula && ((ConditionFormula)function1).ID == BOTTOM) | (function2 instanceof ConditionFormula && ((ConditionFormula)function2).ID == BOTTOM)){
			//FIXME XML replace by function1 == Bottom | function2 == Bottom 
			return Bottom;
		}
		else if(function1 instanceof ConditionFormula && ((ConditionFormula)function1).ID == TOP){//FIXME XML replace by function1 == Top
			return function2;
		}
		else if(function2 instanceof ConditionFormula && ((ConditionFormula)function2).ID == TOP){//FIXME XML replace by function2 == Top
			return function1;
		}
		else{
			assert function1.type.equals(function2.type);
			return apply(LEAST, function1.type, function1, function2);
		}
	}
	
	public static Term<BasicType> minBoundFunctions(final Term<BasicType>... functions){
		ArrayList<Term<BasicType>> nonTrivialFunctions = new ArrayList<Term<BasicType>>();
		for(Term<BasicType> function: functions){
			assert function.type.ID == functions[0].type.ID || (function instanceof ConditionFormula && ( ((ConditionFormula)function).ID == BOTTOM | ((ConditionFormula)function).ID == TOP ));
			if(function instanceof ConditionFormula && ((ConditionFormula)function).ID == BOTTOM){//FIXME XML replace by function == Bottom
				return Bottom;
			}
			else if(!(function instanceof ConditionFormula) || ((ConditionFormula)function).ID != TOP){//FIXME XML replace by function != Top
				//TODO more efficient
				if(!nonTrivialFunctions.contains(function)){
					nonTrivialFunctions.add(function);
				}
			}
		}
		
		if(nonTrivialFunctions.size() == 0){
			return Top;
		}
		else if(nonTrivialFunctions.size() == 1){
			return nonTrivialFunctions.get(0);
		}
		else{
			assert functions.length > 0;
			return apply(LEAST, functions[0].type, nonTrivialFunctions.toArray(new Term<?>[nonTrivialFunctions.size()]));
		}
	}
	
	public ConjunctiveConditionFormula incrementFormula(final ConditionFormula outputSBF, final FixedArrayNavigableMap<String, Term<BasicType>> outputStreamBoundFunctions, final FixedCollection<String> origin) throws TSA_Exception{		
		assert TemporalStreamSchema.isValidBufferStreamBoundFormula(outputSBF);
		
		final ConditionFormula upperBoundFormula = outputSBF.replaceAtomicFormula(new Adapter<AtomicConditionFormula, ConditionFormula>() {			
			@Override
			public ConditionFormula adapt(final AtomicConditionFormula source) {
				if(source instanceof AtomicStreamBoundFormula){
					AtomicStreamBoundFormula sbf = (AtomicStreamBoundFormula) source;
					if(sbf.atomicTerm instanceof BasicTypeAttribute){
						assert sbf.bound instanceof StreamBound;
						StreamBound streamBound = (StreamBound) sbf.bound;
						assert outputStreamBoundFunctions.get(streamBound.qualified_name) != Top;
						assert !(outputStreamBoundFunctions.get(streamBound.qualified_name) instanceof ConditionFormula) || ((ConditionFormula)outputStreamBoundFunctions.get(streamBound.qualified_name)).ID != TOP;//FIXME XML remove
						if(outputStreamBoundFunctions.has(streamBound.qualified_name)){
							if(!(outputStreamBoundFunctions.get(streamBound.qualified_name) instanceof ConditionFormula) || ((ConditionFormula)outputStreamBoundFunctions.get(streamBound.qualified_name)).ID != BOTTOM){
							//FIXME XML replace by if(outputStreamBoundFunctions.get(streamBound.qualified_name) != Bottom){
								Parameter upperBound = new Parameter("output.now:" + streamBound.qualified_name, sbf.bound.type);
								return new BinaryComparissonFormula(LESS_EQUAL, sbf.atomicTerm, upperBound);
							}
						}
						return Bottom;
					}
				}				
				throw new Error();
			}			
		}).flaten();
		
		final ConditionFormula lowerBoundFormula = not(outputSBF.replaceAtomicFormula(new Adapter<AtomicConditionFormula, ConditionFormula>() {			
			@Override
			public ConditionFormula adapt(final AtomicConditionFormula source) {
				if(source instanceof AtomicStreamBoundFormula){
					AtomicStreamBoundFormula sbf = (AtomicStreamBoundFormula) source;
					if(sbf.atomicTerm instanceof BasicTypeAttribute){
						assert sbf.bound instanceof StreamBound;
						StreamBound streamBound = (StreamBound) sbf.bound;
						if(outputStreamBoundFunctions.has(streamBound.qualified_name)){
							assert outputStreamBoundFunctions.get(streamBound.qualified_name) != Top;
							if(outputStreamBoundFunctions.get(streamBound.qualified_name) != Bottom){
								Parameter lowerBound = new Parameter("output.last:" + streamBound.qualified_name, sbf.bound.type);
								return not(new BinaryComparissonFormula(LESS, lowerBound, sbf.atomicTerm));
							}
						}
						return Bottom;
					}
				}				
				throw new Error();
			}			
		})).flaten();
		
		if(upperBoundFormula.flaten().ID == BOTTOM){
			throw new TSA_Exception(origin, "The temporal relations are too weak to derive an incremental evaluation expression for the non-static query.");
		}
		
		return and(upperBoundFormula, lowerBoundFormula);
	}
	
	public ConditionFormula streamBoundFormula(final FixedArray<ConditionFormula> relationConstraints, final FixedCollection<String> origin) throws TSA_Exception{
		ConditionFormula streamBoundFormula = and(relationConstraints).focus(TemporalStreamSchema.streamBoundFilter).flaten();
		try {
			TemporalStreamSchema.checkExpressionStreamBoundFormula(streamBoundFormula);
		} catch (TSA_SchemaException e) {
			throw new TSA_Exception(origin, e);
		}
		return streamBoundFormula;
	}
	
	public DTSA_StreamBuffer differentiateBuffer(final String full_name, final String description, final DTSA_BufferType type, final TemporalStreamSchema schema, final FixedCollection<String> origin) throws TSA_Exception{		
		try{			
			@SuppressWarnings("unchecked")
			FixedArray<ConditionFormula> relationConstraints = FixedArray.concat(
					and(schema.relationConstraints()).focus(notFilter(orFilter(TemporalStreamSchema.streamBoundConstraintFilter, DefaultDefinition.implicitAttributeValuesSpecifications))).subformulas(),
					new FixedArray<ConditionFormula>(TemporalStreamSchema.streamBoundFormulaFromConstraints(full_name, schema.relationConstraints()))
				);
			
			TemporalStreamSchema differentiatedSchema;
			try {
				differentiatedSchema = TemporalStreamSchema.create(schema.ID, schema.attributes, schema.tupleConstraints(), relationConstraints, false);
			} catch (TSA_SchemaException e) {
				throw new TSA_Exception(origin, e);
			}
			
			ConditionFormula streamBoundFormula = this.streamBoundFormula(differentiatedSchema.relationConstraints(), origin);
			assert TemporalStreamSchema.isValidBufferStreamBoundFormula(streamBoundFormula);
			
			return new DTSA_StreamBuffer(full_name, description, type, differentiatedSchema, schema.implicitAttributeSpecifications, origin);
		}
		catch (TSA_SchemaException e) {
			throw new TSA_Exception(origin, e);
		}
		catch (TSA_Exception e) {
			throw e;
		}
		catch (RuntimeException e) {
			throw new TSA_Exception(origin, e);
		}
	}
	
	public DTSA_Expression differentiateExpression(final TSA_Expression tsa_expression, final DTSA_QueryType type, final boolean monotonic, final boolean monotonicInputStreamBounds, final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers) throws TSA_Exception{
		if(tsa_expression instanceof TSA_TemporalStreamExpression){
			TSA_TemporalStreamExpression tsa_temporalStreamExpression = (TSA_TemporalStreamExpression) tsa_expression;
			final DTSA_StreamBuffer dtsa_streamBuffer = streamBuffers.get(tsa_temporalStreamExpression.type).get(tsa_temporalStreamExpression.full_name);		
			
			if(dtsa_streamBuffer == null){
				throw new TSA_Exception(tsa_temporalStreamExpression.origin, "The expression refers to an non-existing stream: " + tsa_temporalStreamExpression.type + ":" + tsa_temporalStreamExpression.full_name);
			}
			else{
				return dtsa_streamBuffer;
			}
		}
		else{assert tsa_expression instanceof TSA_CompositeExpression;
			return this.differentiateExpression((TSA_CompositeExpression)tsa_expression, type, monotonic, monotonicInputStreamBounds, streamBuffers);
		}
				
	}
	
	protected ConditionFormula isNewFormula(final boolean hasIsNewAttribute, final ConditionFormula discardedAttributesFormula){
		ConditionFormula isNewFormula;
		if(discardedAttributesFormula.ID != TOP){
			ConditionFormula isNewPreFormula = this.isNewFormula(discardedAttributesFormula);
				
				final ConditionFormula upperBoundFormula = isNewPreFormula.replaceAtomicFormula(new Adapter<AtomicConditionFormula, ConditionFormula>() {			
					@Override
					public ConditionFormula adapt(final AtomicConditionFormula source) {
						if(source instanceof AtomicStreamBoundFormula){
							AtomicStreamBoundFormula sbf = (AtomicStreamBoundFormula) source;
							if(sbf.atomicTerm instanceof BasicTypeAttribute){
								Parameter upperBound;
								if(sbf.bound instanceof StreamBound){
									upperBound = new Parameter("input.now:" + ((StreamBound)sbf.bound).qualified_name, sbf.bound.type);
								}
								else if(sbf.bound instanceof ClockFunction){
									upperBound = new Parameter("clock.min.now:" + ((ClockFunction)sbf.bound).name, sbf.bound.type);
								}
								else{
									throw new Error("Unexpected case");
								}
								return new BinaryComparissonFormula(LESS_EQUAL, sbf.atomicTerm, upperBound);
							}
						}				
						throw new Error();
					}			
				}).flaten();
				
				final ConditionFormula lowerBoundFormula = not(isNewPreFormula.replaceAtomicFormula(new Adapter<AtomicConditionFormula, ConditionFormula>() {			
					@Override
					public ConditionFormula adapt(final AtomicConditionFormula source) {
						if(source instanceof AtomicStreamBoundFormula){
							AtomicStreamBoundFormula sbf = (AtomicStreamBoundFormula) source;
							if(sbf.atomicTerm instanceof BasicTypeAttribute){
								Parameter lowerBound;
								if(sbf.bound instanceof StreamBound){
									lowerBound = new Parameter("input.last:" + ((StreamBound)sbf.bound).qualified_name, sbf.bound.type);
								}
								else if(sbf.bound instanceof ClockFunction){
									lowerBound = new Parameter("clock.min.last:" + ((ClockFunction)sbf.bound).name, sbf.bound.type);
								}
								else{
									throw new Error("Unexpected case");
								}
								return not(new BinaryComparissonFormula(LESS, lowerBound, sbf.atomicTerm));
							}
						}				
						throw new Error();
					}			
				})).flaten();
			
			
			isNewFormula = and(lowerBoundFormula, upperBoundFormula).flaten();
			if(hasIsNewAttribute){
				isNewFormula = 	or(new BooleanTerm(isNewAttribute) ,isNewFormula).flaten();
			}
		}
		else if(hasIsNewAttribute){
			isNewFormula = 	new BooleanTerm(isNewAttribute);
		}
		else{
			throw new IllegalArgumentException();
		}
		return isNewFormula.flaten();
	}
	
	protected ConditionFormula isNewFormula(final ConditionFormula discardedAttributesFormula){
		ConditionFormula flatenedDiscardedAttributesFormula = discardedAttributesFormula.flaten();
		switch (flatenedDiscardedAttributesFormula.ID) {
		case TOP: return discardedAttributesFormula;
		case BOTTOM: return discardedAttributesFormula;
		case ATOMIC: return discardedAttributesFormula;
		case CONJUNCTIVE:
			ConjunctiveConditionFormula conjunction = (ConjunctiveConditionFormula) discardedAttributesFormula;
			assert conjunction.subformulas().size() > 0;
			FixedArray<ConditionFormula> subFormulas = conjunction.subformulas();
			ConditionFormula[] isNewSubFormulas = new ConditionFormula[subFormulas.size()];
			for(int subFormula = 0; subFormula < isNewSubFormulas.length; ++subFormula){
				isNewSubFormulas[subFormula] = this.isNewFormula(subFormulas.get(subFormula));
			}
			return and(isNewSubFormulas);
		case DISJUNCTIVE:
			DisjunctiveConditionFormula disjunction = (DisjunctiveConditionFormula) discardedAttributesFormula;
			assert disjunction.subformulas().size() > 0;
			return disjunction.subformulas().get(0);
		case NEGATED:
			throw new IllegalArgumentException();
		default:
			throw new Error();
		}
	}
	
	protected Term<BasicType>[] streamBoundFunctions(final FixedArrayNavigableMap<String, StreamBound> streamBounds, final ArithmeticAnalyzer temporalAnalyser, final ConditionFormula expressionSBF, final ConditionFormula outputSBF){
		return this.streamBoundFunctions(this.streamBoundFunctionFormulas(streamBounds, temporalAnalyser, expressionSBF, outputSBF));
	}
	
	protected FixedArrayNavigableMap<String, ConditionFormula> streamBoundFunctionFormulas(final FixedArrayNavigableMap<String, StreamBound> streamBounds, final ArithmeticAnalyzer temporalAnalyser, final ConditionFormula expressionSBF, final ConditionFormula outputSBF){
		assert TemporalStreamSchema.isValidBufferStreamBoundFormula(outputSBF);
		ConditionFormula[] streamBoundFunctionFormulas = new ConditionFormula[streamBounds.size()];
		Arrays.fill(streamBoundFunctionFormulas, Top);
		for(ConditionFormula formula: outputSBF.collectFormulas(atomicFormulaFilter)){
			AtomicStreamBoundFormula streamBoundFormula = (AtomicStreamBoundFormula) formula;
			assert streamBoundFormula.bound instanceof StreamBound;
			int index = streamBounds.keys().indexOf(((StreamBound)streamBoundFormula.bound).qualified_name);
			if(index >= 0){
				assert streamBounds.values().get(index).equals(streamBoundFormula.bound);
				
				streamBoundFunctionFormulas[index] = and(streamBoundFunctionFormulas[index], this.streamBoundFunctionFormulaOutputSBFFirst(temporalAnalyser, expressionSBF, streamBoundFormula));
				//TODO correct/optimal?
			}
		}
		
		for(int streamBoundIndex = 0; streamBoundIndex < streamBoundFunctionFormulas.length; ++streamBoundIndex){
			streamBoundFunctionFormulas[streamBoundIndex] = streamBoundFunctionFormulas[streamBoundIndex].flaten();
		}
		return new FixedArrayNavigableMap<String, ConditionFormula>(streamBounds.keys(), new FixedArray<ConditionFormula>(false, streamBoundFunctionFormulas));
	}
	
	protected ConditionFormula streamBoundFunctionFormulaOutputSBFFirst(final ArithmeticAnalyzer temporalAnalyser, final ConditionFormula expressionSBF, final ConditionFormula outputSBF){
		switch (outputSBF.ID) {
		case TOP: return (temporalAnalyser.isConsistent() & expressionSBF.ID != TOP ? Bottom : Top);
		case BOTTOM: return Top;
		case ATOMIC:
			if(expressionSBF.ID == ConditionFormulaID.ATOMIC){
				if(expressionSBF instanceof AtomicStreamBoundFormula & outputSBF instanceof AtomicStreamBoundFormula){
					return this.streamBoundFunctionFormula(temporalAnalyser, (AtomicStreamBoundFormula) expressionSBF, (AtomicStreamBoundFormula) outputSBF);
				}
				else{
					throw new IllegalArgumentException("Stream bound formulas may only contain instances of " + AtomicStreamBoundFormula.class.getSimpleName() + ".");
				}
			}
			else{
				return this.streamBoundFunctionFormulaInputSBFFirst(temporalAnalyser, expressionSBF, outputSBF);
			}
		case CONJUNCTIVE:
			return or(new CollectionAdapter<ConditionFormula, ConditionFormula>(((ConjunctiveConditionFormula)outputSBF).subformulas()) {
				@Override
				public ConditionFormula adapt(final ConditionFormula outputSBFSubFormula) {
					return TSA_DefaultDifferencer.this.streamBoundFunctionFormulaOutputSBFFirst(temporalAnalyser, expressionSBF, outputSBFSubFormula);
				}
			});
		case DISJUNCTIVE:
			return and(new CollectionAdapter<ConditionFormula, ConditionFormula>(((DisjunctiveConditionFormula)outputSBF).subformulas()) {
				@Override
				public ConditionFormula adapt(final ConditionFormula outputSBFSubFormula) {
					return TSA_DefaultDifferencer.this.streamBoundFunctionFormulaOutputSBFFirst(temporalAnalyser, expressionSBF, outputSBFSubFormula);
				}
			});
		case NEGATED:
			throw new IllegalArgumentException("Stream bound formulas may not contain negation.");
		default: throw new Error("Unexpected case \"" + expressionSBF.ID.name() + "\"");
		}
	}
	
	protected ConditionFormula streamBoundFunctionFormulaInputSBFFirst(final ArithmeticAnalyzer temporalAnalyser, final ConditionFormula inputSBF, final ConditionFormula outputSBF){
		switch (inputSBF.ID) {
		case TOP: return Top;
		case BOTTOM: return (temporalAnalyser.isConsistent() & outputSBF.ID != BOTTOM ? Bottom : Top);//
		case ATOMIC:
			if(outputSBF.ID == ConditionFormulaID.ATOMIC){
				if(inputSBF instanceof AtomicStreamBoundFormula & outputSBF instanceof AtomicStreamBoundFormula){
					return this.streamBoundFunctionFormula(temporalAnalyser, (AtomicStreamBoundFormula) inputSBF, (AtomicStreamBoundFormula) outputSBF);
				}
				else{
					throw new IllegalArgumentException("Stream bound formulas may only contain instances of " + AtomicStreamBoundFormula.class.getSimpleName() + ".");
				}
			}
			else{
				return this.streamBoundFunctionFormulaOutputSBFFirst(temporalAnalyser, inputSBF, outputSBF);
			}
		case CONJUNCTIVE:
			return and(new CollectionAdapter<ConditionFormula, ConditionFormula>(((ConjunctiveConditionFormula)inputSBF).subformulas()) {
				@Override
				public ConditionFormula adapt(final ConditionFormula inputSBFSubFormula) {
					return TSA_DefaultDifferencer.this.streamBoundFunctionFormulaInputSBFFirst(temporalAnalyser, inputSBFSubFormula, outputSBF);
				}
			});
		case DISJUNCTIVE:
			return or(new CollectionAdapter<ConditionFormula, ConditionFormula>(((DisjunctiveConditionFormula)inputSBF).subformulas()) {
				@Override
				public ConditionFormula adapt(final ConditionFormula inputSBFSubFormula) {
					return TSA_DefaultDifferencer.this.streamBoundFunctionFormulaInputSBFFirst(temporalAnalyser, inputSBFSubFormula, outputSBF);
				}
			});
		case NEGATED:
			throw new IllegalArgumentException("Stream bound formulas may not contain negation.");
		default: throw new Error("Unexpected case \"" + inputSBF.ID.name() + "\"");
		}
	}
	
	protected final ArithmeticRelationFormulaExtractor temporalFormulaExtractor = new ArithmeticRelationFormulaExtractor(TIMESTAMP.type);//FIXME
	
	@SuppressWarnings("unchecked")
	protected ConditionFormula streamBoundFunctionFormula(final ArithmeticAnalyzer temporalAnalyser, final AtomicStreamBoundFormula inputSBF, final AtomicStreamBoundFormula outputSBF){
		if(inputSBF.atomicTerm.isTyped & outputSBF.atomicTerm.isTyped){
			assert inputSBF.atomicTerm.type instanceof BasicType;
			assert outputSBF.atomicTerm instanceof BasicTypeAttribute;
			assert outputSBF.bound instanceof StreamBound;
			return this.streamBoundFunctionFormula(temporalAnalyser, (AtomicTerm<? extends BasicType>) inputSBF.atomicTerm, inputSBF.bound, (BasicTypeAttribute) outputSBF.atomicTerm, (StreamBound) outputSBF.bound);
		}
		else{
			throw new IllegalArgumentException();
		}
	}
	
	protected ConditionFormula streamBoundFunctionFormula(final ArithmeticAnalyzer temporalAnalyser, final AtomicTerm<? extends BasicType> inputAttribute, final Term<BasicType> inputStreamBound, final BasicTypeAttribute outputAttribute, final StreamBound outputStreamBound){
		assert inputAttribute.type.ID == inputStreamBound.type.ID;
		assert outputAttribute.type.ID == outputStreamBound.type.ID;
		if(inputStreamBound.type.ID == outputStreamBound.type.ID){
			boolean isAttribute;
			String inputAttributeName;
			if(inputAttribute.ID == ATTRIBUTE){
				isAttribute = true;
				inputAttributeName = ((BasicTypeAttribute)inputAttribute).name;
			}
			else if(inputAttribute.ID == VARIABLE){
				isAttribute = false;
				inputAttributeName = ((BasicTypeVariable)inputAttribute).name;
			}
			else{
				throw new IllegalArgumentException();
			}
			long distance = temporalAnalyser.distance(outputAttribute.name, true, inputAttributeName, isAttribute);
			if(distance == ArithmeticAnalyzer.PLUS_INFTY){
				return Bottom;
			}
			else if(distance == ArithmeticAnalyzer.MINUS_INFTY){
				return Top;
			}
			else{
				long constant = ArithmeticAnalyzer.constant(distance) - (ArithmeticAnalyzer.minusEpsilon(distance) ? 1 : 0);
				if(constant == 0){
					return new BinaryComparissonFormula(LESS_EQUAL, outputStreamBound, inputStreamBound);
				}
				else{
					BasicTypeValue distanceValue = this.temporalFormulaExtractor.value(this.temporalFormulaExtractor.durationType(inputStreamBound.type).ID, constant);
					return new BinaryComparissonFormula(LESS_EQUAL, outputStreamBound, apply(MINUS, inputStreamBound.type, inputStreamBound, distanceValue));
				}
			}
		}
		else{
			return Bottom;
		}		
	}
	
	public DTSA_Expression differentiateExpression(final TSA_CompositeExpression tsa_compositeExpression, final DTSA_QueryType type, final boolean monotonic, final boolean monotonicInputStreamBounds, final FixedEnumMap<DTSA_BufferType, FixedArrayNavigableMap<String, DTSA_StreamBuffer>> streamBuffers) throws TSA_Exception{
		try{
			switch (tsa_compositeExpression.operator.ID) {
			case BASIC: 
			{
				switch (((TSA_BasicOperator)tsa_compositeExpression.operator).ID) {
				case CROSSPRODUCT: 
				{				
					assert tsa_compositeExpression.subExpressions.size() > 0;
					boolean containsAnyIsNewAttribute = false;
					ArrayList<BooleanTerm> isNewAttributes = new ArrayList<BooleanTerm>();
					DTSA_Expression[] differentiatedSubExpressions = new DTSA_Expression[tsa_compositeExpression.subExpressions.size()];
					for (int subExpression = 0; subExpression < differentiatedSubExpressions.length; ++subExpression) {
						differentiatedSubExpressions[subExpression] = this.differentiateExpression(tsa_compositeExpression.subExpressions.get(subExpression), type, monotonic, monotonicInputStreamBounds, streamBuffers);
						if(differentiatedSubExpressions[subExpression].schema().hasAttribute(isNewAttributeName)){
							containsAnyIsNewAttribute = true;
							String isNewAttributeNewName = isNewAttributeName + subExpression;
							isNewAttributes.add(new BooleanTerm(BooleanAttribute.create(isNewAttributeNewName)));
							differentiatedSubExpressions[subExpression] = new DTSA_CompositeExpression(new TSA_RenameOperator(tsa_compositeExpression.operator.origin, new FixedArrayNavigableMap<String, String>(new String[]{isNewAttributeName}, new String[]{isNewAttributeNewName})), differentiatedSubExpressions[subExpression]);
						}
					}
					if(containsAnyIsNewAttribute){
						TSA_ImbeddingOperator isNewImbedding = new TSA_ImbeddingOperator(tsa_compositeExpression.operator.origin, new FixedArrayNavigableMap<TypedAttribute<?>, Term<?>>(new BooleanAttribute[]{isNewAttribute}, new Term<?>[]{and(isNewAttributes)}));
						TSA_ProjectionOperator isNewProjection = new TSA_ProjectionOperator(tsa_compositeExpression.operator.origin, FixedArrayNavigableSet.merge(tsa_compositeExpression.schema().basicAttributes.keys(), new FixedArrayNavigableSet<String>(isNewAttributeName)));
						return new DTSA_CompositeExpression(isNewProjection, new DTSA_CompositeExpression(isNewImbedding, new DTSA_CompositeExpression(tsa_compositeExpression.operator, new FixedArray<DTSA_Expression>(false, differentiatedSubExpressions))));
					}
					else{
						return new DTSA_CompositeExpression(tsa_compositeExpression.operator, new FixedArray<DTSA_Expression>(false, differentiatedSubExpressions));
					}
				}
				case UNION: 
				{				
					assert tsa_compositeExpression.subExpressions.size() > 0;
					DTSA_Expression[] differentiatedSubExpressions = new DTSA_Expression[tsa_compositeExpression.subExpressions.size()];
					for (int subExpression = 0; subExpression < differentiatedSubExpressions.length; ++subExpression) {
						differentiatedSubExpressions[subExpression] = this.differentiateExpression(tsa_compositeExpression.subExpressions.get(subExpression), type, monotonic, monotonicInputStreamBounds, streamBuffers);
					}
					return new DTSA_CompositeExpression(tsa_compositeExpression.operator, new FixedArray<DTSA_Expression>(false, differentiatedSubExpressions));
				}
				case GROUPING:
				{
					assert tsa_compositeExpression.subExpressions.size() == 1;
					assert monotonic == false | tsa_compositeExpression.isStatic;
					return new DTSA_CompositeExpression(tsa_compositeExpression.operator, this.differentiateExpression(tsa_compositeExpression.subExpressions.get(0), type, monotonic, monotonicInputStreamBounds, streamBuffers));
				}
				case IMBEDDING:
				{
					assert tsa_compositeExpression.subExpressions.size() == 1;
					return new DTSA_CompositeExpression(tsa_compositeExpression.operator, this.differentiateExpression(tsa_compositeExpression.subExpressions.get(0), type, monotonic, monotonicInputStreamBounds, streamBuffers));
				}
				case PROJECTION:
				{ 
					assert tsa_compositeExpression.subExpressions.size() == 1;	
					DTSA_Expression differenciatedSubExpression = this.differentiateExpression(tsa_compositeExpression.subExpressions.get(0), type, monotonic, monotonicInputStreamBounds, streamBuffers);
					TSA_ProjectionOperator projection = (TSA_ProjectionOperator) tsa_compositeExpression.operator;
					
					if(monotonic & monotonicInputStreamBounds & type !=  DTSA_QueryType.STATIC & type != DTSA_QueryType.INIT){
						
						TSA_ProjectionOperator relaxedProjection = new TSA_ProjectionOperator(projection.origin, FixedArrayNavigableSet.merge(projection.attributes, new FixedArrayNavigableSet<String>(isNewAttributeName)));
					
						ConditionFormula differentiatedSubExpressionStreamBoundFormula = not(not(this.streamBoundFormula(differenciatedSubExpression.schema().relationConstraints(), differenciatedSubExpression.origin)).focus(monotonicStreamBoundFilterWithVariables)).flaten();
						assert differentiatedSubExpressionStreamBoundFormula.ID != BOTTOM;
						ConditionFormula expressionStreamBoundFormula = not(not(this.streamBoundFormula(tsa_compositeExpression.schema().relationConstraints(), tsa_compositeExpression.origin)).focus(monotonicStreamBoundFilterWithVariables)).flaten();
						assert expressionStreamBoundFormula.ID != BOTTOM;
						
						//replacing variables of discarded attributes by remaining attributes with the same value
						
						FixedArrayNavigableSet<Attribute<?>> availableRelevantAttributes = differentiatedSubExpressionStreamBoundFormula.Attributes();
						FixedArrayNavigableSet<Attribute<?>> remainingRelevantAttributes = expressionStreamBoundFormula.Attributes();
						assert FixedArrayNavigableSet.difference(remainingRelevantAttributes, availableRelevantAttributes).size() == 0;
						
						FixedArrayNavigableSet<Attribute<?>> discardedRelevantAttributes = FixedArrayNavigableSet.difference(availableRelevantAttributes, remainingRelevantAttributes);				
						
						boolean hasIsNewAttribute = differenciatedSubExpression.schema().hasAttribute(isNewAttributeName);
	
						if(discardedRelevantAttributes.size() > 0){
														
							final boolean[] isRelevantType = new boolean[BasicTypeID.values().length];
							final FixedArrayNavigableSet<String> discardedRelevantAttributeNames;
							{
								String[] discardedAttributeNamesArray = new String[discardedRelevantAttributes.size()];
								for(int attributeIndex = 0; attributeIndex < discardedAttributeNamesArray.length; ++attributeIndex){
									assert discardedRelevantAttributes.get(attributeIndex) instanceof BasicTypeAttribute;
									BasicTypeAttribute basicAttribute = (BasicTypeAttribute) discardedRelevantAttributes.get(attributeIndex);
									isRelevantType[basicAttribute.type.ID.ordinal()] = true;
									discardedAttributeNamesArray[attributeIndex] = basicAttribute.name;
								}
								discardedRelevantAttributeNames = new FixedArrayNavigableSet<String>(false, discardedAttributeNamesArray);
								assert discardedAttributeNamesArray.length == discardedRelevantAttributeNames.size();
							}
							
							ConditionFormula tupleConstraints = and(differenciatedSubExpression.schema().tupleConstraints());
							
							FixedArrayNavigableSet<String> potentiallyRelevantAttributes;
							{
								ArrayList<String> potentiallyRelevantAttributeList = new ArrayList<String>();
								for(Attribute<?> attribute: tupleConstraints.Attributes()){
									if(attribute instanceof BasicTypeAttribute){
										if(isRelevantType[((BasicTypeAttribute)attribute).type.ID.ordinal()]){
											potentiallyRelevantAttributeList.add(attribute.name);
										}
									}
								}
								potentiallyRelevantAttributes = new FixedArrayNavigableSet<String>(potentiallyRelevantAttributeList);
							}
							
							ArithmeticAnalyzer equalityAnalyser = this.analyserFactory.make(and(differenciatedSubExpression.schema().tupleConstraints()), FixedArrayNavigableSet.merge(discardedRelevantAttributeNames, potentiallyRelevantAttributes), new FixedArrayNavigableSet<String>());
							
							final FixedArrayNavigableMap<String, String> replaceableDiscardedRelevantAttributesAliasMapping;
							final FixedArrayNavigableSet<String> nonReplaceableDiscaredRelevantAttributeNames;
							//final FixedArrayNavigableMap<String, StreamBound> nonReplaceableDiscardedIdentifierAttributes;
							
							{
								ArrayList<String> replaceableDiscardedAttributesList = new ArrayList<String>();
								ArrayList<String> replaceableDiscardedAttributesAliasList = new ArrayList<String>();
								ArrayList<String> nonReplaceableDiscardedAttributesList = new ArrayList<String>();
								
								for(int discardedIdentifierAttribute = 0; discardedIdentifierAttribute < discardedRelevantAttributes.size(); ++discardedIdentifierAttribute){
									String discardedAttribute = discardedRelevantAttributeNames.get(discardedIdentifierAttribute);
									boolean replaceable = false;
									for(int remainingIdentifierAttribute = 0; remainingIdentifierAttribute < potentiallyRelevantAttributes.size(); ++remainingIdentifierAttribute){
										String remainingAttribute = potentiallyRelevantAttributes.get(remainingIdentifierAttribute);
										if(!discardedAttribute.equals(remainingAttribute) & equalityAnalyser.equal(discardedAttribute, true, remainingAttribute, true)){
											replaceable = true;
											replaceableDiscardedAttributesList.add(discardedAttribute);
											replaceableDiscardedAttributesAliasList.add(remainingAttribute);
											break;
										}
									}
									if(!replaceable){
										nonReplaceableDiscardedAttributesList.add(discardedAttribute);
									}
								}
								
								replaceableDiscardedRelevantAttributesAliasMapping = new FixedArrayNavigableMap<String, String>(new FixedArrayNavigableSet<String>(false, replaceableDiscardedAttributesList), new FixedArray<String>(replaceableDiscardedAttributesAliasList));
								nonReplaceableDiscaredRelevantAttributeNames = new FixedArrayNavigableSet<String>(false, nonReplaceableDiscardedAttributesList);
							}
							
							TemporalStreamSchema modifiedSchema;
							try {
								if(replaceableDiscardedRelevantAttributesAliasMapping.size() > 0){
									modifiedSchema = TemporalStreamSchema.create(differenciatedSubExpression.schema().ID, differenciatedSubExpression.schema().basicAttributes, differenciatedSubExpression.schema().tupleConstraints(), and(differenciatedSubExpression.schema().relationConstraints()).replaceAtomicTerm(new RenameAttributes(replaceableDiscardedRelevantAttributesAliasMapping)).subformulas(), false);
								}
								else{
									modifiedSchema = differenciatedSubExpression.schema();
								}
							} catch (TSA_SchemaException e) {
								throw new TSA_Exception(projection.origin, e);
							}
							
							DTSA_Expression schemaModifiedDifferentiatedSubExpression = differenciatedSubExpression.replaceSchema(modifiedSchema);
							
							Adapter<AtomicConditionFormula, AtomicConditionFormula> nonReplaceableDiscaredRelevantAttributesAdapter = new Adapter<AtomicConditionFormula, AtomicConditionFormula>() {
								
								@Override
								public AtomicConditionFormula adapt(final AtomicConditionFormula element) {
									if(element instanceof AtomicStreamBoundFormula){
										AtomicStreamBoundFormula sbf = (AtomicStreamBoundFormula) element;
										if(sbf.atomicTerm.ID == ATTRIBUTE){
											if(nonReplaceableDiscaredRelevantAttributeNames.contains(((BasicTypeAttribute)sbf.atomicTerm).name)){
												if(sbf.bound instanceof StreamBound){
													return new AtomicStreamBoundFormula((StreamBound) sbf.bound, ((BasicTypeAttribute)sbf.atomicTerm), false);
												}
												else if(sbf.bound instanceof ClockFunction){
													return new AtomicStreamBoundFormula((ClockFunction) sbf.bound, ((BasicTypeAttribute)sbf.atomicTerm), false);
												}
												else{
													throw new Error("Unexpected case");
												}
											}
										}
									}
									return element;
								}
							};
							
							//modified schema after applying the projection
							final boolean requiresAnyUnreplaceableRelevantAttribute = not(not(this.streamBoundFormula(modifiedSchema.relationConstraints(), tsa_compositeExpression.origin).replaceAtomicFormula(nonReplaceableDiscaredRelevantAttributesAdapter)).focus(monotonicStreamBoundFilterWithVariables)).flaten().ID == BOTTOM;
							
							if(requiresAnyUnreplaceableRelevantAttribute){
								ConditionFormula schemaModifiedDifferentiatedSubExpressionStreamBoundFormula = not(not(this.streamBoundFormula(modifiedSchema.relationConstraints(), tsa_compositeExpression.origin)).focus(monotonicStreamBoundFilterWithVariables)).flaten();
								
								final FixedArrayNavigableSet<String> nonReplaceableRequiredDiscaredRelevantAttributeNames;
								final FixedArrayNavigableSet<String> nonReplaceableUnrequiredDiscaredRelevantAttributeNames;
								{									
									ArrayList<String> nonReplaceableRequiredDiscaredRelevantAttributeNamesList = new ArrayList<String>();
									ArrayList<String> nonReplaceableUnrequiredDiscaredRelevantAttributeNamesList = new ArrayList<String>();
									ConditionFormula reducedStreamBoundFormula = schemaModifiedDifferentiatedSubExpressionStreamBoundFormula;
									for(final String nonReplaceableAttribute: nonReplaceableDiscaredRelevantAttributeNames){
										AbstractFilter<AtomicConditionFormula> nonReplaceableAttributeFilter = new AbstractFilter<AtomicConditionFormula>() {
											
											@Override
											public boolean accepts(final AtomicConditionFormula element) {
												if(element instanceof AtomicStreamBoundFormula){
													AtomicStreamBoundFormula sbf = (AtomicStreamBoundFormula) element;
													if(sbf.atomicTerm.ID == ATTRIBUTE){
														return nonReplaceableAttribute.equals(((Attribute<?>)sbf.atomicTerm).name);
													}
												}
												return false;
											}
										};
										
										ConditionFormula furtherReducedStreamBoundFormula = not(not(reducedStreamBoundFormula).focus(notFilter(nonReplaceableAttributeFilter))).flaten();
										if(furtherReducedStreamBoundFormula.ID == BOTTOM){
											nonReplaceableRequiredDiscaredRelevantAttributeNamesList.add(nonReplaceableAttribute);
										}
										else{
											nonReplaceableUnrequiredDiscaredRelevantAttributeNamesList.add(nonReplaceableAttribute);
											reducedStreamBoundFormula = furtherReducedStreamBoundFormula;
										}
									}
									nonReplaceableRequiredDiscaredRelevantAttributeNames = new FixedArrayNavigableSet<String>(false, false, nonReplaceableRequiredDiscaredRelevantAttributeNamesList.toArray(new String[nonReplaceableRequiredDiscaredRelevantAttributeNamesList.size()]));
									nonReplaceableUnrequiredDiscaredRelevantAttributeNames = new FixedArrayNavigableSet<String>(false, false, nonReplaceableUnrequiredDiscaredRelevantAttributeNamesList.toArray(new String[nonReplaceableUnrequiredDiscaredRelevantAttributeNamesList.size()]));
								}								
								
								AbstractFilter<AtomicConditionFormula> nonReplaceableRequiredDiscaredRelevantAttributesFilter = new AbstractFilter<AtomicConditionFormula>() {
									
									@Override
									public boolean accepts(final AtomicConditionFormula element) {
										if(element instanceof AtomicStreamBoundFormula){
											AtomicStreamBoundFormula sbf = (AtomicStreamBoundFormula) element;
											if(sbf.atomicTerm.ID == ATTRIBUTE){
												return nonReplaceableRequiredDiscaredRelevantAttributeNames.contains(((Attribute<?>)sbf.atomicTerm).name);
											}
										}
										return false;
									}
								};
								
								AbstractFilter<AtomicConditionFormula> nonReplaceableUnrequiredDiscaredRelevantAttributesFilter = new AbstractFilter<AtomicConditionFormula>() {
									
									@Override
									public boolean accepts(final AtomicConditionFormula element) {
										if(element instanceof AtomicStreamBoundFormula){
											AtomicStreamBoundFormula sbf = (AtomicStreamBoundFormula) element;
											if(sbf.atomicTerm.ID == ATTRIBUTE){
												return nonReplaceableUnrequiredDiscaredRelevantAttributeNames.contains(((Attribute<?>)sbf.atomicTerm).name);
											}
										}
										return false;
									}
								};
								
								ConditionFormula nonReplaceableRequiredRequiredRelevantAttributesFormula = not(not(schemaModifiedDifferentiatedSubExpressionStreamBoundFormula).focus(notFilter(nonReplaceableUnrequiredDiscaredRelevantAttributesFilter))).focus(nonReplaceableRequiredDiscaredRelevantAttributesFilter).flaten(); 
								ConditionFormula isNewFormula = this.isNewFormula(hasIsNewAttribute, nonReplaceableRequiredRequiredRelevantAttributesFormula).flaten();
								assert isNewFormula != Bottom;
								
								try {
									Adapter<AtomicConditionFormula, AtomicConditionFormula> nonReplaceableUnrequiredDiscaredRelevantAttributesAdapter = new Adapter<AtomicConditionFormula, AtomicConditionFormula>() {
										
										@Override
										public AtomicConditionFormula adapt(final AtomicConditionFormula element) {
											if(element instanceof AtomicStreamBoundFormula){
												AtomicStreamBoundFormula sbf = (AtomicStreamBoundFormula) element;
												if(sbf.atomicTerm.ID == ATTRIBUTE){
													if(nonReplaceableUnrequiredDiscaredRelevantAttributeNames.contains(((BasicTypeAttribute)sbf.atomicTerm).name)){
														if(sbf.bound instanceof StreamBound){
															return new AtomicStreamBoundFormula((StreamBound) sbf.bound, ((BasicTypeAttribute)sbf.atomicTerm), false);
														}
														else if(sbf.bound instanceof ClockFunction){
															return new AtomicStreamBoundFormula((ClockFunction) sbf.bound, ((BasicTypeAttribute)sbf.atomicTerm), false);
														}
														else{
															throw new Error("Unexpected case");
														}
													}
												}
											}
											return element;
										}
									};
									
									modifiedSchema = TemporalStreamSchema.create(modifiedSchema.ID, modifiedSchema.attributes, modifiedSchema.tupleConstraints(), and(modifiedSchema.relationConstraints()).replaceAtomicFormula(nonReplaceableUnrequiredDiscaredRelevantAttributesAdapter).subformulas(), false);
								} catch (TSA_SchemaException e) {
									throw new TSA_Exception(projection.origin, e);
								}
								schemaModifiedDifferentiatedSubExpression = schemaModifiedDifferentiatedSubExpression.replaceSchema(modifiedSchema);
								
								//FIXME modify stream bound formula in schema and 
															
								if(hasIsNewAttribute){
									FixedArrayNavigableMap<String, String> mapping = new FixedArrayNavigableMap<String, String>(new String[]{isNewAttributeName}, new String[]{isNewAttributeAliasName});
									try {
										isNewFormula = isNewFormula.replaceAtomicTerm(new RenameAttributes(mapping));
									} catch (TSA_SchemaException e) {
										throw new TSA_Exception(projection.origin, e);
									}
									TSA_ImbeddingOperator isNewImbedding = new TSA_ImbeddingOperator(projection.origin, new FixedArrayNavigableMap<TypedAttribute<?>, Term<?>>(new BooleanAttribute[]{isNewAttribute}, new Term<?>[]{isNewFormula}));
									return new DTSA_CompositeExpression(relaxedProjection, new DTSA_CompositeExpression(isNewImbedding, new DTSA_CompositeExpression(new TSA_RenameOperator(projection.origin, mapping), schemaModifiedDifferentiatedSubExpression)));
								}
								else{
									TSA_ImbeddingOperator isNewImbedding = new TSA_ImbeddingOperator(projection.origin, new FixedArrayNavigableMap<TypedAttribute<?>, Term<?>>(new BooleanAttribute[]{isNewAttribute}, new Term<?>[]{isNewFormula}));
									return new DTSA_CompositeExpression(relaxedProjection, new DTSA_CompositeExpression(isNewImbedding, schemaModifiedDifferentiatedSubExpression));
								}
							}
							else{								
								try {
									modifiedSchema = TemporalStreamSchema.create(modifiedSchema.ID, modifiedSchema.attributes, modifiedSchema.tupleConstraints(), and(modifiedSchema.relationConstraints()).replaceAtomicFormula(nonReplaceableDiscaredRelevantAttributesAdapter).subformulas(), false);
								} catch (TSA_SchemaException e) {
									throw new TSA_Exception(projection.origin, e);
								}
								schemaModifiedDifferentiatedSubExpression = schemaModifiedDifferentiatedSubExpression.replaceSchema(modifiedSchema);
								
								if(hasIsNewAttribute){								
									return new DTSA_CompositeExpression(relaxedProjection, schemaModifiedDifferentiatedSubExpression);
								}
								else{
									return new DTSA_CompositeExpression(projection, schemaModifiedDifferentiatedSubExpression);
								}
							}		
						}
						else{
							if(hasIsNewAttribute){
								return new DTSA_CompositeExpression(relaxedProjection, differenciatedSubExpression);
							}
							else{
								return new DTSA_CompositeExpression(projection, differenciatedSubExpression);
							}				
						}
					}
					else{
						assert !monotonic | !monotonicInputStreamBounds | type == DTSA_QueryType.STATIC | type == DTSA_QueryType.INIT;
						assert !differenciatedSubExpression.schema().hasAttribute(isNewAttributeName);
						return new DTSA_CompositeExpression(projection, differenciatedSubExpression);
					}
					
				}
				case SELECTION:
				{
					assert tsa_compositeExpression.subExpressions.size() == 1;
					TSA_SelectionOperator selectionOperator = (TSA_SelectionOperator) tsa_compositeExpression.operator;
					DTSA_Expression differenciatedSubExpression = this.differentiateExpression(tsa_compositeExpression.subExpressions.get(0), type, monotonic, monotonicInputStreamBounds, streamBuffers);
					if(and(selectionOperator.conditions).flaten().ID == TOP){
						return differenciatedSubExpression;
					}
					else{
						return new DTSA_CompositeExpression(tsa_compositeExpression.operator, differenciatedSubExpression);
					}
				}	
				case SETDIFFERENCE:
				{
					assert tsa_compositeExpression.subExpressions.size() == 2;
					assert monotonic == false | tsa_compositeExpression.isStatic;
					return new DTSA_CompositeExpression(tsa_compositeExpression.operator, new FixedArray<DTSA_Expression>(false, this.differentiateExpression(tsa_compositeExpression.subExpressions.get(0), type, monotonic, monotonicInputStreamBounds, streamBuffers), this.differentiateExpression(tsa_compositeExpression.subExpressions.get(1), type, monotonic, monotonicInputStreamBounds, streamBuffers)));
				}
				default:
					throw new Error("Unexpected case");
				}
			}
			case COMPLEX: 
			{
				switch (((TSA_ComplexOperator)tsa_compositeExpression.operator).ID) {
				case RENAME: {
					assert tsa_compositeExpression.subExpressions.size() == 1;
					DTSA_Expression differenciatedSubExpression = this.differentiateExpression(tsa_compositeExpression.subExpressions.get(0), type, monotonic, monotonicInputStreamBounds, streamBuffers);
					TSA_RenameOperator renameOperator = (TSA_RenameOperator) tsa_compositeExpression.operator;
					boolean isPrefix = renameOperator.mapping.keys().contains("");
					
					if(isPrefix & differenciatedSubExpression.schema().hasAttribute(isNewAttributeName)){
						FixedArrayNavigableSet<String> attributeNames = differenciatedSubExpression.schema().basicAttributes.keys();
						String[] newAttributeNames = new String[attributeNames.size()];
						String prefix = renameOperator.mapping.get("");
						
						for(int attribute = 0; attribute < newAttributeNames.length; ++attribute){
							newAttributeNames[attribute] = TSA_SyntaxPatterns.prefixAttribute(prefix, attributeNames.get(attribute));
						}
						
						FixedArrayNavigableMap<String, String> modifiedMapping = new FixedArrayNavigableMap<String, String>(attributeNames, new FixedArray<String>(false, newAttributeNames)); 
						TSA_RenameOperator modifiedRenameOperator = new TSA_RenameOperator(renameOperator.origin, modifiedMapping);
						return new DTSA_CompositeExpression(modifiedRenameOperator, differenciatedSubExpression);
					}
					else{
						//rename operator is not allowed to explicitly affect any attribute starting with .tsa
						return new DTSA_CompositeExpression(tsa_compositeExpression.operator, differenciatedSubExpression);
					}
				}
				case GROUPING_AND_AGGREGATION: {
					assert tsa_compositeExpression.subExpressions.size() == 1;
					assert monotonic == false;
					return new DTSA_CompositeExpression(tsa_compositeExpression.operator, this.differentiateExpression(tsa_compositeExpression.subExpressions.get(0), type, monotonic, monotonicInputStreamBounds, streamBuffers));
				}
				case DISTINCT_PROJECTION: {
					assert tsa_compositeExpression.subExpressions.size() == 1;
					assert monotonic == false;
					return new DTSA_CompositeExpression(tsa_compositeExpression.operator, this.differentiateExpression(tsa_compositeExpression.subExpressions.get(0), type, monotonic, monotonicInputStreamBounds, streamBuffers));
				}
				case ANTI_SEMI_JOIN: {
					assert tsa_compositeExpression.subExpressions.size() == 2;
					assert monotonic == false;
					return new DTSA_CompositeExpression(tsa_compositeExpression.operator, this.differentiateExpression(tsa_compositeExpression.subExpressions.get(0), type, monotonic, monotonicInputStreamBounds, streamBuffers));
				}
				case SEMI_JOIN: {
					assert tsa_compositeExpression.subExpressions.size() == 2;
					assert monotonic == false;
					return new DTSA_CompositeExpression(tsa_compositeExpression.operator, this.differentiateExpression(tsa_compositeExpression.subExpressions.get(0), type, monotonic, monotonicInputStreamBounds, streamBuffers));
				}
				default:
					throw new Error("Unexpected case");
				}
			}
			case COMPOSITE: 
			{
				//FIXME
				throw new Error("Not implemented case");
				//break;
			}
			default:
				throw new Error("Unexpected case");
			}
		}
		catch (TSA_Exception e) {
			throw e;
		}
		catch (RuntimeException e) {
			throw new TSA_Exception(tsa_compositeExpression.operator.origin, e);
		}
	}
	
	protected static final Logger logger = LoggerFactory.getLogger(TSA_DefaultDifferencer.class);
	
	@Override
	protected Logger logger(){
		return logger;
	}
	
	/**
	 * Relevance formula's unit
	 */
	
	protected FixedArrayNavigableMap<String, ConditionFormula> relevanceBoundFunctionFormulas(final FixedArrayNavigableMap<String, StreamBound> streamBounds, final ArithmeticAnalyzer temporalAnalyser, final ConditionFormula expressionSBF, final ConditionFormula outputSBF){
		assert TemporalStreamSchema.isValidBufferStreamBoundFormula(outputSBF);
		
		ConditionFormula[] relevanceBoundFunctionFormulas = new ConditionFormula[streamBounds.size()];
		Arrays.fill(relevanceBoundFunctionFormulas, Top);
		for(ConditionFormula formula: expressionSBF.collectFormulas(atomicFormulaFilter)){
			AtomicStreamBoundFormula relevanceBoundFormula = (AtomicStreamBoundFormula) formula;
			assert relevanceBoundFormula.bound instanceof StreamBound;
			
			int index = streamBounds.keys().indexOf(((StreamBound)relevanceBoundFormula.bound).qualified_name);
			if(index >= 0){
				assert streamBounds.values().get(index).equals(relevanceBoundFormula.bound);
				
				relevanceBoundFunctionFormulas[index] = and(relevanceBoundFunctionFormulas[index], this.relevanceBoundFunctionFormulaOutputSBFFirst(temporalAnalyser, relevanceBoundFormula, outputSBF));
				//TODO correct/optimal?
			}
		}
		
		for(int relevanceBoundIndex = 0; relevanceBoundIndex < relevanceBoundFunctionFormulas.length; ++relevanceBoundIndex){
			relevanceBoundFunctionFormulas[relevanceBoundIndex] = relevanceBoundFunctionFormulas[relevanceBoundIndex].flaten();
		}
		return new FixedArrayNavigableMap<String, ConditionFormula>(streamBounds.keys(), new FixedArray<ConditionFormula>(false, relevanceBoundFunctionFormulas));
	}
	
	protected ConditionFormula relevanceBoundFunctionFormulaOutputSBFFirst(final ArithmeticAnalyzer temporalAnalyser, final ConditionFormula expressionSBF, final ConditionFormula outputSBF){
		switch (outputSBF.ID) {
		case TOP: return (temporalAnalyser.isConsistent() & expressionSBF.ID != TOP ? Bottom : Top);
		case BOTTOM: return Top;
		case ATOMIC:
			if(expressionSBF.ID == ConditionFormulaID.ATOMIC){
				if(expressionSBF instanceof AtomicStreamBoundFormula & outputSBF instanceof AtomicStreamBoundFormula){
					return this.relevanceBoundFunctionFormula(temporalAnalyser, (AtomicStreamBoundFormula) expressionSBF, (AtomicStreamBoundFormula) outputSBF);
				}
				else{
					throw new IllegalArgumentException("Stream bound formulas may only contain instances of " + AtomicStreamBoundFormula.class.getSimpleName() + ".");
				}
			}
			else{
				return this.relevanceBoundFunctionFormulaExpressionSBFFirst(temporalAnalyser, expressionSBF, outputSBF);
			}
		case CONJUNCTIVE:
			return and(new CollectionAdapter<ConditionFormula, ConditionFormula>(((ConjunctiveConditionFormula)outputSBF).subformulas()) {
				@Override
				public ConditionFormula adapt(final ConditionFormula outputSBFSubFormula) {
					return TSA_DefaultDifferencer.this.relevanceBoundFunctionFormulaOutputSBFFirst(temporalAnalyser, expressionSBF, outputSBFSubFormula);
				}
			});
		case DISJUNCTIVE:
			return or(new CollectionAdapter<ConditionFormula, ConditionFormula>(((DisjunctiveConditionFormula)outputSBF).subformulas()) {
				@Override
				public ConditionFormula adapt(final ConditionFormula outputSBFSubFormula) {
					return TSA_DefaultDifferencer.this.relevanceBoundFunctionFormulaOutputSBFFirst(temporalAnalyser, expressionSBF, outputSBFSubFormula);
				}
			});
		case NEGATED:
			throw new IllegalArgumentException("Stream bound formulas may not contain negation.");
		default: throw new Error("Unexpected case \"" + expressionSBF.ID.name() + "\"");
		}
	}
	
	protected ConditionFormula relevanceBoundFunctionFormulaExpressionSBFFirst(final ArithmeticAnalyzer temporalAnalyser, final ConditionFormula expressionSBF, final ConditionFormula outputSBF){
		switch (expressionSBF.ID) {
		case TOP: return Top;
		case BOTTOM: return (temporalAnalyser.isConsistent() & outputSBF.ID != BOTTOM ? Bottom : Top);//
		case ATOMIC:
			if(outputSBF.ID == ConditionFormulaID.ATOMIC){
				if(expressionSBF instanceof AtomicStreamBoundFormula & outputSBF instanceof AtomicStreamBoundFormula){
					return this.relevanceBoundFunctionFormula(temporalAnalyser, (AtomicStreamBoundFormula) expressionSBF, (AtomicStreamBoundFormula) outputSBF);
				}
				else{
					throw new IllegalArgumentException("Stream bound formulas may only contain instances of " + AtomicStreamBoundFormula.class.getSimpleName() + ".");
				}
			}
			else{
				return this.relevanceBoundFunctionFormulaOutputSBFFirst(temporalAnalyser, expressionSBF, outputSBF);
			}
		case CONJUNCTIVE:
			return or(new CollectionAdapter<ConditionFormula, ConditionFormula>(((ConjunctiveConditionFormula)expressionSBF).subformulas()) {
				@Override
				public ConditionFormula adapt(final ConditionFormula expressionSBFSubFormula) {
					return TSA_DefaultDifferencer.this.relevanceBoundFunctionFormulaExpressionSBFFirst(temporalAnalyser, expressionSBFSubFormula, outputSBF);
				}
			});
		case DISJUNCTIVE:
			return and(new CollectionAdapter<ConditionFormula, ConditionFormula>(((DisjunctiveConditionFormula)expressionSBF).subformulas()) {
				@Override
				public ConditionFormula adapt(final ConditionFormula expressionSBFSubFormula) {
					return TSA_DefaultDifferencer.this.relevanceBoundFunctionFormulaExpressionSBFFirst(temporalAnalyser, expressionSBFSubFormula, outputSBF);
				}
			});
		case NEGATED:
			throw new IllegalArgumentException("Stream bound formulas may not contain negation.");
		default: throw new Error("Unexpected case \"" + expressionSBF.ID.name() + "\"");
		}
	}
	
	@SuppressWarnings("unchecked")
	protected ConditionFormula relevanceBoundFunctionFormula(final ArithmeticAnalyzer temporalAnalyser, final AtomicStreamBoundFormula expressionSBF, final AtomicStreamBoundFormula outputSBF){
		if(expressionSBF.atomicTerm.isTyped & outputSBF.atomicTerm.isTyped){
			assert expressionSBF.atomicTerm.type instanceof BasicType;
			assert expressionSBF.bound instanceof StreamBound;
			assert outputSBF.atomicTerm instanceof BasicTypeAttribute;
			assert outputSBF.bound instanceof StreamBound;
			return this.relevanceBoundFunctionFormula(temporalAnalyser, (AtomicTerm<? extends BasicType>) expressionSBF.atomicTerm, (StreamBound) expressionSBF.bound, (BasicTypeAttribute) outputSBF.atomicTerm, (StreamBound) outputSBF.bound);
		}
		else{
			throw new IllegalArgumentException();
		}
	}
	
	protected ConditionFormula relevanceBoundFunctionFormula(final ArithmeticAnalyzer temporalAnalyser, final AtomicTerm<? extends BasicType> expressionAttribute, final StreamBound expressionStreamBound, final BasicTypeAttribute outputAttribute, final StreamBound outputStreamBound){
		assert expressionAttribute.type.ID == expressionStreamBound.type.ID;
		assert outputAttribute.type.ID == outputStreamBound.type.ID;
		if(expressionStreamBound.type.ID == outputStreamBound.type.ID){
			boolean isAttribute;
			String expressionAttributeName;
			if(expressionAttribute.ID == ATTRIBUTE){
				isAttribute = true;
				expressionAttributeName = ((BasicTypeAttribute)expressionAttribute).name;
			}
			else if(expressionAttribute.ID == VARIABLE){
				isAttribute = false;
				expressionAttributeName = ((BasicTypeVariable)expressionAttribute).name;
			}
			else{
				throw new IllegalArgumentException();
			}
			long distance = temporalAnalyser.distance(expressionAttributeName, isAttribute, outputAttribute.name, true);
			if(distance == ArithmeticAnalyzer.PLUS_INFTY){
				return Bottom;
			}
			else if(distance == ArithmeticAnalyzer.MINUS_INFTY){
				return Top;
			}
			else{
				long constant = ArithmeticAnalyzer.constant(distance) - (ArithmeticAnalyzer.minusEpsilon(distance) ? 1 : 0);
				if(constant == 0){
					return new BinaryComparissonFormula(LESS_EQUAL, expressionStreamBound, outputStreamBound);
				}
				else{
					BasicTypeValue distanceValue = this.temporalFormulaExtractor.value(this.temporalFormulaExtractor.durationType(expressionStreamBound.type).ID, constant);
					return new BinaryComparissonFormula(LESS_EQUAL, expressionStreamBound, apply(MINUS, outputStreamBound.type, outputStreamBound, distanceValue));
				}
			}
		}
		else{
			return Bottom;
		}		
	}
	
	@SuppressWarnings("unchecked")
	protected Term<BasicType>[] relevanceBoundFunctions(final FixedArrayNavigableMap<String, ConditionFormula> relevanceBoundFunctionFormulas){				
		Term<BasicType>[] relevanceBoundFunctions = (Term<BasicType>[]) new Term<?>[relevanceBoundFunctionFormulas.size()]; 
		
		for(int outputMinTime = 0; outputMinTime < relevanceBoundFunctions.length; ++outputMinTime){
			String streamBound = relevanceBoundFunctionFormulas.keys().get(outputMinTime);
			ConditionFormula flatenedRelevanceBoundFormula = relevanceBoundFunctionFormulas.values().get(outputMinTime).flaten();
			relevanceBoundFunctions[outputMinTime] = this.relevanceBoundFunction(streamBound, flatenedRelevanceBoundFormula);
		}
		
		return relevanceBoundFunctions;
	}
	
	@SuppressWarnings("unchecked")
	protected Term<BasicType> relevanceBoundFunction(final String streamBound, final ConditionFormula outputRelevanceBoundFormula){
		assert outputRelevanceBoundFormula.isFlatened;
		switch(outputRelevanceBoundFormula.ID){
			case ATOMIC: assert outputRelevanceBoundFormula instanceof BinaryComparissonFormula;
				BinaryComparissonFormula bcf = (BinaryComparissonFormula) outputRelevanceBoundFormula;
				assert bcf.operator == LESS_EQUAL;
				assert bcf.secondTerm().isTyped;
				assert ((Term<Type>)bcf.secondTerm()).type.ID == TypeID.BASIC;
				assert bcf.firstTerm() instanceof StreamBound; 
				assert((StreamBound)bcf.firstTerm()).qualified_name.equals(streamBound);
				return (Term<BasicType>) bcf.secondTerm();
			case CONJUNCTIVE:{
				ConjunctiveConditionFormula conj = (ConjunctiveConditionFormula) outputRelevanceBoundFormula;
				Term<BasicType>[] subDefinitions = (Term<BasicType>[])new Term<?>[conj.subformulas().size()];
				for(int subFormula = 0; subFormula < subDefinitions.length; ++subFormula){
					subDefinitions[subFormula] = this.relevanceBoundFunction(streamBound, conj.subformulas().get(subFormula));
				}
				assert subDefinitions.length > 0;
				BasicType type = subDefinitions[0].type;
				return apply(LEAST, type, subDefinitions);
			}
			case DISJUNCTIVE:{
				DisjunctiveConditionFormula disj = (DisjunctiveConditionFormula) outputRelevanceBoundFormula;
				Term<BasicType>[] subDefinitions = (Term<BasicType>[])new Term<?>[disj.subformulas().size()];
				for(int subFormula = 0; subFormula < subDefinitions.length; ++subFormula){
					subDefinitions[subFormula] = this.relevanceBoundFunction(streamBound, disj.subformulas().get(subFormula));
				}
				assert subDefinitions.length > 0;
				BasicType type = subDefinitions[0].type;
				return apply(GREATEST, type, subDefinitions);
			}
			case NEGATED: throw new IllegalArgumentException("Invalid case. ");
			case BOTTOM: return Bottom;
			case TOP: return Top;
		}
		throw new Error("Unexpected Case: " + outputRelevanceBoundFormula);
	}
	
	protected Term<BasicType>[] relevanceBoundFunctions(final FixedArrayNavigableMap<String, StreamBound> streamBounds, final ArithmeticAnalyzer temporalAnalyser, final ConditionFormula expressionSBF, final ConditionFormula outputSBF){
		return this.relevanceBoundFunctions(this.relevanceBoundFunctionFormulas(streamBounds, temporalAnalyser, expressionSBF, outputSBF));
	}
	
	protected Term<BasicType>[] relevanceBoundFunctionsClockIncrease(final FixedArrayNavigableMap<String, StreamBound> streamBounds, final ArithmeticAnalyzer temporalAnalyser, final ConditionFormula expressionSBF, final FixedArrayNavigableMap<String, ClockFunction> clockAttributes, final FixedArrayNavigableMap<String, ClockFunction> clockVariables){
		return this.streamBoundFunctionsClockIncrease(this.relevanceBoundFunctionClockIncreaseFormulas(streamBounds, temporalAnalyser, expressionSBF, clockAttributes, clockVariables));
	}
	
	protected FixedArrayNavigableMap<String, ConditionFormula> relevanceBoundFunctionClockIncreaseFormulas(final FixedArrayNavigableMap<String, StreamBound> streamBounds, final ArithmeticAnalyzer temporalAnalyser, final ConditionFormula expressionSBF, final FixedArrayNavigableMap<String, ClockFunction> clockAttributes, final FixedArrayNavigableMap<String, ClockFunction> clockVariables){
		assert TemporalStreamSchema.isValidExpressionStreamBoundFormula(expressionSBF);
		//TODO stream bound assigned to multiple attributes
		ConditionFormula[] streamBoundFunctionClockFormulas = new ConditionFormula[streamBounds.size()];
	
		Arrays.fill(streamBoundFunctionClockFormulas, Top);
		for(ConditionFormula formula: expressionSBF.collectFormulas(atomicFormulaFilter)){
			AtomicStreamBoundFormula streamBoundFormula = (AtomicStreamBoundFormula) formula;
			assert streamBoundFormula.bound instanceof StreamBound;
			StreamBound streamBound = (StreamBound) streamBoundFormula.bound;
			
			int index = streamBounds.keys().indexOf(streamBound.qualified_name);
			if(index >= 0){
				assert streamBounds.values().get(index).equals(streamBound);

				boolean isAttribute = (streamBoundFormula.atomicTerm.ID == ATTRIBUTE);
				String inputAttributeName = (isAttribute ? ((BasicTypeAttribute) streamBoundFormula.atomicTerm).name : ((BasicTypeVariable) streamBoundFormula.atomicTerm).name);
				
				ArrayList<BinaryComparissonFormula> increasingClocks = new ArrayList<BinaryComparissonFormula>();
				for(String clockAttribute: clockAttributes.keys()){
					long distance = temporalAnalyser.distance(inputAttributeName, isAttribute, clockAttribute, true);
					if(distance < PLUS_INFTY){
						long constant = ArithmeticAnalyzer.constant(distance);
						ClockFunction clock = clockAttributes.get(clockAttribute);
						if(constant == 0){
							increasingClocks.add(new BinaryComparissonFormula(GREATER_EQUAL, streamBound, clock));
						}
						else{
							increasingClocks.add(new BinaryComparissonFormula(GREATER_EQUAL, streamBound, apply(MINUS, clock.type, clock, new DurationValue(constant))));
						}
					}
				}
				for(String clockVariable: clockVariables.keys()){
					long distance = temporalAnalyser.distance(inputAttributeName, isAttribute, clockVariable, false);
					if(distance < ArithmeticAnalyzer.PLUS_INFTY){
						long constant = ArithmeticAnalyzer.constant(distance);
						ClockFunction clock = clockVariables.get(clockVariable);
						if(constant == 0){
							increasingClocks.add(new BinaryComparissonFormula(GREATER_EQUAL, streamBound, clock));
						}
						else{
							increasingClocks.add(new BinaryComparissonFormula(GREATER_EQUAL, streamBound, apply(MINUS, clock.type, clock, new DurationValue(constant))));
						}
					}
				}
				
				if(increasingClocks.size() == 1){
					streamBoundFunctionClockFormulas[index] =  increasingClocks.first();
				}
				else{
					streamBoundFunctionClockFormulas[index] =  and(increasingClocks);//FIXME check which direction is used above	
				}
			}
		}		
		
		return new FixedArrayNavigableMap<String, ConditionFormula>(streamBounds.keys(), new FixedArray<ConditionFormula>(false, streamBoundFunctionClockFormulas));
	}
	
	/**
	 * End of relevance Formula's unit
	 */
}
