package de.lmu.ifi.pms.cep.tsa.conditions;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;

import de.lmu.ifi.pms.cep.tsa.schema.NoType;


public class BinaryComparissonFormula extends RelationFormula {
	
	public final BinaryComparissonOperatorID operator;
	
	protected BinaryComparissonFormula(@Attribute(name="name") final String name, 
			@Element(name="terms") final Tuple<NoType> terms){
		this(BinaryComparissonOperatorID.fromString(name), terms.arg(0), terms.arg(1));
	}

	public BinaryComparissonFormula(final BinaryComparissonOperatorID operator, final Term<?> term1, final Term<?> term2) {
		super(operator.toString(), term1, term2);
		this.operator = operator;
	}
	
	@Override
	protected RelationFormula newFormula(final Term<?>... arguments) {
		assert arguments.length == 2;
		return new BinaryComparissonFormula(this.operator, arguments[0], arguments[1]);
	}
	
	public Term<?> firstTerm(){
		return this.term(0);
	}
	
	public Term<?> secondTerm(){
		return this.term(1);
	}
	
	@Override
	public String toString() {
		return String.format("%s %s %s", this.firstTerm(), this.operator.toString(), this.secondTerm());
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepth) {
		indentTab(sb, indentDepth);
		sb.append(this.firstTerm());		
		sb.append(" ");
		sb.append(this.operator.toString());
		sb.append(" ");
		sb.append(this.secondTerm());
	}
}
