package de.lmu.ifi.pms.cep.tsa.conditions;

import org.simpleframework.xml.Root;

import de.lmu.ifi.pms.cep.tsa.schema.AType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.NamedType;


@Root(name="function")
public abstract class Function<TypeClass extends AType> extends CompositeTerm<TypeClass> {
	
	@SuppressWarnings("hiding")
	public final FunctionID ID;
	public final String name;
	
	public Function(final FunctionID ID, final String name, final TypeClass type, final int minArity, final int maxArity, final Term<?>... arguments) {
		super(TermID.FUNCTION, type, true, arguments);
		this.ID = ID;
		this.name = name;
		if(!(minArity <= arguments.length & arguments.length <= maxArity)){
			throw new IllegalArgumentException();
			//FIXME better exception
		}
	}
	
	@Override
	protected void validate() {
		if(!(this.minArity() <= this.argumentsArray.length & this.argumentsArray.length <= this.maxArity())){
			throw new IllegalArgumentException();
			//TODO better exception
		}
	}
	
	protected abstract int minArity();
	protected abstract int maxArity();
	
	@Override
	public abstract Function<TypeClass> newTerm(Term<?>... arguments);
	
	@Override
	protected int compareHead(final Term<?> o) {
		int cmp = 0;
		if(o instanceof Function<?>){
			Function<?> f = (Function<?>) o;
			cmp = this.name.compareTo(f.name);
		}
		if(cmp == 0){
			cmp = super.compareHead(o);
		}
		return cmp;
	}
	
	@Override
	public int hashCode() {
		return super.hashCode() ^ this.name.hashCode();
	}

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append(this.name);
		sb.append("(");
		if(this.argumentsArray.length > 0){
			for(int argument = 0; argument < this.argumentsArray.length; ++argument){
				sb.append(this.argumentsArray[argument]);
				sb.append(", ");
			}
			sb.setLength(sb.length() - ", ".length());
		}
		sb.append(")");
		if(this.type instanceof NamedType<?>){
			sb.append(" : ");
			sb.append(((NamedType<?>)this.type).name);
		}
		if(this.type instanceof BasicType){
			sb.append(" : ");
			sb.append(((BasicType)this.type).toString());
		}
		return sb.toString();
	}
}
