package de.lmu.ifi.pms.cep.tsa.conditions;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import org.simpleframework.xml.Element;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.filter.Filter;


public class PrimaryKeyConstraint extends AtomicConditionFormula {
	
	@Element(name="attribute")
	public final IdentifierAttribute attribute; 

	public PrimaryKeyConstraint(final IdentifierAttribute attribute) {
		super(true, true);
		this.attribute = attribute;
	}

	@Override
	public ConditionFormula replaceAtomicTerm(final Adapter<? super AtomicTerm<?>, ? extends Term<?>> replacement) {
		Term<?> newAttrinute = replacement.adapt(this.attribute);
		if(newAttrinute instanceof IdentifierAttribute){
			return new PrimaryKeyConstraint((IdentifierAttribute)newAttrinute);
		}
		else{
			throw new IllegalArgumentException("Replacement must be a identifier attribute");			
		}
	}
	
	@Override
	public ConditionFormula replaceTerm(final Adapter<Term<?>, ? extends Term<?>> replacement) {
		return this.replaceAtomicTerm(replacement);
	}
	
	@Override
	protected void collect(final Filter<Term<?>> filter, final ArrayList<Term<?>> terms) {
		this.attribute.collect(filter, terms);
	}

	@Override
	protected int compareToImpl(final AtomicConditionFormula o) {
		int cmp = super.compareToImpl(o);
		if(cmp == 0){
			PrimaryKeyConstraint c = (PrimaryKeyConstraint)o;
			cmp = this.attribute.compareTo(c.attribute);
		}
		return cmp;
	}
	
	@Override
	public int hashCode() {
		return super.hashCode() ^ this.attribute.hashCode();
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepth) {
		indentTab(sb, indentDepth);
		sb.append("PRIMARY KEY{ ");
		this.attribute.writeTo(sb, 0);
		sb.append(" }");
	}
	
}
