package de.lmu.ifi.pms.cep.tsa.conditions;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import org.simpleframework.xml.Root;

import de.lmu.ifi.pms.cep.tsa.conditions.filter.AttributeFilter;
import de.lmu.ifi.pms.cep.tsa.conditions.filter.VariableFilter;
import de.lmu.ifi.pms.cep.tsa.schema.AType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.schema.CompositeType;
import de.lmu.ifi.pms.cep.tsa.schema.NamedType;
import de.lmu.ifi.pms.cep.tsa.schema.NamelessType;
import de.lmu.ifi.pms.cep.tsa.schema.NoType;
import de.lmu.ifi.pms.cep.tsa.schema.Type;
import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.adaption.CollectionAdapter;
import de.lmu.ifi.pms.java.util.collection.AbstractCollection;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.filter.Filter;


@Root(name="term")
public abstract class Term<TypeClass extends AType> implements Comparable<Term<?>>{
	
	public final TermID ID;

	public final TypeClass type;
	
	public final boolean isTyped;

	public Term(final TermID ID, final TypeClass type) {
		super();
		this.ID = ID;
		this.type = type;
		assert type != null;
		this.isTyped = (this.type != NoType.instance);
		assert !(this.type instanceof NoType) | this.type == NoType.instance;
	}

	@SuppressWarnings("unchecked")
	public final FixedArrayNavigableSet<Variable<?>> Variables(){
		return new FixedArrayNavigableSet<Variable<?>>((AbstractCollection<Variable<?>>)(AbstractCollection<?>)this.collect(VariableFilter.instance));
	}

	public FixedArrayNavigableSet<String> VariableNames(){
		return new FixedArrayNavigableSet<String>(new CollectionAdapter<Term<?>, String>(this.collect(VariableFilter.instance)){
			
			@Override
			public String adapt(final Term<?> source) {
				return ((Variable<?>) source).name;
			}
			
		});
	}
	
	@SuppressWarnings("unchecked")
	public final FixedArrayNavigableSet<Attribute<?>> Attributes(){
		return new FixedArrayNavigableSet<Attribute<?>>((AbstractCollection<Attribute<?>>)(AbstractCollection<?>)this.collect(AttributeFilter.instance));
	}
	
	public FixedArrayNavigableSet<String> AttributeNames(){
		return new FixedArrayNavigableSet<String>(new CollectionAdapter<Term<?>, String>(this.collect(AttributeFilter.instance)){
			
			@Override
			public String adapt(final Term<?> source) {
				return ((Attribute<?>) source).name;
			}
			
		});
	}
	
	public abstract Term<?> replaceAtomicTerm(Adapter<? super AtomicTerm<?>, ? extends Term<?>> replacement);
	
	public AbstractCollection<Term<?>> collect(final Filter<Term<?>> filter){
		ArrayList<Term<?>> terms = new ArrayList<Term<?>>();
		this.collect(filter, terms);
		return terms;
	}
	
	protected void collect(final Filter<Term<?>> filter, final ArrayList<Term<?>> terms){
		if(filter.accepts(this)){
			terms.add(this);
		}
	}
		
	@Override
	public final int compareTo(final Term<?> o) {
		if(this == o){
			return 0;
		}
		else{
			return this.compareToImpl(o);
		}
	}
	
	protected int compareToImpl(final Term<?> o) {
		return this.getClass().getName().compareTo(o.getClass().getName());
	}
	

	@Override
	public final boolean equals(final Object o) {
		if(this == o){
			return true;
		}
		else{
			return this.equalsImpl(o);
		}
	}
	
	protected boolean equalsImpl(final Object o) {
		if(this.getClass().isInstance(o)){//succeeds only if obj is at least of type Term
			return this.compareToImpl((Term<?>)o) == 0;
		}
		else{
			return false;
		}
	}
	
	@Override
	public int hashCode() {
		return this.getClass().hashCode();
	}
	
	public void writeTo(final StringBuilder sb, final int indentDepth){
		indentTab(sb, indentDepth);
		sb.append(this.toString());
	}
	
	public void writeTo(final StringBuilder sb){
		this.writeTo(sb, 0);
	}
	
	@Override
	public abstract String toString();
	
	public static UntypedVariable var(final String name){
		return UntypedVariable.create(name);
	}
	
	public static BasicTypeVariable var(final String name, final BasicType type){
		return BasicTypeVariable.create(name, type);
	}
	
	public static DataTypeVariable var(final String name, final CompositeType type){
		return DataTypeVariable.create(name, type);
	}
	
	public static <TypeClass extends NamelessType> NamedTypeVariable<TypeClass> var(final String name, final NamedType<TypeClass> type){
		return NamedTypeVariable.create(name, type);
	}
	
	public static <TypeClass extends Type> TypedVariable<TypeClass> var(final String name, final TypeClass type){
		return TypedVariable.create(name, type);
	}
	
	public static BasicTypeVariable var(final String name, final BasicTypeID typeID){
		return BasicTypeVariable.create(name, typeID);
	}
	
	public static UntypedAttribute attr(final String name){
		return UntypedAttribute.create(name);
	}
	
	
	public static BasicTypeAttribute attr(final String name, final BasicType type){
		return BasicTypeAttribute.create(name, type);
	}
	
	public static DataTypeAttribute attr(final String name, final CompositeType type){
		return DataTypeAttribute.create(name, type);
	}
	
	public static <TypeClass extends NamelessType> NamedTypeAttribute<TypeClass> attr(final String name, final NamedType<TypeClass> type){
		return NamedTypeAttribute.create(name, type);
	}
	
	public static <TypeClass extends Type> TypedAttribute<TypeClass> attr(final String name, final TypeClass type){
		return TypedAttribute.create(name, type);
	}
	
	public static BasicTypeAttribute attr(final String name, final BasicTypeID typeID){
		return BasicTypeAttribute.create(name, typeID);
	}
	
	public static TupleFunction<NoType> apply(final TupleFunctionID ID, final Term<?>... arguments){
		return TupleFunction.apply(ID, arguments);
	}
	
	public static <TypeClass extends AType> TupleFunction<TypeClass> apply(final TupleFunctionID ID, final TypeClass type,final Term<?>... arguments){
		return TupleFunction.apply(ID, type, arguments);
	}
	
	public static AggregationFunction<NoType> apply(final AggregationFunctionID ID, final Term<?> argument){
		return Term.apply(ID, argument);
	}
	
	public static <TypeClass extends AType> AggregationFunction<TypeClass> apply(final AggregationFunctionID ID, final TypeClass type, final Term<?> argument){
		return AggregationFunction.apply(ID, type, argument);
	}
	
}
