package de.lmu.ifi.pms.cep.tsa.exceptions;

import de.lmu.ifi.pms.java.util.collection.fixed.FixedCollection;
import de.lmu.ifi.pms.java.util.collection.fixed.empty.Empty;

public abstract class TSA_AbstractException extends Exception {
	
	public final FixedCollection<String> origin;

	protected TSA_AbstractException() {
		this(Empty.<String>instance());
	}

	protected TSA_AbstractException(final String message) {
		this(Empty.<String>instance(), message);
	}
	
	protected TSA_AbstractException(final String messageFormat, final Object... args) {
		this(String.format(messageFormat, args));
	}
	
	protected TSA_AbstractException(final Throwable cause) {
		this(cause.getMessage(), cause);
		this.setStackTrace(new StackTraceElement[0]);
	}

	protected TSA_AbstractException(final TSA_AbstractException cause) {
		this(cause.origin, cause.getSimpleMessage(), cause);
		this.setStackTrace(new StackTraceElement[0]);
	}

	protected TSA_AbstractException(final String message, final TSA_AbstractException cause) {
		this(Empty.<String>instance(), message, cause);
	}
	
	public TSA_AbstractException(final FixedCollection<String> origin) {
		super();
		this.origin = origin;
	}

	public TSA_AbstractException(final FixedCollection<String> origin, final String message) {
		super(message);
		this.origin = origin;
	}
	
	public TSA_AbstractException(final FixedCollection<String> origin, final String messageFormat, final Object... args) {
		this(origin, String.format(messageFormat, args));
	}
	
	public TSA_AbstractException(final FixedCollection<String> origin, final Throwable cause) {
		super(cause.getMessage(), cause);
		this.origin = origin;
	}

	public TSA_AbstractException(final FixedCollection<String> origin, final TSA_AbstractException cause) {
		super(cause.getSimpleMessage(), cause);
		this.origin = origin;
	}

	public TSA_AbstractException(final FixedCollection<String> origin, final String message, final TSA_AbstractException cause) {
		super(message, cause);
		this.origin = origin;
	}
	
	public String getSimpleMessage(){
		return super.getMessage();
	}
	
	public String getSimpleLocalizedMessage(){
		return super.getLocalizedMessage();
	}
	
	@Override
	public String getMessage() {
		String message = this.getSimpleMessage();
		if(message == null){
			message = "null";
		}
		StringBuilder sb = new StringBuilder(message);
		for(String o: this.origin){
			sb.append("\n");
			sb.append(o);
		}
		return sb.toString();
	}

}
