package de.lmu.ifi.pms.cep.tsa.operators;

import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;

import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.fixed.FixedCollection;

public class TSA_CompositeOperator extends TSA_Operator
{
	
	public final TSA_Operator operator;
	public final FixedArray<TSA_Operator> subOperators;
	protected final int inputArity;

	public TSA_CompositeOperator(final FixedCollection<String> origin, final TSA_Operator operator, final FixedArray<TSA_Operator> subOperators)
	{
		super(TSA_OperatorID.COMPOSITE, origin);
		this.operator = operator;
		this.subOperators = subOperators;
		int inputArity = 0;
		for(TSA_Operator subOperator: subOperators){
			inputArity += subOperator.inputArity();
		}
		this.inputArity = inputArity;
		
		//FIXME check
	}
	
	public TSA_CompositeOperator(final FixedCollection<String> origin, final TSA_Operator operator, final TSA_Operator... subOperators)
	{
		this(origin, operator, new FixedArray<TSA_Operator>(subOperators));
	}
	
	@Override
	public int inputArity() {
		return this.inputArity;
	}
	
	@Override
	public TemporalStreamSchema apply(final Collection<TemporalStreamSchema> inputs) throws TSA_Exception{
		//FIXME check
		
		FixedArray<TemporalStreamSchema> inputsArray = new FixedArray<TemporalStreamSchema>(inputs);
		int start = 0;
		TemporalStreamSchema[] subSchemas = new TemporalStreamSchema[this.subOperators.size()];
		for(int subOperator = 0; subOperator < this.subOperators.size(); ++subOperator){
			subSchemas[subOperator] = this.subOperators.get(subOperator).apply(inputsArray.subArray(start, this.subOperators.get(subOperator).inputArity()));
		}
		return this.operator.apply(new FixedArray<TemporalStreamSchema>(false, subSchemas));
	}
	
	
	@Override
	public boolean isMonotonic(final Collection<TemporalStreamSchema> inputs){
		FixedArray<TemporalStreamSchema> inputsArray = new FixedArray<TemporalStreamSchema>(inputs);
		int start = 0;
		TemporalStreamSchema[] subSchemas = new TemporalStreamSchema[this.subOperators.size()];
		for(int subOperatorIndex = 0; subOperatorIndex < this.subOperators.size(); ++subOperatorIndex){
			try {
				TSA_Operator subOperator = this.subOperators.get(subOperatorIndex);
				FixedArray<TemporalStreamSchema> subInputs = inputsArray.subArray(start, subOperator.inputArity());
				subSchemas[subOperatorIndex] = subOperator.apply(subInputs);
				if(!subOperator.isMonotonic(subInputs)){
					return false;
				}
			} catch (TSA_Exception e) {
				return true;
			}
		}
		return this.operator.isMonotonic(new FixedArray<TemporalStreamSchema>(false, subSchemas));
	}
	
	@Override
	public void check(final Collection<TemporalStreamSchema> inputs) {
		//FIXME implement
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepthFirstLine, final int indentDepth) {
		//FIXME implement
	}

}
