package de.lmu.ifi.pms.cep.tsa.operators;

import org.simpleframework.xml.ElementList;
import org.simpleframework.xml.Root;

import de.lmu.ifi.pms.cep.tsa.DTSA_Expression;
import de.lmu.ifi.pms.cep.tsa.TSA_CompositeExpression;
import de.lmu.ifi.pms.cep.tsa.TSA_Expression;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;

import de.lmu.ifi.pms.java.util.adaption.CollectionAdapter;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.fixed.FixedCollection;


@Root(name="operator")
public abstract class TSA_Operator
{
	
	public final FixedCollection<String> origin;
	
	@ElementList(entry="origin", inline=true)
	java.util.ArrayList<String> getOriginList(){
		return new java.util.ArrayList<String>(this.origin.javaCollection());
	}
	
	
	public final TSA_OperatorID ID;
	
	public TSA_Operator(final TSA_OperatorID iD, final FixedCollection<String> origin)
	{
		super();
		this.ID = iD;
		
		this.origin = origin;
	}
	
	public abstract int inputArity();
	
	public abstract boolean isMonotonic(Collection<TemporalStreamSchema> inputs);
	
	public boolean isMonotonicTSA(final Collection<TSA_Expression> inputs){
		return this.isMonotonic(new CollectionAdapter<TSA_Expression, TemporalStreamSchema>(inputs) {
			
			@Override
			public TemporalStreamSchema adapt(final TSA_Expression source) {
				return source.schema();
			}
			
		});
	}
	
	public boolean isMonotonicDTSA(final Collection<DTSA_Expression> inputs){
		return this.isMonotonic(new CollectionAdapter<DTSA_Expression, TemporalStreamSchema>(inputs) {
			
			@Override
			public TemporalStreamSchema adapt(final DTSA_Expression source) {
				return source.schema();
			}
			
		});		
	}
	
	public abstract void check(Collection<TemporalStreamSchema> inputs) throws TSA_Exception;

	public abstract TemporalStreamSchema apply(Collection<TemporalStreamSchema> inputs) throws TSA_Exception;
	
	public TSA_CompositeExpression apply(final FixedArray<TSA_Expression> subExpressions) throws TSA_Exception{
		return new TSA_CompositeExpression(this, subExpressions);
	}
	
	public TSA_CompositeExpression apply(final TSA_Expression... subExpressions) throws TSA_Exception{
		return new TSA_CompositeExpression(this, new FixedArray<TSA_Expression>(subExpressions));
	}
	
	public TSA_CompositeExpression apply(final TSA_Expression subExpression) throws TSA_Exception{
		return new TSA_CompositeExpression(this, subExpression);
	}
	
	public TemporalStreamSchema apply(final TemporalStreamSchema... inputs) throws TSA_Exception{
		return this.apply(new FixedArray<TemporalStreamSchema>(inputs));
	}
	
	public void checkInputs(final Collection<TemporalStreamSchema> inputs) throws TSA_Exception{
		if(inputs.size() != this.inputArity()){
			throw new TSA_Exception(this.origin, "The number of provided inputs (%s) does not match the input arity (%s).", inputs.size(), this.inputArity());
		}
	}
	
	public abstract void writeTo(final StringBuilder sb, final int indentDepthFirstLine, final int indentDepth);
	
	public void writeTo(final StringBuilder sb, final int indentDepth){
		this.writeTo(sb, indentDepth, indentDepth);
	}
	
	public void writeTo(final StringBuilder sb){
		this.writeTo(sb, 0);
	}
	
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		this.writeTo(sb);
		return sb.toString();
	}

}
