package de.lmu.ifi.pms.cep.tsa.operators;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;
import static de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula.*;
import static de.lmu.ifi.pms.cep.tsa.conditions.Term.var;
import static de.lmu.ifi.pms.cep.tsa.conditions.TupleFunctionID.PLUS;
import static de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID.*;

import java.util.Iterator;

import org.simpleframework.xml.ElementList;

import de.lmu.ifi.pms.cep.tsa.conditions.AtomicTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.Attribute;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeVariable;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.BinaryComparissonOperatorID;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.DurationValue;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.conditions.TypedAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.Variable;
import de.lmu.ifi.pms.cep.tsa.conditions.adaptor.PrefixVariables;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_Exception;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_SchemaException;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;
import de.lmu.ifi.pms.cep.tsa.time.functions.TimeConstants;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.collection.fixed.FixedCollection;
import de.lmu.ifi.pms.java.util.map.Entry;



public class TSA_UnionOperator extends TSA_BasicOperator
{
	@org.simpleframework.xml.Attribute(name="input-arity")
	protected final int inputArity;
	
	public TSA_UnionOperator(final String origin, final int inputArity){
		this(new FixedArray<String>(false, origin), inputArity);
	}
	
	public TSA_UnionOperator(@ElementList(entry="origin", inline=true) final java.util.ArrayList<String> origin
			, @org.simpleframework.xml.Attribute(name="input-arity")final int inputArity){
		this(new FixedArray<String>(origin), inputArity);
	}

	public TSA_UnionOperator(final FixedCollection<String> origin, final int inputArity)
	{
		super(TSA_BasicOperatorID.UNION, origin);
		if(inputArity >= 2){
			this.inputArity = inputArity;
		}
		else{
			throw new IllegalArgumentException("The crossproduct expects a minimum arity of 2. (" + inputArity + ")");
		}
	}
	
	@Override
	public int inputArity() {
		return this.inputArity;
	}
	
	@Override
	public TemporalStreamSchema apply(final Collection<TemporalStreamSchema> inputs) throws TSA_Exception {
		this.checkInputs(inputs);
		
		TemporalStreamSchema[] inputsArray = inputs.toArray(new TemporalStreamSchema[0]);
		assert inputsArray.length >= 1;
		
		final ArrayList<Attribute<?>> identifierAttributes = new ArrayList<Attribute<?>>();
		final ArrayList<Attribute<?>> timestampAttributes = new ArrayList<Attribute<?>>();
		
		for(Entry<String, BasicType> attribute: inputsArray[0].basicAttributes()){
			if(attribute.value().ID == IDENTIFIER){
				identifierAttributes.add(BasicTypeAttribute.create(attribute.key(), attribute.value()));
			}
			if(attribute.value().ID == TIMESTAMP){
				timestampAttributes.add(BasicTypeAttribute.create(attribute.key(), attribute.value()));
			}
		}
		
		ConditionFormula[] tupleConstraints = new ConditionFormula[inputsArray.length];
		ConditionFormula[] relationConstraints = new ConditionFormula[inputsArray.length];
		ConditionFormula[] dontCareFormulaElements = new ConditionFormula[inputsArray.length];
		ConditionFormula[] doCareFormulas = new ConditionFormula[inputsArray.length];
		ConditionFormula[] dontCareFormulas = new ConditionFormula[inputsArray.length];
		
		for(int inputIndex = 0; inputIndex < inputsArray.length; ++inputIndex){
			TemporalStreamSchema input = inputsArray[inputIndex];
			
			//prefix variables to ensure free names
			
			PrefixVariables prefixer = new PrefixVariables("$$" + (inputIndex + 1));
					
			//convert removed attributes to variables
			final Adapter<AtomicTerm<?>, AtomicTerm<?>> removedAttributeReplacer = new Adapter<AtomicTerm<?>, AtomicTerm<?>>(){
				@Override
				public AtomicTerm<?> adapt(final AtomicTerm<?> source) {
					if(source instanceof Attribute){
						Attribute<?> attribute = (Attribute<?>) source;
						if(attribute instanceof TypedAttribute<?>){
							return var("$" + attribute.name, ((TypedAttribute<?>)attribute).type);
						}
						else{
							return var("$" + attribute.name);
						}
					}
					return source;
				}
			};
			
			tupleConstraints[inputIndex] = and(input.tupleConstraints()).replaceAtomicTerm(prefixer).replaceAtomicTerm(removedAttributeReplacer);
			relationConstraints[inputIndex] = and(input.relationConstraints()).replaceAtomicTerm(prefixer).replaceAtomicTerm(removedAttributeReplacer);
			
			final ArrayList<Variable<?>> streamBoundIdentifierVariables = new ArrayList<Variable<?>>();
			final ArrayList<Variable<?>> streamBoundTimestampVariables = new ArrayList<Variable<?>>();
			
			for(Variable<?> streamBoundVariable: and(relationConstraints[inputIndex]).Variables()){
				assert streamBoundVariable instanceof BasicTypeVariable;
				if(streamBoundVariable.type == BasicTypeID.IDENTIFIER.type){
					streamBoundIdentifierVariables.add(streamBoundVariable);
				}
				if(streamBoundVariable.type == BasicTypeID.TIMESTAMP.type){
					streamBoundTimestampVariables.add(streamBoundVariable);
				}
			}
			
			BinaryComparissonFormula[] localDontCareFormulaElements = new BinaryComparissonFormula[
			                      identifierAttributes.size() * streamBoundIdentifierVariables.size() +
			                      timestampAttributes.size() * streamBoundTimestampVariables.size()
			                      ];
			
			BinaryComparissonFormula[] localDoCareFormulas = new BinaryComparissonFormula[identifierAttributes.size() + timestampAttributes.size()];
			
			int dontCareIndex = 0;
			int doCareIndex = 0;
			
			for(Attribute<?> attribute: identifierAttributes){
				for(Variable<?> sbVariable2: streamBoundIdentifierVariables){
					localDontCareFormulaElements[dontCareIndex] = new BinaryComparissonFormula(BinaryComparissonOperatorID.LESS_EQUAL, sbVariable2, Term.apply(PLUS, IDENTIFIER.type, attribute, new DurationValue(TimeConstants.MINUS_INFINITY)));
					++dontCareIndex;
				}
				localDoCareFormulas[doCareIndex] = new BinaryComparissonFormula(BinaryComparissonOperatorID.EQUAL, removedAttributeReplacer.adapt(attribute), attribute);
				++doCareIndex;
			}
			for(Attribute<?> attribute: timestampAttributes){
				for(Variable<?> sbVariable2: streamBoundTimestampVariables){
					localDontCareFormulaElements[dontCareIndex] = new BinaryComparissonFormula(BinaryComparissonOperatorID.LESS_EQUAL, sbVariable2, Term.apply(PLUS, TIMESTAMP.type, attribute, new DurationValue(TimeConstants.MINUS_INFINITY)));
					++dontCareIndex;
				}
				localDoCareFormulas[doCareIndex] = new BinaryComparissonFormula(BinaryComparissonOperatorID.EQUAL, removedAttributeReplacer.adapt(attribute), attribute);
				++doCareIndex;
			}
			
			dontCareFormulaElements[inputIndex] = and(localDontCareFormulaElements);
			doCareFormulas[inputIndex] = and(localDoCareFormulas);
		}
		
		for(int inputIndex = 0; inputIndex < inputsArray.length; ++inputIndex){
			ConditionFormula[] localDontCareFormulaElements = new ConditionFormula[inputsArray.length - 1];
			System.arraycopy(dontCareFormulaElements, 0, localDontCareFormulaElements, 0, inputIndex);
			System.arraycopy(dontCareFormulaElements, inputIndex + 1, localDontCareFormulaElements, inputIndex, inputsArray.length - (inputIndex + 1));
			dontCareFormulas[inputIndex] = and(localDontCareFormulaElements);
		}
		
		for(int inputIndex = 0; inputIndex < inputsArray.length; ++inputIndex){
			tupleConstraints[inputIndex] = and(tupleConstraints[inputIndex], doCareFormulas[inputIndex], dontCareFormulaElements[inputIndex]);
		}
		
		try {
			return new TemporalStreamSchema(inputsArray[0].attributes, new FixedArray<ConditionFormula>(false, new ConditionFormula[]{or(tupleConstraints)}), new FixedArray<ConditionFormula>(false, relationConstraints));
		} catch (TSA_SchemaException e) {
			throw new TSA_Exception(this.origin, e);
		}
	}
	
	@Override
	public void checkInputs(final Collection<TemporalStreamSchema> inputs) throws TSA_Exception {
		super.checkInputs(inputs);

		assert inputs.size() >= 1;
		
		Iterator<TemporalStreamSchema> inputsIterator = inputs.iterator();
		assert inputsIterator.hasNext();
		TemporalStreamSchema input1 = inputsIterator.next();
		int index = 1;
		int numberOfAttributes = input1.basicAttributes.size();
				
		TemporalStreamSchema input;
		while(inputsIterator.hasNext()){
			input = inputsIterator.next();
			++index;
		
			if(FixedArrayNavigableSet.symmetricDifference(input1.basicAttributes.keys(), input.basicAttributes.keys()).size() > 0){
				throw new TSA_Exception(this.origin, "The attributes of the 1. and the " + index + ". input do not match.");
			}
			
			for(int attribute = 0; attribute < numberOfAttributes; ++attribute){
				if(!input1.basicAttributes.values().get(attribute).equals(input.basicAttributes.values().get(attribute))){
					String attributeName = input1.basicAttributes.keys().get(attribute);
					assert attributeName.equals(input.basicAttributes.keys().get(attribute));
					throw new TSA_Exception(this.origin, "The types for attribute \"" + attributeName + "\" do not match for inputs 1 and " + index + ".");
				}
			}
		}
	}
	
	@Override
	public boolean isMonotonic(Collection<TemporalStreamSchema> inputs){
		return true;
	}
	
	@Override
	public void check(final Collection<TemporalStreamSchema> inputs) throws TSA_Exception {
		this.checkInputs(inputs);
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepthFirstLine, final int indentDepth) {
		indentTab(sb, indentDepthFirstLine);
		sb.append(this.ID.toString());
	}

}
