package de.lmu.ifi.pms.cep.tsa.schema;

import org.simpleframework.xml.Text;
import org.simpleframework.xml.core.Resolve;

import de.lmu.ifi.pms.cep.tsa.TSA_SyntaxPatterns;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_TypeException;

import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.filter.Filter;

public final class BasicType extends NamelessType {

	@SuppressWarnings("hiding")
	public final BasicTypeID ID;
	
	BasicType(final BasicTypeID id) {
		super(TypeID.BASIC);
		this.ID = id;
	}
	
	public static BasicType instance(final BasicTypeID id){
		return id.type;
	}
	
	@Override
	protected int compareToImpl(final Type o) {		
		int cmp = super.compareToImpl(o);
		if(cmp == 0){
			cmp = this.ID.ordinal() - ((BasicType)o).ID.ordinal();
		}
		return cmp;
	}
	
	@Override
	protected boolean equalsImpl(final Object o) {
		if(o instanceof BasicType){
			return this.ID == ((BasicType)o).ID;
		}
		else{
			return false;
		}
	}
	
	@Override
	public boolean structuralEqualsImpl(final Type type) {
		return this.equals(type);
	}
	
	@Override
	public int hashCode() {
		return super.hashCode() ^ this.ID.hashCode();
	}
	
	@Override
	protected Type retain(final String[][] retainedAttributes, final int startIndex, final int endIndex, final int level) {
		if(startIndex < endIndex){
			if(retainedAttributes[startIndex].length == level){
				return this;
			}
		}
		return null;
	}
	
	@Override
	protected Type extend(final String[][] newAttributeNames, final Type[] newAttributeTypes, final int startIndex, final int endIndex, final int level) throws TSA_TypeException {
		if(startIndex < endIndex){
			throw new TSA_TypeException("A basic type cannot be extended.");
		}
		else{
			return this;
		}
	}
	
	@Override
	protected Type typeOf(final String[] name, final int index) throws TSA_TypeException{
		assert TSA_SyntaxPatterns.isValidAttributeName(name);
		assert index >= 0;
		if(index < name.length){
			throw new TSA_TypeException("No Attribute with this name: " + TSA_SyntaxPatterns.attributeName(name));
		}
		else{
			assert index == name.length;
			return this;
		}
	}
	
	@Override
	public boolean isEmpty() {
		return false;
	}
	
	@Override
	protected void basicAttributes(final StringBuilder namePrefix, final ArrayList<String> nameAcc, final ArrayList<BasicType> basicTypeAcc) {
		nameAcc.add(namePrefix.toString());
		basicTypeAcc.add(this);
	}
	
	@Override
	protected void tupleConstraints(final String ownName, final ArrayList<ConditionFormula> constraintAcc, final Filter<String> attributeFilter) {
		if(attributeFilter.accepts(ownName)){
			//do nothing as there are no constraints
			//TODO allow restricted types: e.g. int type with value > 10
		}
	}
	
	BasicType(@Text final String name) {
		this(BasicTypeID.valueOf(name));
	}
	
	@Text
	String getName(){
		return this.ID.name();
	}
	
	@Resolve
	BasicType resolve(){
		return instance(this.ID);
	}
	
	@Override
	public String toString() {
		return this.ID.toString();
	}
}
