package de.lmu.ifi.pms.cep.tsa.schema;

import org.simpleframework.xml.Attribute;
import org.simpleframework.xml.Element;
import org.simpleframework.xml.core.Resolve;

import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.exceptions.TSA_TypeException;

import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.filter.Filter;

public class NamedType<TypeSort extends NamelessType> extends Type {
	
	@Attribute(name="name")
	public final String name;
	
	@Element(name="type")
	public final TypeSort type;
	
	NamedType(@Attribute(name="name") final String name, @Element(name="type") final TypeSort type){
		super(TypeID.NAMED);
		this.name = name;
		this.type = type;
	}
	
	@Resolve
	NamedType<TypeSort> resolve(){
		try{
			return get(this.name, this.type);
		}
		catch (Exception e) {
			throw new Error("Should not be possible.");
		}
	}
	
	public static <TypeSort extends NamelessType> NamedType<TypeSort> get(final String name, final TypeSort type) {
		return new NamedType<TypeSort>(name, type);
		//TODO cache and return unique object;
	}
	
	@Override
	protected Type retain(final String[][] retainedAttributes, final int startIndex, final int endIndex, final int level) throws TSA_TypeException {
		Type newType = this.type.retain(retainedAttributes, startIndex, endIndex, level);
		if(newType.equals(this.type)){
			return this;
		}
		else{
			return newType;
		}
	}
	
	@Override
	protected Type extend(final String[][] newAttributeNames, final Type[] newAttributeTypes, final int startIndex, final int endIndex, final int level) throws TSA_TypeException {
		if(startIndex < endIndex){
			return this.type.extend(newAttributeNames, newAttributeTypes, startIndex, endIndex, level);
		}
		else{
			return this;
		}
	}

	@Override
	protected Type typeOf(final String[] name, final int index)  throws TSA_TypeException{
		if(index < name.length){
			return this.type.typeOf(name, index);
		}
		else{
			assert index == name.length;
			return this;
		}
	}
	
	@Override
	public boolean isEmpty() {
		return this.type.isEmpty();
	}

	@Override
	protected void basicAttributes(final StringBuilder namePrefix, final ArrayList<String> nameAcc, final ArrayList<BasicType> basicTypeAcc) {
		this.type.basicAttributes(namePrefix, nameAcc, basicTypeAcc);
	}

	@Override
	protected void tupleConstraints(final String ownName, final ArrayList<ConditionFormula> constraintAcc, final Filter<String> attributeFilter) {
		this.type.tupleConstraints(ownName, constraintAcc, attributeFilter);
	}
	
	@Override
	protected int compareToImpl(final Type o) {
		int cmp = super.compareToImpl(o);
		if(cmp == 0){
			NamedType<?> type = (NamedType<?>) o;
			cmp = this.name.compareTo(type.name);
			if(cmp == 0){
				cmp = this.type.compareTo(type.type);
			}
		}
		return cmp;
	}
	
	@Override
	protected boolean equalsImpl(final Object o) {
		if(o instanceof NamedType<?>){
			NamedType<?> type = (NamedType<?>) o;		
			if(this.name.equals(type.name)){
				return this.type.equals(type.type);
			}
		}
		return false;
	}
	
	@Override
	public int hashCode() {
		return super.hashCode() ^ this.name.hashCode() ^ this.type.hashCode();
	}

	@Override
	protected boolean structuralEqualsImpl(final Type type) {
		return this.type.structuralEquals(type);
	}
	
	@Override
	public String toString() {
		return this.name;
	}

}
