package de.lmu.ifi.pms.cep.tsa.time.analysis;


public abstract class ArithmeticAnalyzer {
	
	public static final long PLUS_INFTY = Long.MAX_VALUE - 1;
	public static final long ZERO = encode(0, false);
	public static final long MINUS_INFTY = Long.MIN_VALUE;
	
	public static final long CONSTANT_MAX_SIZE = Long.MAX_VALUE >> 2;
	public static final long CONSTANT_MIN_SIZE = Long.MIN_VALUE >> 2;
	
	public static final long CONST_MASK = -1 << 2;
	public static final long EPSILON_MASK = -1 >>> 62;
	
	public static long constant(final long encodedConstant){
		return encodedConstant >> 2;
	}
	
	public static boolean minusEpsilon(final long encodedConstant){
		return (encodedConstant & EPSILON_MASK) == 0;
	}
	
	public static long encode(final long value, final boolean minusEpsilon){
		assert CONSTANT_MIN_SIZE <= value;
		assert value <= CONSTANT_MAX_SIZE;
		
		return value << 2 | (minusEpsilon ? 0 : 1);
	}
	
	public static long encodedPlus(final long m, final long n){
		long k = m + n;
		long overflow = ((m ^ k) & (n ^ k)) >> 63;
		
		//overflow occured if overflow = -1 (only 1 bits) else overflow = 0 (only 0 bits);		
		long plusInfty = (m << 62 | n << 62 | (overflow & k)) >> 63;
		//is -1 if n or m are PLUS_INFTY or an overflow occurs else 0
		//if neither m nor n is PLUS_INFTY then the second bit of n and m is 0 and therefore the last two bits of k are either 00 or 01 or 10 
		//in the case of 00 or 01 either a overflow occured or k is smaller than PLUS_INFTY
		//in the case of 10 either a overflow occured or k is smaller or equal than PLUS_INFTY. 
		//If k is equal to PLUS_INFTY then the last two bits (01) mean that both input values were not reduced by epsilon. 
		//Thus the last two bits of the result should be 01, i.e. the result is PLUS_INFTY - 1.		
		long minusInfty = overflow & m & MINUS_INFTY;//(overflow & m) >> 63 & MINUS_INFTY would be independent from the value of MINUS_INFTY
		//is MINUS_INFTY if an underflow occurs else 0 		
		return (plusInfty  & PLUS_INFTY) | minusInfty | (~overflow & ~plusInfty & ((k & CONST_MASK) | (k & EPSILON_MASK) >>> 1));
	}
	
	public boolean equal(final String name1, final boolean isAttribute1, final String name2, final boolean isAttribute2){
		long distance1 = this.distance(name1, isAttribute1, name2, isAttribute2);
		long distance2 = this.distance(name2, isAttribute2, name1, isAttribute1);
		return constant(distance1) == 0 & !minusEpsilon(distance1) & constant(distance2) == 0 & !minusEpsilon(distance2);
	}
	
	public abstract long distance(String name1, boolean isAttribute1, String name2, boolean isAttribute2);
	
	public abstract boolean isConsistent();
	
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		this.writeTo(sb);
		return sb.toString();
	}
	
	public abstract void writeTo(StringBuilder sb);

}
