package de.lmu.ifi.pms.cep.tsa.time.analysis;

import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;

import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.map.Entry;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;
import de.lmu.ifi.pms.java.util.map.fixed.FixedNavigableMap;


public class CombinedArithmeticAnalyzer extends ArithmeticAnalyzer{
	
	boolean isComputed;
	
	final FixedArrayNavigableSet<String> relevantAttributeNames;
	final BasicTypeID[] relevantAttributeTypes;
	final FixedArrayNavigableSet<String> relevantVariableNames;
	final BasicTypeID[] relevantVariableTypes;
	boolean isConsitent = false;
	
	final SimpleArithmeticAnalyzer[] singleTypeAnalysers;

	public CombinedArithmeticAnalyzer(final FixedNavigableMap<BasicTypeID, SimpleArithmeticAnalyzer, ? extends Entry<BasicTypeID, SimpleArithmeticAnalyzer>> analysers, final FixedArrayNavigableMap<String, BasicTypeID> relevantAttributes, final FixedArrayNavigableMap<String, BasicTypeID> relevantVariables){
		this.isComputed = false;
		
		this.relevantAttributeNames = relevantAttributes.keys();
		this.relevantAttributeTypes = relevantAttributes.values().toArray(new BasicTypeID[relevantAttributes.size()]);
		this.relevantVariableNames = relevantVariables.keys();
		this.relevantVariableTypes = relevantVariables.values().toArray(new BasicTypeID[relevantVariables.size()]);
		
		this.singleTypeAnalysers = new SimpleArithmeticAnalyzer[BasicTypeID.values().length];
		for(Entry<BasicTypeID, SimpleArithmeticAnalyzer> analyser: analysers){
			this.singleTypeAnalysers[analyser.key().ordinal()] = analyser.value();
		}

	}
	
	public void calculate(){		
		if(!this.isComputed){
			this.isConsitent = true;
			for(int analyser = 0; analyser < this.singleTypeAnalysers.length; ++analyser){
				SimpleArithmeticAnalyzer singleTypeAnalyser = this.singleTypeAnalysers[analyser];
				if(singleTypeAnalyser != null){
					singleTypeAnalyser.calculate();
					assert this.singleTypeAnalysers[analyser].isComputed;
					if(!this.singleTypeAnalysers[analyser].isConsitent){
						this.isConsitent = false;
						break;
					}
				}
			}
			
			this.isComputed = true;
		}
	}
	
	@Override
	public long distance(final String name1, final boolean isAttribute1, final String name2, final boolean isAttribute2){
		if(!this.isComputed){
			this.calculate();
		}
		int index1 = (isAttribute1 ? this.relevantAttributeNames.indexOf(name1) : this.relevantVariableNames.indexOf(name1));
		int index2 = (isAttribute2 ? this.relevantAttributeNames.indexOf(name2) : this.relevantVariableNames.indexOf(name2));
		if(index1 < 0 | index2 < 0){
			if((isAttribute1 == isAttribute2) && name1.equals(name2)){
				return ZERO;
			}
			else{
				return PLUS_INFTY;
			}
		}
		else{
			assert 0 <= index1 & (isAttribute1 ? index1 < this.relevantAttributeTypes.length : index1 < this.relevantVariableTypes.length);
			assert 0 <= index2 & (isAttribute2 ? index2 < this.relevantAttributeTypes.length : index2 < this.relevantVariableTypes.length);
			BasicTypeID type1 = (isAttribute1 ? this.relevantAttributeTypes[index1] : this.relevantVariableTypes[index1]);
			BasicTypeID type2 = (isAttribute2 ? this.relevantAttributeTypes[index2] : this.relevantVariableTypes[index2]);
			if(type1 != type2 | type1 == null /*| type2 == null already implied*/){
				if(((isAttribute1 == isAttribute2) & (type1 == type2)) && name1.equals(name2)){
					return ZERO;
				}
				else{
					return PLUS_INFTY;
				}
			}
			else{assert type2 != null;
				return this.distance(type1, name1, isAttribute1, name2, isAttribute2);
			}
		}
	}
	
	public long distance(final BasicTypeID type, final String name1, final boolean isAttribute1, final String name2, final boolean isAttribute2){
		ArithmeticAnalyzer singleTypeAnalyser = this.singleTypeAnalysers[type.ordinal()];
		if(singleTypeAnalyser != null){
			return singleTypeAnalyser.distance(name1, isAttribute1, name2, isAttribute2);
		}
		else{
			return PLUS_INFTY;
		}
	}
	
	@Override
	public boolean isConsistent(){
		if(!this.isComputed){
			this.calculate();
		}
		return this.isConsitent;
	}
	
	@Override
	public void writeTo(final StringBuilder sb) {
		BasicTypeID[] types = BasicTypeID.values();
		for(int analyser = 0; analyser < this.singleTypeAnalysers.length; ++analyser){
			SimpleArithmeticAnalyzer singleTypeAnalyser = this.singleTypeAnalysers[analyser];
			if(singleTypeAnalyser != null){
				sb.append(types[analyser]);
				sb.append("\n");
				singleTypeAnalyser.writeTo(sb);
				sb.append("\n");
			}
		}
	}

}
