package de.lmu.ifi.pms.cep.tsa.time.analysis;

import static de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID.*;

import de.lmu.ifi.pms.cep.tsa.conditions.Attribute;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.Variable;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;
import de.lmu.ifi.pms.cep.tsa.schema.Type;
import de.lmu.ifi.pms.cep.tsa.schema.TypeID;

import de.lmu.ifi.pms.java.util.collection.array.Arrays;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.map.Entry;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;
import de.lmu.ifi.pms.java.util.map.fixed.FixedEntry;
import de.lmu.ifi.pms.java.util.map.fixed.FixedNavigableMap;

public class CombinedArithmeticAnalyzerFactory extends ArithmeticAnalyzerFactory{
	
	final FixedNavigableMap<BasicTypeID, SimpleArithmeticAnalyzerFactory, ? extends FixedEntry<BasicTypeID, SimpleArithmeticAnalyzerFactory>> analysers;

	public CombinedArithmeticAnalyzerFactory() {
		this(new FixedArrayNavigableMap<BasicTypeID, SimpleArithmeticAnalyzerFactory>(new BasicTypeID[]{IDENTIFIER, TIMESTAMP}, new SimpleArithmeticAnalyzerFactory[]{new SimpleArithmeticAnalyzerFactory(IDENTIFIER), new SimpleArithmeticAnalyzerFactory(TIMESTAMP)}));
	}
	
	public CombinedArithmeticAnalyzerFactory(final FixedNavigableMap<BasicTypeID, SimpleArithmeticAnalyzerFactory, ? extends FixedEntry<BasicTypeID, SimpleArithmeticAnalyzerFactory>> analysers) {
		super();
		this.analysers = analysers;
	}
	
	@Override
	public ArithmeticAnalyzer make(final ConditionFormula formula) {
		final BasicTypeID[] types = new BasicTypeID[this.analysers.size()];
		final SimpleArithmeticAnalyzer[] analysers = new SimpleArithmeticAnalyzer[types.length];
		int type = 0;
		for(Entry<BasicTypeID, SimpleArithmeticAnalyzerFactory> entry: this.analysers){
			types[type] = entry.key();
			analysers[type] = entry.value().make(formula);
			++type;
		}
		return new CombinedArithmeticAnalyzer(new FixedArrayNavigableMap<BasicTypeID, SimpleArithmeticAnalyzer>(types, analysers), new FixedArrayNavigableMap<String, BasicTypeID>(), new FixedArrayNavigableMap<String, BasicTypeID>());
	}
	
	@Override
	public ArithmeticAnalyzer make(final ConditionFormula formula, final FixedArrayNavigableSet<String> relevantAttributes, final FixedArrayNavigableSet<String> relevantVariables) {
		final BasicTypeID[] types = new BasicTypeID[this.analysers.size()];
		final SimpleArithmeticAnalyzer[] analysers = new SimpleArithmeticAnalyzer[types.length];
		FixedArrayNavigableSet<String>[] relevantAttributesByType = this.relevantAttributes(formula, relevantAttributes);
		FixedArrayNavigableSet<String>[] relevantVariablesByType = this.relevantVariables(formula, relevantVariables);
		
		
		int type = 0;
		for(Entry<BasicTypeID, SimpleArithmeticAnalyzerFactory> entry: this.analysers){
			types[type] = entry.key();
			analysers[type] = entry.value().make(formula, relevantAttributesByType[types[type].ordinal()], relevantVariablesByType[types[type].ordinal()]);
			++type;
		}
		return new CombinedArithmeticAnalyzer(new FixedArrayNavigableMap<BasicTypeID, SimpleArithmeticAnalyzer>(types, analysers), this.relevantAttributes(relevantAttributesByType, relevantAttributes), this.relevantAttributes(relevantVariablesByType, relevantVariables));
	}
	
	protected FixedArrayNavigableMap<String, BasicTypeID> relevantAttributes (final FixedArrayNavigableSet<String>[] relevantAttributeTypes, final FixedArrayNavigableSet<String> relevantAttributes){
		BasicTypeID[] basicTypes = BasicTypeID.values();
		BasicTypeID[] attributeTypes = new BasicTypeID[relevantAttributes.size()];
		for(int type = 0; type < basicTypes.length; ++type){
			for(String attribute: relevantAttributeTypes[type]){
				int index = relevantAttributes.indexOf(attribute);
				if(index >= 0){
					attributeTypes[index] = basicTypes[type];
				}
			}
		}
		return new FixedArrayNavigableMap<String, BasicTypeID>(relevantAttributes, new FixedArray<BasicTypeID>(false, attributeTypes));
	}
	
	protected FixedArrayNavigableSet<String>[] relevantAttributes (final ConditionFormula formula, final FixedArrayNavigableSet<String> relevantAttributes){
		@SuppressWarnings("unchecked")
		ArrayList<String>[] relevantAttributesList = (ArrayList<String>[]) new ArrayList<?>[BasicTypeID.values().length];
		Arrays.fillFactory(relevantAttributesList, new Arrays.ValueFactory<ArrayList<String>>() {			
			@Override
			public ArrayList<String> make(final int index) {
				return new ArrayList<String>();
			}
		});
		for(Attribute<?> attribute: formula.Attributes()){
			if(relevantAttributes.contains(attribute.name)){
				if(attribute.isTyped){
					Type type = (Type) attribute.type;
					if(type.ID == TypeID.BASIC){
						relevantAttributesList[((BasicType) type).ID.ordinal()].add(attribute.name);
					}
				}
			}
		}
		@SuppressWarnings("unchecked")
		FixedArrayNavigableSet<String>[] relevantAttributesArray = (FixedArrayNavigableSet<String>[]) new FixedArrayNavigableSet<?>[relevantAttributesList.length];
		for(int type = 0; type < relevantAttributesArray.length; ++type){
			relevantAttributesArray[type] = new FixedArrayNavigableSet<String>(false, false, relevantAttributesList[type].toArray(new String[0]));
		}
		return relevantAttributesArray;
	}
	
	protected FixedArrayNavigableSet<String>[] relevantVariables (final ConditionFormula formula, final FixedArrayNavigableSet<String> relevantVariables){
		@SuppressWarnings("unchecked")
		ArrayList<String>[] relevantVariablesList = (ArrayList<String>[]) new ArrayList<?>[BasicTypeID.values().length];
		Arrays.fillFactory(relevantVariablesList, new Arrays.ValueFactory<ArrayList<String>>() {			
			@Override
			public ArrayList<String> make(final int index) {
				return new ArrayList<String>();
			}
		});
		for(Variable<?> variable: formula.Variables()){
			if(relevantVariables.contains(variable.name)){
				if(variable.isTyped){
					Type type = (Type) variable.type;
					if(type.ID == TypeID.BASIC){
						relevantVariablesList[((BasicType) type).ID.ordinal()].add(variable.name);
					}
				}
			}
		}
		@SuppressWarnings("unchecked")
		FixedArrayNavigableSet<String>[] relevantVariablesArray = (FixedArrayNavigableSet<String>[]) new FixedArrayNavigableSet<?>[relevantVariablesList.length];
		for(int type = 0; type < relevantVariablesArray.length; ++type){
			relevantVariablesArray[type] = new FixedArrayNavigableSet<String>(false, false, relevantVariablesList[type].toArray(new String[0]));
		}
		return relevantVariablesArray;
	}
}
