package de.lmu.ifi.pms.cep.tsa.time.analysis;

import de.lmu.ifi.pms.cep.tsa.conditions.AtomicConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.conditions.analysis.SimpleFocusedFormulaAnalyser;
import de.lmu.ifi.pms.cep.tsa.schema.Type;
import de.lmu.ifi.pms.cep.tsa.time.constraints.ConjunctiveTemporalRelationFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.DisjunctiveTemporalRelationFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.NegatedTemporalRelationFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.SimpleBoundTemporalFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.SimpleTemporalFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.TemporalRelationFormula;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.AbstractCollection;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArrayNavigableSet;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.filter.AbstractFilter;
import de.lmu.ifi.pms.java.util.filter.Filter;

public class TemporalSimpleFocusedFormulaAnalyser extends SimpleFocusedFormulaAnalyser {
	
	public final ArithmeticRelationFormulaExtractor formulaExtractor;
	
	//FIXME What about types other than TIMESTAMP (IDENTIFIER)
	public final Adapter<AtomicConditionFormula, TemporalRelationFormula> localType;
		
	public final AbstractFilter<AtomicConditionFormula> preAdaptfocus;
	
	public final AbstractFilter<AtomicConditionFormula> postAdaptfocus;
	
	
	public TemporalSimpleFocusedFormulaAnalyser(final Type periodType) {
		super();
		
		this.formulaExtractor = new ArithmeticRelationFormulaExtractor(periodType);
		
		this.localType =
				new Adapter<AtomicConditionFormula, TemporalRelationFormula>(){
				
					@Override
					public TemporalRelationFormula adapt(final AtomicConditionFormula atom) {
						try{
							if(TemporalSimpleFocusedFormulaAnalyser.this.formulaExtractor.focus.accepts(atom)){
								ConditionFormula normalizedFormula = TemporalSimpleFocusedFormulaAnalyser.this.formulaExtractor.adapt(atom);
								return TemporalSimpleFocusedFormulaAnalyser.this.formulaExtractor.convert(normalizedFormula);
							}
							else{
								return null;
							}
						}
						catch (IllegalArgumentException e) {
							return null;
						}
						
					}
				};
				
		this.preAdaptfocus = this.formulaExtractor.focus;
		
		this.postAdaptfocus = new AbstractFilter<AtomicConditionFormula>() {
			
			@Override
			public boolean accepts(final AtomicConditionFormula element) {
				return element instanceof TemporalRelationFormula;
			}
			
		};
	}

	@Override
	public Adapter<AtomicConditionFormula, TemporalRelationFormula> localType() {
		return this.localType;
	}
	
	@Override
	public Filter<AtomicConditionFormula> postAdaptFocus() {
		return this.postAdaptfocus;
	}

	@Override
	public boolean isConsistent(final AbstractCollection<AtomicConditionFormula> positiveLiterals, final AbstractCollection<AtomicConditionFormula> negativeLiterals) {
		@SuppressWarnings("unchecked")
		TemporalRelationFormula temporalRelationFormula = new ConjunctiveTemporalRelationFormula(new ConjunctiveTemporalRelationFormula((AbstractCollection<TemporalRelationFormula>)(AbstractCollection<?>)positiveLiterals),new NegatedTemporalRelationFormula(new DisjunctiveTemporalRelationFormula((AbstractCollection<TemporalRelationFormula>)(AbstractCollection<?>)negativeLiterals))).simpleForm().flaten();
		//TODO .simpleForm() could be omited if adapter aready guarantees simpleForm (same for flaten())
		
		
		//TODO reuse matrix, do not need to collect attribute names / variable names -> init function telling about attributes and variables
		
		final FixedArrayNavigableSet<String> attributeNames = temporalRelationFormula.AttributeNames();
		int numberOfAttributes = attributeNames.size();
		final FixedArrayNavigableSet<String> variableNames = temporalRelationFormula.VariableNames();
		int size = numberOfAttributes + variableNames.size();
		long[][] matrix = new long[size][size];
		
		for(ConjunctiveTemporalRelationFormula conjunction :temporalRelationFormula.tempDNFiteration()){
			
			ArrayList<SimpleTemporalFormula> simpleFormulas = new ArrayList<SimpleTemporalFormula>();
			ArrayList<SimpleBoundTemporalFormula> boundedFormulas = new ArrayList<SimpleBoundTemporalFormula>();
			for(TemporalRelationFormula atrf: conjunction.subFormulas()){
				if(atrf instanceof SimpleTemporalFormula){
					simpleFormulas.add((SimpleTemporalFormula)atrf);
				}
				else if(atrf instanceof SimpleBoundTemporalFormula){
					boundedFormulas.add((SimpleBoundTemporalFormula)atrf);
				}
				else{
					throw new Error();
				}
			}
			SimpleArithmeticAnalyzer.setDistanceMatrix(matrix, simpleFormulas, 0, 0, numberOfAttributes, new FixedArrayNavigableSet<String>(false, false, new String[0]), attributeNames, new FixedArrayNavigableSet<String>(false, false, new String[0]), variableNames);
			SimpleFloydWarshall floydWarshall = new SimpleFloydWarshall(matrix, 0);
			floydWarshall.calculate();
			if(floydWarshall.isConsistent()){
				return true;
			}
		}
		return false;
	}

}
