package de.lmu.ifi.pms.cep.tsa.time.constraints;

import static de.lmu.ifi.pms.cep.event_mill.sql.creation.SQL_QueryBuildHelper.indentTab;

import org.simpleframework.xml.Element;

import de.lmu.ifi.pms.cep.tsa.conditions.AtomicTerm;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeAttribute;
import de.lmu.ifi.pms.cep.tsa.conditions.BasicTypeVariable;
import de.lmu.ifi.pms.cep.tsa.conditions.Term;
import de.lmu.ifi.pms.cep.tsa.schema.BasicType;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;

import de.lmu.ifi.pms.java.util.adaption.Adapter;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.filter.Filter;

public class SimpleBoundTemporalFormula extends AtomicTemporalRelationFormula{
	
	@Element(name="name")
	public final String v;
	@Element(name="is-Attribute")
	public final boolean isAttribute;

	public SimpleBoundTemporalFormula(@Element(name="type") final BasicTypeID type, @Element(name="name") final String v, @Element(name="is-Attribute") final boolean isAttribute){
		super(type);
		this.v = v;
		this.isAttribute = isAttribute;
	}
	
	@Override
	protected void VariableNames(final ArrayList<String> variables) {
		if(!this.isAttribute){
			variables.add(this.v);
		}
	}
	
	@Override
	protected void AttributeNames(final ArrayList<String> attributes) {
		if(this.isAttribute){
			attributes.add(this.v);
		}
	}
	
	@Override
	public void writeTo(final StringBuilder sb, final int indentDepth) {
		indentTab(sb, indentDepth);
		sb.append("bounded( ");
		sb.append((this.isAttribute ? "attr " : "var "));
		sb.append(this.v);		
		sb.append(" )");
	}
	
	@Override
	public TemporalRelationFormula flaten() {
		return this;
	}
	
	@Override
	protected TemporalRelationFormula simpleForm(final boolean positivePolarity){
		if(positivePolarity){
			return this;
		}
		else{
			throw new Error();
		}
	}
	

	public TemporalRelationFormula negatedForm(){
		throw new Error();
	}
	
	protected AtomicTerm<BasicType> arg(){
		if(this.isAttribute){
			return BasicTypeAttribute.create(this.v, this.type);
		}
		else{
			return BasicTypeVariable.create(this.v, this.type);
		}
	}
	
	@Override
	protected void collect(final Filter<Term<?>> filter, final ArrayList<Term<?>> terms) {
		AtomicTerm<BasicType> arg1 = this.arg();
		if(filter.accepts(arg1)){
			terms.add(arg1);
		}
	}
	
	@Override
	public TemporalRelationFormula replaceAtomicTerm(final Adapter<? super AtomicTerm<?>, ? extends Term<?>> replacement) {
		Term<?> t = replacement.adapt(this.arg());
		String v;
		boolean isAttribute;
		BasicTypeID type;
		if(t instanceof BasicTypeAttribute){
			v = ((BasicTypeAttribute) t).name;
			type = ((BasicTypeAttribute) t).type.ID;
			isAttribute = true;
		}
		else if(t instanceof BasicTypeVariable){
			v = ((BasicTypeVariable) t).name;
			type = ((BasicTypeVariable) t).type.ID;
			isAttribute = false;
		}
		else{
			throw new IllegalArgumentException();
			//TODO better exception
		}
		if(v.equals(this.v) & type == this.type & isAttribute == this.isAttribute){
			return this;
		}
		else{
			return new SimpleBoundTemporalFormula(type, v, isAttribute);
		}
				
	}
	
	@Override
	public TemporalRelationFormula replaceTerm(final Adapter<Term<?>, ? extends Term<?>> replacement) {
		return this.replaceAtomicTerm(replacement);
	}
	
}
