package de.lmu.ifi.pms.java.util.collection;

import java.lang.reflect.Array;

import de.lmu.ifi.pms.java.lang.iteration.Iterator;
import de.lmu.ifi.pms.java.util.collection.javaWrapper.FromJavaCollectionWrapper;
import de.lmu.ifi.pms.java.util.collection.javaWrapper.ToJavaCollectionWrapper;
import de.lmu.ifi.pms.java.util.string.StringBuilder;



public abstract class AbstractCollection<Type> implements Collection<Type> {
	
	@Override
	public Object[] toArray() {
		return toArray(this);
	}

	@Override
	public <T> T[] toArray(final T[] a) {
		return toArray(a, this);
	}
	
	@Override
	public boolean contains(final Object o) {
		return contains(o, this);
	}
	
	@SuppressWarnings({ "unchecked", "rawtypes" })
	@Override
	public boolean containsAll(final java.util.Collection<?> c) {
		return this.containsAll(new FromJavaCollectionWrapper(c));
	}

	@Override
	public boolean containsAll(final Collection<?> c) {
		return containsAll(c, this);
	}

	@Override
	public boolean isEmpty() {
		return this.size() == 0;
	}
	
	@Override
	public java.util.AbstractCollection<Type> javaCollection() {
		return new ToJavaCollectionWrapper<Type>(this);
	}

	@Override
	public abstract Iterator<Type> iterator();

	@Override
	public abstract int size();
	
	protected String openingBracket(){
		return "[";
	}
	
	protected String closingBracket(){
		return "]";
	}
	
	@Override
	public String toString() {//TODO more efficeint implementation
		StringBuilder sb = new StringBuilder();
		sb.append(this.openingBracket());
		boolean first = true;
		for(Type type: this){
			if(first){
				first = false;
			}
			else{
				sb.append(", ");
			}
			sb.append(type.toString());
		}
		sb.append(this.closingBracket());
		return sb.toString();
	}
	
	public final static <Type> Object[] toArray(final Collection<Type> c) {
		Object[] result = new Object[c.size()];
		int i = 0;
		for(Type t: c){
			result[i] = t;
			++i;
		}
		return result;
	}

	@SuppressWarnings("unchecked")
	public final static <T, Type> T[] toArray(final T[] a, final Collection<Type> c) {
		T[] result;
		if(a.length >= c.size()){
			result = a;
		}
		else{
			result = (T[]) Array.newInstance(a.getClass().getComponentType(), c.size());			
		}
		int i = 0;
		for(Type t: c){
			result[i] = (T)t;
			++i;
		}
		return result;
	}
	
	public static <Type> boolean contains(final Object o, final Collection<Type> c) {
		for(Type t: c){
			if(o == null ? t == null : o.equals(t)){
				return true;
			}
		}
		return false;
	}

	public static <Type> boolean containsAll(final Collection<?> c1, final Collection<Type> c) {
		for(Object o: c1){
			if(!c.contains(o)){
				return false;
			}
		}
		return true;
	}
	
	public static final <Type> byte toByte(final Type value){
		return ((Number)value).byteValue();
	}
	
	public static final <Type> short toShort(final Type value){
		return ((Number)value).shortValue();
	}
	
	public static final <Type> int toInt(final Type value){
		return ((Number)value).intValue();
	}
	
	public static final <Type> long toLong(final Type value){
		return ((Number)value).longValue();
	}
	
	public static final <Type> float toFloat(final Type value){
		return ((Number)value).floatValue();
	}
	
	public static final <Type> double toDouble(final Type value){
		return ((Number)value).doubleValue();
	}
	
	public static final <Type> char toChar(final Type value){
		return (Character) value; 
	}
	
	public static final <Type> boolean toBoolean(final Type value){
		return (Boolean) value;
	}

}
