package de.lmu.ifi.pms.java.util.collection.array;

import java.util.TreeMap;

import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;
import de.lmu.ifi.pms.java.util.map.javaWrapper.FromJavaNavigableMapExtendibleWrapper;



public enum TypeConstant {

	Integer{
		public final java.lang.Integer zero = java.lang.Integer.valueOf(0);
		
		@Override
		public void initArray(final Object array) {
			Arrays.fill((java.lang.Integer[])array, this.zero);
		}
		
	},
	
	Long{
		public final java.lang.Long zero = java.lang.Long.valueOf(0l);
		
		@Override
		public void initArray(final Object array) {
			Arrays.fill((java.lang.Long[])array, this.zero);
		}
	},
	
	Float{
		public final java.lang.Float zero = java.lang.Float.valueOf(0f);
		
		@Override
		public void initArray(final Object array) {
			Arrays.fill((java.lang.Float[])array, this.zero);
		}
	},
	
	Double{
		public final java.lang.Double zero = java.lang.Double.valueOf(0d);
		
		@Override
		public void initArray(final Object array) {
			Arrays.fill((java.lang.Double[])array, this.zero);
		}
	},
	
	Short{
		public final java.lang.Short zero = java.lang.Short.valueOf((short)0);
		
		@Override
		public void initArray(final Object array) {
			Arrays.fill((java.lang.Short[])array, this.zero);
		}
	},
	
	Byte{
		public final java.lang.Byte zero = java.lang.Byte.valueOf((byte)0);
		
		@Override
		public void initArray(final Object array) {
			Arrays.fill((java.lang.Byte[])array, this.zero);
		}
	},
	
	Character{
		public final java.lang.Character zero = java.lang.Character.valueOf('\u0000');
		
		@Override
		public void initArray(final Object array) {
			Arrays.fill((java.lang.Character[])array, this.zero);
		}
	},
	
	Boolean{
		public final java.lang.Boolean zero = java.lang.Boolean.valueOf(false);
		
		@Override
		public void initArray(final Object array) {
			Arrays.fill((java.lang.Boolean[])array, this.zero);
		}
	},
	
	Primitive,
	
	Other
	
	;
	
	public void initArray(final Object array){
		
	}
	
	public static final Package java_lang_package = Package.getPackage("java.lang");
	
	public static final FixedArrayNavigableMap<String, TypeConstant> typeMapping = new FixedArrayNavigableMap<String, TypeConstant>(true);
	
	static{
		TreeMap<String, TypeConstant> map = new TreeMap<String, TypeConstant>();
		TypeConstant[] values = TypeConstant.values();
		for(int i = 0 ; i < 8; ++i){
			map.put(values[i].name(), values[i]);
		}
		typeMapping.fillBody(new FromJavaNavigableMapExtendibleWrapper<String, TypeConstant>(map));
	}
	
	public static TypeConstant get(final Class<?> type){
		if(type.isPrimitive()){
			return Primitive;
		}
		else{
			if(type.getPackage().equals(java_lang_package)){
				int index = typeMapping.keys().indexOf(type.getSimpleName());
				if(index < 0){
					return Other;
				}
				else{
					return typeMapping.values().get(index);
				}
			}
			else{
				return Other;
			}
		}
	}
	
}
