package de.lmu.ifi.pms.java.util.collection.array.extendible;

import java.util.NoSuchElementException;

import de.lmu.ifi.pms.development.UnimplementedMethodError;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.Queue;
import de.lmu.ifi.pms.java.util.collection.array.AbstractExtendibleArray;
import de.lmu.ifi.pms.java.util.collection.array.AbstractModifyableArray;
import de.lmu.ifi.pms.java.util.collection.array.ModifyableArray;


public class GenericIArrayQueue<Type> extends AbstractExtendibleArray<Type> implements Queue<Type>{
	
	protected final AbstractModifyableArray<AbstractModifyableArray<Type>> arrays;
	
	private int capacityExp;
	
	private int size;
	
	public GenericIArrayQueue(final int size, final Class<Type> type) {
		super();
		//baseCapaciyExp = 0;
		//adminCapacity = 32;
		AbstractModifyableArray<Type> baseArray = this.newModifyableArray(1, type);		
		this.arrays = new ModifyableArray<AbstractModifyableArray<Type>>(32/*adminCapacity*/, AbstractModifyableArray.class);
		this.arrays.set(31/*array.length - 1*/, baseArray);
		
		this.capacityExp = 0;
		
		this.initialize(size);
	}
	
	@SuppressWarnings("unchecked")
	public GenericIArrayQueue(final int size) {
		this(size, (Class<Type>)Object.class);
	}
	
	public GenericIArrayQueue() {
		this(0);
	}
	
	public GenericIArrayQueue(final Collection<? extends Type> collection) {
		this(collection.size());
		for(Type e: collection){
			this.add(e);
		}
	}
	
	protected AbstractModifyableArray<Type> newModifyableArray(final int size, final Class<? super Type> type){
		return new ModifyableArray<Type>(size, type);
	}
	
	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractArray#ensuredType()
	 */
	@Override
	public Class<? super Type> ensuredType() {
		assert this.arrays.get(31/*array.length - 1*/) != null;
		return this.arrays.get(31/*array.length - 1*/).ensuredType();
	}

	private void addArray(final int j){
		assert j > 0 && j < 31;
		this.arrays.set(j, this.newModifyableArray(1 << j, this.ensuredType()));
	}
	
	private void ensureArrays(final int j){
		if(j < 31){
			while(this.capacityExp <= j){
				this.addArray(this.capacityExp);
				++this.capacityExp;
			}
		}
	}
	
	private void dropArrays(final int j){
		assert j >= 0;
		while(this.capacityExp > j){
			--this.capacityExp;
			this.arrays.set(this.capacityExp, null);
		}
	}
	
	protected void initialize(final int size){
		assert size >= this.size();
		if(size > 0){
			int j = (63 - Integer.numberOfLeadingZeros(size-1)) & 31;// x & 31 = x % 32
			this.ensureArrays(j);
			this.size = size;
		}
		else{
			this.size = 0;
		}
	}
	
	@Override
	public Type get(int i){
		this.checkIndex(i);
		int j = (63 - Integer.numberOfLeadingZeros(i)) & 31;// x & 31 = x % 32
		i -= ((1 << j) & 2147483647);//(1 << 31) & Integer.MaxValue = 0 ; 2147483647 = Integer.MAX_VALUE
		return this.arrays.get(j).get(i);
	}
	
	@Override
	public void set(int i, final Type value){
		int j = (63 - Integer.numberOfLeadingZeros(i)) & 31;// x & 31 = x % 32
		if(i >= this.size()){
			this.ensureArrays(j);
			this.size = i + 1;
		}
		i -= ((1 << j) & 2147483647);//(1 << 31) & Integer.MaxValue = 0 ; 2147483647 = Integer.MAX_VALUE
		this.arrays.get(j).set(i, value);
	}

	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractExtendibleArray#insert(int, java.lang.Object)
	 */
	@Override
	public void insert(final int i, final Type value) {
		if(i < this.size()){
			Type last = this.last();
			this.insertIndexDropLastUnchecked(i, value);
			this.addLast(last);
		}
		else{
			this.set(i, value);
		}
	}
	
	protected void insertIndexDropLastUnchecked(int i, final Type value) {
		assert i < this.size();
		int j = (63 - Integer.numberOfLeadingZeros(i)) & 31;// x & 31 = x % 32
		if(j == 31){
			this.arrays.get(0/*second array; size 1*/).set(0, this.arrays.get(31/*first array; size 1*/).get(0));
			for(int greater_j = this.capacityExp-1; greater_j > 0;){
				this.arrays.get(greater_j).addFirstDropLast(this.arrays.get(--greater_j).last());
			}
			assert i == 0;
			this.arrays.get(31/*first array; size 1*/).set(0/* size 1 => i == 0*/, value);
		}
		else{
			for(int greater_j = this.capacityExp-1; greater_j > j;){
				this.arrays.get(greater_j).addFirstDropLast(this.arrays.get(--greater_j).last());
			}
			i -= ((1 << j) & 2147483647);//(1 << 31) & Integer.MaxValue = 0 ; 2147483647 = Integer.MAX_VALUE
			this.arrays.get(j).set(i, value);
		}
	}
	
	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractModifyableArray#insertIndexDropLast(int, java.lang.Object)
	 */
	@Override
	public void insertIndexDropLast(final int i, final Type value) {
		 if(i < this.size()){
			this.insertIndexDropLastUnchecked(i, value);
		 }
		 else{
			 this.set(i, value);
		 }
	}
	
	@Override
	public void addLast(final Type value){
		this.set(this.size(), value);
	}
	
	@Override
	public void removeLast(){
		this.removeLast(1);
	}
	
	@Override
	public void removeLast(final int n){
		this.setSize(this.size() - n);
	}
	
	@Override
	public void setSize(final int size){
		if(size <= 0){
			this.dropArrays(1);
			this.size = 0;
		}
		else if(size > this.size()){
			this.initialize(size);
		}
		else{
			int j = (63 - Integer.numberOfLeadingZeros(size-1)) & 31;// x & 31 = x % 32
			this.dropArrays(j);
			//FIXME blocks garbage collection of Objects! Needs to set to null on Object array
			//FIXME dereasing size + increasing size means values at positions where no/default values are expected
			this.size = size;
		}
	}
	
	
	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractExtendibleArray#setExpSize(int)
	 */
	@Override
	public void setExpSize(final int exp) {
		 if(exp > this.capacityExp){
			 this.ensureArrays((exp + 32) & 31);
		 }
		 else{
			 this.dropArrays(exp);
		 }
	}
	
	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractArray#minIndex()
	 */
	@Override
	public int minIndexInc() {
		return 0;
	}
	
	@Override
	public int size(){
		return this.size;
	}
	
	@Override
	public ExtendibleArrayIterator<Type, GenericIArrayQueue<Type>> iterator(){
		return new ExtendibleArrayIterator<Type, GenericIArrayQueue<Type>>(this){
			
			int index = 0;
			
			int i = 0;
			int j = 31;
			
			@Override
			public boolean hasNext() {
				return this.index < GenericIArrayQueue.this.size();
			}
			
			@Override
			public Type next() {
				if(this.hasNext()){
					Type next = GenericIArrayQueue.this.arrays.get(this.j).get(this.i);
					++this.i;
					if(this.i == GenericIArrayQueue.this.arrays.get(this.j).size()){
						this.i = 0;
						this.j = (this.j + 1) & 31;
					}
					++this.index;
					
					++this.position;
					return next;
				}
				else{
					throw new NoSuchElementException();//TODO better exception
				}
			}
			
		};
	}
	
	@Override
	public Type peek() {
		if(this.isEmpty()){
			return null;
		}
		else{
			return this.element();
		}
	}
	
	@Override
	public Type element() {
		return this.last();
	}
	
	public boolean offer(final Type e) {
		this.addLast(e);
		return true;
	}
	
	public Type poll() {
		if(this.isEmpty()){
			return null;
		}
		else{
			return this.remove();
		}
	}
	
	public Type remove() {
		Type head = this.last();
		this.removeLast();
		return head;
	}
	
	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractExtendibleArray#remove(int)
	 */
	@Override
	public void remove(final int index) {
		 //FIXME implement (see insertFirstRemoveLast , will uses insertLastReoveFirst
		throw new UnimplementedMethodError();
	}
	
	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractExtendibleArray#subArray(int, int)
	 */
	@Override
	public AbstractExtendibleArray<Type> subArray(final int offset, final int length) {
		//FIXME implement
		throw new UnimplementedMethodError();
	}
	
	@Override
	public <T> void copyTo(final int sourceStart, final T[] destination, final int destinationStart, final int count) {
		int sourceEnd = sourceStart + count;
		if(count > 0){
			this.checkIndex(sourceStart);		
			this.checkIndex(sourceEnd-1);
			int j_start = (63 - Integer.numberOfLeadingZeros(sourceStart));// x & 31 = x % 32
			int j_end = ((63 - Integer.numberOfLeadingZeros(sourceEnd - 1))) + 1;// x & 31 = x % 32
			int i_src = sourceStart - ((1 << j_start) & 2147483647);
			int i_dest = destinationStart;
			int remainingCount = count;
			int local_count;
			AbstractModifyableArray<Type> localArray;
			for(int j = j_start; j < j_end; ++j, i_src = 0){
				localArray = this.arrays.get(j & 31);
				local_count = localArray.size();
				if(remainingCount < local_count){
					local_count = remainingCount;
				}			
				localArray.copyTo(i_src, destination, i_dest, local_count);
				remainingCount -= local_count;
				i_dest += local_count;
			}
		}
	}
	
	@Override
	public void copyTo(final int sourceStart, final int[] destination, final int destinationStart, final int count) {
		int sourceEnd = sourceStart + count;
		if(count > 0){
			this.checkIndex(sourceStart);		
			this.checkIndex(sourceEnd-1);
			int j_start = (63 - Integer.numberOfLeadingZeros(sourceStart));// x & 31 = x % 32
			int j_end = ((63 - Integer.numberOfLeadingZeros(sourceEnd - 1))) + 1;// x & 31 = x % 32
			int i_src = sourceStart - ((1 << j_start) & 2147483647);
			int i_dest = destinationStart;
			int remainingCount = count;
			int local_count;
			AbstractModifyableArray<Type> localArray;
			for(int j = j_start; j < j_end; ++j, i_src = 0){
				localArray = this.arrays.get(j & 31);
				local_count = localArray.size();
				if(remainingCount < local_count){
					local_count = remainingCount;
				}			
				localArray.copyTo(i_src, destination, i_dest, local_count);
				remainingCount -= local_count;
				i_dest += local_count;
			}
		}
	}
	
	public static void main(final String[] args) {
		/*
		Collection<Integer> ia = (args.length == 0 ? new GenericIArrayQueue<Integer>(0,Integer.TYPE) : new Array<Integer>(1));
		int b = (1 <<  6)-1;
		
		Collection<Integer> ints = new Array<Integer>(b);
		for(int i = 0; i <= b; ++i){
			ints.add(i);
		}
		for(Integer i: ints){
			ia.add(i);
		}
		for(Integer i: ia){
			System.out.println(i);
		}
		*/
	}

}
