package de.lmu.ifi.pms.java.util.collection.array.extendible;

import static de.lmu.ifi.pms.java.util.collection.array.Arrays.getArray;

import java.util.NoSuchElementException;

import de.lmu.ifi.pms.development.UnimplementedMethodError;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.Queue;
import de.lmu.ifi.pms.java.util.collection.array.AbstractExtendibleArray;
import de.lmu.ifi.pms.java.util.collection.array.Arrays;


public class IArrayQueue<Type> extends AbstractExtendibleArray<Type> implements Queue<Type>{
	
	protected final Type[][] arrays;
	
	private int capacityExp;
	
	private int size;
	
	public IArrayQueue(final Class<? super Type> ensuredType) {
		this(0, ensuredType);
	}
	
	@SuppressWarnings("unchecked")
	public IArrayQueue(final int size, final Class<? super Type> ensuredType) {
		super();
		//baseCapaciyExp = 0;
		//adminCapacity = 32;
		Type[] baseArray = getArray((Class<Type>)ensuredType, 1);		
		this.arrays = getArray((Class<Type[]>)baseArray.getClass(), 32/*adminCapacity*/);
		this.arrays[31/*arrays.length - 1*/] = baseArray;
		
		this.capacityExp = 0;
		
		this.initialize(size);
	}
	
	public IArrayQueue(final int size) {
		this(size, Object.class);
	}
	
	public IArrayQueue() {
		this(0);
	}
	
	public IArrayQueue(final Collection<? extends Type> collection) {
		this(collection.size());
		int index = 0;
		for(Type e: collection){
			this.set(index, e);
			++index;
		}
	}
	
	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractArray#ensuredType()
	 */
	@Override
	@SuppressWarnings("unchecked")
	public Class<? super Type> ensuredType() {
		return (Class<? super Type>)this.arrays.getClass().getComponentType().getComponentType();
	}

	private void addArray(final int j){
		assert j >= 0 && j < 31 && this.arrays[31/*arrays.length - 1*/] != null;
		this.arrays[j] = getArray(this.arrays[31/*arrays.length - 1*/], 1 << j);
	}
	
	private void ensureArrays(final int j){
		if(j < 31){
			while(this.capacityExp <= j){
				this.addArray(this.capacityExp);
				++this.capacityExp;
			}
		}
	}
	
	private void dropArrays(int j){
		assert j >= 0 && j <= 31;		
		++j;
		j &= 31;
		while(this.capacityExp > j){
			--this.capacityExp;
			this.arrays[this.capacityExp] = null;
		} 
	}
	
	protected void initialize(final int size){
		assert size >= this.size();
		if(size > 0){
			int j = (63 - Integer.numberOfLeadingZeros(size-1)) & 31;// x & 31 = x % 32
			this.ensureArrays(j);
			this.size = size;
		}
		else{
			this.size = 0;
		}
	}
	
	@Override
	public Type get(int i){
		this.checkIndex(i);
		int j = (63 - Integer.numberOfLeadingZeros(i)) & 31;// x & 31 = x % 32
		i -= ((1 << j) & 2147483647);//(1 << 31) & Integer.MaxValue = 0 ; 2147483647 = Integer.MAX_VALUE
		return this.arrays[j][i];
	}
	
	@Override
	public void set(int i, final Type value){
		int j = (63 - Integer.numberOfLeadingZeros(i)) & 31;// x & 31 = x % 32
		if(i >= this.size()){
			this.ensureArrays(j);
			this.size = i + 1;
		}
		i -= ((1 << j) & 2147483647);//(1 << 31) & Integer.MaxValue = 0 ; 2147483647 = Integer.MAX_VALUE
		this.arrays[j][i] = value;
	}
	
	@Override
	public void addLast(final Type value){
		this.set(this.size(), value);
	}
	
	@Override
	public void removeLast(){
		this.removeLast(1);
	}
	
	@Override
	public void removeLast(final int n){
		this.setSize(this.size() - n);
	}
	
	@Override
	public void setSize(final int size){
		if(size <= 0){
			this.dropArrays(1);
			this.size = 0;
		}
		else if(size > this.size()){
			this.initialize(size);
		}
		else{
			int j = (63 - Integer.numberOfLeadingZeros(size - 1)) & 31;// x & 31 = x % 32
			this.dropArrays(j);
			int offset = ((1 << j) & 2147483647);//(1 << 31) & Integer.MaxValue = 0 ; 2147483647 = Integer.MAX_VALUE
			int newLocalSize = size - offset;
			int oldLocalSize = this.size - offset;
			Type[] lastArray = this.arrays[j];
			if(oldLocalSize > lastArray.length){
				oldLocalSize = lastArray.length;
			}
			for(;newLocalSize < oldLocalSize; ++newLocalSize){//TODO improve
				lastArray[newLocalSize] = null;
			}
			this.size = size;
		}
	}
	
	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractExtendibleArray#setExpSize(int)
	 */
	@Override
	public void setExpSize(final int exp) {
		 if(exp > this.capacityExp){
			 this.ensureArrays((exp + 32) & 31);
		 }
		 else{
			 this.dropArrays(exp);
		 }
	}
	
	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractArray#minIndex()
	 */
	@Override
	public int minIndexInc() {
		return 0;
	}
	
	@Override
	public int size(){
		return this.size;
	}
	
	@Override
	public ExtendibleArrayIterator<Type, IArrayQueue<Type>> iterator(){
		return new ExtendibleArrayIterator<Type, IArrayQueue<Type>>(this){
			
			int index = 0;
			
			int i = 0;
			int j = 31;
			
			@Override
			public boolean hasNext() {
				return this.index < IArrayQueue.this.size();
			}
			
			@Override
			public Type next() {
				if(this.hasNext()){
					Type next = IArrayQueue.this.arrays[this.j][this.i];
					++this.i;
					if(this.i == IArrayQueue.this.arrays[this.j].length){
						this.i = 0;
						this.j = (this.j + 1) & 31;
					}
					++this.index;
					
					++this.position;
					return next;
				}
				else{
					throw new NoSuchElementException();//TODO better exception
				}
			}
			
		};
	}
	
	@Override
	public Type peek() {
		if(this.isEmpty()){
			return null;
		}
		else{
			return this.element();
		}
	}
	
	@Override
	public Type element() {
		return this.last();
	}
	
	public boolean offer(final Type e) {
		this.addLast(e);
		return true;
	}
	
	public Type poll() {
		if(this.isEmpty()){
			return null;
		}
		else{
			return this.remove();
		}
	}
	
	public Type remove() {
		Type head = this.last();
		this.removeLast();
		return head;
	}
	
	@Override
	public void insert(final int i, final Type value) {
		if(i < this.size()){
			Type last = this.last();
			this.insertIndexDropLastUnchecked(i, value);
			this.addLast(last);
		}
		else{
			this.set(i, value);
		}
	}
	
	protected void insertIndexDropLastUnchecked(int i, final Type value) {
		assert i < this.size();
		int j = (63 - Integer.numberOfLeadingZeros(i)) & 31;// x & 31 = x % 32
		if(j == 31){
			this.arrays[0/*second array; size 1*/][0] = this.arrays[31/*first array; size 1*/][0];
			for(int greater_j = this.capacityExp-1; greater_j > 0;){
				Arrays.addFirstDropLast(this.arrays[greater_j], Arrays.last(this.arrays[--greater_j]));
			}
			assert i == 0;
			this.arrays[31/*first array; size 1*/][0/* size 1 => i == 0*/] = value;
		}
		else{
			for(int greater_j = this.capacityExp-1; greater_j > j;){
				Arrays.addFirstDropLast(this.arrays[greater_j], Arrays.last(this.arrays[--greater_j]));
			}
			i -= ((1 << j) & 2147483647);//(1 << 31) & Integer.MaxValue = 0 ; 2147483647 = Integer.MAX_VALUE
			this.arrays[j][i] = value;
		}
	}
	
	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractModifyableArray#insertIndexDropLast(int, java.lang.Object)
	 */
	@Override
	public void insertIndexDropLast(final int i, final Type value) {
		 if(i < this.size()){
			this.insertIndexDropLastUnchecked(i, value);
		 }
		 else{
			 this.set(i, value);
		 }
	}
	
	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractExtendibleArray#remove(int)
	 */
	@Override
	public void remove(final int index) {
		//FIXME implement
		throw new UnimplementedMethodError();
	}
	
	/* (non-Javadoc)
	 * @see extension.java.util.collection.array.AbstractExtendibleArray#subArray(int, int)
	 */
	@Override
	public AbstractExtendibleArray<Type> subArray(final int offset, final int length) {
		//FIXME implement
		throw new UnimplementedMethodError();
	}
	
	@Override
	public <T> void copyTo(final int sourceStart, final T[] destination, final int destinationStart, final int count) {
		int sourceEnd = sourceStart + count;
		if(count > 0){
			this.checkIndex(sourceStart);		
			this.checkIndex(sourceEnd-1);
			int j_start = (63 - Integer.numberOfLeadingZeros(sourceStart));// x & 31 = x % 32
			int j_end = ((63 - Integer.numberOfLeadingZeros(sourceEnd - 1))) + 1;// x & 31 = x % 32
			int i_src = sourceStart - ((1 << j_start) & 2147483647);
			int i_dest = destinationStart;
			int remainingCount = count;
			int local_count;
			Type[] localArray;
			for(int j = j_start; j < j_end; ++j, i_src = 0){
				localArray = this.arrays[j & 31];
				local_count = localArray.length;
				if(remainingCount < local_count){
					local_count = remainingCount;
				}			
				System.arraycopy(localArray, i_src, destination, i_dest, local_count);
				remainingCount -= local_count;
				i_dest += local_count;
			}
		}
	}
	
	@Override
	public void copyTo(final int sourceStart, final int[] destination, final int destinationStart, final int count) {
		int sourceEnd = sourceStart + count;
		if(count > 0){
			this.checkIndex(sourceStart);		
			this.checkIndex(sourceEnd-1);
			int j_start = (63 - Integer.numberOfLeadingZeros(sourceStart));// x & 31 = x % 32
			int j_end = ((63 - Integer.numberOfLeadingZeros(sourceEnd - 1))) + 1;// x & 31 = x % 32
			int i_src = sourceStart - ((1 << j_start) & 2147483647);
			int i_dest = destinationStart;
			int remainingCount = count;
			int local_count;
			Type[] localArray;
			for(int j = j_start; j < j_end; ++j, i_src = 0){
				localArray = this.arrays[j & 31];
				local_count = localArray.length;
				if(remainingCount < local_count){
					local_count = remainingCount;
				}			
				System.arraycopy(localArray, i_src, destination, i_dest, local_count);
				remainingCount -= local_count;
				i_dest += local_count;
			}
		}
	}
	
	public static void main(final String[] args) {
		/*
		Collection<Integer> ia = (args.length == 0 ? new IArrayQueue<Integer>(0,Integer.TYPE) : new ArrayList<Integer>(1));
		int b = (1 <<  16)-1;
		
		Collection<Integer> ints = new ArrayList<Integer>(b);
		for(int i = 1; i <= b; ++i){
			ints.add(i);
		}
		for(Integer i: ints){
			ia.add(i);
		}
		for(Integer i: ia){
			System.out.println(i);
		}*/
	}

}
