package de.lmu.ifi.pms.java.util.collection.composition;

import de.lmu.ifi.pms.java.lang.iteration.AbstractIterator;
import de.lmu.ifi.pms.java.lang.iteration.ComposedIterator;
import de.lmu.ifi.pms.java.util.collection.AbstractCollection;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.array.fixed.FixedArray;

public abstract class ComposedCollection<Type, CollectionType extends Collection<Type>> extends AbstractCollection<Type> {
	
	protected final FixedArray<CollectionType> collections;
	
	
	public ComposedCollection(final CollectionType... collections){
		this(new FixedArray<CollectionType>(collections));
	}
	
	@SuppressWarnings("unchecked")
	public ComposedCollection(final Collection<? extends CollectionType> collections){
		this(new FixedArray<CollectionType>((Collection<CollectionType>)collections));		
	}
	
	public ComposedCollection(final FixedArray<CollectionType> collections){
		super();
		this.collections = collections;
	}

	@Override
	public AbstractIterator<Type> iterator() {
		return new ComposedIterator<Type>(this.collections, true);
	}
	
	@Override
	public int size() {
		int size = 0;
		for(Collection<? extends Type> collection: this.collections){
			size = size + collection.size();
		}
		return size;
	}

}
