package de.lmu.ifi.pms.java.util.collection.fixed.empty;

import java.util.Comparator;
import java.util.NoSuchElementException;

import de.lmu.ifi.pms.java.lang.iteration.AbstractFixedIterator;
import de.lmu.ifi.pms.java.util.collection.fixed.set.AbstractFixedNavigableSet;
import de.lmu.ifi.pms.java.util.collection.fixed.set.FixedNavigableSet;
import de.lmu.ifi.pms.java.util.collection.fixed.set.FixedSortedSet;


public class Empty<Type> extends AbstractFixedNavigableSet<Type> {
	
	@SuppressWarnings("rawtypes")
	private static Empty empty = new Empty();
	
	@SuppressWarnings("unchecked")
	public static final <Type> Empty<Type> instance(){
		return empty;
	}
	
	public static final <Type> Empty<Type> instance(Comparator<? super Type> comparator){
		return new Empty<Type>(comparator);
	}
	
	private final Comparator<? super Type> comparator;

	private Empty() {
		this(null);
	}
	
	private Empty(final Comparator<? super Type> comparator) {
		super();
		this.comparator = comparator;
	}

	@Override
	public final Type ceiling(Type e) {
		return null;
	}

	@Override
	public Comparator<? super Type> comparator() {
		return this.comparator;
	}

	@Override
	public AbstractFixedIterator<Type> descendingIterator() {
		return EmptyIterator.instance();
	}

	@Override
	public FixedNavigableSet<Type> descendingSet() {
		return this;
	}

	@Override
	public Type first() {
		throw new NoSuchElementException(); //TODO better exception
	}

	@Override
	public final Type floor(Type e) {
		return null;
	}

	@Override
	public final FixedSortedSet<Type> headSet(Type toElement) {
		return this;
	}

	@Override
	public final FixedNavigableSet<Type> headSet(Type toElement, boolean inclusive) {
		return this;
	}

	@Override
	public final Type higher(Type e) {
		return null;
	}

	@Override
	public AbstractFixedIterator<Type> iterator() {
		return EmptyIterator.instance();
	}

	@Override
	public final Type last() {
		throw new NoSuchElementException(); //TODO better exception
	}

	@Override
	public final Type lower(Type e) {
		return null;
	}

	@Override
	public final int size() {
		return 0;
	}

	@Override
	public FixedSortedSet<Type> subSet(Type fromElement, Type toElement) {
		return this;
	}

	@Override
	public FixedNavigableSet<Type> subSet(Type fromElement, boolean fromInclusive, Type toElement, boolean toInclusive) {
		return this;
	}

	@Override
	public FixedSortedSet<Type> tailSet(Type fromElement) {
		return this;
	}

	@Override
	public FixedNavigableSet<Type> tailSet(Type fromElement, boolean inclusive) {
		return this;
	}
	
	@Override
	public final boolean isEmpty() {
		return true;
	}

}
