package de.lmu.ifi.pms.java.util.collection.javaWrapper;

import java.util.Comparator;

import de.lmu.ifi.pms.java.lang.iteration.ExtendibleIterator;
import de.lmu.ifi.pms.java.lang.iteration.JavaIteratorExtendibleWrapper;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.extendible.set.ExtendibleSortedSet;
import de.lmu.ifi.pms.java.util.collection.set.AbstractSortedSet;
import de.lmu.ifi.pms.java.util.collection.set.SortedSet;


public class FromJavaExtendibleSortedSetWrapper<Type> extends AbstractSortedSet<Type> implements ExtendibleSortedSet<Type>{
	
	private final java.util.SortedSet<Type> javaCollection;
	
	public FromJavaExtendibleSortedSetWrapper(final java.util.SortedSet<Type> javaCollection) {
		super();
		this.javaCollection = javaCollection;
	}

	@Override
	public boolean contains(final Object o) {
		return this.javaCollection.contains(o);
	}
	
	@Override
	public boolean containsAll(final de.lmu.ifi.pms.java.util.collection.Collection<?> c) {
		return this.javaCollection.containsAll(c.javaCollection());
	}
	
	@Override
	public boolean isEmpty() {
		return this.javaCollection.isEmpty();
	}
	
	@Override
	public AbstractJavaSortedSet<Type> javaCollection() {
		if(this.javaCollection instanceof AbstractJavaSortedSet){
			return (AbstractJavaSortedSet<Type>) this.javaCollection;
		}
		else{
			return new ToJavaSortedSetWrapper<Type>(this);
		}
	}
	
	@Override
	public boolean add(final Type value) {
		return this.javaCollection.add(value);
	}
	
	@Override
	public void clear() {
		this.javaCollection.clear();
	}
	
	@Override
	public boolean remove(final Object o) {
		return this.javaCollection.remove(o);
	}
	
	@Override
	public Object[] toArray() {
		return this.javaCollection.toArray();
	}
	
	@Override
	public <T> T[] toArray(final T[] a) {
		return this.javaCollection.toArray(a);
	}

	@Override
	public ExtendibleIterator<Type> iterator() {
		return new JavaIteratorExtendibleWrapper<Type>(this.javaCollection);
	}

	@Override
	public final int size() {
		return this.javaCollection.size();
	}

	@Override
	public boolean addAll(final java.util.Collection<? extends Type> c) {
		return this.javaCollection.addAll(c);
	}

	@Override
	public boolean addAll(final de.lmu.ifi.pms.java.util.collection.Collection<? extends Type> c) {
		return this.addAll(c.javaCollection());
	}

	@Override
	public boolean removeAll(final java.util.Collection<?> c) {
		return this.javaCollection.removeAll(c);
	}
	
	@Override
	public boolean removeAll(final Collection<?> c) {
		return this.removeAll(c.javaCollection());
	}
	
	@Override
	public boolean retainAll(final java.util.Collection<?> c) {
		return this.javaCollection.retainAll(c);
	}

	@Override
	public boolean retainAll(final Collection<?> c) {
		return this.retainAll(c.javaCollection());
	}

	@Override
	public Comparator<? super Type> comparator() {
		return this.javaCollection.comparator();
	}

	@Override
	public Type first() {
		return this.javaCollection.first();
	}

	@Override
	public Type last() {
		return this.javaCollection.last();
	}

	@Override
	public SortedSet<Type> headSet(final Type toElement) {
		return new FromJavaExtendibleSortedSetWrapper<Type>(this.javaCollection.headSet(toElement));
	}

	@Override
	public SortedSet<Type> tailSet(final Type fromElement) {
		return new FromJavaExtendibleSortedSetWrapper<Type>(this.javaCollection.tailSet(fromElement));
	}

	@Override
	public SortedSet<Type> subSet(final Type fromElement, final Type toElement) {
		return new FromJavaExtendibleSortedSetWrapper<Type>(this.javaCollection.subSet(fromElement, toElement));
	}
	
	

}
