/**
 * 
 */
package de.lmu.ifi.pms.java.util.map.javaWrapper;

import java.util.NoSuchElementException;

import de.lmu.ifi.pms.java.lang.iteration.AbstractModifyableIterator;
import de.lmu.ifi.pms.java.lang.iteration.ModifyableIterator;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.fixed.wrapper.FixedCollectionWrapper;
import de.lmu.ifi.pms.java.util.collection.fixed.wrapper.FixedSetWrapper;
import de.lmu.ifi.pms.java.util.collection.javaWrapper.FromJavaCollectionWrapper;
import de.lmu.ifi.pms.java.util.collection.set.Set;
import de.lmu.ifi.pms.java.util.map.AbstractModifyableMap;




/**
 * @author Simon Matthias Brodt
 *
 */
public class FromJavaMapModifyableWrapper<KeyType, ValueType>
		extends AbstractModifyableMap<KeyType, ValueType, FromJavaEntryModifyableWrapper<KeyType, ValueType>> {

	final java.util.Map<KeyType, ValueType> javaMap;

	public FromJavaMapModifyableWrapper(java.util.Map<KeyType, ValueType> javaMap) {
		super();
		this.javaMap = javaMap;
	}

	@Override
	public FromJavaEntryModifyableWrapper<KeyType, ValueType> entry(KeyType key) {
		for(java.util.Map.Entry<KeyType, ValueType> javaEntry: this.javaMap.entrySet()){
			if(javaEntry.getKey().equals(key)){
				return new FromJavaEntryModifyableWrapper<KeyType, ValueType>(javaEntry);
			}
		}
		throw new NoSuchElementException();//TODO better exception
	}

	@Override
	public ValueType get(KeyType key) {
		return this.javaMap.get(key);
	}

	@Override
	public boolean has(KeyType key) {
		return this.javaMap.containsKey(key);
	}

	@Override
	public ModifyableIterator<FromJavaEntryModifyableWrapper<KeyType, ValueType>> iterator() {
		final java.util.Iterator<java.util.Map.Entry<KeyType, ValueType>> javaIterator = this.javaMap.entrySet().iterator();
		return new AbstractModifyableIterator<FromJavaEntryModifyableWrapper<KeyType,ValueType>>() {
			
			FromJavaEntryModifyableWrapper<KeyType, ValueType> current;

			@Override
			public boolean hasNext() {
				return javaIterator.hasNext();
			}

			@Override
			public FromJavaEntryModifyableWrapper<KeyType, ValueType> next() {
				this.current = new FromJavaEntryModifyableWrapper<KeyType, ValueType>(javaIterator.next());
				return this.current;
			}

			@Override
			public void replace(FromJavaEntryModifyableWrapper<KeyType, ValueType> newValue) {
				if(this.current == null){
					throw new IllegalStateException();//TODO better exception
				}
				else{
					if(this.current.key().equals(newValue)){
						try{
							this.current.setValue(newValue.value());
						}
						catch (UnsupportedOperationException e) {
							set(newValue.key(), newValue.value());
						}
					}
					else{
						throw new IllegalArgumentException();//TODO better exception
					}
				}
			}

		};
	}

	@Override
	public Set<KeyType> keys() {
		return new FixedSetWrapper<KeyType>(this.javaMap.keySet());
	}


	@Override
	public int size() {
		return this.javaMap.size();
	}

	@Override
	public Collection<ValueType> values() {
		return new FixedCollectionWrapper<ValueType>(new FromJavaCollectionWrapper<ValueType>(this.javaMap.values()));
	}

	@Override
	public void set(KeyType key, ValueType value) {
		if(this.has(key)){
			this.javaMap.put(key, value);
		}
		else{
			throw new NoSuchElementException();//TODO better exception
		}
	}
	
}
