/**
 * 
 */
package de.lmu.ifi.pms.java.util.map.javaWrapper;

import java.util.Comparator;
import java.util.NoSuchElementException;

import de.lmu.ifi.pms.java.lang.iteration.ExtendibleIterator;
import de.lmu.ifi.pms.java.util.adaption.ExtendibleIteratorAdapter;
import de.lmu.ifi.pms.java.util.collection.Collection;
import de.lmu.ifi.pms.java.util.collection.fixed.wrapper.FixedCollectionWrapper;
import de.lmu.ifi.pms.java.util.collection.fixed.wrapper.FixedNavigableSetWrapper;
import de.lmu.ifi.pms.java.util.collection.javaWrapper.FromJavaCollectionWrapper;
import de.lmu.ifi.pms.java.util.collection.set.NavigableSet;
import de.lmu.ifi.pms.java.util.map.AbstractNavigableMap;
import de.lmu.ifi.pms.java.util.map.Entry;
import de.lmu.ifi.pms.java.util.map.ExtendibleNavigableMap;
import de.lmu.ifi.pms.java.util.map.Map;
import de.lmu.ifi.pms.java.util.map.NavigableMap;



/**
 * @author Simon Matthias Brodt
 *
 */
public class FromJavaNavigableMapExtendibleWrapper<KeyType, ValueType>
		extends AbstractNavigableMap<KeyType, ValueType, FromJavaEntryExtendibleWrapper<KeyType, ValueType>> implements ExtendibleNavigableMap<KeyType, ValueType, FromJavaEntryExtendibleWrapper<KeyType, ValueType>>{

	final java.util.NavigableMap<KeyType, ValueType> javaMap;

	public FromJavaNavigableMapExtendibleWrapper(final java.util.NavigableMap<KeyType, ValueType> javaMap) {
		super();
		this.javaMap = javaMap;
	}

	@Override
	public Entry<KeyType, ValueType> ceilingEntry(final KeyType key) {
		return new FromJavaEntryExtendibleWrapper<KeyType, ValueType>(this.javaMap.ceilingEntry(key), this.javaMap);
	}

	@Override
	public NavigableMap<KeyType, ValueType, ?> descendingMap() {
		return new FromJavaNavigableMapExtendibleWrapper<KeyType, ValueType>(this.javaMap.descendingMap());
	}

	@Override
	public FromJavaEntryExtendibleWrapper<KeyType, ValueType> entry(final KeyType key) {
		if(this.has(key)){
			return new FromJavaEntryExtendibleWrapper<KeyType, ValueType>(this.javaMap.floorEntry(key), this.javaMap);
		}
		else{
			throw new NoSuchElementException();//TODO better exception
		}
	}

	@Override
	public Entry<KeyType, ValueType> firstEntry() {
		return new FromJavaEntryExtendibleWrapper<KeyType, ValueType>(this.javaMap.firstEntry(), this.javaMap);
	}

	@Override
	public Entry<KeyType, ValueType> floorEntry(final KeyType key) {
		return new FromJavaEntryExtendibleWrapper<KeyType, ValueType>(this.javaMap.floorEntry(key), this.javaMap);
	}

	@Override
	public KeyType floorKey(final KeyType key) {
		return this.javaMap.floorKey(key);
	}

	@Override
	public ValueType get(final KeyType key) {
		return this.javaMap.get(key);
	}

	@Override
	public boolean has(final KeyType key) {
		return this.javaMap.containsKey(key);
	}

	@Override
	public NavigableMap<KeyType, ValueType, ?> headMap(final KeyType toKey, final boolean inclusive) {
		return new FromJavaNavigableMapExtendibleWrapper<KeyType, ValueType>(this.javaMap.headMap(toKey, inclusive));
	}

	@Override
	public Entry<KeyType, ValueType> higherEntry(final KeyType key) {
		return new FromJavaEntryExtendibleWrapper<KeyType, ValueType>(this.javaMap.higherEntry(key), this.javaMap);
	}

	/* (non-Javadoc)
	 * @see extension.java.util.map.AbstractNavigableMap#iterator()
	 */
	@Override
	public ExtendibleIterator<FromJavaEntryExtendibleWrapper<KeyType, ValueType>> iterator() {
		return new ExtendibleIteratorAdapter<java.util.Map.Entry<KeyType, ValueType>, FromJavaEntryExtendibleWrapper<KeyType,ValueType>>(this.javaMap.entrySet()){
			
			/* (non-Javadoc)
			 * @see extension.java.lang.iteration.IteratorAdapter#adapt(java.lang.Object)
			 */
			@Override
			public FromJavaEntryExtendibleWrapper<KeyType, ValueType> adapt(final java.util.Map.Entry<KeyType, ValueType> source) {
				return new FromJavaEntryExtendibleWrapper<KeyType, ValueType>(source, FromJavaNavigableMapExtendibleWrapper.this.javaMap);
			}
			
			@Override
			public java.util.Map.Entry<KeyType, ValueType> inverseAdapt(final FromJavaEntryExtendibleWrapper<KeyType, ValueType> source) {
				return source.javaEntry;
			}
			
		};
	}

	@Override
	public NavigableSet<KeyType> keys() {
		return new FixedNavigableSetWrapper<KeyType>(this.javaMap.navigableKeySet());
	}

	@Override
	public Entry<KeyType, ValueType> lastEntry() {
		return new FromJavaEntryExtendibleWrapper<KeyType, ValueType>(this.javaMap.lastEntry(), this.javaMap);
	}

	@Override
	public Entry<KeyType, ValueType> lowerEntry(final KeyType key) {
		return new FromJavaEntryExtendibleWrapper<KeyType, ValueType>(this.javaMap.lowerEntry(key), this.javaMap);
	}
	
	@Override
	@Deprecated
	public final boolean put(final KeyType key, final ValueType value) {
		return this.add(key, value);
	}
	
	@Override
	public boolean add(final KeyType key, final ValueType value) {
		ValueType oldValue = this.javaMap.put(key, value);
		return oldValue != null;
	}
	
	@Override
	public boolean add(final Entry<KeyType, ValueType> entry){
		return this.add(entry.key(), entry.value());
	}
	
	@Override
	public boolean add(final FromJavaEntryExtendibleWrapper<KeyType,ValueType> entry){
		return this.add((Entry<KeyType, ValueType>)entry);
	}
	
	@Override
	public boolean addAll(final java.util.Collection<? extends FromJavaEntryExtendibleWrapper<KeyType,ValueType>> c) {
		return this.addAll(new FromJavaCollectionWrapper<FromJavaEntryExtendibleWrapper<KeyType,ValueType>>(c));
	}
	
	@Override
	public boolean addAll(final Collection<? extends FromJavaEntryExtendibleWrapper<KeyType,ValueType>> c){
		boolean modified = false;
		for(FromJavaEntryExtendibleWrapper<KeyType,ValueType> entry: c){
			if(this.add(entry)){
				modified = true;
			}
		}
		return modified;
	}
	
	@Override
	public boolean add(final java.util.Map<KeyType, ValueType> map) {
		return this.add(new FromJavaMapFixedWrapper<KeyType, ValueType>(map));
	}
	
	@Deprecated
	@Override
	public final boolean putAll(final java.util.Map<KeyType, ValueType> map) {
		return this.add(map);
	}
	
	@Override
	public boolean add(final Map<KeyType, ValueType, ?> map){
		boolean modified = false;
		for(Entry<KeyType, ValueType> entry: map){
			if(this.add(entry)){
				modified = true;
			}
		}
		return modified;
	}
	
	@Override
	public void set(final KeyType key, final ValueType value) {
		this.javaMap.put(key, value);
	}
	
	@Deprecated
	@SuppressWarnings("unchecked")
	@Override
	public final boolean remove(final Object o){
		return this.removeKey((KeyType)o);
	}
	
	@Override
	public boolean removeKey(final KeyType key) {
		ValueType value = this.javaMap.remove(key);
		return value != null;
	}
	
	@SuppressWarnings({ "unchecked", "rawtypes" })
	@Override
	public boolean removeAll(final java.util.Collection<?> c) {
		return this.removeAll(new FromJavaCollectionWrapper(c));
	}
	
	@Override
	public boolean removeAll(final Collection<?> c) {
		boolean modified = false;
		ExtendibleIterator<FromJavaEntryExtendibleWrapper<KeyType, ValueType>> e = this.iterator();
		while (e.hasNext()) {
		    if (c.contains(e.next())) {
			e.remove();
			modified = true;
		    }
		}
		return modified;
	}
	
	@SuppressWarnings({ "unchecked", "rawtypes" })
	@Override
	public boolean retainAll(final java.util.Collection<?> c) {
		return this.retainAll(new FromJavaCollectionWrapper(c));
	}

	@Override
	public boolean retainAll(final Collection<?> c) {
		boolean modified = false;
		ExtendibleIterator<FromJavaEntryExtendibleWrapper<KeyType, ValueType>> e = this.iterator();
		while (e.hasNext()) {
		    if (!c.contains(e.next())) {
			e.remove();
			modified = true;
		    }
		}
		return modified;
	}
	
	@Override
	public void clear() {
		this.javaMap.clear();
	}
	
	@Override
	public boolean remove(final Entry<KeyType, ValueType> o) {
		KeyType key = o.key();
		if(this.has(key)){
			FromJavaEntryExtendibleWrapper<KeyType, ValueType> entry = this.entry(key);
			if(o.value().equals(entry.value())){
				entry.remove();
				return true;
			}
			else{
				return false;
			}
		}
		else{
			return false;
		}
	}

	@Override
	public int size() {
		return this.javaMap.size();
	}

	@Override
	public NavigableMap<KeyType, ValueType, ?> subMap(final KeyType fromKey, final boolean fromInclusive, final KeyType toKey, final boolean toInclusive) {
		return new FromJavaNavigableMapExtendibleWrapper<KeyType, ValueType>(this.javaMap.subMap(fromKey, fromInclusive, toKey, toInclusive));
	}

	@Override
	public NavigableMap<KeyType, ValueType, ?> tailMap(final KeyType fromKey, final boolean inclusive) {
		return new FromJavaNavigableMapExtendibleWrapper<KeyType, ValueType>(this.javaMap.tailMap(fromKey, inclusive));
	}

	@Override
	public Collection<ValueType> values() {
		return new FixedCollectionWrapper<ValueType>(new FromJavaCollectionWrapper<ValueType>(this.javaMap.values()));
	}
	
	@Override
	public Comparator<? super KeyType> comparator() {
		return this.javaMap.comparator();
	}
	
}
