package org.postgresql.event_mill;

import java.util.Properties;

import de.lmu.ifi.pms.cep.event_mill.configuration.SQL_DatabaseDriver;
import de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID;

public class PostGreDriver extends SQL_DatabaseDriver {

	public PostGreDriver() throws ClassNotFoundException {
		super("org.postgresql.Driver", "postgresql");
	}	

	@Override
	protected boolean isSetLogging(final Properties properties) {
		return properties.containsKey("loglevel");
	}

	@Override
	protected void setLoggingProperty(final Properties properties, final LogLevel logLevel, final String logDestination) {
		if(logLevel.ordinal() <= LogLevel.DEBUG.ordinal()){
			properties.setProperty("loglevel", "2");
		}
		else if(logLevel.ordinal() <= LogLevel.INFO.ordinal()){
			properties.setProperty("loglevel", "1");
		}
	}
	
	@Override
	public String escape(final String s){
		//FIXME move to Driver
		return s.replaceAll("[']", "''");
	}
	
	@Override
	public boolean supportsExceptCorresponding(){
		return false;
	}
	
	@Override
	public String SQLType(final BasicTypeID basicType, final boolean autoincrement){
		if(autoincrement){
			switch (basicType) {
				case TIMESTAMP: throw new IllegalArgumentException();
				case DURATION: throw new IllegalArgumentException();
				case INT: return "SERIAL";
				case LONG: return "BIGSERIAL";
				case FLOAT: throw new IllegalArgumentException();
				case DOUBLE: throw new IllegalArgumentException();
				case BOOLEAN: throw new IllegalArgumentException();
				case STRING: throw new IllegalArgumentException();
				case IDENTIFIER: throw new IllegalArgumentException();
				case REFERENCE: throw new IllegalArgumentException();
			}
			throw new Error();
		}
		else{
			return this.SQLType(basicType);
		}
		//TODO better exception
	}
	
	@Override
	public String SQLType(final BasicTypeID basicType){
		switch (basicType) {
		case TIMESTAMP: return "BIGINT";
		case DURATION: return "BIGINT";
		case INT: return "INT";
		case LONG: return "BIGINT";
		case FLOAT: return "REAL";
		case DOUBLE: return "DOUBLE PRECISION";
		case BOOLEAN: return "BOOLEAN";
		case STRING: return "TEXT";
		case IDENTIFIER: return "BIGINT";
		case REFERENCE: return "BIGINT";
		}
		throw new Error();
		//TODO better exception
	}
	
	@Override
	public String sqlCreateSequence(final String program_meta_schema, final String sequenceName){
		return "CREATE SEQUENCE \"" + program_meta_schema + "\".\"" + sequenceName + "\" MINVALUE 0 START WITH 0;";
	}
	
	@Override
	public String sqlInitSequence(final String program_meta_schema, final String sequenceName){
		return "SELECT " + this.sqlGetSequenceNextValueFunction(program_meta_schema, sequenceName) + ";";
	}

	@Override
	public String sqlGetSequenceNextValueFunction(final String parameter){
		return "NEXTVAL(" + parameter + ")"; 
	}
	
	@Override
	public String sqlGetSequenceNextValueParameter(final String program_meta_schema, final String sequenceName){
		return "'\"" + program_meta_schema + "\".\"" + sequenceName + "\"'"; 
	}
	
	@Override
	public String sqlGetSequenceCurrentValueFunction(final String parameter){
		return "(CURRVAL(" + parameter + ") + 1)"; 
	}
	
	@Override
	public String sqlGetSequenceCurrentValueParameter(final String program_meta_schema, final String sequenceName){
		return "'\"" + program_meta_schema + "\".\"" + sequenceName + "\"'"; 
	}
}
