package de.lmu.ifi.pms.cep.unitTests;

import static de.lmu.ifi.pms.cep.tsa.schema.BasicTypeID.*;
import static de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchemaTypeID.*;

import java.sql.Connection;
import java.sql.DriverManager;

import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import de.lmu.ifi.pms.cep.event_mill.sql.SQL_StreamBuffer;
import de.lmu.ifi.pms.cep.event_mill.sql.SQL_StreamBufferStore;
import de.lmu.ifi.pms.cep.tsa.conditions.ConditionFormula;
import de.lmu.ifi.pms.cep.tsa.schema.TemporalStreamSchema;
import de.lmu.ifi.pms.cep.tsa.schema.Type;


@SuppressWarnings("unqualified-field-access")
public class SQL_StreamBuffer_Test
{
	
	Connection con;
	String catalog = "demo";
	SQL_StreamBufferStore streamBufferStore;
	
	Type int32  = INT.type;
	Type int64  = LONG.type;
	Type string = STRING.type;
	
	TemporalStreamSchema bidType = EVENT.schema(
			new String[]{"auction", "value"}, 
			new Type[]  {this.int64    , this.int32},
			new ConditionFormula[]{},
			new ConditionFormula[]{}
		);
	
	TemporalStreamSchema auctionType = STATE.schema(
			new String[]{"startime", "endtime", "product"}, 
			new Type[]  {this.int64     , this.int64    , this.string},
			new ConditionFormula[]{},
			new ConditionFormula[]{}
		);
	
	TemporalStreamSchema auctionStateType = STATE.schema(
			new String[]{"startime", "endtime", "product"}, 
			new Type[]  {this.int64     , this.int64    , this.string},
			new ConditionFormula[]{},
			new ConditionFormula[]{}
		);
	
	TSA_MaterialisationPoint bid = new TSA_MaterialisationPoint("bid", this.bidType);
	TSA_MaterialisationPoint auction = new TSA_MaterialisationPoint("auction", this.auctionType);
	TSA_MaterialisationPoint auctionState = new TSA_MaterialisationPoint("auctionstate", this.auctionStateType);

	@BeforeClass
	public static void setUpBeforeClass() throws Exception
	{

	}

	@AfterClass
	public static void tearDownAfterClass() throws Exception
	{
	}

	@Before
	public void setUp() throws Exception
	{
		Class.forName("nl.cwi.monetdb.jdbc.MonetDriver");
		this.con = DriverManager.getConnection("jdbc:monetdb://localhost/demo", "monetdb", "monetdb");
		
		this.streamBufferStore = new SQL_StreamBufferStore(this.con, this.catalog);
	}

	@After
	public void tearDown() throws Exception
	{
	}

	@Test
	public void testSQL_StreamBuffer()
	{
		SQL_StreamBuffer bidBuffer = new SQL_StreamBuffer(this.bid, this.streamBufferStore);
		SQL_StreamBuffer auctionBuffer = new SQL_StreamBuffer(this.auction, this.streamBufferStore);
		SQL_StreamBuffer auctionStateBuffer = new SQL_StreamBuffer(this.auctionState, this.streamBufferStore);
		
		bidBuffer.clear();
		auctionBuffer.clear();
		auctionStateBuffer.clear();
		
		bidBuffer.delete();
		auctionBuffer.delete();
		auctionStateBuffer.delete();
	}

}
