package de.lmu.ifi.pms.cep.unitTests.time.constraints.tests;
import java.util.Arrays;

import org.junit.Test;

import de.lmu.ifi.pms.cep.tsa.time.analysis.SimpleArithmeticAnalyzer;
import de.lmu.ifi.pms.cep.tsa.time.constraints.ConjunctiveTemporalRelationFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.DisjunctiveTemporalRelationFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.SimpleTemporalFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.TemporalRelationFormula;
import de.lmu.ifi.pms.cep.tsa.time.constraints.enums.BinaryComparisonOperator;
import de.lmu.ifi.pms.java.util.collection.compatibility.ArrayList;
import de.lmu.ifi.pms.java.util.map.fixed.FixedArrayNavigableMap;


public class SimpleTemporalAnalyzerTest {
	
	FixedArrayNavigableMap<String, String> attributes = new FixedArrayNavigableMap<String, String>();
	
	@Test
	public void testSimpleFloydWarshall(){
		SimpleTemporalFormula stf = new SimpleTemporalFormula("v", BinaryComparisonOperator.LESS_EQUAL, "w",0);
		SimpleTemporalFormula stf2 = new SimpleTemporalFormula("w", BinaryComparisonOperator.LESS_EQUAL, "y",10000);
		TemporalRelationFormula trf = new ConjunctiveTemporalRelationFormula(new ArrayList<TemporalRelationFormula>(Arrays.asList(stf,stf2)));
		SimpleArithmeticAnalyzer sta = new SimpleArithmeticAnalyzer(trf);
		sta.calculate();
		SimpleTemporalFormula stf3 = new SimpleTemporalFormula("v", BinaryComparisonOperator.LESS_EQUAL, "y",100);
		TemporalRelationFormula trf2 = new ConjunctiveTemporalRelationFormula(new ArrayList<TemporalRelationFormula>(Arrays.asList(stf,stf2,stf3)));
		SimpleArithmeticAnalyzer sta2 = new SimpleArithmeticAnalyzer(trf2);
		sta2.calculate();
	}
	
	@Test
	public void testEpsilonCombinations(){
		System.out.println("Testing epsilon combinations");
		SimpleTemporalFormula stf = new SimpleTemporalFormula("v", BinaryComparisonOperator.LESS_EQUAL, "w",0);
		SimpleTemporalFormula stf2 = new SimpleTemporalFormula("w", BinaryComparisonOperator.LESS, "y",10000);
		SimpleTemporalFormula stf3 = new SimpleTemporalFormula("v", BinaryComparisonOperator.LESS, "y",10000);
		SimpleTemporalFormula stf4 = new SimpleTemporalFormula("w", BinaryComparisonOperator.LESS_EQUAL, "y",10000);
		SimpleTemporalFormula stf5 = new SimpleTemporalFormula("v", BinaryComparisonOperator.LESS, "y",10000);
		SimpleTemporalFormula stf6 = new SimpleTemporalFormula("v", BinaryComparisonOperator.LESS_EQUAL, "y",10000);
		TemporalRelationFormula trf = new ConjunctiveTemporalRelationFormula(stf,stf2,stf3);
		TemporalRelationFormula trf2 = new ConjunctiveTemporalRelationFormula(stf,stf4,stf5,stf6);
		ArrayList<TemporalRelationFormula> queue = new ArrayList<TemporalRelationFormula>(trf,trf2);
		for(TemporalRelationFormula tf : queue){
			SimpleArithmeticAnalyzer sta = new SimpleArithmeticAnalyzer(tf);
			sta.calculate();
			printOut(sta.distances());
		}
	}
	
	@Test 
	public void testNegativeNumbers(){
		System.out.println("Testing negative Numbers");
		SimpleTemporalFormula stf = new SimpleTemporalFormula("v", BinaryComparisonOperator.LESS_EQUAL, "w",-10);
		SimpleTemporalFormula stf2 = new SimpleTemporalFormula("w", BinaryComparisonOperator.LESS, "y",10000);
		SimpleTemporalFormula stf3 = new SimpleTemporalFormula("v", BinaryComparisonOperator.LESS, "y",10000);
		TemporalRelationFormula trf = new ConjunctiveTemporalRelationFormula(stf,stf2,stf3);
		SimpleArithmeticAnalyzer sta = new SimpleArithmeticAnalyzer(trf);
		sta.calculate();
		printOut(sta.distances());
	}
	@Test
	public void testDisjunction(){
		System.out.println("Test Disjunction");
		SimpleTemporalFormula stf = new SimpleTemporalFormula("v", BinaryComparisonOperator.LESS_EQUAL, "w",-10);
		SimpleTemporalFormula stf2 = new SimpleTemporalFormula("w", BinaryComparisonOperator.LESS, "y",10000);
		SimpleTemporalFormula stf3 = new SimpleTemporalFormula("v", BinaryComparisonOperator.LESS, "y",10000);
		DisjunctiveTemporalRelationFormula disj = new DisjunctiveTemporalRelationFormula(stf,stf2);
		ConjunctiveTemporalRelationFormula conj = new ConjunctiveTemporalRelationFormula(disj,stf3);
		SimpleArithmeticAnalyzer sta = new SimpleArithmeticAnalyzer(conj);
		sta.calculate();
		printOut(sta.distances());
	}
	//@Test
	public void testLargeFormula(){
		System.out.println("Large Formula");
		ArrayList<TemporalRelationFormula> formulas = new ArrayList<TemporalRelationFormula>();
		for(int i=0;i<100000;i++){
			SimpleTemporalFormula stf = new SimpleTemporalFormula("v", BinaryComparisonOperator.LESS_EQUAL, "w",-10);
			formulas.add(stf);
		}
		SimpleArithmeticAnalyzer sta = new SimpleArithmeticAnalyzer(new ConjunctiveTemporalRelationFormula(formulas));
		sta.calculate();
		printOut(sta.distances());
	}
	//@Test
	public void testComplexFormula(){
		TemporalRelationFormula stf1 = new SimpleTemporalFormula("w", BinaryComparisonOperator.LESS,"v",5);
		TemporalRelationFormula stf2 = new SimpleTemporalFormula("x", BinaryComparisonOperator.LESS,"y",10);
		TemporalRelationFormula stf3 = new SimpleTemporalFormula("z", BinaryComparisonOperator.LESS,"t",11);
		TemporalRelationFormula stf4 = new SimpleTemporalFormula("m", BinaryComparisonOperator.LESS,"n",11);
		TemporalRelationFormula stf5 = new SimpleTemporalFormula("a", BinaryComparisonOperator.LESS,"b",11);
		TemporalRelationFormula stf6 = new SimpleTemporalFormula("c", BinaryComparisonOperator.LESS,"d",11);
		TemporalRelationFormula disj = new DisjunctiveTemporalRelationFormula(stf1,stf2);
		TemporalRelationFormula disj2 = new DisjunctiveTemporalRelationFormula(stf3,stf4);
		TemporalRelationFormula disj3 = new DisjunctiveTemporalRelationFormula(stf5,stf6);
		TemporalRelationFormula conj = new ConjunctiveTemporalRelationFormula(disj,disj2,disj3);
		SimpleArithmeticAnalyzer sta = new SimpleArithmeticAnalyzer(conj);
		sta.calculate();
		printOut(sta.distances());
	}
	
	public static void printOut(final long[][] out){
		for(int x=0;x<out.length;x++){
			for(int y=0;y<out.length;y++){
				if((out[x][y]>>2)!=(Long.MAX_VALUE>>2)){
					System.out.print((out[x][y]>>2)+" "+(out[x][y]&0x3)+"\t");
				}else{
					System.out.print("N/A"+"\t");
				}
			}
			System.out.println();
		}
	}
}
